
/* fpg.h --
 *
 *     George Carvalho, Esq.
 *     Wei-ming Dai
 *     (Advising Professor - E. S. Kuh)
 *
 * This module defines and manipulates floor plan graphs.
 * A floor plan graph can be efficiently derived from a
 * pair of tile planes: horizontal and vertical tile planes.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 *
 * Needs: magic.h, geometry.h, tile.h, bear.h
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Include/fpg.h,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $
 */

#ifndef _FPG
#define _FPG /* once-only latch */

#ifndef _MAGIC
int err1 = Need_to_include_magic_header;
#endif _MAGIC

#ifndef _GEOMETRY
int err2 = Need_to_include_geometry_header;
#endif _GEOMETRY

#ifndef _TILES
int err3 = Need_to_include_tiles_header;
#endif _TILES

#ifndef _BEAR
int err4 = Need_to_include_bear_header;
#endif _BEAR


#define FPG_MOD_NAME "fpg"


/*
 * These definitions are used as array indicies.
 */
#define FPG_NORTH               0       /* North compass direction          */
#define FPG_EAST                1       /* East  compass direction          */
#define FPG_SOUTH               2       /* South compass direction          */
#define FPG_WEST                3       /* West  compass direction          */
#define FPG_MAX_COMPASS_DIR     4

typedef enum _SegmentDir
{
    FPG_HORIZONTAL,                     /* Indicate horizontal segment      */
    FPG_VERTICAL                        /* Indicate vertical segment        */
} SegmentDir;

#define FPG_NW                  1       /* north west wall junction point   */
#define FPG_CENTER              2       /* center wall junction point       */
#define FPG_SE                  3       /* south east wall junction point   */

typedef struct _Fpg_Room 
{
    struct _GenericCell *gcellPtr;
    Rect                rect;
    lsList              wallList[4];
} Fpg_Room;

typedef struct _WallJunction
{
    int         itrMark;                /* This mark is for the exclusive use
                                         * of an interator algorithum.
                                         */
    int         mark;                   /* This mark is intended for the user
                                         * and its default value of is
                                         * BEAR_UNMARKED.
                                         */
    /*
     * 'centerPoint' will contain the location of the center of the
     * intersection of the horizontal and vertical dominant tiles
     * that define a wall junction.  'nwPoint' and 'sePoint' will be
     * placed a bit to the north-west and south-east of 'centerPoint'
     * respectively.
     */
    Point       nwPoint, centerPoint, sePoint;

    int         degree;                 /* The number of wall segments 
                                         * incident to the wall junction.
                                         */
    struct _WallSegment *wallSegPtr[FPG_MAX_COMPASS_DIR];
                                        /* The wall segments incident 
                                         * to the junction.
                                         */
    Fpg_Room    *roomPtr;

    BearGeneric greModulePtr;           /* Reserved for the global routing
                                         * editor module.
                                         */
    BearGeneric ngrModulePtr;           /* Reserved for the new global
                                         * routing module.
                                         */
    BearGeneric nplModulePtr;           /* Reserved for the new placement
                                         * module.
                                         */

    ClientData  clientData;             /* This space for rent.             */
} WallJunction;

typedef struct _PseudoJunct
{
    BearId      id;
    Point       absPos;
    ClientData  clientData;             /* This space for rent.             */
} PseudoJunct;

typedef struct _LinkedWallJunction
{
    WallJunction               *junctPtr;
    struct _LinkedWallJunction *nextPtr;
} LinkedWallJunction;

#define FPG_WALL_JUNCT          1
#define FPG_PSEUDO_JCT          2
#define FPG_PIN                 3
#define FPG_PSEUDO_PIN          4
#define FPG_ROUTE_PIN           5
#define FPG_BBL_TERM            6
#define FPG_BBL_JUNCT           7

typedef struct _PinOrJunct {
    int unType;
    union {
        WallJunction        *junctPtr;      /* unType = FPG_WALL_JUNCT      */
        PseudoJunct         *pjunctPtr;     /* unType = FPG_PSEUDO_JCT      */
        struct _PinInst     *pinPtr;        /* unType = FPG_PIN             */
        struct _PseudoPin   *ppinPtr;       /* unType = FPG_PSEUDO_PIN      */
        struct _RoutePin    *rPinPtr;       /* unType = FPG_ROUTE_PIN       */
        struct term         *bblTermPtr;    /* unType = FPG_BBL_TERM        */
        struct srjun        *bblJunctPtr;   /* unType = FPG_BBL_JUNCT       */
    } unValue;
    Tile        *bnTilePtr;
    BearGeneric genericPtr;
} PinOrJunct;

typedef struct _WallSegment
{
    int          mark;        /* Marking of the wall segment: default 
                               * value is BEAR_UNMARKED 
                               */
    SegmentDir   direction;   /* Segment direction */

    struct _BottleNeck *bnPtr;
                              /* Pointer to the corresponding bottleneck.
                               * which is either a tile or a line.
                               *
                               * Note that more than one wall segment
                               * will point to the same bottleneck tile
                               * in case of an empty room.
                               */
    WallJunction *NWJunctPtr; /* North or west wall junction incident to the
                               * wall segment.
                               */
    WallJunction *SEJunctPtr; /* South or east wall junction incident to the
                               * wall segment.
                               */

    BearGeneric ngrModulePtr; /* Reserved for the new global
                               * routing module.
                               */

    ClientData   clientData;  /* This space for rent.                        */
} WallSegment;

typedef struct _LinkedWallSegment
{
    WallSegment                 *segPtr;
    struct _LinkedWallSegment   *nextPtr;
} LinkedWallSegment;

typedef struct _FloorPlanGraph {
    BearId              id;              /* Identification number.           */
    WallJunction        ***junctTable;
    int                 numHorzDomTiles; /* The number of horizontal dominant
                                          * tiles on the tile plane, which is
                                          * equal to the number of horizontal
                                          * walls in the floor plan.
                                          */
    int                 numVertDomTiles; /* The number of vertical dominant
                                          * tiles on the tile plane, which is
                                          * equal to the number of vertical
                                          * walls in the floor plan.
                                          */
    int                 numJuncts;       /* The number of wall junctions in 
                                          * the floor plan graph.
                                          */
    int                 numSegs;         /* The number of wall segments in 
                                          * the floor plan graph.
                                          */
    Fpg_Room            *mainRoomPtr;
    lsList              roomList;

    WallJunction        *ulJunctPtr;
    WallJunction        *urJunctPtr;
    WallJunction        *lrJunctPtr;
    WallJunction        *llJunctPtr;     /* The lower left wall junction in
                                          * the floor plan graph.
                                          */

    ClientData          clientData;      /* This space for rent.             */
} FloorPlanGraph;

/* ---------- Floor Plan Graph Primitives ---------- */

extern bool Fpg_IsFpgBuilt(/* planeDataPtr */);
/*
 * PlaneData *planeDataPtr;
 */

extern WallJunction *Fpg_GetLLJunctOfFpg(/* fpgPtr */);
/*
 * FloorPlanGraph *fpgPtr;
 */

extern int Fpg_GetNumOfWallJunctsOfFpg(/* fpgPtr */);
/*
 * FloorPlanGraph *fpgPtr;
 */

extern int Fpg_GetNumOfWallSegsOfFpg(/* fpgPtr */);
/*
 * FloorPlanGraph *fpgPtr;
 */

extern BearId Fpg_GetIdOfFpg(/* fpgPtr */);
/*
 * FloorPlanGraph *fpgPtr;
 */

extern ClientData Fpg_GetClientDataOfFpg(/* fpgPtr */);
/*
 * FloorPlanGraph *fpgPtr;
 */

extern void Fpg_PutClientDataOfFpg(/* clientData, fpgPtr */);
/*
 * ClientData   clientData;
 * FloorPlanGraph *fpgPtr;
 */


/* ---------- Wall Junction Primitives ---------- */

extern void Fpg_MarkWallJunct(/* junctPtr */);
/*
 * WallJunction *junctPtr;
 */

extern void Fpg_UnmarkWallJunct(/* junctPtr */);
/*
 * WallJunction *junctPtr;
 */

extern bool Fpg_IsWallJunctMarked(/* junctPtr */);
/*
 * WallJunction *junctPtr;
 */

extern int Fpg_GetDegreeOfWallJunct(/* junctPtr */);
/*
 * WallJunction *junctPtr;
 */

extern Point *Fpg_GetPointOfWallJunct(/* whichPoint, junctPtr */);
/*
 * int          whichPoint;
 * WallJunction *junctPtr;
 */

extern WallSegment *Fpg_GetNorthSegOfWallJunct(/* junctPtr */);
/*
 * WallJunction *junctPtr;
 */

extern WallSegment *Fpg_GetSouthSegOfWallJunct(/* junctPtr */);
/*
 * WallJunction *junctPtr;
 */

extern WallSegment *Fpg_GetEastSegOfWallJunct(/* junctPtr */);
/*
 * WallJunction *junctPtr;
 */

extern WallSegment *Fpg_GetWestSegOfWallJunct(/* junctPtr */);
/*
 * WallJunction *junctPtr;
 */

extern ClientData Fpg_GetClientDataOfWallJunct(/* junctPtr */);
/*
 * WallJunction *junctPtr;
 */

extern void Fpg_PutClientDataOfWallJunct(/* clientData, junctPtr */);
/*
 * ClientData   clientData;
 * WallJunction *junctPtr;
 */

extern WallJunction *
Fpg_GetWallJunctWhereDomTilesCross(/* fpgPtr, horzDomTilePtr, vertDomTilePtr*/);
/*
 * FloorPlanGraph  *fpgPtr;
 * Tile            *horzDomTilePtr, *vertDomTilePtr;
 */


/* ---------- Wall Segment Primitives ---------- */

extern void Fpg_MarkWallSeg(/* segPtr */);
/*
 * WallSegment *segPtr;
 */

extern void Fpg_UnmarkWallSeg(/* segPtr */);
/*
 * WallSegment *segPtr;
 */

extern bool Fpg_IsWallSegMarked(/* segPtr */);
/*
 * WallSegment *segPtr;
 */

extern WallJunction *Fpg_GetSEJunctOfWallSeg(/* segPtr */);
/*
 * WallSegment *segPtr;
 */

extern WallJunction *Fpg_GetNWJunctOfWallSeg(/* segPtr */);
/*
 * WallSegment *segPtr;
 */

extern struct _BottleNeck *Fpg_GetBottleNeckOfWallSeg(/* segPtr */);
/*
 * WallSegment *segPtr;
 */

extern ClientData Fpg_GetClientDataOfWallSeg(/* segPtr */);
/*
 * WallSegment *segPtr;
 */

extern void Fpg_PutClientDataOfWallSeg(/* clientData, segPtr */);
/*
 * ClientData   clientData;
 * WallSegment *segPtr;
 */

extern BearStatus Fpg_FindWallSeg(/* junct1Ptr, junct2Ptr, segPtrPtr */);
/*
 * WallJunction *junct1Ptr, *junct2Ptr;
 * WallSegment  **segPtrPtr;
 */


/* ---------- Basic Primitives ---------- */

extern BearStatus
Fpg_CreateFloorPlanGraph(/* planeDataPtr, llJunctPtrPtr */);
/*
 * PlaneData    *planeDataPtr;
 * WallJunction **llJunctPtrPtr;
 */

extern void
Fpg_DeleteFloorPlanGraph(/* planeDataPtr */);
/*
 * PlaneData *planeDataPtr;
 */

extern int Fpg_OppositeDir(/* direction */);
/*
 * int direction;
 */

extern int Fpg_MacPinDirToFpgDir(/* pinDir */);
/*
 * PinDirection pinDir;
 */

extern BearStatus
Fpg_GetAbsPosOfPinOrJunct(/* planeDataPtr, pinOrJunctPtr, absPosPtr */);
/*
 * PlaneData   *planeDataPtr;                      (* IN       *)
 * PinOrJunct  *pinOrJunctPtr;                     (* IN       *)
 * Point       *absPosPtr;                         (* OUT      *)
 */

extern BearStatus Fpg_GenRooms(/* cellMasterPtr */);
/*
 * CellMaster  *coreMasterPtr;                     (* IN       *)
 */

extern PinOrJunct  *Fpg_MakePinOrJunct();

extern PseudoJunct *Fpg_MakePseudoJunct();


/* ---------- Interators ---------- */

extern BearStatus Fpg_ItrWallSeg(/* llJunctPtr, func, clientData */);
/*
 * WallJunction    *llJunctPtr; 
 * BearStatus      (*func)();
 * ClientData      clientData;
 */

extern BearStatus Fpg_ItrWallSegSeq(/* fpgPtr, func, clientData */);
/*
 * FloorPlanGraph  *fpgPtr;  
 * BearStatus      (*func)();
 * ClientData      clientData;
 */

extern BearStatus Fpg_ItrWallJunct(/* llJunctPtr, func, clientData */);
/*
 * WallJunction    *llJunctPtr;  
 * BearStatus      (*func)();
 * ClientData      clientData;
 */

extern BearStatus Fpg_ItrWallJunctSeq(/* fpgPtr, func, clientData */);
/*
 * FloorPlanGraph  *fpgPtr;  
 * BearStatus      (*func)();
 * ClientData      clientData;
 */

extern BearStatus Fpg_ItrBottleNeckOfFpg(/* fpgPtr, func, clientData */);
/*
 * FloorPlanGraph  *fpgPtr;  
 * BearStatus      (*func)();
 * ClientData      clientData;
 */


/* Macros */

#define Fpg_GetNorthJunctOfWallJunct(junct) \
    ((Fpg_GetNorthSegOfWallJunct(junct) == NIL(WallSegment)) \
    ? (NIL(WallJunction)) \
    : (Fpg_GetNWJunctOfWallSeg(Fpg_GetNorthSegOfWallJunct(junct))))

#define Fpg_GetSouthJunctOfWallJunct(junct) \
    ((Fpg_GetSouthSegOfWallJunct(junct) == NIL(WallSegment)) \
    ? NIL(WallJunction) \
    : (Fpg_GetSEJunctOfWallSeg(Fpg_GetSouthSegOfWallJunct(junct))))

#define Fpg_GetEastJunctOfWallJunct(junct) \
    ((Fpg_GetEastSegOfWallJunct(junct) == NIL(WallSegment)) \
    ? NIL(WallJunction) \
    : (Fpg_GetSEJunctOfWallSeg(Fpg_GetEastSegOfWallJunct(junct))))

#define Fpg_GetWestJunctOfWallJunct(junct) \
    ((Fpg_GetWestSegOfWallJunct(junct) == NIL(WallSegment)) \
    ? NIL(WallJunction) \
    : (Fpg_GetNWJunctOfWallSeg(Fpg_GetWestSegOfWallJunct(junct))))

#endif _FPG

