
/* newCInt.h --
 *
 *    Bernhard Eschermann / Yasushi Ogawa / Massoud Pedram
 *    George Carvalho, Esq.
 *    (Advising Professor - E. S. Kuh)
 *
 *    Contains the internal structure definitions for the clustering algorithm.
 *
 * Copyright (C) 1991 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Installed/newC/newCInt.h,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $
 */

#ifndef _NEW_C_INT_H
#define _NEW_C_INT_H /* once-only latch */

#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "newC.h"


#define BLOCK 0

/*
 * NCL_REAL_MAX is used as maximum value for both double and float
 * variables.  For our purposes it should be bigger then the square
 * of the largest posible int variable, i.e. 2147483647 * 2147483647.
 */
#define NCL_REAL_MAX 1.0e+25

#define NCL_DEF_NUM_OF_SHAPES 30

#define MAX_CORNERS 3
#define LOWER       0
#define MIDDLE      1
#define UPPER       2


/* ---------- Internal structure definitions ---------- */


typedef struct _GraphNode GraphNode;
struct _GraphNode {
    int                 gn_pos;
    int                 numOfSubnodes;      /* # of dependent leaf nodes */
    CellInst            *gn_cellInstPtr;    /* pointer to cell in BEAR */
    Point               gn_dim;             /* x and y width of bounding box */
    double              gn_area;            /* sum of area of leaf nodes */
    double              intraConn;          /* conn. within cluster */
    double              interConn;          /* conn. with other clusters */
    double              maxArea;            /* max. blocksize in cluster */
    double              minArea;            /* min. blocksize in cluster */
    int                 gn_maxDim;          /* max of extreme width/length
                                             * in leaves 
                                             */
    GraphNode           *belongsToPtr;      /* cluster to which this node
                                             * is attached 
                                             */
    struct _ConnEdge    *edgeList;          /* pointers to connecting edges */
    double              totalConn;          /* sum of connectivities of all
                                             * edges 
                                             */
    int                 gn_numOfChildren;   /* number of clustered nodes */
    GraphNode           *gn_children[MAX_CHILDREN];
                                            /* pointers to clustered nodes */
    Ncl_Node            *gn_members[MAX_CHILDREN];
                                            /*   -"-     in cluster_Tree */
    int                 pinCount[4];        /* # of pins on the 4 sides */
    GraphNode           *nextPtr;           /* next list element */
    GraphNode           *prevPtr;           /* previous list element */
    Ncl_ShapeSet        *gn_shapeSet;    /* shape function */
    Ncl_LengthSet       *gn_lengthSet;   /* length function */
    Ncl_LinkedCritNetInNode *gn_lnkCritNetList;
};

typedef struct _GraphEdge GraphEdge;
struct _GraphEdge {
    double              connect1;           /* weight of this edge during clustering */
    double              connect2;           /* weight during placement */
    double              cycleWeight;        /* weight of cycles to which
                                             * this edge belongs */
    double              sortingCrit;        /* determines the sequence to
                                             * process the edges */
    int                 edgeType;           /* edge to I/O pad or to cell ? */
    GraphNode           *node1Ptr,          /* nodes connected by this edge */
                        *node2Ptr;
    struct _Ncl_LinkedCritNetOnEdge *critOnEdgePtr;   /* timing-driven */
    GraphEdge           *nextPtr;           /* next list element */
};

typedef struct _ConnEdge ConnEdge;
struct _ConnEdge { /* linked list of GraphEdge */
    GraphEdge   *edgePtr;
    ConnEdge    *nextPtr;
};

typedef struct _ConnectedNode ConnectedNode;
struct _ConnectedNode {
    GraphNode       *nodePtr;
    ConnectedNode   *nextPtr;
};

typedef struct ncl_PositionData {
     CellInst  *cellInstPtr;
     Ncl_Node  *nodePtr;
     Rect      *boxPtr;
     Point     position;
} PositionData;

typedef struct ncl_MatrixHolder {
    Point matrix[MAX_CHILDREN][3];
} MatrixHolder;

/*
 * Yet another type of node !
 */
typedef struct _InputNode Ncl_InputNode;
struct _InputNode {
    int             in_pos;
    char            *in_name;
    int             in_level;

    bool            in_isRootFlag;
    bool            in_isParentFlag;
    bool            in_isChildFlag;
    Ncl_InputNode   *in_parentPtr;

    int             in_numOfChildren;
    Ncl_InputNode   *in_child[MAX_CHILDREN];
    CellInst        *in_cellInstPtr;
};

typedef struct _InputTree Ncl_InputTree;
struct _InputTree {
    Ncl_InputNode   *it_rootPtr;    /* Root node of the tree. */
    int             it_depth;       /* Depth of the tree. */
};

typedef struct _Ncl_LinkedBinaryMerged Ncl_LinkedBinaryMerged;
struct _Ncl_LinkedBinaryMerged {
    float                   bm_x, bm_y;
    Ncl_ShapeData           *src1, *src2;
    Ncl_LinkedBinaryMerged  *nextPtr;
};


/* ---------- Internal procedures for the newC module ---------- */

extern BearStatus NclBuildTree(/* cellMasterPtr */);
/*
 * Function takes a pointer to a cellMaster, clusters all the cells in
 * cellMaster and stores the generated information in the global variable
 * ncl_ClusterTree.
 *
 * CellMaster   *cellMasterPtr;
 */

extern  Ncl_LengthData  *NclInitLengths();
extern  void            NclPrintLengthSet();
extern  void            NclCopyLengthSet();
extern  void            NclReInitLengths();
extern  BearStatus      NclMergeLengthSets();

extern  float           NclGetIntraBlockL();
extern  float           NclGetInterBlockL();
extern  BearStatus      NclGetLowerBnds();

extern  double          NclPtime();
extern  void            NclEvaluateClustering();

extern  void            NclSetUsedLengthsForCritNets();
extern  void            NclMakeConnectivityForCritNet();
extern  void            NclAssignTimingWeights();
extern  void            NclRemoveTimingWeights();

/*
 * nclInput.c
 */
extern BearStatus       NclGetNetFromBEAR();

/*
 * nclMain.c
 */
extern BearStatus       NclClusterGen2(); 

/*
 * nclCluster.c
 */
extern bool             NclClusterMerge();
extern bool             NclMergeTest();
extern void             NclSortEdges();
extern double           NclCycleDetect(), NclEval();
extern double           NclSumConn();

/*
 * nclUtilities.c
 */
extern BearStatus       NclTidyUpCells();
extern Ncl_Node         *NclMakeNode();
extern char             *NclGetAlgString();

/*
 * nclReadData.c
 */
extern BearStatus       NclReadTimingData();

/*
 * nclInputTree.c
 */
extern BearStatus       NclReadCTree();
extern BearStatus       NclGenLevelLists();
extern void             NclFreeInputTree();

/*
 * nclWriteData.c
 */
extern BearStatus       NclWriteCTree();


/* ---------- Internal global variables for the newC module ---------- */
extern BearString   nclCTreeFileName;   /* cluster tree file name */
extern BearString   nclStFileName;      /* storage file name      */
extern BearString   nclTmFileName;      /* timing file name       */
extern double       nclAreaFact, nclAreaRelax;
extern double       nclRatioMax, nclRatioRelax;
extern double       nclConnMin, nclRoutAdjust;
extern int          nclAlgorithum;

#include "nclMacros.h"
#include "nclTemplates.h"

/* ---------- Declarations of external procedures ---------- */
extern char *util_strsav();

#endif _NEW_C_INT_H


