
/* gtGraphItr.c --
 *
 *     George Carvalho, Esq.
 *     Wei-Ming Dai
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains the iterators related to a graph.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/gtGraph/gtGraphItr.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include <stdio.h>
#include "gtBearStuff.h"
#include "gtGraph.h"
#include "gtGraphInt.h"


/*
 * ----------------------------------------------------------------------------
 *
 * Gt_ItrNodesOfGraph() --
 *
 * Visit each node in the graph, and apply the given procedure to it. 
 * The procedure should be of the following form:
 *
 *  BearStatus
 *  userProc(node, clientData)
 *  Gt_GraphNode    node;
 *  ClientData      clientData;
 *  {
 *  }
 *
 * userProc normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus
Gt_ItrNodesOfGraph(graph, userProc, clientData)
Gt_Graph    graph;                              /* IN       */
BearStatus  (*userProc)();                      /* IN       */
ClientData  clientData;                         /* IN       */
{
    lsGen       gen;
    vertex_t    v;

#ifdef DEBUG
    if (graph == (Gt_Graph) NULL) {
        Bear_PrintFatalErr("Gt_ItrNodesOfGraph() failed: no graph given");
    } 
    else if (userProc == (BearStatus (*)()) NULL) {
        Bear_PrintFatalErr("Gt_ItrNodesOfGraph() failed: no procedure given");
    }
#endif DEBUG

    foreach_vertex(((Gt_GraphInt *) graph)->g, gen, v) {
        if (((*userProc)((Gt_GraphNode) v->user_data, clientData)) ==
            BEAR_STOP_SEARCH) {

            lsFinish(gen);
            return (BEAR_STOP_SEARCH);
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Gt_ItrEdgesOfGraph() --
 *
 * Visit each edge in the graph, and apply the given procedure to it. 
 * The procedure should be of the following form:
 *
 *  BearStatus
 *  userProc(edge, clientData)
 *  Gt_GraphEdge    edge;
 *  ClientData      clientData;
 *  {
 *  }
 *
 * userProc normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus
Gt_ItrEdgesOfGraph(graph, userProc, clientData)
Gt_Graph    graph;                              /* IN       */
BearStatus  (*userProc)();                      /* IN       */
ClientData  clientData;                         /* IN       */
{
    lsGen       gen;
    edge_t      e;

#ifdef DEBUG
    if (graph == (Gt_Graph) NULL) {
        Bear_PrintFatalErr("Gt_ItrEdgesOfGraph() failed: no graph given");
    } 
    else if (userProc == (BearStatus (*)()) NULL) {
        Bear_PrintFatalErr("Gt_ItrEdgesOfGraph() failed: no procedure given");
    }
#endif DEBUG

    foreach_edge(((Gt_GraphInt *) graph)->g, gen, e) {
        if (((*userProc)((Gt_GraphEdge) e->user_data, clientData)) ==
            BEAR_STOP_SEARCH) {

            lsFinish(gen);
            return (BEAR_STOP_SEARCH);
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Gt_ItrEdgesOfNode() --
 *
 * Visit each edge connected to the given node, and apply the given
 * procedure to it. The procedure should be of the following form:
 *
 *  BearStatus
 *  userProc(edge, clientData)
 *  Gt_GraphEdge    edge;
 *  ClientData      clientData;
 *  {
 *  }
 *
 * userProc normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus
Gt_ItrEdgesOfNode(node, userProc, clientData)
Gt_GraphNode node;                              /* IN       */
BearStatus   (*userProc)();                     /* IN       */
ClientData   clientData;                        /* IN       */
{
    edge_t      e;
    lsGen       gen;
    vertex_t    v;

#ifdef DEBUG
    if (userProc == (BearStatus (*)()) NULL) {
        Bear_PrintFatalErr("Gt_ItrEdgesOfNode() failed: no procedure given");
    }
#endif DEBUG

    v = ((Gt_GraphNodeInt *) node)->v;
    foreach_in_edge(v, gen, e) {
        if (((*userProc)((Gt_GraphEdge) e->user_data, clientData)) ==
            BEAR_STOP_SEARCH) {

            lsFinish(gen);
            return (BEAR_STOP_SEARCH);
        }
    }
    foreach_out_edge(v, gen, e) {
        if (((*userProc)((Gt_GraphEdge) e->user_data, clientData)) ==
            BEAR_STOP_SEARCH) {

            lsFinish(gen);
            return (BEAR_STOP_SEARCH);
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



