
/* cpaMain.c --
 *
 *     Benjamin Chen
 *     Electronics Research Laboratory
 *     University of California, Berkeley
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/koalaBear/cpaMain.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#ifdef CPA_BEAR
#include "bearCopyRight.h"
#endif CPA_BEAR
#include <stdio.h>
#include "cpaList.h"
#include "cpa.h"
#include "cpaInt.h"

int AssignPinColumn();
int AssignMatch();
void AssignSlots();
void AssignAvailSlots();


List            *globalPinList;
NetClassList    *topNetList;
NetClassList    *bottomNetList;
char            *cpaInFiName;       /* Input file name  */
char            *cpaOutFiName;      /* Output file name */


/*
 * ArrangePins -- 
 *      Algorithm: Optimal Pin Arrangement for a Basic Channel
 *                 Jingsheng (Jason) Cong
 * Input: A channel containing pins on the top edge and the bottom edge of
 *        the channel, end exits on the left side and the right side of the
 *        channel.  The nets must have already been classified. (see cpa.h)
 *
 * Step 1: Starting from the left side of the channel, assign all matched LS
 *         nets pair-by-pair, one pair per column; Similarly, starting from
 *         the right right side of the channel, assign all the RP nets
 *         pair-by-pair, one pair per column.
 * Step 2: Starting from the leftmost available position, assign all the
 *         LP nets one-by-one, one net per column; Similarly, starting from
 *         the rightmost available position, assign all the RP nets 
 *         one-by-one, one net per column.
 * Step 3: Starting from the leftmost available position, assign unmatched
 *         LS nets one-by-one, one net per column.
 * Step 4: Starting from the leftmost available position, assign all unmatched
 *         IP nets one-by-one, one net per column.
 * Step 5: Starting from the rightmost available position, assign all unmatched
 *         RS nets one-by-one, some of them may share a column with unmatched 
 *         LS nets assigned in Step 3.
 * Step 6: Assign BS nets and BP nets at any available columns.  They may
 *         share columns with unmatched LS nets or RS nets.
 */
void ArrangePins(channel, topNetList, bottomNetList)
     BChannel *channel;
     NetClassList *topNetList, *bottomNetList;
{
    int leftColumn, rightColumn,  numSlots, i;
    Pin **topSlots;
    Pin **bottomSlots;
    /* This is temporary, as we assume a rectangular channel for now. */

    numSlots = C_TOP(channel) / C_SPACING(channel);
    leftColumn = 1;
    rightColumn = numSlots;
    MALLOC(Pin **, topSlots, sizeof(Pin) * numSlots);
    MALLOC(Pin **, bottomSlots, sizeof(Pin) * numSlots);
    for(i = 0; i < numSlots; i++) {
    topSlots[i] = (Pin *) NULL;
    bottomSlots[i] = (Pin *) NULL;
    }
    
    AssignMatch(&LS_NETS(topNetList), &LS_NETS(bottomNetList), 
        topSlots, bottomSlots, &leftColumn, LEFT_EXIT);
    AssignMatch(&RS_NETS(topNetList), &RS_NETS(bottomNetList), 
        topSlots, bottomSlots, &rightColumn, RIGHT_EXIT);
    /* The following nets exist in both netlists. */
    AssignSlots(&LP_NETS(topNetList), &leftColumn, 
           topSlots, bottomSlots, LEFT_EXIT);
    AssignSlots(&RP_NETS(topNetList), &rightColumn, 
           topSlots, bottomSlots, RIGHT_EXIT);
    AssignSlots(&LS_NETS(topNetList), &leftColumn, 
           topSlots, bottomSlots, LEFT_EXIT); 
    AssignSlots(&LS_NETS(bottomNetList), &leftColumn, 
           topSlots, bottomSlots, LEFT_EXIT); /* Assign rest of LS nets */
    AssignSlots(&IP_NETS(topNetList), &leftColumn, 
           topSlots, bottomSlots, LEFT_EXIT);
    AssignSlots(&RS_NETS(topNetList), &rightColumn, 
           topSlots, bottomSlots, RIGHT_EXIT);
    AssignSlots(&RS_NETS(bottomNetList), &rightColumn, 
           topSlots, bottomSlots, RIGHT_EXIT); /* Assign rest of RS nets */
    AssignAvailSlots(&BP_NETS(topNetList), &leftColumn, &rightColumn, numSlots,
             topSlots, bottomSlots); 
    AssignAvailSlots(&BS_NETS(topNetList), &leftColumn, &rightColumn, numSlots,
             topSlots, bottomSlots);
    AssignAvailSlots(&BS_NETS(bottomNetList), &leftColumn, &rightColumn,
             numSlots, topSlots, bottomSlots);
}

/*
 * AssignPinColumn --
 *     This procedure assigns the pin contained by the entry to the column.
 *     The original pin location on the channel, i.e. top/bottom edge
 *     is preserved (i.e. positive or negative value for the column.
 *     Return 1 if successful, 0 otherwise.
 */
int AssignPinColumn(pinEntry, column, topSlots, bottomSlots)
     List *pinEntry;
     int column;
     Pin **topSlots, **bottomSlots;
{
    Pin *pin;
    
    pin = (Pin *) L_ITEM(pinEntry);
    
    if (P_COLUMN(pin) > 0) {
    if (topSlots[column] != (Pin *) NULL) return 0;
    P_COLUMN(pin) = column;
    topSlots[column] = pin;
    }
    else {
    if (bottomSlots[column] != (Pin *) NULL) return 0;
    P_COLUMN(pin) = -column;
    bottomSlots[column] = pin;
    }
    return 1;
}

/*
 * AssignMatchLeft --
 *     This procedure matches each entry of both lists to the same column
 *     of the target channel, starting from the column indicated by 
 *     columnStart  As entries are assigned, they are permanently
 *     deleted from the list.  Pins in netList1 and netList2 should be
 *     single pin nets.
 */
int AssignMatch(netList1, netList2, topSlots, bottomSlots, columnStart, exit)
     List **netList1, **netList2;
     Pin **topSlots, **bottomSlots;
     int *columnStart;
     Exit exit;  
{
    List *current1, *current2;
    BNet *net1, *net2;
    int column;

    column = *columnStart;
    current1 = *netList1;
    current2 = *netList2;
    while(current1 != (List *) NULL && current2 != (List *) NULL) {
    net1 = (BNet *) L_ITEM(current1);
    net2 = (BNet *) L_ITEM(current2);
    /* There should be only one pin for each net */
    (void) AssignPinColumn(N_PINS(net1), column, topSlots, bottomSlots);
    (void) AssignPinColumn(N_PINS(net2), column, topSlots, bottomSlots);
    DeleteEntry(&current1);
    DeleteEntry(&current2);
    if (exit == LEFT_EXIT) column++;
    else column--;
    }
    *columnStart = column;
    *netList1 = current1;
    *netList2 = current2;
}

/*
 * AssignSlots --
 *     This procedure assigns all pins in the given netlist, starting
 *     from the column indicated by columnStart.  As entries are assigned,
 *     they are permanently deleted from the list. 
 */
void AssignSlots(netList, columnStart, topSlots, bottomSlots, exit)
     List **netList;
     int *columnStart;
     Pin **topSlots, **bottomSlots;
     Exit exit;
{
    List *current;
    BNet *net;
    int column;

    column = *columnStart;
    current = *netList;
    while(current != (List *) NULL) {
    net = (BNet *) L_ITEM(current);
    (void) AssignPinColumn(N_PINS(net), column, topSlots, bottomSlots);
    if (NEXT(N_PINS(net)) != (List *) NULL)
        (void) AssignPinColumn(NEXT(N_PINS(net)), column, 
                   topSlots, bottomSlots);
    DeleteEntry(&current);
    if (exit == LEFT_EXIT) column++;
    else column--;
    }
    *columnStart = column;
    *netList = current;
}

/*
 * AssignAvailSlots --
 *     This procedure assigns pins in the netlist to any available slots.
 *     Empty columns are searched first. When those are exhausted, empty
 *     slots are searched from left to right.
 */
void AssignAvailSlots(netList, leftStart, rightStart, numSlots, topSlots,
              bottomSlots)
     List **netList;
     int *leftStart, *rightStart;
     int numSlots;
     Pin **topSlots, **bottomSlots;
{
    List *current, *pinList;
    BNet *net;
    Pin *pin;
    int topIndex, botIndex, i;

    current = *netList;
    for (i = *leftStart; i <= *rightStart && current != (List *) NULL; i++) {
    net = (BNet *) L_ITEM(current);
    (void) AssignPinColumn(N_PINS(net), i, topSlots, bottomSlots);
    if (NEXT(N_PINS(net)) != (List *) NULL)
        (void) AssignPinColumn(NEXT(N_PINS(net)), i, 
                   topSlots, bottomSlots);
    DeleteEntry(&current);
    }
    *leftStart = i;
    if(current != (List *) NULL) {
    topIndex = botIndex = 1;
    while (topIndex < numSlots+1 && botIndex < numSlots +1
           && current != (List *) NULL) {
        net = (BNet *) L_ITEM(current);
        pinList = N_PINS(net);
        while (pinList != (List *) NULL) {
        pin = (Pin *) L_ITEM(pinList);
        if (P_COLUMN(pin) > 0) {
            while(AssignPinColumn(pinList, topIndex, topSlots, 
                      bottomSlots) == 0) topIndex++;
        }
        else {
            while(AssignPinColumn(N_PINS(net), botIndex, topSlots, 
                      bottomSlots) == 0) botIndex++;
        }
        pinList = NEXT(pinList);
        }
        DeleteEntry(&current);
    }
    }
}
