
/* macCellGroup.c --
 *
 *      June Wang
 *      (Advising Professor - E. S. Kuh)
 *
 * It has the subroutines to make a linked cell, and to create a cell group.
 * The only iterator is also in this file.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macCellGroup.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"

/*
 * ----------------------------------------------------------------------------
 *  MacFindEndOfCellGroup() --
 *
 *  It's called by Mac_MergeCellGroups().
 *  If the linked cell is the last one in the cell group then
 *  that pointer is stored as endCellPtr. 
 *
 * Results: 
 *   BEAR_COMPLETE_SEARCH, if the end of cell group is found.
 *   BEAR_CONTINUE_SEARCH, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
MacFindEndOfCellGroup(linkedCellPtr, endCellPtrPtr)

LinkedCell    *linkedCellPtr; 
LinkedCell    **endCellPtrPtr;
{
    if (linkedCellPtr->nextPtr == NIL(LinkedCell))
    {
        *endCellPtrPtr = linkedCellPtr;
        return (BEAR_STOP_SEARCH);
    }
    return (BEAR_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *  MacChangeCellGroupPtr() --
 *
 *  This is called by Mac_MergeCellGroups.
 *  It changes the CellGroup pointer of a linked cell to the one
 *  given by cellGroupPtr.  
 *
 *  Depending on the cell type of the given linked cell, the 
 *  cellGroupPtr field of that corresponding cell pointer will have 
 *  the value of cellGroupPtr.
 *
 * Returns: 
 *   BEAR_CONTINUE_SEARCH.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */
BearStatus
MacChangeCellGroupPtr(linkedCellPtr, cellGroupPtr)
LinkedCell  *linkedCellPtr;
CellGroup  *cellGroupPtr;

{
    switch (linkedCellPtr->cellPtr->unType)
    {
        case CELL_INST:
            linkedCellPtr->cellPtr->unValue.cellInstPtr->cellGroupPtr = cellGroupPtr;
            break;
        case ROUTE_CELL:
            linkedCellPtr->cellPtr->unValue.routeCellPtr->cellGroupPtr = cellGroupPtr;
    }
    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 * Mac_IsGroupedCell() --
 *
 *  It determines whether a cell instance or a route cell belongs to 
 *  a cell group.
 *
 * Results: 
 *   TRUE, if the cell belongs to a cell group.
 *   FALSE, if not.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

bool
Mac_IsGroupedCell(cellClass, cellPtr)
CellClass    cellClass;
ClientData   cellPtr;
{
    switch (cellClass) {
        case CELL_INST:
            return (((CellInst *) cellPtr)->cellGroupPtr != NIL(CellGroup));
        case ROUTE_CELL:
            return (((RouteCell *) cellPtr)->cellGroupPtr != NIL(CellGroup));
        default:
            MacPrintFatalErr("Mac_IsGroupedCell() failed : illegal cell class");
            break;
    }
    /* NOTREACHED */
}



/*
 * ----------------------------------------------------------------------------
 * Mac_IsMemberOfCellGroup() --
 *
 *  It determines whether a cell instance or a route cell belongs to 
 *  a given cell group.
 *
 * Results: 
 *   TRUE, if the cell belongs to the cell group.
 *   FALSE, if not.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

bool
Mac_IsMemberOfCellGroup(cellClass, cellPtr, cellGroupPtr)
CellClass   cellClass;
ClientData  *cellPtr;
CellGroup   *cellGroupPtr;
{
    switch(cellClass) {
        case CELL_INST:
            return (((CellInst *)cellPtr)->cellGroupPtr == cellGroupPtr);
        case ROUTE_CELL:
            return (((RouteCell *)cellPtr)->cellGroupPtr == cellGroupPtr);
        default:
    MacPrintFatalErr("Mac_IsMemberOfCellGroup() failed : illegal cell class");
            break;
    }
    /* NOTREACHED */
}



/*
 * ----------------------------------------------------------------------------
 * Mac_MakeLinkedCell() --
 *
 *  It allocates a memory space for a pointer to LinkedCell.
 *
 * Results: 
 *   A new pointer to LinkedCell. 
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */
LinkedCell *
Mac_MakeLinkedCell()
{
    LinkedCell  *newLinkedCellPtr;

    MALLOC(LinkedCell *, newLinkedCellPtr, sizeof (LinkedCell));
    newLinkedCellPtr->cellPtr = NIL(GenericCell);
    newLinkedCellPtr->nextPtr = NIL(LinkedCell);
    return (newLinkedCellPtr);
}


/*
 * ----------------------------------------------------------------------------
 * Mac_CreateCellGroup() --
 *
 *  It creates a new cell group containing the given cell instance
 *  or route cell.
 *
 * Results: 
 *   BEAR_OK
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */
BearStatus
Mac_CreateCellGroup(cellClass, cellPtr, cGroupPtrPtr)

CellClass   cellClass;
ClientData  cellPtr;
CellGroup  **cGroupPtrPtr;
{
    GenericCell  *gCellPtr;
    LinkedCell   *newLinkedCellPtr;
    CellGroup    *newCGroupPtr;

#ifdef DEBUG
    if (cellPtr == (ClientData) NULL)
    {
        MacPrintFatalErr("Mac_CreateCellGroup() failed: no cell is given");
    }
#endif DEBUG

    (void) Mac_CreateGCell(cellClass, cellPtr, &gCellPtr);
    newLinkedCellPtr = Mac_MakeLinkedCell();
    newLinkedCellPtr->cellPtr = gCellPtr;
    MALLOC(CellGroup *, newCGroupPtr, sizeof (CellGroup));
    newCGroupPtr->cellList = newLinkedCellPtr;
    switch (newLinkedCellPtr->cellPtr->unType) 
    {
    case CELL_INST:
        newLinkedCellPtr->cellPtr->unValue.cellInstPtr->cellGroupPtr = newCGroupPtr;
        break;
    case ROUTE_CELL:
        newLinkedCellPtr->cellPtr->unValue.routeCellPtr->cellGroupPtr = newCGroupPtr;
        break;
        default:
        break;
    }
    *cGroupPtrPtr = newCGroupPtr;
    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 * Mac_MergeCellGroups() --
 *
 *  It connects cellGroup2 at the end of cellGroup1, and passes
 *  the resulting new cell group to the calling subroutine as newGroup,
 *  which is a pointer to a pointer to CellGroup.
 *
 *  MacFindEndOfCellGroup is called to find the end of cellGroup1,
 *  and MacChangeCellGroup is called to update all of the cell group
 *  pointer of cellGroup2. 
 *
 * Results: 
 *   BEAR_OK, if the operation is successful.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */
BearStatus
Mac_MergeCellGroups(cellGroup1, cellGroup2, newGroup)

CellGroup    *cellGroup1; 
CellGroup    *cellGroup2;
CellGroup    **newGroup;
{
    LinkedCell *endCellPtr;

#ifdef DEBUG
    if (cellGroup1 == NIL(CellGroup))
    {
        MacPrintFatalErr("Mac_MergeCellGroups() failed: the first cell group is nil");
    }
    else if (cellGroup2 == NIL(CellGroup))
    {
        MacPrintFatalErr("Mac_MergeCellGroups() failed: the second cell group is nil");
    }
#endif DEBUG

    endCellPtr = NIL(LinkedCell);
    if (Mac_ItrCellGroup(cellGroup1, MacFindEndOfCellGroup, 
    (ClientData) &endCellPtr) == BEAR_CONTINUE_SEARCH)
    {
        return (BEAR_ERROR);
    }
    endCellPtr->nextPtr = cellGroup2->cellList;
    if (Mac_ItrCellGroup(cellGroup2, MacChangeCellGroupPtr, 
    (ClientData) cellGroup1) == BEAR_STOP_SEARCH)
    {
        return (BEAR_ERROR);
    }
    *newGroup = cellGroup1;

    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 * Mac_ItrCellGroup() --
 *
 *  It iterates all the linked cells in the given
 *  cellGroupPtr, and applies the given function to each
 *  linked cell.
 *
 * Result:  
 *   BEAR_STOP_SEARCH, if the function returns 
 *      BEAR_STOP_SEARCH.
 *   BEAR_COMPLETE_SEARCH, if either the given cellGroupPtr
 *      has no linked cell or all of the linked cells are
 *      walked through.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */
BearStatus
Mac_ItrCellGroup(cellGroupPtr, func, clientData)
CellGroup *cellGroupPtr;
BearStatus  (*func)();
ClientData  clientData;

{
    LinkedCell *curLinkedCell;

#ifdef DEBUG
    if (cellGroupPtr == NIL(CellGroup))
    {
        MacPrintFatalErr("Mac_ItrCellGroup() failed: no cell group is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
        MacPrintFatalErr("Mac_ItrCellGroup() failed: no function is given");
    }
#endif DEBUG
    if (cellGroupPtr->cellList == NIL(LinkedCell))
    {
        return (BEAR_COMPLETE_SEARCH);
    }
    else
    {
        curLinkedCell = cellGroupPtr->cellList;
        while (curLinkedCell != NIL(LinkedCell))
        {
            if (((*func)(curLinkedCell, clientData)) == BEAR_STOP_SEARCH)
            {
                return (BEAR_STOP_SEARCH);
            }
            curLinkedCell = curLinkedCell->nextPtr;
        }
        return (BEAR_COMPLETE_SEARCH);
    }
}


