
/* macDCell.c --
 *
 *      June Wang
 *      George Carvalho
 *      (Advising Professor - E. S. Kuh)
 *
 * It contains the operations to make and create a dummy cell as well
 * as insert and delete pins to a dummy cell.
 * 
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macDCell.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


/* ---------- Local Procedures ---------- */
static DummyCell *MakeDummyCell();
static BearStatus FreeDCell();


/*
 * ----------------------------------------------------------------------------
 *
 * FreeDCell() --
 *
 * This procedure is designed to be used by an "interator".  When its called
 * all of the memory associated with the given dummy cell is freed.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
static BearStatus
FreeDCell(dcellPtr, clientData)
DummyCell   *dcellPtr;
ClientData  clientData;
{
    GenericCell gcell;
    PseudoPin   *ppinPtr, *ppinList;

    gcell.unType = DUMMY_CELL;
    gcell.unValue.dummyCellPtr = dcellPtr;

    /*
     * Dummy cells are by definition a single rectangle, which means
     * each of the following lists will only contain a single element.
     */
    FREE((char *) Mac_GetHorzRectListOfGCell(&gcell));
    FREE((char *) Mac_GetVertRectListOfGCell(&gcell));

    /*
     * Free the psuedo pin list on the dummy cells top side.
     */
    Mac_RemovePPinListOfDCell(dcellPtr, (SideDirection) TOP_SIDE, &ppinList);
    while ((ppinPtr = ppinList) != NIL(PseudoPin)) {
        ppinList = ppinPtr->nextPtr;
        FREE((char *) ppinPtr);
    }

    /*
     * Free the psuedo pin list on the dummy cells bottom side.
     */
    Mac_RemovePPinListOfDCell(dcellPtr, (SideDirection) BOTTOM_SIDE, &ppinList);
    while ((ppinPtr = ppinList) != NIL(PseudoPin)) {
        ppinList = ppinPtr->nextPtr;
        FREE((char *) ppinPtr);
    }

    /*
     * Free the psuedo pin list on the dummy cells left side.
     */
    Mac_RemovePPinListOfDCell(dcellPtr, (SideDirection) LEFT_SIDE, &ppinList);
    while ((ppinPtr = ppinList) != NIL(PseudoPin)) {
        ppinList = ppinPtr->nextPtr;
        FREE((char *) ppinPtr);
    }

    /*
     * Free the psuedo pin list on the dummy cells right side.
     */
    Mac_RemovePPinListOfDCell(dcellPtr, (SideDirection) RIGHT_SIDE, &ppinList);
    while ((ppinPtr = ppinList) != NIL(PseudoPin)) {
        ppinList = ppinPtr->nextPtr;
        FREE((char *) ppinPtr);
    }

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 * MakeDummyCell() --
 *
 *  Allocates a memory space for a dummy cell.
 *  Initializes the fields of that dummy cell.
 *
 * Results:
 *   A pointer to the newly created dummy cell.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

static DummyCell *
MakeDummyCell()

{
    DummyCell *dcellPtr;

    MALLOC(DummyCell *, dcellPtr, sizeof (DummyCell));
    dcellPtr->id = 0;
    dcellPtr->horzRectList = NIL(LinkedRect);
    dcellPtr->vertRectList = NIL(LinkedRect);
    Mac_CopyTransform(&dcellPtr->transform, &GeoIdentityTransform);
    dcellPtr->topPPinList = NIL(PseudoPin);
    dcellPtr->bottomPPinList = NIL(PseudoPin);
    dcellPtr->leftPPinList = NIL(PseudoPin);
    dcellPtr->rightPPinList = NIL(PseudoPin);
    dcellPtr->nextPtr = NIL(DummyCell);

    return (dcellPtr);
}



/*
 * ----------------------------------------------------------------------------
 * Mac_CreateDCell() --
 *
 *  Creates a new dummy cell, puts the original list of dummy cells
 *  after the new dummy cell, and makes the new dummy cell as the head
 *  of the new list of dummy cells.  An identification number is also
 *  assigned to the dummy cell.
 *
 *  Notice that both of the list of the dummy cells and the new dummy cell 
 *  is passed back to the calling subroutine as a pointer to a pointer 
 *  to DummyCell.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_CreateDCell(dcellListPtr, dcellPtrPtr)

DummyCell  **dcellListPtr;
DummyCell  **dcellPtrPtr;
{
    BearId dcellId;
    
    *dcellPtrPtr = MakeDummyCell();

    dcellId = Mac_GenerateId();
    Mac_PutIdOfDCell(dcellId, *dcellPtrPtr);

    (*dcellPtrPtr)->nextPtr = *dcellListPtr;
    *dcellListPtr = *dcellPtrPtr;
    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeleteDCell() --
 *
 *  Delete a dummy cell from the MAC databese (but not from the tile plane),
 *  and then free the memory associated with it.
 *
 * Results:
 *  BEAR_OK:    when done.
 *  BEAR_ERROR: if NULL list, NULL cell, or the cell is not on the list. 
 *
 * Side effects:
 *  The dummy cell list is updated.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeleteDCell(dcellListPtr, dcellPtr)
DummyCell **dcellListPtr;
DummyCell *dcellPtr;

{
    DummyCell *curDCellPtr;

#ifdef DEBUG
    if (dcellPtr == NIL(DummyCell)) 
    {
        MacPrintFatalErr("Mac_DeleteDCell() failed: no dummy cell is given");
    }
    else if (dcellListPtr == (DummyCell **) NULL)
    {
    MacPrintFatalErr("Mac_DeleteDCell() failed: the dummy cell list is empty");
    }
#endif DEBUG

    if (dcellPtr == *dcellListPtr)
    {
        (*dcellListPtr) = (*dcellListPtr)->nextPtr;
    }
    else
    {
    curDCellPtr = *dcellListPtr;
    while (curDCellPtr->nextPtr != NIL(DummyCell) &&
        curDCellPtr->nextPtr != dcellPtr) 
    {
        curDCellPtr = curDCellPtr->nextPtr;
    }
    if (curDCellPtr->nextPtr == dcellPtr)
    {
        curDCellPtr->nextPtr = curDCellPtr->nextPtr->nextPtr;
    }
    else
    {
    Bear_PrintErr("Mac_DeleteCellMaster() failed: dummy cell is not in the list");
        return (BEAR_ERROR);
    }
    }
    (void) FreeDCell(dcellPtr, (ClientData) NULL);

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeleteDCellList() --
 *
 * Delete an entire dummy cell list from the MAC database (but not from the
 * tile plane).  Note that this includes freeing all of the respective memory.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

void
Mac_DeleteDCellList(dcellListPtr)
DummyCell **dcellListPtr;
{
    DummyCell *dcellPtr;

    /*
     * Free all of the memory associated with each dummy cell.
     */
    (void) Mac_ItrDCell(*dcellListPtr, FreeDCell, (ClientData) NULL);

    /*
     * Free each dummy cell.
     */
    while ((dcellPtr = *dcellListPtr) != NIL(DummyCell)) {
        (*dcellListPtr) = dcellPtr->nextPtr;
        FREE((char *) dcellPtr);
    }
}



BearStatus
Mac_InsertPPinOfDCell(dcellPtr, side, ppinPtr)
DummyCell *dcellPtr;
SideDirection side;
PseudoPin *ppinPtr;

{
#ifdef DEBUG
    if (dcellPtr == NIL(DummyCell) || ppinPtr == NIL(PseudoPin))
    {
        MacPrintFatalErr("Mac_InsertPPinOfDCell() failed: no dummy cell or no pseudo pin is given");
    }
#endif DEBUG

    switch (side)
    {
    case (SideDirection) LEFT_SIDE:
        if (dcellPtr->leftPPinList != NIL(PseudoPin))
        {
            ppinPtr->nextPtr = dcellPtr->leftPPinList;
        }
        dcellPtr->leftPPinList = ppinPtr;
        break;
    case (SideDirection) RIGHT_SIDE:
        if (dcellPtr->rightPPinList != NIL(PseudoPin))
        {
            ppinPtr->nextPtr = dcellPtr->rightPPinList;
        }
        dcellPtr->rightPPinList = ppinPtr;
        break;
    case (SideDirection) TOP_SIDE:
        if (dcellPtr->topPPinList != NIL(PseudoPin))
        {
            ppinPtr->nextPtr = dcellPtr->topPPinList;
        }
        dcellPtr->topPPinList = ppinPtr;
        break;
    case (SideDirection) BOTTOM_SIDE:
        if (dcellPtr->bottomPPinList != NIL(PseudoPin))
        {
            ppinPtr->nextPtr = dcellPtr->bottomPPinList;
        }
        dcellPtr->bottomPPinList = ppinPtr;
        break;
    default:
MacPrintFatalErr("Mac_InsertPPinOfDCell() failed : side has illegal value");
        break;
    }
    ppinPtr->tilePtr = Mac_GetTileOfDCell(dcellPtr);
}



/*
 * ----------------------------------------------------------------------------
 * Mac_DeletePPinOfDCell() --
 *
 *  Given a dummy cell, the side direction of the dummy cell, and
 *  the pseudo pin to be deleted, it fetches the list of pseudo pins
 *  from the designated side of the dummy cell, and searches for the
 *  pseudo pin.
 *  When found, the pin is deleted from the list of pseudo pins.
 *
 * Results:
 *   BEAR_OK, when successful.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeletePPinOfDCell(dcellPtr, side, ppinPtr)
DummyCell *dcellPtr;
SideDirection side;
PseudoPin *ppinPtr;

{
    PseudoPin *curPPinPtr;

#ifdef DEBUG
    if (dcellPtr == NIL(DummyCell) || ppinPtr == NIL(PseudoPin))
    {
        MacPrintFatalErr("Mac_InsertPPinOfDCell() failed: no dummy cell or no pseudo pin is given");
    }
#endif DEBUG

    switch (side)
    {
        case (SideDirection) LEFT_SIDE:
            curPPinPtr = dcellPtr->leftPPinList;
            break;
        case (SideDirection) RIGHT_SIDE:
            curPPinPtr = dcellPtr->rightPPinList;
            break;
        case (SideDirection) TOP_SIDE:
            curPPinPtr = dcellPtr->topPPinList;
            break;
        case (SideDirection) BOTTOM_SIDE:
            curPPinPtr = dcellPtr->bottomPPinList;
            break;
    }
    if (curPPinPtr == NIL(PseudoPin))
    {
        Bear_PrintErr("Mac_DeletePPinOfDCell() failed: the ppin list is empty");
        return (BEAR_ERROR);
    }
    else
    {
        if (curPPinPtr == ppinPtr)
        {
            curPPinPtr = ppinPtr->nextPtr;
        }
        else
        {
            while (curPPinPtr->nextPtr != ppinPtr)
            {
                curPPinPtr = curPPinPtr->nextPtr;
            }
            curPPinPtr  = curPPinPtr->nextPtr->nextPtr;
        }
    }

    return (BEAR_OK);
}



void
Mac_RemovePPinListOfDCell(dcellPtr, side, ppinListPtr)
DummyCell       *dcellPtr;
SideDirection   side;
PseudoPin       **ppinListPtr;
/*
 * Return the head of the pseudo pin list to the user and then
 * make the head of the list stored in the dummy cell NULL.
 */
{
#ifdef DEBUG
    if (dcellPtr == NIL(DummyCell))
    {
    MacPrintFatalErr("Mac_RemovePPinListOfDcell() failed: no dummy cell given");
    }
#endif DEBUG

    switch (side) {
    case (SideDirection) LEFT_SIDE:
        *ppinListPtr = dcellPtr->leftPPinList;
        dcellPtr->leftPPinList = NIL(PseudoPin);
        break;
    case (SideDirection) RIGHT_SIDE:
        *ppinListPtr = dcellPtr->rightPPinList;
        dcellPtr->rightPPinList = NIL(PseudoPin);
        break;
    case (SideDirection) TOP_SIDE:
        *ppinListPtr = dcellPtr->topPPinList;
        dcellPtr->topPPinList = NIL(PseudoPin);
        break;
    case (SideDirection) BOTTOM_SIDE:
        *ppinListPtr = dcellPtr->bottomPPinList;
        dcellPtr->bottomPPinList = NIL(PseudoPin);
        break;
    default:
MacPrintFatalErr("Mac_RemovePPinListOfDcell() failed: illegal side direction");
        break;
    }
}



void
Mac_AppendPPinListOfDCell(ppinList, dcellPtr, side)
PseudoPin       *ppinList;
DummyCell       *dcellPtr;
SideDirection   side;
/*
 * Append a pseudo pin list onto the given side of a dummy cell.
 */
{
    PseudoPin *ppinPtr, **ppinPtrPtr;

#ifdef DEBUG
    if (ppinList == NIL(PseudoPin))
    {
MacPrintFatalErr("Mac_RemovePPinListOfDcell() failed: no pseudo pin list given");
    }
    else if (dcellPtr == NIL(DummyCell))
    {
    MacPrintFatalErr("Mac_RemovePPinListOfDcell() failed: no dummy cell given");
    }
#endif DEBUG

    for (ppinPtr = ppinList; ppinPtr != NIL(PseudoPin);
     ppinPtr = ppinPtr->nextPtr) 
    {
    ppinPtr->tilePtr = Mac_GetTileOfDCell(dcellPtr);
    }

    switch (side) {
    case (SideDirection) LEFT_SIDE:
        ppinPtrPtr = &dcellPtr->leftPPinList;
        break;
    case (SideDirection) RIGHT_SIDE:
        ppinPtrPtr = &dcellPtr->rightPPinList;
        break;
    case (SideDirection) TOP_SIDE:
        ppinPtrPtr = &dcellPtr->topPPinList;
        break;
    case (SideDirection) BOTTOM_SIDE:
        ppinPtrPtr = &dcellPtr->bottomPPinList;
        break;
    default:
MacPrintFatalErr("Mac_RemovePPinListOfDcell() failed: illegal side direction");
        break;
    }

    if (*ppinPtrPtr == NIL(PseudoPin)) {
        *ppinPtrPtr = ppinList;
    } else {
        /*
         * Find the end of the pseudo pin list that is currently 
         * attached to this side of the dummy cell.
         */
         for (ppinPtr = *ppinPtrPtr;
            /* infinite loop */;
            ppinPtr = ppinPtr->nextPtr) {

            if (ppinPtr->nextPtr == NIL(PseudoPin)) {
                break;
            }
        }

        /* Append the new list onto the end of the old one.  */
        ppinPtr->nextPtr = ppinList;
    }
}



