
/* macPin.h --
 *
 *     George Carvalho, Esq.
 *     Wei-ming Dai
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * Defines the data structures related to Pins. 
 * Lists the subroutnes used to create, insert, and delete pins.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macPin.h,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $
 */

#ifndef _MAC_PIN
#define _MAC_PIN /* once-only latch */

typedef enum _PinKind PinKind;
enum _PinKind {
    MAC_BORDER_PIN,     /* Border pins are the ones on the chips border     */
    MAC_INSIDE_PIN      /* Inside pins are all the others                   */
};

typedef enum _PinDirection PinDirection;
enum _PinDirection {
    NORTH_PIN,          /* The pin's routing direction is north             */
    SOUTH_PIN,          /* The pin's routing direction is south             */
    EAST_PIN,           /* The pin's routing direction is east              */
    WEST_PIN,           /* The pin's routing direction is west              */
    NORTH_SOUTH_PIN,    /* The pin's routing direction is north and south   */
    EAST_WEST_PIN,      /* The pin's routing direction is east and west     */
    ALL_DIRECTION_PIN   /* The pin's routing direction is all               */
};

typedef struct _PinMaster PinMaster;
typedef struct _PinInst PinInst;

struct _PinMaster {
    BearString      name;           /* Pin master name                      */
    int             order;          /* Relative ordering of the pin         */
    LayerVector     layer;          /* The layer of the pin                 */

    bool            posDefFlag;     /* Flag indicating whether or not the
                                     * position of this pin is defined or
                                     * not.  Note pins may start life with
                                     * no position and rely on the placement
                                     * pin assignment code to give them one.
                                     */
    Flexibility     flexType;       /* Type of flexibility                  */
    union {
        Point       position;       /* Position related to cell origin      */
                                    /* for fixed pins                       */
        BearString  sideConstr;     /* Side constraints for flexible pins: 
                                     * a string composed of cell boundary
                                     * identifies (intergers) separated by
                                     * spaces, tabs, or newlines.
                                     */
    } posUn;
    PortMaster      *portPtr;       /* The port master to which the pin 
                                     * belongs.
                                     */ 
    PinInst         *instList;      /* The list of the pin instances.       */
    PinMaster       *nextPtr;       /* The next pin of the port master.     */
    ClientData      clientData;     /* This space for rent.                 */
};

struct _PinInst {
    BearString      pi_name;        /* Pin instance name                    */
    BearId          pi_id;          /* Identification number that is unique
                                     * within the scope of the chip and is 
                                     * only intended to be used internally.
                                     */
    BearId          intNetId;       /* Internal net id.                     */
    BearId          extNetId;       /* External net id.                     */
    BearId          horzExtNetId;   /* External net id after horzontal
                                     * unification.
                                     */
    BearId          vertExtNetId;   /* External net id after vertical
                                     * unification.
                                     */
    PinKind         pinKind;        /* Kind of pin.                         */
    PinDirection    pinDir;         /* Pin direction relative to cell bound */
    PinMaster       *masterPtr;     /* The master of the pin instance       */
    PortInst        *portPtr;       /* The port instance to which 
                                     * the pin belongs 
                                     */ 
    float           pinWidth;       /* Width of the pin                     */
    float           segmWidth;      /* Width of the routing segment         */
    float           widthVect[8];   /* Width vector of the pin            
                                     *    0: poly-silicon; 1-7: Al
                                     */
    struct _RdbPinInfo      *BBLTermPtr;
    PinInst         *nextPtr;       /* The next pin of the port instance    */
    PinInst         *nextInMListPtr;/* The next pin instance in the
                                     * master list.
                                     */
    PinInst         *nextInCellPtr; /* The next pin instance on the         */
                                    /* boundary of the cell instance        */
    BearGeneric     ngrModulePtr;   /* Reserved for the new global
                                     * routing module.
                                     */
    ClientData      clientData;     /* This space for rent                  */
};

typedef struct _LinkedPinInst LinkedPinInst;
struct _LinkedPinInst {
    PinInst         *pinInstPtr;    /* Pointer to a pin inst.              */
    LinkedPinInst   *nextPtr;       /* Pointer to another linked pin inst. */
};


extern BearStatus Mac_CreatePinMaster(/* cellMasterPtr, portMasterPtrPtr,
                                       * portMasterName, pinMasterName,
                                       * pinMasterPtrPtr
                                       */);

extern BearStatus Mac_CreatePinOfMaster(/* cellInstPtr, portInstName, 
                                         * pinMasterPtr,
                                         * pinInstName, pinInstPtrPtr
                                         */);

extern BearStatus Mac_CreatePin();
/*
 * CellInst   *cellPtr;
 * BearString portMasterName;
 * BearString portInstName;
 * BearString pinMasterName;
 * PinInst    **pinInstPtrPtr;
 */

extern BearStatus Mac_DeletePin(/* pinInstPtr */);
/*
 * PinInst *pinInstPtr;
 */

extern BearStatus Mac_DuplicatePin();
extern BearStatus Mac_DuplicatePinMaster();

extern BearStatus Mac_InsertPinMaster(/* portMasterPtr, pinMasterPtr*/);
/*
 * PortMaster *portMasterPtr;
 * PinMaster  *pinMasterPtr;
 */

extern BearStatus Mac_DeletePinMaster(/* pinMasterPtr */);
/*
 * PinMaster *pinMasterPtr;
 */

extern BearStatus Mac_InsertPinInstOfPinMaster(/* pinInstPtr, pinMasterPtr */);
/*
 * PinInst   *pinInstPtr;
 * PinMaster *pinMasterPtr;
 */

extern BearStatus Mac_DeletePinInstOfPinMaster(/* pinInstPtr, pinMasterPtr */);
/*
 * PinInst   *pinInstPtr;
 * PinMaster *pinMasterPtr;
 */

extern BearStatus Mac_InsertPinInstOfPort(/* portPtr, pinInstPtr*/);
/*
 * PortInst *portPtr;
 * PinInst  *pinInstPtr;
 */

extern BearStatus Mac_DeletePinInstOfPort(/* portPtr, pinInstPtr*/);
/*
 * PortInst *portPtr;
 * PinInst  *pinInstPtr;
 */

extern BearStatus Mac_ConvertToBorderPin(/* pinInstPtr, clientData */);
/*
 * PinInst     *pinInstPtr;
 * ClientData  clientData;
 */

#endif _MAC_PIN

