
/* macPinPrim.c --
 *
 *     George Carvalho, Esq.
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains all of the operations to retrieve and update the
 * information stored in pin masters and pin instances.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macPinPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


/* ---------- Pin Primitives ---------- */

BearString
Mac_GetNameOfPinMaster(pinMasterPtr) 
PinMaster *pinMasterPtr;
{
#ifdef DEBUG
    if (pinMasterPtr == NIL(PinMaster)) {
    MacPrintFatalErr("Mac_GetNameOfPinMaster() failed: no master is given");
    }
#endif /* DEBUG */

    return (pinMasterPtr->name);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_PutNameOfPinMaster() --
 *
 * Each pin master must have a unique name.
 *
 * Results:
 *   BEAR_OK is returned if the name is not already in the hash table,
 *   BEAR_ERROR otherwise.
 *
 * Side effects:
 *   The name of the pin master is stored in the hash table.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_PutNameOfPinMaster(pinMastName, pinMastPtr) 
BearString  pinMastName;
PinMaster   *pinMastPtr;
{
#ifdef DEBUG
    if (pinMastPtr == NIL(PinMaster)) {
   MacPrintFatalErr("Mac_PutNameOfPinMaster() failed: no pin master is given");
    }
#endif /* DEBUG */

    if (Mac_IsMember(PIN_MASTER_HT, pinMastName)) {
        sprintf(bear_ErrorMsg, "Cannot have two pin masters named \"%s\"", 
            pinMastName);
        Bear_PrintErr(bear_ErrorMsg);

        return (BEAR_ERROR);
    } 
    pinMastPtr->name = pinMastName;
    Mac_Insert(PIN_MASTER_HT, pinMastName, (char *) pinMastPtr);

    return (BEAR_OK);
}


BearId
Mac_GetIdOfPinInst(pinInstPtr) 
PinInst *pinInstPtr;
/*
 * The identification number of the pin instance is generated by us 
 * and is for our use only.
 */
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_GetIdOfPinInst() failed: no pin is given");
    }
#endif /* DEBUG */

    return (pinInstPtr->pi_id);
}


BearString
Mac_GetNameOfPinInst(pinInstPtr) 
PinInst *pinInstPtr;
/*
 * The name of the pin instance is supplied by the user.
 */
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_GetNameOfPinInst() failed: no pin is given.");
    }
#endif /* DEBUG */

    return (pinInstPtr->pi_name);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_PutNameOfPinInst() --
 *
 * Each pin instance must have a unique name.
 *
 * Results:
 *   BEAR_OK is returned if the name is not already in the hash table,
 *   BEAR_ERROR otherwise.
 *
 * Side effects:
 *   The entry name is stored in the hash table.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_PutNameOfPinInst(pinInstName, pinInstPtr) 
BearString  pinInstName;
PinInst     *pinInstPtr;
{
    st_table *tablePtr;

#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr(
           "Mac_PutNameOfPinInst() failed: no pin instance is given");
    }
    else if (pinInstPtr->portPtr == NIL(PortInst)) {
        MacPrintFatalErr(
           "Mac_PutNameOfPinInst() failed: pin is not connected to a port");
    }
    else if (pinInstPtr->portPtr->cellInstPtr == NIL(CellInst)) {
        MacPrintFatalErr(
           "Mac_PutNameOfPinInst() failed: port is not connected to a cell");
    }
    else if (pinInstPtr->portPtr->cellInstPtr->masterPtr == NIL(CellMaster)) {
        MacPrintFatalErr(
           "Mac_PutNameOfPinInst() failed: cell is not connected to a master");
    }
#endif /* DEBUG */

    tablePtr = pinInstPtr->portPtr->cellInstPtr->masterPtr->pinInstTablePtr;

    if (st_is_member(tablePtr, pinInstName)) {
        (void) sprintf (bear_ErrorMsg, 
            "Cell named \"%s\" has two pin instances named \"%s\".", 
            Mac_GetNameOfCellInst(pinInstPtr->portPtr->cellInstPtr),
            pinInstName);
        Bear_PrintErr(bear_ErrorMsg);

        return (BEAR_ERROR);
    } 
    pinInstPtr->pi_name = pinInstName;
    (void) st_insert(tablePtr, pinInstName, (char *) pinInstPtr);

    return (BEAR_OK);
}



bool
Mac_IsFixedPin(pinInstPtr)  
PinInst *pinInstPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))  {
        MacPrintFatalErr("Mac_IsFixedPin() failed : no pin given");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster)) {
        MacPrintFatalErr(
            "Mac_IsFixedPin() failed : pin is not connected to a master");
    }
#endif /* DEBUG */

    return ((pinInstPtr->masterPtr->flexType == FIXED) ? TRUE : FALSE);
}


Flexibility
Mac_GetFlexOfPin(pinInstPtr)
PinInst *pinInstPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_GetFlexOfPin() failed : no pin given");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster)) {
        MacPrintFatalErr(
            "Mac_GetFlexOfPin() failed : pin is not connected to a master");
    }
#endif /* DEBUG */

    return (pinInstPtr->masterPtr->flexType);
}


void
Mac_PutFlexOfPin(flexType, pinInstPtr)
Flexibility flexType;
PinInst     *pinInstPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_PutFlexOfPin() failed : no pin given");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster)) {
        MacPrintFatalErr(
            "Mac_PutFlexOfPin() failed : pin is not connected to a master");
    }
#endif /* DEBUG */

    pinInstPtr->masterPtr->flexType = flexType;
}


void
Mac_PutPinPosDefFlag(posDefFlag, pinInstPtr)
bool    posDefFlag;
PinInst *pinInstPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_PutPinPosDef() failed : no pin given");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster)) {
        MacPrintFatalErr(
            "Mac_PutPinPosDef() failed : pin is not connected to a master");
    }
#endif /* DEBUG */

    pinInstPtr->masterPtr->posDefFlag = posDefFlag;
}


bool
Mac_IsPinPosDef(pinInstPtr)
PinInst *pinInstPtr;
/*
 * Is the given pin's position defined ?
 */
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_IsPinPosDef() failed : no pin given");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster)) {
        MacPrintFatalErr(
            "Mac_IsPinPosDef() failed : pin is not connected to a master");
    }
#endif /* DEBUG */

    return (pinInstPtr->masterPtr->posDefFlag ? TRUE : FALSE);
}


Point *
Mac_GetRelPosOfPin(pinInstPtr)
PinInst *pinInstPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_GetRelPosOfPin() failed: no pin is given");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster)) {
        MacPrintFatalErr(
            "Mac_GetRelPosOfPin() failed : pin is not connected to a master");
    }
#endif /* DEBUG */

    return (&pinInstPtr->masterPtr->posUn.position);
}


void
Mac_PutRelPosOfPin(relPosPtr, pinInstPtr)
Point   *relPosPtr;
PinInst *pinInstPtr;

{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_PutRelPosOfPin() failed: no pin is given");
    }
    else if (relPosPtr == NIL(Point)) {
        MacPrintFatalErr("Mac_PutRelPosOfPin() failed: no point is given");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster)) {
        MacPrintFatalErr(
            "Mac_PutRelPosOfPin() failed: pin is not connected to a master");
    }
#endif /* DEBUG */

    Mac_CopyPoint(&pinInstPtr->masterPtr->posUn.position, relPosPtr);
}


/* 
 * ----------------------------------------------------------------------------
 *
 * Mac_GetAbsPosOfPin() --
 *
 * Figure out the coordinates of the pins absolute position
 * on the chip, by knowing where it is relative to the cell.
 * 
 * IMPORTANT : The pointer to the point structure must point
 *     to a piece of memory.
 *
 *     PinInst  *pinInstPtr;
 *     Point    absPos;
 *
 *     Mac_GetAbsPosOfPin(pinInstPtr, &absPos);
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

void
Mac_GetAbsPosOfPin(pinInstPtr, absPosPtr)
PinInst *pinInstPtr;
Point   *absPosPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))  {
        MacPrintFatalErr("Mac_GetAbsPosOfPin() failed: no pin is given");
    } 
    else if (pinInstPtr->portPtr == NIL(PortInst)) {
        MacPrintFatalErr(
            "Mac_GetAbsPosOfPin() failed: pin is not connected to a port");
    }
    else if (pinInstPtr->portPtr->cellInstPtr == NIL(CellInst)) {
        MacPrintFatalErr(
            "Mac_GetAbsPosOfPin() failed: port is not connected to a cell");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster)) {
        MacPrintFatalErr(
            "Mac_GetAbsPosOfPin() failed: pin is not connected to a master");
    }
    else if (absPosPtr == NIL(Point)) {
        MacPrintFatalErr("Mac_GetAbsPosOfPin() failed: absPosPtr is nil");
    }
#endif /* DEBUG */

    GeoTransPoint(&pinInstPtr->portPtr->cellInstPtr->transform,
        &pinInstPtr->masterPtr->posUn.position, absPosPtr);
}



LinkedSegID *
Mac_GetPosConstrOfCell(pinInstPtr)
PinInst *pinInstPtr;
{
    LinkedLineSeg   *currentSeg;
    LinkedSegID     *linkedSegIDList;
    LinkedSegID     *linkedSegID;
    LinkedSegID     *nextLinkedSegID;

#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_GetPorConstrOfCell() failed: no pin in given");
    }
#endif /* DEBUG */

    currentSeg = pinInstPtr->portPtr->cellInstPtr->boundary; 
    MALLOC(LinkedSegID *, linkedSegIDList, sizeof (LinkedSegID)); 
    MALLOC(LinkedSegID *, linkedSegID, sizeof (LinkedSegID)); 
    linkedSegIDList->nextPtr = linkedSegID;

    while (currentSeg != NIL(LinkedLineSeg)) {
        linkedSegID->segID = currentSeg->lineSeg->segID;
        currentSeg = currentSeg->nextPtr;
        MALLOC(LinkedSegID  *, nextLinkedSegID, sizeof (LinkedSegID)); 
        linkedSegID->nextPtr = nextLinkedSegID;
        linkedSegID = linkedSegID->nextPtr;
    }

    return (linkedSegIDList);
}


PinMaster *
Mac_GetMasterOfPin(pinInstPtr)
PinInst  *pinInstPtr;

{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst ))
    {
        MacPrintFatalErr("Mac_GetMasterOfPin() failed: no pin inst is given");
    }
    if (pinInstPtr->masterPtr == NIL(PinMaster))
    {
    MacPrintFatalErr("Mac_GetMasterOfPin() failed: the given pin is not\
           connected to a master");
    }
#endif /* DEBUG */

    return (pinInstPtr->masterPtr);
}


float
Mac_GetPinWidthOfPin(pinInstPtr)
PinInst  *pinInstPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_GetWidthOfPin() failed:  no pin is given");
    }
#endif /* DEBUG */

    return (pinInstPtr->pinWidth);
}


void
Mac_PutPinWidthOfPin(pinWidth, pinInstPtr)
float    pinWidth;
PinInst  *pinInstPtr;

{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_PutWidthOfPin() failed:  no pin is given");
    }
#endif /* DEBUG */

    pinInstPtr->pinWidth = pinWidth;
}


float
Mac_GetSegmWidthOfPin(pinInstPtr)
PinInst *pinInstPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_GetSegmWidthOfPin() failed:  no pin is given");
    }
#endif /* DEBUG */

    return (pinInstPtr->segmWidth);
}


void
Mac_PutSegmWidthOfPin(segmWidth, pinInstPtr)
float    segmWidth;
PinInst  *pinInstPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_PutSegmWidthOfPin() failed:  no pin is given");
    }
#endif /* DEBUG */

    pinInstPtr->segmWidth = segmWidth;
}

float *
Mac_GetWidthVectOfPin(pinInstPtr)
PinInst  *pinInstPtr;

{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetWidthVectOfPin() failed:  no pin is given");
    }
#endif /* DEBUG */

    return (pinInstPtr->widthVect);
}


void
Mac_PutWidthVectOfPin(widthVect, pinInstPtr)
float    *widthVect;
PinInst  *pinInstPtr;

{
    int i;

#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_PutWidthVectOfPin() failed:  no pin is given");
    }
#endif /* DEBUG */

    for (i = 0; i < 8; i++)
    {
        pinInstPtr->widthVect[i] = widthVect[i];
    }
}

PinDirection
Mac_GetDirectionOfPin(pinInstPtr)
PinInst  *pinInstPtr;

{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetPinDirOfPin() failed:  no pin is given");
    }
#endif /* DEBUG */

    return (pinInstPtr->pinDir);
}

void
Mac_PutDirectionOfPin(pinDir, pinInstPtr)

PinDirection    pinDir;
PinInst     *pinInstPtr;
{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_PutPinDirOfPin() failed:  no pin is given");
    }
#endif /* DEBUG */

    pinInstPtr->pinDir = pinDir;
}

LayerVector
Mac_GetLayerOfPin(pinInstPtr)
PinInst  *pinInstPtr;

{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetLayerOfPin() failed: no pin instance is given");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster))
    {
        MacPrintFatalErr("Mac_GetLayerOfPin() failed: the pin is not connected to a master");
    }
#endif /* DEBUG */

    return (pinInstPtr->masterPtr->layer);
}


void
Mac_PutLayerOfPin(layer, pinInstPtr)
LayerVector layer;
PinInst   *pinInstPtr;

{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_PutLayerOfPin() failed: no pin instance is given");
    }
    else if (pinInstPtr->masterPtr == NIL(PinMaster))
    {
        MacPrintFatalErr("Mac_PutLayerOfPin() failed: the pin is not connected to a master");
    }
#endif /* DEBUG */

    pinInstPtr->masterPtr->layer = layer;
}


PortInst *
Mac_GetPortOfPin(pinInstPtr)
PinInst *pinInstPtr;

{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetPortOfPin() failed: no pin instance is given");
    }
#endif /* DEBUG */

    return (pinInstPtr->portPtr);
}


Net *
Mac_GetNetOfPin(pinInstPtr)
PinInst *pinInstPtr;

{
#ifdef DEBUG
    if (pinInstPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetNetOfPin() failed: no pin instance is given");
    }
    else if (pinInstPtr->portPtr == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_GetNetOfPin() failed: pin instance has no port");
    }
#endif /* DEBUG */

    return (pinInstPtr->portPtr->netPtr);
}


BearString
Mac_GetNetNameOfPin(pinPtr)
PinInst  *pinPtr;

{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetNetNameOfPin() failed: no pin is given");
    }
#endif /* DEBUG */

    return (pinPtr->portPtr->netPtr->name);
}


void
Mac_PutNetNameOfPin(namePtr, pinPtr)
BearString  namePtr;
PinInst    *pinPtr;
{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst)) {
        MacPrintFatalErr("Mac_PutNetNameOfPin() failed: no pin is given");
    }
    if (pinPtr->portPtr == NIL(PortInst)) {
        MacPrintFatalErr("Mac_PutNetNameOfPin() failed: the given pin is not connected to a port");
    }
    if (pinPtr->portPtr->netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_PutNetNameOfPin() failed: the related port is not connected to a net");
    }
#endif /* DEBUG */

    pinPtr->portPtr->netPtr->name = namePtr;
}


BearId
Mac_GetIntNetIdOfPin(pinPtr)
PinInst  *pinPtr;

{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetIntNetIdOfPin() failed: no pin is given");
    }
#endif /* DEBUG */

    return (pinPtr->intNetId);
}


void 
Mac_PutIntNetIdOfPin(id, pinPtr)
BearId  id;
PinInst  *pinPtr;

{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_PutIntNetIdOfPin() failed: no pin is given");
    }
#endif /* DEBUG */

    pinPtr->intNetId = id;
}


BearId 
Mac_GetExtNetIdOfPin(pinPtr)
PinInst  *pinPtr;

{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetIntNetIdOfPin() failed: no pin is given");
    }
#endif /* DEBUG */

    return (pinPtr->extNetId);
}


void
Mac_PutExtNetIdOfPin(id, pinPtr)
BearId  id;
PinInst *pinPtr;

{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_PutIntNetIdOfPin() failed: no pin is given");
    }
#endif /* DEBUG */

    pinPtr->extNetId = id;
}


BearId 
Mac_GetHExtNetIdOfPin(pinPtr)
PinInst  *pinPtr;

{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetHExtNetIdOfPin() failed: no pin is given");
    }
#endif /* DEBUG */

    return (pinPtr->horzExtNetId);
}


void
Mac_PutHExtNetIdOfPin(id, pinPtr)
BearId  id;
PinInst *pinPtr;

{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_PutHExtNetIdOfPin() failed: no pin is given");
    }
#endif /* DEBUG */

    pinPtr->horzExtNetId = id;
}


BearId 
Mac_GetVExtNetIdOfPin(pinPtr)
PinInst  *pinPtr;

{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_GetIntNetIdOfPin() failed: no pin is given");
    }
#endif /* DEBUG */

    return (pinPtr->vertExtNetId);
}


void
Mac_PutVExtNetIdOfPin(id, pinPtr)
BearId  id;
PinInst *pinPtr;

{
#ifdef DEBUG
    if (pinPtr == NIL(PinInst))
    {
        MacPrintFatalErr("Mac_PutIntNetIdOfPin() failed: no pin is given");
    }
#endif /* DEBUG */

    pinPtr->vertExtNetId = id;
}


/* ARGSUSED */
BearStatus
MacCountNumOfPins(pinPtr, numPtr)
PinInst *pinPtr;
int     *numPtr;
{
    (*numPtr)++;
    return (BEAR_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_NumOfPinsOfPort() --
 *
 * Given a port instance, it counts the total number of pin instances
 * connected to that port by using iterator.  The sum is passed back
 * to the calling subroutine as a pointer to integer.
 *
 * Results:
 *   BEAR_OK, if the iterator returns BEAR_COMPLETE_SEARCH.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus 
Mac_NumOfPinsOfPort(portPtr, numPtr)
PortInst    *portPtr;
int         *numPtr;

{
#ifdef DEBUG
    if (portPtr == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_NumOfPinsOfPort() failed: no port is given");
    }
#endif /* DEBUG */

    *numPtr = 0;
    return ((Mac_ItrPinOfPortInst(portPtr, MacCountNumOfPins, 
        (ClientData) numPtr) == BEAR_COMPLETE_SEARCH) ?
        BEAR_OK : BEAR_ERROR);
}



PinKind
Mac_GetKindOfPin(pinInstPtr)
PinInst *pinInstPtr;
{
    return (pinInstPtr->pinKind);
}



void
Mac_PutKindOfPin(pinKind, pinInstPtr)
PinKind pinKind;
PinInst *pinInstPtr;
{
    pinInstPtr->pinKind = pinKind;
}



