
/* macStats.c --
 *
 *     George Carvalho, Esq. (code to write new global routing statistics)
 *     (Advising Professor - E. S. Kuh)
 *
 * Modification History
 *
 * bug fix (18 April 91)
 *
 *     The variables used to calculate the core area, total cell area, and
 *     chip area have been changed from integer to double in order to 
 *     accomodate a larger range of values.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macStats.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "newG.h"
#include "routeDB.h"

#ifdef USE_OCT_CODE
#include "vov.h"
#endif /* USE_OCT_CODE */


typedef struct _EdgeForPins     EdgeForPins;
typedef struct _LinkedEdgeList  LinkedEdgeList;
typedef struct _PinInfo         PinInfo;
typedef struct _RoutingStats    RoutingStats;

struct _PinInfo {
    Point           position;
    bool            finishFlag;
    LinkedEdgeList  *lnkEdgeList;
};

struct _EdgeForPins {
    PinInfo     *node1Ptr;
    PinInfo     *node2Ptr;
    int         total_length,
                x_length,
                y_length;
    bool        finishFlag;
    EdgeForPins *nextPtr;
};

struct _LinkedEdgeList {
    EdgeForPins     *edgeForPinsPtr;
    LinkedEdgeList  *nextPtr;
};

struct _RoutingStats {
    FILE                *filePtr;

    int                 totalLength;
    int                 totalLayer1;
    int                 totalLayer2;
    int                 totalPNetArea;  /* Total power net area             */
    int                 totalGNetArea;  /* Total ground net area            */
    int                 totalNetArea;   /* Total net area                   */
    int                 totalNetVias;

    int                 netCount;
    int                 netLength;
    int                 netLayer1;
    int                 netLayer2;
    int                 netArea;
    int                 netVias;

    double              totalCellArea;
    double              coreArea;
    double              chipArea;

    PinInfo             *pins;
    int                 size;

    PlaneData           *planeDataPtr;  /* Only used with WriteStatsThree() */
    Ngr_BldTreeDataPtr  bldTreePtr;     /* Only used with WriteStatsThree() */
};


/* ---------- Local Procedures ---------- */
static BearStatus   AddCellToList();
static BearStatus   AddNetToList();
static BearStatus   AddPinToList();
static int          CompareCells();
static int          CompareNets();
static int          ComparePins();

static BearStatus   AddCellAreaToTotal();
static void         CalcCellAreaStats();
static void         WriteCellAreaStats();

static BearStatus   PrintNet();
static BearStatus   WriteStatsZero();

static BearStatus   SetPinPos();
static BearStatus   EstimateNetLen();
static void         CreateEdge();
static void         GenerateCompGraph();
static void         FindMST();
static void         FreeSpace();
static void         WriteInfoOne();
static BearStatus   WriteStatsOne();

#ifdef USE_DR_CODE
static BearStatus   FindRRootLen();
static BearStatus   FindRSegLen();
static BearStatus   FindRJunCount();
static BearStatus   FindNetLen();
static void         WriteInfoTwo();
static BearStatus   WriteStatsTwo();
#endif /* USE_DR_CODE */

static BearStatus   CalcNetLen();
static BearStatus   WriteStatsThree();

static BearStatus   WriteStatsFour();


static BearStatus
AddCellToList(portInstPtr, cellList)
PortInst    *portInstPtr;
lsList      cellList;
{
    lsNewEnd(cellList, (lsGeneric) Mac_GetCellInstOfPort(portInstPtr), LS_NH);

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddNetToList(netPtr, netList)
Net     *netPtr;
lsList  netList;
{
    (void) Mac_NumOfPinsInPortListOfNet(netPtr, &netPtr->numOfPin);

    lsNewEnd(netList, (lsGeneric) netPtr, LS_NH);

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddPinToList(pinPtr, pinList)
PinInst     *pinPtr;
lsList      pinList;
{
    lsNewEnd(pinList, (lsGeneric) pinPtr, LS_NH);

    return (BEAR_CONTINUE_SEARCH);
}


static int
CompareCells(item1, item2)
lsGeneric item1, item2;
{
    CellInst *cellInst1Ptr = (CellInst *) item1;
    CellInst *cellInst2Ptr = (CellInst *) item2;

    /*
     * Note that strcmp() will not produce correct results if both strings
     * contain only digits.
     */
    return (strcmp(Mac_GetNameOfCellInst(cellInst1Ptr),
        Mac_GetNameOfCellInst(cellInst2Ptr)));
}


static int
CompareNets(item1, item2)
lsGeneric item1, item2;
{
    register Net *net1Ptr = (Net *) item1;
    register Net *net2Ptr = (Net *) item2;

    if (net1Ptr->numOfPin < net2Ptr->numOfPin) {
        return (-1);
    }
    else if (net1Ptr->numOfPin == net2Ptr->numOfPin) {
        /*
         * Note that strcmp() will not produce correct results if both strings
         * contain only digits.
         */
        return (strcmp(net1Ptr->name, net2Ptr->name));
    }
    else {
        return (1);
    }
    /* NOTREACHED */
}


static int
ComparePins(item1, item2)
lsGeneric item1, item2;
{
    PinInst *pin1Ptr = (PinInst *) item1;
    PinInst *pin2Ptr = (PinInst *) item2;

    /*
     * Note that strcmp() will not produce correct results if both strings
     * contain only digits.
     */
    return (strcmp(Mac_GetNameOfPinInst(pin1Ptr),
        Mac_GetNameOfPinInst(pin2Ptr)));
}


/* ARGSUSED */
static BearStatus
AddCellAreaToTotal(cellInstPtr, totalCellAreaPtr)
CellInst        *cellInstPtr;
register double *totalCellAreaPtr;
{
    GenericCell         gcell;
    register LinkedRect *lnRectPtr;
    double              w, h;

    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = cellInstPtr;

    for (lnRectPtr = Mac_GetHorzRectListOfGCell(&gcell);
        lnRectPtr != NIL(LinkedRect);
        lnRectPtr = lnRectPtr->r_next) {

        w = lnRectPtr->r_r.r_xtop - lnRectPtr->r_r.r_xbot;
        h = lnRectPtr->r_r.r_ytop - lnRectPtr->r_r.r_ybot;
        *totalCellAreaPtr += (w * h);
    }

    return (BEAR_CONTINUE_SEARCH);
}


static void
CalcCellAreaStats(cellMasterPtr, routStatsPtr)
CellMaster      *cellMasterPtr;
RoutingStats    *routStatsPtr;
{
    PlaneData   *planeDataPtr;
    Rect        bbox, *horzRectPtr;
    double      w, h;

    routStatsPtr->totalCellArea = 0.0;
    (void) Mac_ItrChildrenOfCell(cellMasterPtr, AddCellAreaToTotal,
        (ClientData) &routStatsPtr->totalCellArea);

    Mac_FindBBoxForAllCells(cellMasterPtr, &bbox);
    w = bbox.r_xtop - bbox.r_xbot;
    h = bbox.r_ytop - bbox.r_ybot;
    routStatsPtr->coreArea = w * h;

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    horzRectPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
    w = horzRectPtr->r_xtop - horzRectPtr->r_xbot;
    h = horzRectPtr->r_ytop - horzRectPtr->r_ybot;
    routStatsPtr->chipArea = w * h;
}


#ifdef USE_DR_CODE
static BearStatus
FindNetLen(netPtr, routStatsPtr)
Net             *netPtr;
RoutingStats    *routStatsPtr;
{
    routStatsPtr->netLength = 0;
    routStatsPtr->netLayer1 = 0;
    routStatsPtr->netLayer2 = 0;
    routStatsPtr->netVias = 0;
    routStatsPtr->netArea = 0;

    fprintf (routStatsPtr->filePtr, "    Net %8s :\n",
        Mac_GetNameOfNet(netPtr));

    (void) Mac_ItrRRootOfNet(netPtr, FindRRootLen, (ClientData) routStatsPtr);

    fprintf (routStatsPtr->filePtr, "        Layer 1 length is %d\n", 
        routStatsPtr->netLayer1);
    fprintf (routStatsPtr->filePtr, "        Layer 2 length is %d\n", 
        routStatsPtr->netLayer2);
    fprintf (routStatsPtr->filePtr, "        Total length is %d\n", 
        routStatsPtr->netLayer1 + routStatsPtr->netLayer2);
    fprintf (routStatsPtr->filePtr, "        Total vias is %d\n", 
        routStatsPtr->netVias);

    fprintf (routStatsPtr->filePtr, "        Total ");
    if (Mac_GetKindOfNet(netPtr) == (NetKind) POWER_NET) {
        fprintf (routStatsPtr->filePtr, "(power) ");
        routStatsPtr->totalPNetArea += routStatsPtr->netArea;
    }
    else if (Mac_GetKindOfNet(netPtr) == (NetKind) GROUND_NET) {
        fprintf (routStatsPtr->filePtr, "(ground) ");
        routStatsPtr->totalGNetArea += routStatsPtr->netArea;
    }
    fprintf (routStatsPtr->filePtr, "area is %d\n",
        routStatsPtr->netArea);

    (void) fflush (routStatsPtr->filePtr);

    routStatsPtr->netCount++;
    routStatsPtr->totalLength += routStatsPtr->netLength;
    routStatsPtr->totalLayer1 += routStatsPtr->netLayer1;
    routStatsPtr->totalLayer2 += routStatsPtr->netLayer2;
    routStatsPtr->totalNetVias += routStatsPtr->netVias;
    routStatsPtr->totalNetArea += routStatsPtr->netArea;

    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_DR_CODE */



#ifdef USE_DR_CODE
static BearStatus
FindRRootLen(rRootPtr, routStatsPtr)
root *rRootPtr;
RoutingStats *routStatsPtr;
{
    (void) Rdb_ItrRSegmOfRRoot(rRootPtr, FindRSegLen,
        (ClientData) routStatsPtr);
    (void) Rdb_ItrRJunctOfRRoot(rRootPtr, FindRJunCount,
        (ClientData) routStatsPtr);

    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_DR_CODE */



#ifdef USE_DR_CODE
static BearStatus
FindRSegLen(rSegPtr, routStatsPtr)
rseg            *rSegPtr;
RoutingStats    *routStatsPtr;
{
    gterm *gtp;
    int x0, y0, x1, y1;
    int length, abs();

    if (rSegPtr != NIL(rseg))
    {
        gtp = rSegPtr->hsr->mp->geop->gtp;
        if (rSegPtr->t0fsr == 1) {
            x0 = gtp[rSegPtr->j0sr.rt->tnum].loc.xy[0];
            y0 = gtp[rSegPtr->j0sr.rt->tnum].loc.xy[1];
        } else {
            x0 = rSegPtr->j0sr.rj->locjr.xy[0];
            y0 = rSegPtr->j0sr.rj->locjr.xy[1];
        }
        if (rSegPtr->t1fsr == 1) {
            x1 = gtp[rSegPtr->j1sr.rt->tnum].loc.xy[0];
            y1 = gtp[rSegPtr->j1sr.rt->tnum].loc.xy[1];
        } else {
            x1 = rSegPtr->j1sr.rj->locjr.xy[0];
            y1 = rSegPtr->j1sr.rj->locjr.xy[1];
        }
        if (x0 == x1) {
            length = abs(y0 - y1);
        } else if (y0 = y1) {
            length = abs(x0 - x1);
        }
        if (rSegPtr->msklvl == 2) {
            routStatsPtr->netLayer2 += length;
        } else {
            routStatsPtr->netLayer1 += length;
        }
        routStatsPtr->netArea += length * 2 * rSegPtr->hwidth;
    }
    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_DR_CODE */



#ifdef USE_DR_CODE
static BearStatus
FindRJunCount(rJunPtr, routStatsPtr)
rjun            *rJunPtr;
RoutingStats    *routStatsPtr;
{
    bool    needVia; 
    int     layer;
    rseg    *rSegPtr;

    if (rJunPtr != NIL(rjun))
    {
        needVia  = FALSE;
        rSegPtr = rJunPtr->sljr;
        layer = rSegPtr->msklvl;
        while (rSegPtr != NIL(rseg) &&  needVia == FALSE)
        {
            if (rSegPtr->msklvl != layer)
            {
                needVia = TRUE;
            }
            rSegPtr = nearslsr(rJunPtr, rSegPtr);
        }

        /*
         * 22 means the via had been eliminated by via minimizer, 
         * defined in localSpacerInt.h
         */
        if (needVia == TRUE || rJunPtr->conjr != 22)
        {
            routStatsPtr->netVias++;
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_DR_CODE */



static BearStatus
SetPinPos(portIndexInstPtr, routStatsPtr)
PortInst *portIndexInstPtr;
RoutingStats *routStatsPtr;
{
    Point absPinPos;

    Mac_GetAbsPosOfPin(Mac_GetAssignedPinOfPort(portIndexInstPtr), &absPinPos);
    routStatsPtr->pins[routStatsPtr->size].position.p_x = absPinPos.p_x;
    routStatsPtr->pins[routStatsPtr->size].position.p_y = absPinPos.p_y;
    routStatsPtr->pins[routStatsPtr->size].finishFlag = FALSE;
    routStatsPtr->pins[routStatsPtr->size].lnkEdgeList = NIL(LinkedEdgeList);
    routStatsPtr->size ++;

    return (BEAR_CONTINUE_SEARCH);
}

 

static void
CreateEdge(node1, node2, x_length, y_length, routStatsPtr, edgeListPtr)
int   node1, node2, x_length, y_length;
RoutingStats *routStatsPtr;
EdgeForPins **edgeListPtr;

{
    EdgeForPins *edgePtr;
    LinkedEdgeList *lnkEdge1Ptr, *lnkEdge2Ptr;

    MALLOC(EdgeForPins *, edgePtr, sizeof (EdgeForPins));
    edgePtr->node1Ptr = &(routStatsPtr->pins[node1]);
    edgePtr->node2Ptr = &(routStatsPtr->pins[node2]);
    edgePtr->total_length = x_length+y_length;
    edgePtr->x_length = x_length;
    edgePtr->y_length = y_length;
    edgePtr->finishFlag = FALSE;
    edgePtr->nextPtr  = *edgeListPtr;
    *edgeListPtr = edgePtr;

    MALLOC(LinkedEdgeList *, lnkEdge1Ptr, sizeof (LinkedEdgeList));
    lnkEdge1Ptr->edgeForPinsPtr = edgePtr;
    lnkEdge1Ptr->nextPtr = edgePtr->node1Ptr->lnkEdgeList;
    edgePtr->node1Ptr->lnkEdgeList = lnkEdge1Ptr;
    MALLOC(LinkedEdgeList *, lnkEdge2Ptr, sizeof (LinkedEdgeList));
    lnkEdge2Ptr->edgeForPinsPtr = edgePtr;
    lnkEdge2Ptr->nextPtr = edgePtr->node2Ptr->lnkEdgeList;
    edgePtr->node2Ptr->lnkEdgeList = lnkEdge2Ptr;
}

    
    
static void
GenerateCompGraph(netPtr, routStatsPtr, edgeListPtr)
Net         *netPtr;
RoutingStats *routStatsPtr;
EdgeForPins **edgeListPtr;
{
    int  i,j, x_length, y_length;

    for(i = 0; i < routStatsPtr->size; i++) {
        for(j = i+1; j < routStatsPtr->size; j++) {
            x_length = abs(routStatsPtr->pins[i].position.p_x-
                           routStatsPtr->pins[j].position.p_x);
            y_length = abs(routStatsPtr->pins[i].position.p_y-
                           routStatsPtr->pins[j].position.p_y);
            CreateEdge(i, j, x_length, y_length, routStatsPtr, edgeListPtr);
        }
    }
}



static void
FindMST(edgeListPtr, routStatsPtr)
EdgeForPins **edgeListPtr;
RoutingStats *routStatsPtr;
{
    EdgeForPins     *edgePtr, *edgePtrForMinCost;
    LinkedEdgeList  *lnkEdgeList; 
    int             pinCount, i, initFlag;

    edgePtrForMinCost = *edgeListPtr;
    for(edgePtr = *edgeListPtr; edgePtr != NIL(EdgeForPins); 
        edgePtr = edgePtr->nextPtr) {
        if (edgePtr->total_length < edgePtrForMinCost->total_length)
                edgePtrForMinCost = edgePtr;
    }
    edgePtrForMinCost->finishFlag = TRUE;
    edgePtrForMinCost->node1Ptr->finishFlag = TRUE; 
    edgePtrForMinCost->node2Ptr->finishFlag = TRUE;
    pinCount = 2;
    routStatsPtr->netLength = edgePtrForMinCost->total_length;
    routStatsPtr->netLayer1 = edgePtrForMinCost->x_length;
    routStatsPtr->netLayer2 = edgePtrForMinCost->y_length;

    while(pinCount < routStatsPtr->size) {
        initFlag = 0;
        for(i = 0; i < routStatsPtr->size; i++) {
            if (routStatsPtr->pins[i].finishFlag) {
                for(lnkEdgeList = routStatsPtr->pins[i].lnkEdgeList;
                       lnkEdgeList != NIL(LinkedEdgeList);
                       lnkEdgeList = lnkEdgeList ->nextPtr) {
                    edgePtr = lnkEdgeList->edgeForPinsPtr;
                    if (!edgePtr->finishFlag && 
                       !(edgePtr->node1Ptr->finishFlag && 
                         edgePtr->node2Ptr->finishFlag)) {
                        if (initFlag == 0) {
                             edgePtrForMinCost = edgePtr;
                             initFlag = 1;
                        }
                        else {
                            if (edgePtr->total_length < 
                                edgePtrForMinCost->total_length) {
                                edgePtrForMinCost = edgePtr;
                            }
                        }
                    }
                }
            }
        }
        if (!edgePtrForMinCost->node1Ptr->finishFlag) {
            edgePtrForMinCost->node1Ptr->finishFlag = TRUE;
            pinCount++;
        }
        if (!edgePtrForMinCost->node2Ptr->finishFlag) {
            edgePtrForMinCost->node2Ptr->finishFlag = TRUE;
            pinCount++;
        }
        routStatsPtr->netLength += edgePtrForMinCost->total_length;
        routStatsPtr->netLayer1 += edgePtrForMinCost->x_length;
        routStatsPtr->netLayer2 += edgePtrForMinCost->y_length;
    }
}

 

static void
FreeSpace(edgeListPtr, pinInfo, numOfPorts)
EdgeForPins *edgeListPtr;
PinInfo     *pinInfo;
int         numOfPorts;
{
    EdgeForPins *edgePtr, *oldEdgePtr;
    LinkedEdgeList *lnkEdgePtr;
    int   i;

    edgePtr = edgeListPtr;
    while(edgePtr != NIL(EdgeForPins)) { 
        oldEdgePtr = edgePtr; 
        edgePtr = edgePtr->nextPtr;
    FREE((EdgeForPins *) oldEdgePtr);
    }
   
    for(i = 0; i < numOfPorts; i++) {
        lnkEdgePtr = pinInfo[i].lnkEdgeList;
    FREE((LinkedEdgeList *) lnkEdgePtr);
    }
        
}



static BearStatus
EstimateNetLen(netPtr, routStatsPtr)
Net          *netPtr;
RoutingStats *routStatsPtr;
{
    EdgeForPins *edgeListPtr;
    int         numOfPorts;
    PinInfo     *pinInfo;

    routStatsPtr->netLength = 0;
    routStatsPtr->netLayer1 = 0;
    routStatsPtr->netLayer2 = 0;

    (void) fprintf (routStatsPtr->filePtr,
        "    Net %8s ( %4d pins ) :\n",
        Mac_GetNameOfNet(netPtr), netPtr->numOfPin);

    (void) Mac_NumOfPortsOfNet(netPtr, &numOfPorts);
    if (numOfPorts > 1) {
        MALLOC(PinInfo *, pinInfo, numOfPorts * sizeof (PinInfo));
        routStatsPtr->pins = pinInfo;
        routStatsPtr->size = 0;

        (void) Mac_ItrAllPortOfNet(netPtr, SetPinPos,
            (ClientData) routStatsPtr);

        edgeListPtr = NIL(EdgeForPins);
        GenerateCompGraph(netPtr, routStatsPtr, &edgeListPtr);

        FindMST(&edgeListPtr, routStatsPtr);

        fprintf (routStatsPtr->filePtr, "        Layer 1 length is %d\n", 
            routStatsPtr->netLayer1);
        fprintf (routStatsPtr->filePtr, "        Layer 2 length is %d\n", 
            routStatsPtr->netLayer2);
        fprintf (routStatsPtr->filePtr, "        Total length is %d\n", 
            routStatsPtr->netLayer1 + routStatsPtr->netLayer2);

        (void) fflush (routStatsPtr->filePtr);

        routStatsPtr->netCount++;
        routStatsPtr->totalLayer1 += routStatsPtr->netLayer1;
        routStatsPtr->totalLayer2 += routStatsPtr->netLayer2;

        FreeSpace(edgeListPtr, pinInfo, numOfPorts);
        FREE((PinInfo *) pinInfo);
    }

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
CalcNetLen(netPtr, routStatsPtr)
Net             *netPtr;
RoutingStats    *routStatsPtr;
{
    Gt_Graph baseGraph, compGraph;

    if (Ngr_BuildBaseGraph(routStatsPtr->bldTreePtr, netPtr, &baseGraph) == 
        BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }
    if (Ngr_BuildCompGraph(routStatsPtr->bldTreePtr, &compGraph) == 
        BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }
    if (Ngr_FindMST(routStatsPtr->bldTreePtr) == BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }
    if (Ngr_FindSteinerTree(routStatsPtr->bldTreePtr) == BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }
    Ngr_DeleteCompGraph(&compGraph);

    routStatsPtr->netLayer1 = 0;
    routStatsPtr->netLayer2 = 0;
    if (Ngr_CalcEdgeLenOfGraph(baseGraph, routStatsPtr->planeDataPtr,
        &routStatsPtr->netLayer1, &routStatsPtr->netLayer2) == BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }
    Ngr_DeleteBaseGraph(&baseGraph);

    (void) fprintf (routStatsPtr->filePtr,
        "    Net %8s ( %4d pins ) :\n",
        Mac_GetNameOfNet(netPtr), netPtr->numOfPin);
    (void) fflush (routStatsPtr->filePtr);

    (void) fprintf (routStatsPtr->filePtr,
        "        Layer 1 length is %d\n", 
        routStatsPtr->netLayer1);
    (void) fprintf (routStatsPtr->filePtr,
        "        Layer 2 length is %d\n", 
        routStatsPtr->netLayer2);
    (void) fprintf (routStatsPtr->filePtr,
        "        Total length is %d\n",
        routStatsPtr->netLayer1 + routStatsPtr->netLayer2);

    routStatsPtr->netCount++;
    routStatsPtr->totalLayer1 += routStatsPtr->netLayer1;
    routStatsPtr->totalLayer2 += routStatsPtr->netLayer2;

    return (BEAR_CONTINUE_SEARCH);
}



static void
WriteInfoOne(routStatsPtr)
RoutingStats *routStatsPtr;
{
    (void) fprintf (routStatsPtr->filePtr,
        "Total layer 1 length is %d\n",
        routStatsPtr->totalLayer1);
    (void) fprintf (routStatsPtr->filePtr,
        "Average layer 1 length per net is %.2f\n\n",
        (double) routStatsPtr->totalLayer1 / (double) routStatsPtr->netCount);

    (void) fprintf (routStatsPtr->filePtr,
        "Total layer 2 length is %d\n",
        routStatsPtr->totalLayer2);
    (void) fprintf (routStatsPtr->filePtr,
        "Average layer 2 length per net is %.2f\n\n",
        (double) routStatsPtr->totalLayer2 / (double) routStatsPtr->netCount);

    (void) fprintf (routStatsPtr->filePtr,
        "Total routing length is %d\n",
        routStatsPtr->totalLayer1 + routStatsPtr->totalLayer2);

    (void) fprintf (routStatsPtr->filePtr,
        "Average routing length per net is %.2f\n\n",
        (double) (routStatsPtr->totalLayer1 + routStatsPtr->totalLayer2) /
            (double) routStatsPtr->netCount);

    (void) fflush (routStatsPtr->filePtr);
}


#ifdef USE_DR_CODE
static void
WriteInfoTwo(routStatsPtr)
RoutingStats *routStatsPtr;
{
    (void) fprintf (routStatsPtr->filePtr,
        "Total layer 1 length is %d\n",
        routStatsPtr->totalLayer1);
    (void) fprintf (routStatsPtr->filePtr,
        "Average layer 1 length per net is %.2f\n\n",
        (double) routStatsPtr->totalLayer1 / (double) routStatsPtr->netCount);

    (void) fprintf (routStatsPtr->filePtr,
        "Total layer 2 length is %d\n",
        routStatsPtr->totalLayer2);
    (void) fprintf (routStatsPtr->filePtr,
        "Average layer 2 length per net is %.2f\n\n",
        (double) routStatsPtr->totalLayer2 / (double) routStatsPtr->netCount);

    (void) fprintf (routStatsPtr->filePtr,
        "Total routing length is %d\n",
        routStatsPtr->totalLayer1 + routStatsPtr->totalLayer2);

    (void) fprintf (routStatsPtr->filePtr,
        "Average routing length per net is %.2f\n\n",
        (double) (routStatsPtr->totalLayer1 + routStatsPtr->totalLayer2) /
        (double) routStatsPtr->netCount);

    (void) fprintf (routStatsPtr->filePtr,
        "Total number of vias is %d\n",
        routStatsPtr->totalNetVias);
    (void) fprintf (routStatsPtr->filePtr,
        "Average number of vias per net is %.2f\n\n",
        (double) routStatsPtr->totalNetVias / (double) routStatsPtr->netCount);

    (void) fprintf (routStatsPtr->filePtr,
        "Area of power nets is %d\n",
        routStatsPtr->totalPNetArea);
    (void) fprintf (routStatsPtr->filePtr,
        "Area of ground nets is %d\n",
        routStatsPtr->totalGNetArea);
    (void) fprintf (routStatsPtr->filePtr,
        "Total area of nets is %d\n",
        routStatsPtr->totalNetArea);
    (void) fprintf (routStatsPtr->filePtr,
        "Average area per net is %.2f\n\n",
        (double) routStatsPtr->totalNetArea / (double) routStatsPtr->netCount);

    (void) fflush (routStatsPtr->filePtr);
}
#endif /* USE_DR_CODE */


static void
WriteCellAreaStats(routStatsPtr)
RoutingStats *routStatsPtr;
{
    double percent;

    percent = (routStatsPtr->totalCellArea * 100.0) / routStatsPtr->chipArea;
    (void) fprintf (routStatsPtr->filePtr, "%-40s = %15.0f (%6.2f%%)\n",
        "Area of cells", 
        routStatsPtr->totalCellArea, percent);

    percent = (routStatsPtr->coreArea * 100.0) / routStatsPtr->chipArea;
    (void) fprintf (routStatsPtr->filePtr, "%-40s = %15.0f (%6.2f%%)\n",
        "Area of bounding box around all cells",
        routStatsPtr->coreArea, percent);

    (void) fprintf (routStatsPtr->filePtr, "%-40s = %15.0f (100.00%%)\n\n",
        "Total chip area",
        routStatsPtr->chipArea);

    (void) fflush (routStatsPtr->filePtr);
}


static BearStatus
PrintNet(netPtr, routStatsPtr)
Net             *netPtr;
RoutingStats    *routStatsPtr;
{
    lsGen       gen;
    lsList      cellList, pinList;
    CellInst    *cellInstPtr;
    PinInst     *pinPtr;
    PortInst    *portPtr;
    int         i, pinCount;

    cellList = lsCreate();
    (void) Mac_ItrAllPortOfNet(netPtr, AddCellToList, (ClientData) cellList);
    lsSort(cellList, CompareCells);
    lsUniq(cellList, CompareCells, (void (*)()) NULL);

    pinList = lsCreate();
    (void) Mac_ItrAllPinOfNet(netPtr, AddPinToList, (ClientData) pinList);
    pinCount = lsLength(pinList);
    lsSort(pinList, ComparePins);

    (void) fprintf (routStatsPtr->filePtr, 
        "net name        : %s\n", Mac_GetNameOfNet(netPtr));
    (void) fprintf (routStatsPtr->filePtr, 
        "number of cells : %d\n", lsLength(cellList));
    (void) fprintf (routStatsPtr->filePtr, 
        "number of pins  : %d\n", pinCount);

    i = 1;
    (void) fprintf (routStatsPtr->filePtr, 
        "list of cells :\n");
    lsForeachItem(cellList, gen, cellInstPtr) {
        (void) fprintf (routStatsPtr->filePtr, 
            "%3d) %s\n", i, Mac_GetNameOfCellInst(cellInstPtr));
        i++;
    }

    i = 1;
    (void) fprintf (routStatsPtr->filePtr, 
        "list of pins :\n");
    lsForeachItem(pinList, gen, pinPtr) {
        portPtr = Mac_GetPortOfPin(pinPtr);
        cellInstPtr = Mac_GetCellInstOfPort(portPtr);

        (void) fprintf (routStatsPtr->filePtr, 
            "%3d) %s ( %s )\n", i, 
            Mac_GetNameOfPinInst(pinPtr),
            Mac_GetNameOfCellInst(cellInstPtr));
        i++;
    }

    (void) fprintf (routStatsPtr->filePtr, "\n");

    lsDestroy(cellList, (void (*)()) NULL);
    lsDestroy(pinList, (void (*)()) NULL);

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
WriteStatsZero(cellMasterPtr, fileNamePtr)
CellMaster  *cellMasterPtr;
char        *fileNamePtr;
{
    char            *cmFileNamePtr; /* Cell Master file name. */
    RoutingStats    routStats;

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((routStats.filePtr = fopen (fileNamePtr, "w")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg, 
            "Could not open file named \"%s\" for writing\n", fileNamePtr);
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_ERROR_3);
    }

    cmFileNamePtr = Mac_GetFileNameOfCellMaster(cellMasterPtr);
    (void) fprintf (routStats.filePtr,
        "List of nets for \"%s\" sorted by decreasing pin count.\n\n",
        cmFileNamePtr);

    (void) Mac_ItrNetOfCell(cellMasterPtr, PrintNet, (ClientData) &routStats);

    (void) fprintf (routStats.filePtr, "\n");

    (void) fclose (routStats.filePtr);

    return (BEAR_OK);
}



static BearStatus
WriteStatsOne(cellMasterPtr, fileNamePtr)
CellMaster  *cellMasterPtr;
char        *fileNamePtr;
{
    char         *cmFileNamePtr; /* Cell Master file name. */
    Net          *netListPtr;
    RoutingStats routStats;

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((routStats.filePtr = fopen (fileNamePtr, "w")) == NIL(FILE)) {
        sprintf(bear_ErrorMsg, "Could not open file named \"%s\" for writing\n",   
            fileNamePtr);
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_ERROR_3);
    }

    cmFileNamePtr = Mac_GetFileNameOfCellMaster(cellMasterPtr);
    (void) fprintf (routStats.filePtr,
        "Routing statistics for \"%s\" based on bounding boxes.\n\n",
        cmFileNamePtr);

    routStats.netCount = 0;
    routStats.totalLayer1 = 0;
    routStats.totalLayer2 = 0;

    routStats.totalLength = 0;
    routStats.totalNetVias = 0;
    routStats.totalPNetArea = 0;
    routStats.totalGNetArea = 0;
    routStats.totalNetArea = 0;

    netListPtr = Mac_GetNetListOfCell(cellMasterPtr);
    (void) Mac_ItrNet(netListPtr, EstimateNetLen, (ClientData) &routStats);
    (void) fprintf (routStats.filePtr, "\n");

    CalcCellAreaStats(cellMasterPtr, &routStats);

    WriteInfoOne(&routStats);
    WriteCellAreaStats(&routStats);
    (void) fclose (routStats.filePtr);

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        routStats.filePtr = bearout;
        (void) fprintf (routStats.filePtr,
            "Routing statistics for \"%s\" based on bounding boxes.\n\n",
            cmFileNamePtr);

        WriteInfoOne(&routStats);
        WriteCellAreaStats(&routStats);
    }
#endif BEAR_BLACK_BOX

    return (BEAR_OK);
}



#ifdef USE_DR_CODE
static BearStatus
WriteStatsTwo(cellMasterPtr, fileNamePtr)
CellMaster  *cellMasterPtr;
char        *fileNamePtr;
{
    Net             *netListPtr;
    RoutingStats    routStats;
    char            *cmFileNamePtr; /* Cell Master file name. */

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((routStats.filePtr = fopen (fileNamePtr, "w")) == NIL(FILE)) {
        sprintf(bear_ErrorMsg, "Could not open file named \"%s\" for writing\n",   
            fileNamePtr);
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_ERROR_3);
    }

    cmFileNamePtr = Mac_GetFileNameOfCellMaster(cellMasterPtr);
    (void) fprintf (routStats.filePtr,
        "Statistics for \"%s\" after ring routing.\n\n", cmFileNamePtr);

    routStats.netCount = 0;
    routStats.totalLength = 0;
    routStats.totalLayer1 = 0;
    routStats.totalLayer2 = 0;
    routStats.totalNetVias = 0;
    routStats.totalPNetArea = 0;
    routStats.totalGNetArea = 0;
    routStats.totalNetArea = 0;

    netListPtr = Mac_GetNetListOfCell(cellMasterPtr);
    (void) Mac_ItrNet(netListPtr, FindNetLen, (ClientData) &routStats);
    (void) fprintf (routStats.filePtr, "\n");

    CalcCellAreaStats(cellMasterPtr, &routStats);

    WriteInfoTwo(&routStats);
    WriteCellAreaStats(&routStats);
    (void) fclose (routStats.filePtr);

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        routStats.filePtr = bearout;
        (void) fprintf (routStats.filePtr,
            "Statistics for \"%s\" after ring routing.\n\n",
            cmFileNamePtr);

        WriteInfoTwo(&routStats);
        WriteCellAreaStats(&routStats);
    }
#endif BEAR_BLACK_BOX

    return (BEAR_OK);
}
#endif /* USE_DR_CODE */



static BearStatus
WriteStatsThree(cellMasterPtr, fileNamePtr)
CellMaster  *cellMasterPtr;
char        *fileNamePtr;
{
    char            *cmFileNamePtr; /* Cell Master file name. */
    RoutingStats    routStats;

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((routStats.filePtr = fopen (fileNamePtr, "w")) == NIL(FILE)) {
        sprintf(bear_ErrorMsg, "Could not open file named \"%s\" for writing\n",   
            fileNamePtr);
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_ERROR_3);
    }

    routStats.netCount = 0;
    routStats.totalLayer1 = 0;
    routStats.totalLayer2 = 0;

    routStats.totalLength = 0;      /* not calculated */
    routStats.totalNetVias = 0;     /* not calculated */
    routStats.totalPNetArea = 0;    /* not calculated */
    routStats.totalGNetArea = 0;    /* not calculated */
    routStats.totalNetArea = 0;     /* not calculated */

    routStats.planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    if (Ngr_BldTreeBegin(cellMasterPtr, NIL(Ngr_BldOptions),
        &routStats.bldTreePtr) != BEAR_OK) {
      Bear_PrintErr("Unable to compute routing statistics based on steiner trees.");

        (void) fclose (routStats.filePtr);
        return (BEAR_ERROR);
    }

    cmFileNamePtr = Mac_GetFileNameOfCellMaster(cellMasterPtr);
    (void) fprintf (routStats.filePtr,
        "Routing statistics for \"%s\" based on steiner trees.\n\n",
        cmFileNamePtr);

    if (Mac_ItrNetOfCell(cellMasterPtr, CalcNetLen,
        (ClientData) &routStats) == BEAR_STOP_SEARCH) {
 Bear_PrintErr("Unable to compute routing statistics based on steiner trees.");

        (void) fclose (routStats.filePtr);
        return (BEAR_ERROR);
    }
    (void) fprintf (routStats.filePtr, "\n");

    (void) Ngr_BldTreeEnd(routStats.bldTreePtr);

    CalcCellAreaStats(cellMasterPtr, &routStats);

    WriteInfoOne(&routStats);
    WriteCellAreaStats(&routStats);
    (void) fclose (routStats.filePtr);

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        routStats.filePtr = bearout;
        (void) fprintf (routStats.filePtr,
            "Routing statistics for \"%s\" based on steiner trees.\n\n",
            cmFileNamePtr);

        WriteInfoOne(&routStats);
        WriteCellAreaStats(&routStats);
    }
#endif BEAR_BLACK_BOX

    return (BEAR_OK);
}



static BearStatus
WriteStatsFour(cellMasterPtr)
CellMaster *cellMasterPtr;
{
    RoutingStats routStats;

    routStats.filePtr = stdout;
    CalcCellAreaStats(cellMasterPtr, &routStats);
    WriteCellAreaStats(&routStats);
    (void) fflush (stdout);

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_WriteStats() --
 *
 * type 0 : List of nets sorted by decreasing pin count.
 *          Written to the given file.
 * type 1 : Routing statistics based on bounding boxes, and cell statistics.
 *          Written to the given file.
 * type 2 : Routing statistics based on steiner trees, and cell statistics.
 *          Written to the given file.
 * type 3 : Routing statistics based on ring routing results, and cell
 *          statistics.
 *          Written to the given file.
 * type 4 : Cell statistics.
 *          Written to the standard output.
 *
 * Results:
 *   BEAR_OK is returned if everything is spiffy.
 *   BEAR_ERROR is returned
 *       if (type == 2 and there was a problem with the calculation),
 *       or if (type == 3 and the ring routing has not been done) 
 *   BEAR_ERROR_2 is returned if type does not equal 0, 1, 2, 3, or 4.
 *   BEAR_ERROR_3 is returned if the output file could not be opened
 *       for writing.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_WriteStats(type, cellMasterPtr, fileNamePtr)
unsigned int    type;
CellMaster      *cellMasterPtr;
char            *fileNamePtr;
{
    lsList  netList;
    lsGen   gen;
    Net     *netPtr;

    /*
     * The first thing we do is to count the number of pins in each
     * net, and then re-do the MAC net list so that nets appear in
     * order of decreasing pin count.
     */
    netList = lsCreate();
    (void) Mac_ItrNetOfCell(cellMasterPtr, AddNetToList, (ClientData) netList);
    lsSort(netList, CompareNets);
    cellMasterPtr->contents.netList = NIL(Net);
    lsForeachItem(netList, gen, netPtr) {
        netPtr->nextPtr = cellMasterPtr->contents.netList;
        cellMasterPtr->contents.netList = netPtr;
    }
    lsDestroy(netList, (void (*)()) NULL);

    switch (type) {
    case 0:
        return (WriteStatsZero(cellMasterPtr, fileNamePtr));

    case 1:
        return (WriteStatsOne(cellMasterPtr, fileNamePtr));

    case 2:
        return (WriteStatsThree(cellMasterPtr, fileNamePtr));

#ifdef USE_DR_CODE
    case 3:
        if (Mac_IsRingRoutingDone(cellMasterPtr) == FALSE) {
            Bear_PrintErr("Complete ring routing and try again.\n");
            return (BEAR_ERROR);
        }
        return (WriteStatsTwo(cellMasterPtr, fileNamePtr));
#endif /* USE_DR_CODE */

    case 4:
        return (WriteStatsFour(cellMasterPtr));
        break;

    default:
        Bear_PrintErr("Mac_WriteStats() error : unknown type.\n");
        return (BEAR_ERROR_2);
    }
    /* NOTREACHED */
}



