
/* nclCluster.c --
 *
 *    Bernhard Eschermann
 *    (Advising Professor - E. S. Kuh)
 *
 *    Contains all the procedures that are called from ClusterGen in
 *    nclMain.c, with exception of the procedure to sort the edgeList.
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclCluster.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "newCInt.h"



/* ---------- Local Procedures ---------- */
static bool MatchBlocks();


/*
 * ----------------------------------------------------------------------------
 *
 * NclSumConn() --
 *
 *    Needed to compute one part of the sortingCrit for the edges.
 *    g(i) = sum(j) c(i, j)
 *
 * Results:
 *    Returns sum of the connectivities of all edges in edgeList.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

double
NclSumConn(edgeList)
ConnEdge *edgeList;
{
    ConnEdge    *connEdgePtr;
    double      sum = 0.0;

    for (connEdgePtr = edgeList; 
        connEdgePtr != NIL(ConnEdge);
        NEXT(connEdgePtr)) {

        sum += connEdgePtr->edgePtr->connect1;
    }

    return (sum);
}



/*
 * ----------------------------------------------------------------------------
 *
 * NclCycleDetect() --
 *
 *    Another part of the calculation of the sortingCrit for the edges.
 *    At the moment only detects cycles of length 3.
 *
 * Results:
 *    Returns the sum of the "weakest" edgeWeights of all cycles 
 *    to which the edge between node1 and node2 belongs.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

double
NclCycleDetect(node1, node2)
GraphNode *node1, *node2;
{
    ConnEdge *edgePtr1, *edgePtr2;
    GraphNode *neighb1, *neighb2;
    double cycleWeight = 0.0;

    for (edgePtr1 = node1->edgeList; edgePtr1 != NIL(ConnEdge);
        NEXT(edgePtr1)) {
        neighb1 = (edgePtr1->edgePtr->node1Ptr == node1) ?
            edgePtr1->edgePtr->node2Ptr : edgePtr1->edgePtr->node1Ptr;

        for (edgePtr2 = node2->edgeList; edgePtr2 != NIL(ConnEdge);
            NEXT(edgePtr2)) {
            neighb2 = (edgePtr2->edgePtr->node1Ptr == node2) ?
                edgePtr2->edgePtr->node2Ptr : edgePtr2->edgePtr->node1Ptr;

            if (edgePtr1->edgePtr->edgeType == edgePtr2->edgePtr->edgeType &&
                neighb1 == neighb2) {
                cycleWeight += MIN(edgePtr1->edgePtr->connect1,
                    edgePtr2->edgePtr->connect1);
            }
        }
    }
    
    return (cycleWeight);
}



/*
 * ----------------------------------------------------------------------------
 *
 * NclEval() --
 *
 *    The final part of the computation of sortingCrit, this score is
 *    later on used to sort the edges.
 *    c'(i, j) = f(i) * (c(i, j) + w) / (g(i) - c(i,j)) + the same for j.
 *    g(i) is provided by NclSumConn, w by NclCycleDetect, f(i) is just a
 *    normalizing factor 1/numOfSubnodes.
 *
 * Results:
 *    Returns the score of the edge to which edgePtr points.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

double
NclEval(edgePtr)
GraphEdge *edgePtr;
{
    double score1, score2, finalScore;

    if (edgePtr->node1Ptr->totalConn == edgePtr->connect1 ||
        edgePtr->node2Ptr->totalConn == edgePtr->connect1)
    {
        /* only connection of this node */
        return (99999.9);
    }

    score1 = 1 / (edgePtr->node1Ptr->totalConn - edgePtr->connect1)
               / edgePtr->node1Ptr->numOfSubnodes;
    score2 = 1 / (edgePtr->node2Ptr->totalConn - edgePtr->connect1)
               / edgePtr->node2Ptr->numOfSubnodes;
    finalScore = (score1 + score2) *
        (edgePtr->connect1 + edgePtr->cycleWeight);
    /*
    if (edgePtr->node1Ptr->belongsToPtr->gn_numOfChildren +
        edgePtr->node2Ptr->belongsToPtr->gn_numOfChildren == 5)
    {
        finalScore -= PENALTY (???);
    }
    */
    return (finalScore);
}



/*
 * ----------------------------------------------------------------------------
 *
 * MatchBlocks() --
 *
 *    Checks if merging of the two blocks in one cluster should be rejected
 *    because of geometric considerations.
 *
 * Results:
 *    Returns TRUE if it is acceptable to merge the two nodes, otherwise FALSE.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

static bool
MatchBlocks(node1Ptr, node2Ptr, areaBound, ratioBound)
register GraphNode  *node1Ptr, *node2Ptr;
double              areaBound, ratioBound;
{
    double ratio = (double) node1Ptr->gn_maxDim / (double) node2Ptr->gn_maxDim;

    return ((node1Ptr->gn_area < areaBound) && 
        (node2Ptr->gn_area < areaBound) &&
        (ratio < ratioBound) && 
        (ratio > (1.0 / ratioBound)));
}



/*
 * ----------------------------------------------------------------------------
 *
 * NclMergeTest() --
 *
 *    Checks geometrical constraints and cluster size constraints for
 *    the merging of two clusters.
 *
 * Results:
 *    TRUE if the nodes should be merged, FALSE if they shouldn't.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

bool
NclMergeTest(edgePtr, areaBound, ratioBound)
GraphEdge   *edgePtr;
double      areaBound, ratioBound;
{
    bool geoMatch, sizeOK;

    geoMatch = MatchBlocks(edgePtr->node1Ptr, edgePtr->node2Ptr, areaBound,
        ratioBound);
    sizeOK = ClusterSizeMc(edgePtr->node1Ptr) +
        ClusterSizeMc(edgePtr->node2Ptr) <= MAX_CHILDREN;

    return (geoMatch && sizeOK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * NclClusterMerge() --
 *
 *    Adds the numerical information from the clusterNodes of node1 and node2
 *    and stores it in the clusterNode to which node1 belongs; the clusterNode
 *    of node2 is eliminated from the list of clusterNodes, nodes connected to
 *    this clusterNode are now connected to the clusterNode of node1.
 *
 * Results:
 *    TRUE if the nodes have been actually merged, FALSE if they were
 *    already in one cluster.
 *
 * Side effects:
 *    The clusterNodes list is changed to represent the new situation.
 *
 * ----------------------------------------------------------------------------
 */

bool
NclClusterMerge(node1, node2, clusterNodesPtr)
GraphNode   *node1; 
GraphNode   *node2;
GraphNode   **clusterNodesPtr;
{
    GraphNode   *survivor = node1->belongsToPtr;
    GraphNode   *victim = node2->belongsToPtr;
    /* for efficiency later on: cluster with fewer nd_children = victim */
    int         i;

    if (survivor == victim) {
        return (FALSE);
    }

    survivor->numOfSubnodes += victim->numOfSubnodes;
    survivor->gn_area += victim->gn_area;
    survivor->maxArea = MAX(survivor->maxArea, victim->maxArea);
    survivor->minArea = MIN(survivor->minArea, victim->minArea);
    survivor->gn_maxDim = MAX(survivor->gn_maxDim, victim->gn_maxDim);

#ifdef DEBUG
    if ((survivor->gn_numOfChildren + victim->gn_numOfChildren) >
        MAX_CHILDREN) {
        Bear_PrintFatalErr("NclClusterMerge() : memory troubles");
    }
#endif DEBUG
    for (i = 0; i < victim->gn_numOfChildren; i++) {
        survivor->gn_children[survivor->gn_numOfChildren] =
            victim->gn_children[i];
        survivor->gn_numOfChildren++;

        victim->gn_children[i]->belongsToPtr = survivor;
    }

    if (victim->prevPtr != NIL(GraphNode)) {
        victim->prevPtr->nextPtr = victim->nextPtr;
    }
    else {
        *clusterNodesPtr = victim->nextPtr;
    }

    if (victim->nextPtr != NIL(GraphNode)) {
        victim->nextPtr->prevPtr = victim->prevPtr;
    }

    FREE((char *) victim);

    return (TRUE);
}


