
/* nclInterface.c --
 *
 *    George Carvalho, Esq.
 *    Bernhard Eschermann
 *    (Advising Professor - E. S. Kuh)
 *
 * This file contains the interface to the clustering program
 * of Mallek for the hierarchical floorplanner.  This file also
 * contains the interface for the polarBear routines.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclInterface.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "newCInt.h"

#define USE_SIM_ANN_CODE 1

#ifndef BEAR_BLACK_BOX
#include "bearIV.h"
#endif BEAR_BLACK_BOX

/* ---------- External Procedures ---------- */
extern void         DoMatching();                             
extern void         DoSimAnnealing();

/* ---------- Local Procedures ---------- */
static int          CompareNode();
static int          Interface1(); 
static BearStatus   ReadTmpCTree();

static BearStatus   DoFileIn();
static int          DoRandom();

/* ---------- Local Global Variables ---------- */
#if !defined(BEAR_BLACK_BOX)
static IVWindow *globalIVWinPtr;
#endif /* BEAR_BLACK_BOX */

static int      *areaRatioPtr, *dimRatioPtr;
static bool     promptFlag;



static int
CompareNode(item1, item2)
lsGeneric item1, item2;
{
    register int a = ((GraphNode *) item1)->gn_pos;
    register int b = ((GraphNode *) item2)->gn_pos;

    return ((a == b) ? 0 : ((a < b) ? -1 : 1));
}


/*
 * ----------------------------------------------------------------------------
 *
 * NclClusterGen2() --
 *
 *    Transforms the data structure of the C program into an easier data
 *    structure that can be used by the fortran clustering program to
 *    produce the next clustering level. This information is then again
 *    transformed to fit into the C tree structure.
 *
 * Results:
 *   BEAR_ERROR is returned if the cluster file could not be opened.
 *   BEAR_OK otherwise.
 *
 * Side effects:
 *    Changes clusterNodesPtr to represent the next higher hierarchical
 *    level.
 *    Places the resulting number of remaining nodes (clusters) in
 *    numOfClustersPtr.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
NclClusterGen2(netNodes, netEdgesPtr, levelList,
    clusterNodesPtr, numOfClustersPtr)

GraphNode   *netNodes;          /* IN : nodes of net to be clustered */
GraphEdge   **netEdgesPtr;      /* IN : edges of net to be clustered */
lsList      levelList;          /* IN  */
GraphNode   **clusterNodesPtr;  /* OUT */
int         *numOfClustersPtr;  /* OUT */
/*
 * The array of integers named dimList contains three fields for each element 
 * and is used as follows:
 *
 *     1.) The first field ( dimList[3*i] ) contains the pointer to a node
 *         and is used by the random, matching, annealing, and file-in
 *         algorithums.  
 *     2.) The second field ( dimList[3*i + 1] ) is only used by the matching 
 *         and annealing algorithums.
 *     3.) The third field ( dimList[3*i + 2] ) is not used at all anymore but
 *         it is allocated and must be skipped over.  It should actually be
 *         removed but it is embedded to deeply in the code.
 * 
 * The array of doubles named dimAreaList contains one field for each element
 * and is used by the matching and annealing algorithums as a replacement for
 * the third field of dimList.
 */
{
    GraphEdge   *edgePtr;
    GraphNode   *nodePtr, *node1Ptr, *node2Ptr;
    double      *dimAreaList = NIL(double);
    double      *weights = NIL(double);
    int         *dimList = NIL(int);
    int         *netList = NIL(int); 
    int         *nodeList = NIL(int); 
    int         nodeListSize, numOfNets, numOfNodes, numOfClusters;
    int         cn; /* cluster number */
    int         i, j;
    
    numOfNodes = 0;
    for (nodePtr = netNodes; nodePtr != NIL(GraphNode); NEXT(nodePtr)) {
        nodePtr->totalConn = NclSumConn(nodePtr->edgeList);
        numOfNodes++;
    }

    numOfNets = 0;
    for (edgePtr = *netEdgesPtr; edgePtr != NIL(GraphEdge); NEXT(edgePtr)) {
        edgePtr->sortingCrit = (edgePtr->edgeType != BLOCK) ? 0.0 : 1.0;
        numOfNets++;
    }

    /* 
     * After NclSortEdges() all I/O pad connections are at the end
     * of the netEdges list.
     */
    NclSortEdges(numOfNets, netEdgesPtr);

    MALLOC(int *, dimList, (3 * numOfNodes) * sizeof (int));

    if ((nclAlgorithum == MATCHING) || (nclAlgorithum == SIM_ANNEAL)) {
        MALLOC(int *, netList, (2 * numOfNets) * sizeof (int));
        MALLOC(double *, weights, numOfNets * sizeof (double));
        MALLOC(double *, dimAreaList, numOfNodes * sizeof (double));
    }

    /*
     * Note that each node could theoretically be in a single cluster.
     */
    nodeListSize = numOfNodes * MAX_CHILDREN;
    MALLOC(int *, nodeList, nodeListSize * sizeof (int));
    for (i = 0; i < nodeListSize; i++) {
        nodeList[i] = 0;
    }

    /* 
     * Generate array data structure for Mallek's algorithum.
     */

    numOfNets = 0;
    for (edgePtr = *netEdgesPtr; edgePtr != NIL(GraphEdge); NEXT(edgePtr)) {

        if (edgePtr->edgeType == BLOCK) {
            if ((nclAlgorithum == MATCHING) || (nclAlgorithum == SIM_ANNEAL)) {
                i = 2 * numOfNets;
                netList[i]     = (int) edgePtr->node1Ptr;
                netList[i + 1] = (int) edgePtr->node2Ptr;

                weights[numOfNets] = edgePtr->connect1;
            }

            numOfNets++;
        }
    }

    for (nodePtr = netNodes, numOfNodes = 0; 
        nodePtr != NIL(GraphNode); 
        NEXT(nodePtr), numOfNodes++) {

        i = 3 * numOfNodes;
        dimList[i] = (int) nodePtr;

        if ((nclAlgorithum == MATCHING) || (nclAlgorithum == SIM_ANNEAL)) {
            dimList[i + 1] = nodePtr->gn_maxDim;

            dimAreaList[numOfNodes] = nodePtr->gn_area;
        }
    }

    if (numOfNets == 0) {
        /* MAJOR HACK (massoud) */
        (void) fprintf (stderr, 
            "                    *** WARNING WILL ROBINSON ***\n");
       (void) fprintf (stderr, 
      "Number of nets = 0 so we will do clustering by simulated annealing !\n");
       (void) fprintf (stderr,
            "                    *** WARNING WILL ROBINSON ***\n");

        DoSimAnnealing(dimList, dimAreaList, netList, weights, nodeList, 
            numOfNodes, numOfNets, MAX_CHILDREN, 
            &numOfClusters, 
            *areaRatioPtr, *dimRatioPtr);
    }
    else {
        switch (nclAlgorithum) {
        case RANDOM:
            numOfClusters = DoRandom(dimList, nodeList, numOfNodes);
            break;
        case MATCHING:
        case SIM_ANNEAL:
            numOfClusters = Interface1(netList, dimList, dimAreaList, weights, 
                nodeList, numOfNodes, numOfNets, 
                ncl_ClusterTreeDepth, nclAlgorithum);
            break;
        case FILE_IN:
            if (DoFileIn(levelList, nodeList, nodeListSize, dimList, 
                &numOfClusters) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            break;
        default:
            Bear_PrintErr("NclClusterGen2() failed : unexpected algorithum.");
            return (BEAR_ERROR);
            break;
        }
    }

    /* 
     * Transform lists into next level of tree.
     */
    for (cn = 0; cn < numOfClusters; cn++) {
        j = cn * MAX_CHILDREN;

        node1Ptr = (GraphNode *) nodeList[j];
        if (nclAlgorithum == FILE_IN) {
            node1Ptr->belongsToPtr->gn_pos = cn;
        }

        for (i = 1; i < MAX_CHILDREN; i++) {
            if ((node2Ptr = (GraphNode *) nodeList[j + i]) != NIL(GraphNode)) {
                (void) NclClusterMerge(node1Ptr, node2Ptr, clusterNodesPtr);
            }
        }
    }

    if (nclAlgorithum == FILE_IN) {
        GraphNode   *prevNodePtr;
        lsGen       gen;
        lsList      tmpList;

        tmpList = lsCreate();
        for (nodePtr = *clusterNodesPtr; nodePtr != NIL(GraphNode); 
            NEXT(nodePtr)) {

            lsNewEnd(tmpList, (lsGeneric) nodePtr, LS_NH);
        }
        lsSort(tmpList, CompareNode);

        lsFirstItem(tmpList, (lsGeneric *) &nodePtr, LS_NH);
        *clusterNodesPtr = nodePtr;

        prevNodePtr = NIL(GraphNode);
        lsForeachItem(tmpList, gen, nodePtr) {
            nodePtr->prevPtr = prevNodePtr;

            nodePtr->nextPtr = NIL(GraphNode);
            if (prevNodePtr != NIL(GraphNode)) {
                prevNodePtr->nextPtr = nodePtr;
            }

            prevNodePtr = nodePtr;
        }

        lsDestroy(tmpList, (void (*)()) NULL);
    }

    if (dimList != NIL(int)) {
        FREE((char *) dimList);
    }
    if (dimAreaList != NIL(double)) {
        FREE((char *) dimAreaList);
    }
    if (netList != NIL(int)) {
        FREE((char *) netList);
    }
    if (nodeList != NIL(int)) {
        FREE((char *) nodeList);
    }
    if (weights != NIL(double)) {
        FREE((char *) weights);
    }

    *numOfClustersPtr = numOfClusters;

    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * DoRandom() --
 *
 *    Does a random clustering.
 *
 * Results:
 *    Returns the resulting number of remaining nodes (clusters).
 *
 * Side effects:
 *    The array nodeList afterwards contains the information about the
 *    clusters.
 *    nodeList = [child11, child12, ..., child1m, child21, ...]
 *    If there are less than m = MAX_CHILDREN children per cluster, the
 *    remaining array elements up to the next group of MAX_CHILDREN or
 *    the end of the array are 0.
 *
 * ----------------------------------------------------------------------------
 */

static int
DoRandom(dimList, nodeList, numOfNodes)
int *dimList;       /* in: description of nodes */
int *nodeList;      /* out: result of next level clustering */
int numOfNodes;     /* in: # of nodes */ 
{
    int i, numOfClusters, randomCluster, *assignments;

    numOfClusters = (numOfNodes + MAX_CHILDREN - 1) / MAX_CHILDREN;

    MALLOC(int *, assignments, numOfClusters * sizeof (int));
    for (i = 0; i < numOfClusters; i++) {
        assignments[i] = 0;
    }

    for (i = 0; i < numOfNodes; i++)  {
        do {
            randomCluster = (int) random() % numOfClusters;
        }
        while (assignments[randomCluster] >= MAX_CHILDREN);

        nodeList[randomCluster*MAX_CHILDREN + assignments[randomCluster]] = 
            dimList[3 * i];
        assignments[randomCluster]++;
    }

    FREE((char *) assignments);

    return (numOfClusters);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Interface1() --
 *
 *    This function is given several arrays with information about a graph
 *    and produces a list that describes the next level of the clustering
 *    tree. It calls the fortran function that actually does the clustering.
 *
 *    netList = [node, connected node, node, connected node, ...]
 *    dimList = [node, maxDim, "not used", node, maxDim, "not used", ...]
 *    dimAreaList = [area, area, ...]
 *    weights = [weight, weight, ...]
 *
 * Results:
 *    Returns the resulting number of remaining nodes (clusters).
 *
 * Side effects:
 *    The array nodeList afterwards contains the information about the
 *    clusters.
 *    nodeList = [child11, child12, ..., child1m, child21, ...]
 *    If there are less than m = MAX_CHILDREN children per cluster, the
 *    remaining array elements up to the next group of MAX_CHILDREN or the
 *    end of the array are 0.
 *
 * ----------------------------------------------------------------------------
 */

static int
Interface1(netList, dimList, dimAreaList, weights, nodeList,
    numOfNodes, numOfNets, level, algorithum)

int     netList[];             /* in:  nets */
int     dimList[];             /* in:  node ptr, maxDim, "not used" for nodes */
double  dimAreaList[];         /* in:  area for nodes */


double  weights[];             /* in:  weights of nets in netList */
int     nodeList[];            /* out: result of next level clustering */
int     numOfNodes, numOfNets; /* in:  # of nodes / nets */ 
int     level;                 /* in:  cluster level */ 
int     algorithum;
{
    static double error, inner;

    double  stor1, stor2;
    double  minArea, maxArea;
    int     cnvr = 1, method = 1;
    int     maxChildren = MAX_CHILDREN;
    int     minDim, maxDim;
    int     numOfClusters;
    int     i;

    if (numOfNodes > MAX_CHILDREN) {
        minDim = maxDim = dimList[1];
        minArea = maxArea = dimAreaList[0];
        for (i = 0; i < numOfNodes; i++) {
            if (maxDim < dimList[3*i + 1]) {
                maxDim = dimList[3*i + 1];
            }
            if (minDim > dimList[3*i + 1]) {
                minDim = dimList[3*i + 1];
            }

            if (maxArea < dimAreaList[i]) {
                maxArea = dimAreaList[i];
            }
            if (minArea > dimAreaList[i]) {
                minArea = dimAreaList[i];
            }
        }

        stor1 = maxDim;
        stor2 = minDim;
        *dimRatioPtr = (int) (stor1 / stor2);

        *areaRatioPtr = (int) (maxArea / minArea);

        if (promptFlag == TRUE) {
/*
 * (7 feb 91) This iv code doesn't seem to work.
 *
 *          iv_PromptVar(globalIVWinPtr, dimRatioPtr, 1);
 *          iv_PromptVar(globalIVWinPtr, areaRatioPtr, 1);
 */
            (void) fprintf (bearout, "\n");
            (void) fprintf (bearout, "Current dimension ratio = %d\n",
                *dimRatioPtr);
            (void) fprintf (bearout, "Enter dimension ratio : ");
            (void) fscanf (stdin, "%d", dimRatioPtr);
            (void) fprintf (bearout, "\n");

            (void) fprintf (bearout, "Current area ratio = %d\n",
                *areaRatioPtr);
            (void) fprintf (bearout, "Enter area ratio : ");
            (void) fscanf (stdin, "%d", areaRatioPtr);
            (void) fprintf (bearout, "\n");
        }
    }

    switch (algorithum) {
    case MATCHING:
        /*
         * (gyorgy 17 April 91)
         *
         * The fortran code that did the matching should not be used
         * anymore.  Firstly because we don't want any fortran code, and
         * secondly because the integer elements of the array dimList that
         * it relies on are known to overflow thus producing spurious results.
         */
        DoMatching(dimList, dimAreaList, netList, weights, nodeList, 
            &numOfNodes, &numOfNets, &maxChildren, 
            &level, &method, &cnvr, 
            &numOfClusters, 
            &error, &inner, 
            areaRatioPtr, dimRatioPtr);
        break;
    case SIM_ANNEAL:
        DoSimAnnealing(dimList, dimAreaList, netList, weights, nodeList, 
            numOfNodes, numOfNets, maxChildren, 
            &numOfClusters, 
            *areaRatioPtr, *dimRatioPtr);
        break;
    default:
        Bear_PrintFatalErr("Interface1() failed : unexpected algorithum.");
        break;
    }

    return (numOfClusters);
}



/*
 * ----------------------------------------------------------------------------
 *
 * DoFileIn() --
 *
 *    The numbers are offsets into the array 'dimList' which
 *    was allocated earlier. If this is not the case the program will
 *    totally screw up.
 *
 * Results:
 *    BEAR_OK is returned if everything is spiffy.
 *    BEAR_ERROR otherwise.
 *
 * Side effects:
 *    The array nodeList afterwards contains the information about the
 *    clusters.
 *    nodeList = [child11, child12, ... , child1m, child21, ...]
 *    If there are less than m = MAX_CHILDREN children per cluster, the
 *    remaining array elements up to the next group of MAX_CHILDREN or the
 *    end of the array are 0.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
DoFileIn(levelList, nodeList, nodeListSize, dimList, numOfClustersPtr)
lsList  levelList;          /* IN  */
int     nodeList[];         /* OUT */
int     nodeListSize;       /* IN : number of elements in nodeList[] */
int     dimList[];          /* IN  */
int     *numOfClustersPtr;  /* OUT */
{
    lsGen           levelGen;
    Ncl_InputNode   *inputNodePtr;
    int             cn, i, j, pos, len;

    len = lsLength(levelList);

    if ((len * MAX_CHILDREN) > nodeListSize) {
        Bear_PrintErr("ReadTmpCTree() failed : nodeList array is to small.");
        return (BEAR_ERROR);
    }

    cn = 0;
    lsForeachItem(levelList, levelGen, inputNodePtr) {
        j = cn * MAX_CHILDREN;
        for (i = 0; i < inputNodePtr->in_numOfChildren; i++) {
            pos = inputNodePtr->in_child[i]->in_pos;
            nodeList[j + i] = dimList[3 * pos];
        }

        cn++;
    }

    *numOfClustersPtr = len;

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Ncl_BearInterface() --
 *    This procedure provides an interface between polarBear and the cluster
 *    routines.  It reads from the parameter structure and sets the
 *    corresponding global variables.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Global variables are set to values in parameter structure. (newC.h)
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Ncl_BearInterface(cellMasterPtr, parmsPtr, ivWinPtr)
CellMaster      *cellMasterPtr;
Ncl_UserParms   *parmsPtr;
BearGeneric     ivWinPtr;
{
    bool storeFlag;

    nclAlgorithum = parmsPtr->algorithm;

    areaRatioPtr = parmsPtr->areaRatioPtr;
    dimRatioPtr = parmsPtr->dimRatioPtr;
    promptFlag = parmsPtr->promptFlag;

    nclAreaFact = parmsPtr->areaFact;
    nclAreaRelax = parmsPtr->areaRelax;
    nclRatioMax = parmsPtr->ratioMax;
    nclRatioRelax = parmsPtr->ratioRelax;

    nclConnMin = parmsPtr->connMin;
    ncl_RouteAdjust = parmsPtr->routAdjust;

    ncl_QuickRouteEstFlag = parmsPtr->quickRtEstFlag;
    ncl_ChannelFreeFlag = parmsPtr->chanFreeFlag;
    ncl_TimingFlag = parmsPtr->timingFlag;
    storeFlag = parmsPtr->storeFlag;

    nclCTreeFileName = parmsPtr->clusterFileName;
    nclStFileName = parmsPtr->storageFileName;
    nclTmFileName = parmsPtr->timingFileName;

    ncl_MaxNumOfShapes = parmsPtr->maxNumOfShapes;
    if (ncl_MaxNumOfShapes < 2) {
        Bear_PrintErr("MaxNumOfShapes < 2\n");
        return (BEAR_ERROR);
    }
     
    ncl_WireWidth = Mac_CalcMinPinSpacing();

#ifndef BEAR_BLACK_BOX
    globalIVWinPtr = (IVWindow *) ivWinPtr;
#endif /* BEAR_BLACK_BOX */

#ifdef DEBUG
    (void) fprintf (bearout, "----------\n");
    (void) fprintf (bearout, "New clustering code global variables.\n\n");

    (void) fprintf (bearout, "%-33s : \"%s\"\n", "nclAlgorithum",          
        NclGetAlgString(nclAlgorithum));

    (void) fprintf (bearout, "%-33s : %d\n", "MaxAreaRatio",    
        *areaRatioPtr);
    (void) fprintf (bearout, "%-33s : %d\n", "MaxDimRatio",     
        *dimRatioPtr);
    (void) fprintf (bearout, "%-33s : %s\n", "promptFlag",
        Bear_GetBooleanString(promptFlag));

    (void) fprintf (bearout, "%-33s : %f\n", "nclAreaFact",     
        nclAreaFact);
    (void) fprintf (bearout, "%-33s : %f\n", "nclAreaRelax",    
        nclAreaRelax);
    (void) fprintf (bearout, "%-33s : %f\n", "nclRatioMax",     
        nclRatioMax);
    (void) fprintf (bearout, "%-33s : %f\n", "nclRatioRelax",   
        nclRatioRelax);
    (void) fprintf (bearout, "%-33s : %f\n", "nclConnMin",      
        nclConnMin);
    (void) fprintf (bearout, "%-33s : %f\n", "ncl_RouteAdjust",   
        ncl_RouteAdjust);

    (void) fprintf (bearout, "%-33s : %s\n", "ncl_QuickRouteEstFlag",
        Bear_GetBooleanString(ncl_QuickRouteEstFlag));
    (void) fprintf (bearout, "%-33s : %s\n", "ncl_ChannelFreeFlag",
        Bear_GetBooleanString(ncl_ChannelFreeFlag));
    (void) fprintf (bearout, "%-33s : %s\n", "ncl_TimingFlag",
        Bear_GetBooleanString(ncl_TimingFlag));
    (void) fprintf (bearout, "%-33s : %u\n", "ncl_MaxNumOfShapes", 
        ncl_MaxNumOfShapes);

    (void) fprintf (bearout, "%-33s : %s\n", "storeFlag",
        Bear_GetBooleanString(storeFlag));

    (void) fprintf (bearout, "%-33s : \"%s\"\n", "nclCTreeFileName",  
        (nclCTreeFileName != NIL(char)) ?  nclCTreeFileName : "No file given");
    (void) fprintf (bearout, "%-33s : \"%s\"\n", "nclStFileName",  
        (nclStFileName != NIL(char))    ?  nclStFileName    : "No file given");
    (void) fprintf (bearout, "%-33s : \"%s\"\n", "nclTmFileName",  
        (nclTmFileName != NIL(char))    ?  nclTmFileName    : "No file given");

    (void) fprintf (bearout, "%-33s : %f\n", "ncl_WireWidth",
        ncl_WireWidth);

    (void) fprintf (bearout, "----------\n\n");
    fflush (bearout);
#endif

    if (NclBuildTree(cellMasterPtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    if (storeFlag == TRUE) {
        if (NclWriteCTree(ncl_ClusterTree, nclStFileName) != BEAR_OK) {
            return (BEAR_ERROR);
        }
    }

    return (BEAR_OK);
}


