
/* nclTiming.c
 *
 *     Yasushi Ogawa
 *     (Advising Professor - E. S. Kuh)
 *
 * This file contains the procedures associated with clustering for 
 * timing-driven placement.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclTiming.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "newCInt.h"

#ifdef DEBUG
bool timing_printStat = FALSE;
#endif DEBUG

static float rho[] = { 0.0, 
    0.0,  1.0,  1.0,  1.5,  1.5,  1.67, 1.75, 1.83, 2.0,  2.0,
    2.16, 2.23, 2.30, 2.37, 2.44, 2.50, 2.56, 2.62, 2.68, 2.74 }; /* constant */


/* ---------- Local Procedures ---------- */
static double   CalculateWeights();
static double   EstimateWireLength();
static void     EstimateArea();


/*
 * ----------------------------------------------------------------------------
 *
 * NclAssignTimingWeights() --
 *
 *    Assign timing-driven weights to the natural connectivity for clustering. 
 *    The weight is calculated by comparing the estimated wire length and the 
 *    correponding required wire length for each critical net.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Connectivity is changed by weighting.
 *
 * ----------------------------------------------------------------------------
 */

void
NclAssignTimingWeights(netEdgesPtr, critNetList, maxConnect)
GraphEdge           *netEdgesPtr; /* in: edges of current cluster graph */
Ncl_LinkedCritNet   *critNetList; /* in: critical net list */
double              *maxConnect;  /* out: max. of connectivity in all edges */
{
    Ncl_LinkedCritNetOnEdge *critOnEdgePtr;
    Ncl_LinkedCritNet       *lnkCritNetPtr;
    GraphEdge               *edgePtr; 
    double                  weightOfConnectivity, estimatedWireLength;
    int                     biasFlag, nodesNumber;
 
    /* 
     * For all critical nets, weight for timing is calculated and 
     * stored in timeWeight. Additionally, maxConnect is set. 
     */
    *maxConnect = 0.0;

    for (edgePtr = netEdgesPtr; edgePtr != NIL(GraphEdge); NEXT(edgePtr)) {
        if (*maxConnect < edgePtr->connect1) {
            *maxConnect = edgePtr->connect1;
        }
        /* for all critical nets that edge belongs to  */
        for (critOnEdgePtr = edgePtr->critOnEdgePtr;
                critOnEdgePtr != NIL(Ncl_LinkedCritNetOnEdge); 
                NEXT(critOnEdgePtr)) {
            lnkCritNetPtr = critOnEdgePtr->critNetPtr;
            /*  examine that the net was not handled  */
            if (lnkCritNetPtr->timeWeight < (0.0)) {
                estimatedWireLength = 
                    EstimateWireLength(lnkCritNetPtr->netPtr, 
                        netEdgesPtr, &nodesNumber);
                weightOfConnectivity = 
                    CalculateWeights(lnkCritNetPtr, estimatedWireLength);
                lnkCritNetPtr->timeWeight = weightOfConnectivity;

#ifdef WEIGHT_PROP
                lnkCritNetPtr->numPseudoPins = nodesNumber;
#endif

            }
        }
    }

    /*  for all edges of cluster graph, connectivity for clustering is
     *  weighted by summing the associated timing weights. */
    for (edgePtr = netEdgesPtr; edgePtr != NIL(GraphEdge); NEXT(edgePtr)) {
        biasFlag = 0;
        /*  for all critical nets that the edge belongs to  */
        for (critOnEdgePtr = edgePtr->critOnEdgePtr;
                critOnEdgePtr != NIL(Ncl_LinkedCritNetOnEdge); 
                NEXT(critOnEdgePtr)) {
            lnkCritNetPtr = critOnEdgePtr->critNetPtr;
            weightOfConnectivity = lnkCritNetPtr->timeWeight;

            /*  add a weight for the edge  */
            if (weightOfConnectivity != (0.0)) {
                edgePtr->connect1 += weightOfConnectivity;

                /* overflow check for clster routine. */
                if (edgePtr->connect1 > (21400.0)) {
                    printf ("WARNING: connect1 exceeds max integer. \n");
                    printf ("Specified maximum length is too small. \n");
                    edgePtr->connect1 = (21000.0);
                }

                biasFlag = 1;
            }

        }
        /*  add max value to the old connectivity as a bias for the edge  */
        if (biasFlag == 1) {
            edgePtr->connect1 += *maxConnect;

            /* overflow check for clster routine. */
            if (edgePtr->connect1 > (21400.0)) {
                printf ("WARNING: connectivity exceeds max integer. \n");
                printf ("Specified maximum length is too small. \n");
                edgePtr->connect1 = (21000.0);
            }

        }
    }
}



/*
 * ----------------------------------------------------------------------------
 *
 * CalculateWeights() --
 *
 *    Given a critical net, weight for adding to natural connectivity is 
 *    calculated.
 *
 * Results:
 *    Return weight of connectivity.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

static double
CalculateWeights(critNetPtr, estimatedLength)
Ncl_LinkedCritNet  *critNetPtr;  /*in: a critical net */
double  estimatedLength;     /*in: estimated wire length for the net */
{
#define WEIGHT ((double) 1.0) /*  heuristic value */ 

    double requiredMaxLength, weight, x;
   
    requiredMaxLength = critNetPtr->netPtr->maxLength - critNetPtr->used_l;
    if (estimatedLength <= requiredMaxLength) {
        weight = (0.0);
    } else {
        x = (estimatedLength-requiredMaxLength) / requiredMaxLength;
        weight = WEIGHT * x * x;
    }
#ifdef DEBUG
    if (timing_printStat) {
        printf("estimated= %e, required= %e, weight= %e \n", 
               estimatedLength, requiredMaxLength, weight);
    }
#endif DEBUG
    return (weight);

#undef WEIGHT
}



/*
 * ----------------------------------------------------------------------------
 *
 * EstimateWireLength() --
 *
 *    Estimate wire length of a specified critical net.
 *
 * Results:
 *    returns estimated wire length.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

static double
EstimateWireLength(netPtr, netEdgesPtr, nodesNumber)
Net  *netPtr;            /*in: the critical net */
GraphEdge *netEdgesPtr;  /*in: edges of current cluster graph */
int  *nodesNumber;
{
    float RhoValue ();
    
    double wireLength, activeArea, wiringArea;
    
    EstimateArea(netPtr, netEdgesPtr, nodesNumber, &activeArea, &wiringArea);

#ifdef DEBUG
    if (timing_printStat) {
        printf ("netPtr.name = %s \n", netPtr->name);
        printf ("activeArea = %e, wiringArea = %e \n", activeArea, wiringArea);
    }
#endif DEBUG

    wireLength = (double) RhoValue(*nodesNumber) * (2.0) * 
            sqrt(activeArea + wiringArea);

    return (wireLength);
}



float
RhoValue(n)
int n;
{
    float rhoValue;

    if (n > 20) {
        rhoValue = (float) ((sqrt((double) n) + 1.0) / 2.0);
    } else {
        rhoValue = rho[n];
    }

    return (rhoValue);
}



/*
 * ----------------------------------------------------------------------------
 *
 * EstimateArea() --
 *
 *    Calculate both active area and wiring area of cluster nodes strongly
 *    connected to a specified critical net for estimating wire length of
 *    the net. Also find the number of nodes directly connected with the 
 *    net.
 *
 * Results:
 *    activeArea, wiringArea, nodesnumber.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

static void
EstimateArea(netPtr, netEdgesPtr, nodesNumber, activeArea, wiringArea)
Net  *netPtr;            /*in: the critical net */
GraphEdge *netEdgesPtr;  /*in: edges of current cluster graph */
int  *nodesNumber;       /*out: number of directly connected nodes */
double  *activeArea;     /*out: active area of strongly connected nodes */
double  *wiringArea;     /*out: wiring area of strongly connected nodes */
{
#define SpaceX ((float) 1.0)
#define SpaceY ((float) 1.0)

    ConnectedNode           *nodePtr, *newNodePtr; 
    ConnectedNode           *setOfNodesPtr = NIL(ConnectedNode); 
    GraphEdge               *edgePtr; 
    GraphNode               *node;
    Ncl_LinkedCritNetOnEdge *critOnEdgePtr;
    double thresholdConnect = (214000.0); /* bignumber = max value */
    double                  connect, totalConnectivity = 0.0;
    int                     flag1, flag2, count, numberOfNeighborNodes;
    int                     sumOfWidth = 0, sumOfHeight = 0;
 
    *nodesNumber = 0;
    *activeArea = *wiringArea = (0.0);
    /* Find a set of nodes directly connected to edges with the critical net. */
    for (edgePtr = netEdgesPtr; edgePtr != NIL(GraphEdge); NEXT(edgePtr)) {
        for (critOnEdgePtr = edgePtr->critOnEdgePtr;
                critOnEdgePtr != NIL(Ncl_LinkedCritNetOnEdge); 
                NEXT(critOnEdgePtr)) {
            if (critOnEdgePtr->critNetPtr->netPtr == netPtr) {
                totalConnectivity += edgePtr->connect1;
                flag1 = flag2 = 0;
                for (nodePtr = setOfNodesPtr; 
                     nodePtr != NIL(ConnectedNode); NEXT(nodePtr)) {
                    if (nodePtr->nodePtr == edgePtr->node1Ptr) flag1 = 1;
                    if (nodePtr->nodePtr == edgePtr->node2Ptr) flag2 = 1;
                }
                /* Append the new nodes to setOfNodesPtr. */    
                if (flag1 == 0 && edgePtr->node1Ptr != NIL(GraphNode)) {
                    GetStorMc(newNodePtr, 1, ConnectedNode);
                    newNodePtr->nextPtr = setOfNodesPtr;
                    setOfNodesPtr = newNodePtr;
                    newNodePtr->nodePtr = edgePtr->node1Ptr;
                    (*nodesNumber)++; 
                }
                if (flag2 == 0 && edgePtr->node2Ptr != NIL(GraphNode)) {
                    GetStorMc(newNodePtr, 1, ConnectedNode);
                    newNodePtr->nextPtr = setOfNodesPtr;
                    setOfNodesPtr = newNodePtr;
                    newNodePtr->nodePtr = edgePtr->node2Ptr;
                    (*nodesNumber)++; 
                }
                /* Calculate the max. connectivity, thresholdConnect. */
                if (edgePtr->connect1 < thresholdConnect ) 
                     thresholdConnect = edgePtr->connect1;
            }
        }
    }

    /* Traverse all edges directly connected to the set of nodes.
     * And find a set of nodes strongly connected to the edges 
     * whose connectivities are greater than thresholdConnect.
     */
#ifdef DEBUG
    if (timing_printStat) {
        printf ("nodesNumber (direct) = %d \n", *nodesNumber);
    }
#endif DEBUG
    nodePtr = setOfNodesPtr;
    numberOfNeighborNodes = *nodesNumber;
    for (count = 0; count < *nodesNumber; count++) {
        for (edgePtr = nodePtr->nodePtr->edgeList->edgePtr; 
             edgePtr != NIL(GraphEdge); NEXT(edgePtr)) {
            if (edgePtr->connect1 > thresholdConnect) {
                if ((node = edgePtr->node1Ptr) != nodePtr->nodePtr) { 
                    node = edgePtr->node2Ptr;
                }
                if (node != NIL(GraphNode)) {
                    totalConnectivity += edgePtr->connect1;
                }
                flag1 = 0;
                for (newNodePtr = setOfNodesPtr; 
                        newNodePtr != NIL(ConnectedNode); 
                        NEXT(newNodePtr)) {
                    if (newNodePtr->nodePtr == node) {
                        flag1 = 1;
                    }
                }
                if (flag1 == 0 && node != NIL(GraphNode)) {
                    GetStorMc(newNodePtr, 1, ConnectedNode);
                    newNodePtr->nextPtr = setOfNodesPtr; 
                    setOfNodesPtr = newNodePtr;
                    newNodePtr->nodePtr = node;
                    numberOfNeighborNodes++; 
                }    
            }
        }
        NEXT(nodePtr);
    }

#ifdef DEBUG
    if (timing_printStat) {
        printf ("numberOfNeighborNodes= %d \n", numberOfNeighborNodes);
    }
#endif DEBUG
    for (nodePtr = setOfNodesPtr; nodePtr != NIL(ConnectedNode); 
            NEXT(nodePtr)) {
        *activeArea += nodePtr->nodePtr->gn_area;
        sumOfWidth += nodePtr->nodePtr->gn_dim.p_x;  
        sumOfHeight += nodePtr->nodePtr->gn_dim.p_y;
    }
    if (setOfNodesPtr != NIL(ConnectedNode)) {
        *wiringArea = (SpaceY*(double)sumOfWidth+SpaceX*(double)sumOfHeight)
                      * (sqrt((double)numberOfNeighborNodes)/(3.0))
                      * totalConnectivity/(double)numberOfNeighborNodes;
#ifdef DEBUG
    if (timing_printStat) {
        printf ("sumOfWidth = %d, sumOfHeight = %d, Connect = %e \n", 
                 sumOfWidth, sumOfHeight, totalConnectivity);
    }
#endif DEBUG
        FREE((GraphNode *) setOfNodesPtr);
    } 

#undef SpaceY 
#undef SpaceX 
}


/*
 * ------------------------------------------------------------------------------
 *
 * NclRemoveTimingWeights --
 *
 *    Remove weights assigned by NclAssignTimingWeights from the connectivity.
 *    Restore it to natural connectivity before weighting. Moreover, all
 *    weights are initialized to -1.0 in the critical net list.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Connectivity each edge has is restored to the state before weighting.
 *
 * ------------------------------------------------------------------------------
 */

void
NclRemoveTimingWeights (netEdgesPtr, critNetList, maxConnect)

    GraphEdge *netEdgesPtr;        /*in: edges of current cluster graph */
    Ncl_LinkedCritNet *critNetList; /*in: critical net list */
    double  *maxConnect;           /*in: max. of connectivity in previous level */

{
    Ncl_LinkedCritNetOnEdge  *critOnEdgePtr;
    Ncl_LinkedCritNet        *lnkCritNetPtr;
    GraphEdge               *edgePtr; 
    int                     flag;
   
    for (edgePtr = netEdgesPtr; edgePtr != NIL(GraphEdge); NEXT(edgePtr)) {
        flag = 0;
        for (critOnEdgePtr = edgePtr->critOnEdgePtr; 
                critOnEdgePtr != NIL(Ncl_LinkedCritNetOnEdge); 
                NEXT(critOnEdgePtr)) {
            if (critOnEdgePtr->critNetPtr->timeWeight > (0.0)) {
                flag = 1;
                edgePtr->connect1 -= critOnEdgePtr->critNetPtr->timeWeight;
            }
        }
       if (flag == 1) {
           edgePtr->connect1 -= *maxConnect;
       }
    }
    for (lnkCritNetPtr = critNetList;
            lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
            NEXT(lnkCritNetPtr)) {
        lnkCritNetPtr->timeWeight = (-1.0);
        lnkCritNetPtr->used_l = (0.0);
    }

}



void
NclSetWeight2ForCritNet(clusterNodes)
GraphNode *clusterNodes;  /* in: */
{
    Ncl_LinkedCritNetOnEdge  *onEdgePtr;
    Ncl_LinkedCritNet        *lnkCritNetPtr;
    GraphNode               *gNodePtr, *gChildPtr, *anotherGNodePtr;
    ConnEdge                *connPtr;
    int                     i;

    for (gNodePtr = clusterNodes; gNodePtr != NIL(GraphNode); 
            NEXT(gNodePtr)) {

        for (lnkCritNetPtr = ncl_CritNetList;
                lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                NEXT(lnkCritNetPtr)) {
            lnkCritNetPtr->numPseudoPins = 0;
        }

        for (i = 0; i < gNodePtr->gn_numOfChildren; i++) {
            gChildPtr = gNodePtr->gn_children[i];

            for (connPtr = gChildPtr->edgeList; 
                connPtr != NIL(ConnEdge); NEXT(connPtr)) {

                if (connPtr->edgePtr->edgeType == BLOCK) {

                    anotherGNodePtr = 
                            connPtr->edgePtr->node1Ptr == gChildPtr ?
                            connPtr->edgePtr->node2Ptr : 
                            connPtr->edgePtr->node1Ptr;

                    if (anotherGNodePtr->belongsToPtr == gNodePtr) {

                        for (onEdgePtr = connPtr->edgePtr->critOnEdgePtr;
                                onEdgePtr != NIL(Ncl_LinkedCritNetOnEdge); 
                                NEXT(onEdgePtr)) {
                            (onEdgePtr->critNetPtr->numPseudoPins)++;
                        }

                    }
                }

            }
        }

        for (lnkCritNetPtr = ncl_CritNetList;
                lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                NEXT(lnkCritNetPtr)) {

            switch (lnkCritNetPtr->numPseudoPins) {
            case 0:
                break;
            case 2:
                lnkCritNetPtr->numPseudoPins = 2;
                break;
            case 6:
                lnkCritNetPtr->numPseudoPins = 3;
                break;
            case 12:
                lnkCritNetPtr->numPseudoPins = 4;
                break;
            case 40:
                lnkCritNetPtr->numPseudoPins = 5;
                break;
            default:
                break;
            }
        }
        for (i = 0; i < gNodePtr->gn_numOfChildren; i++) {
            gChildPtr = gNodePtr->gn_children[i];

            for (connPtr = gChildPtr->edgeList; 
                    connPtr != NIL(ConnEdge); NEXT(connPtr)) {
                if (connPtr->edgePtr->edgeType == BLOCK) {

                    anotherGNodePtr = 
                            connPtr->edgePtr->node1Ptr == gChildPtr ?
                            connPtr->edgePtr->node2Ptr : 
                            connPtr->edgePtr->node1Ptr;

                    if (anotherGNodePtr->belongsToPtr == gNodePtr) {

                        for (onEdgePtr = connPtr->edgePtr->critOnEdgePtr;
                                onEdgePtr != NIL(Ncl_LinkedCritNetOnEdge); 
                                NEXT(onEdgePtr)) {
                            onEdgePtr->weight2 = TIMES100(2.0 / 
                                    onEdgePtr->critNetPtr->numPseudoPins);
                        }

                    }
                }
            }
        }
    }
}
        


/*
 * Applicable to timing driven placement / floorplanning only.
 */
void
NclMakeConnectivityForCritNet(clusterNodes)
GraphNode *clusterNodes;
{
    GraphNode               *nodePtr, *gNode;
    Ncl_Node                *tNode;
    Ncl_LinkedCritNetInNode *critInNodePtr, *otherInNodePtr;
    Ncl_Edge                *netPtr, *edgePtr;
    Ncl_LinkedCritNetOnEdge *critOnEdgePtr;
    Ncl_LinkedEdge          *edgeForCritNetPtr, *newLinkedEdgePtr;
    int                     i, j, k, matchFlag, sumConnectivity;
    int                     interConnectivity, intraConnectivity;
    
    for (nodePtr = clusterNodes; nodePtr != NIL(GraphNode); NEXT(nodePtr)) {
        for (i = 0; i < nodePtr->gn_numOfChildren; i++) {
            gNode = nodePtr->gn_children[i];
            tNode = nodePtr->gn_members[i];
            for (critInNodePtr = tNode->nd_lnkCritNetList;
                critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                NEXT(critInNodePtr)) {

                for (j = 0; j < MAX_CHILDREN; j++) {
                    for (k = 0; k < MAX_CHILDREN+4; k++) {
                        critInNodePtr->connectivity[j][k] = 0;
                    }
                }
                critInNodePtr->numOfChildren = 0;

    /*
     * Begin indentation violation.
     */
    for (j = 0; j < tNode->nd_numOfChildren; j++) {
            critInNodePtr->edgeList[j] = NIL(Ncl_LinkedEdge);
        for (netPtr = tNode->nd_children[j]->nd_netList;
                netPtr != NIL(Ncl_Edge); NEXT(netPtr)) {
            for (critOnEdgePtr = netPtr->critOnEdgePtr;
                    critOnEdgePtr != NIL(Ncl_LinkedCritNetOnEdge); 
                    NEXT(critOnEdgePtr)) {
                if (critOnEdgePtr->critNetPtr == critInNodePtr->critNetPtr) {
                    if (netPtr->edgeType == BLOCK) {
                        if (netPtr->connectedNode->nd_parentPtr == tNode) {
                            /* both endpoints of edge in same cluster */
                            int pos;

                            ArrayPosMc(netPtr->connectedNode,
                                tNode->nd_children);
                            critInNodePtr->connectivity[j][pos] = 1;
                        }
                        else {
                            matchFlag = 0;
                            for (edgeForCritNetPtr = critInNodePtr->edgeList[j];
                                  edgeForCritNetPtr != NIL(Ncl_LinkedEdge);
                                  NEXT(edgeForCritNetPtr)) {
                                if (edgeForCritNetPtr->edgePtr->connectedNode ==
                                      netPtr->connectedNode->nd_parentPtr) {
                                    matchFlag = 1;
                                    break;
                                }
                            }
                            if (matchFlag == 0) {
                                GetStorMc(newLinkedEdgePtr, 
                                          1, Ncl_LinkedEdge);
                                newLinkedEdgePtr->nextPtr = critInNodePtr->edgeList[j];
                                critInNodePtr->edgeList[j] = newLinkedEdgePtr;
                                GetStorMc(edgePtr, 1, Ncl_Edge);
                                newLinkedEdgePtr->edgePtr = edgePtr;
                                edgePtr->connectedNode = netPtr->connectedNode->nd_parentPtr;
                                edgePtr->critOnEdgePtr = NIL(Ncl_LinkedCritNetOnEdge);
                                edgePtr->edgeType = BLOCK;
                                edgePtr->nextPtr = NIL(Ncl_Edge);
                            }
                        }
                    }
                }
            }
        }
        /* 
         * deleted a check here.
         */
        for (otherInNodePtr = gNode->gn_members[j]->nd_lnkCritNetList;
             otherInNodePtr != NIL(Ncl_LinkedCritNetInNode);
             NEXT(otherInNodePtr)) {
            if (otherInNodePtr->critNetPtr == critInNodePtr->critNetPtr) {
                critInNodePtr->children[(critInNodePtr->numOfChildren)++] =
                    otherInNodePtr;
            }
        }
    }
    /*
     * End indentation violation.
     */

                sumConnectivity = 0;
                for (j = 0; j < tNode->nd_numOfChildren; j++) {
                    for (k = j+1; k < tNode->nd_numOfChildren; k++) {
                        sumConnectivity += critInNodePtr->connectivity[j][k];
                    }
                }

                switch (sumConnectivity) {
                case 1:
                    intraConnectivity = TIMES100(1.0);  /* = 2/2 */
                    interConnectivity = TIMES100(0.5);  /* = 1/2 */
                    break;
                case 3:
                    intraConnectivity = TIMES100(2.0/3.0);
                    interConnectivity = TIMES100(1.0/3.0);
                    break;
                case 6:
                    intraConnectivity = TIMES100(0.5);  /* = 2/4 */
                    interConnectivity = TIMES100(0.25); /* = 1/4 */
                    break;
                case 20:
                    intraConnectivity = TIMES100(0.4);  /* = 2/5 */
                    interConnectivity = TIMES100(0.2);  /* = 1/5 */
                    break;
                case 0:
                    intraConnectivity = 0;
                    interConnectivity = TIMES100(1.0);  /* = 1/1 */
                    break;
                }
                for (j = 0; j < tNode->nd_numOfChildren; j++) {
                    for (k = 0; k < tNode->nd_numOfChildren; k++) {
                        if (critInNodePtr->connectivity[j][k] > 0) {
                            critInNodePtr->connectivity[j][k] = intraConnectivity;
                        }
                    }
                    for (edgeForCritNetPtr = critInNodePtr->edgeList[j];
                            edgeForCritNetPtr != NIL(Ncl_LinkedEdge);
                            NEXT(edgeForCritNetPtr)) {
                        edgeForCritNetPtr->edgePtr->weight = interConnectivity;
                    }
                }
            }
        }
    }
}



/*
 * Applicable to timing driven placement / floorplanning only.
 */
void
NclSetUsedLengthsForCritNets(gNodes)
GraphNode   *gNodes;
{
    GraphNode               *gNodePtr;
    Ncl_LinkedCritNetInNode *critInNodePtr;
    double                  usedLength;
    int                     lengthNr;

    for (gNodePtr = gNodes; gNodePtr != NIL(GraphNode); NEXT(gNodePtr)) {
        for (critInNodePtr = gNodePtr->gn_lnkCritNetList;
            critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
            NEXT(critInNodePtr)) {

            usedLength = NCL_REAL_MAX;     
            if (critInNodePtr->ld.size != 0) {
                for (lengthNr = 0; 
                    lengthNr < critInNodePtr->ld.size; 
                    lengthNr++) {

                    if (critInNodePtr->ld.length[lengthNr].l < usedLength) {
                        usedLength = critInNodePtr->ld.length[lengthNr].l;
                    }
                }
            } 
            else {
                usedLength = 0.0;
            }
            critInNodePtr->critNetPtr->used_l += usedLength;
        }
    }
}


