
/* nplEnum.c --
 *
 *     Massoud Pedram
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1989 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplEnum.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newPInt.h"
#include "nplArrays.h"

typedef struct _FloorplanInfo FloorplanInfo;
struct _FloorplanInfo {
    int numUpdates, pattern, *label, orient, shapeNr, leafO[MAX_CHILDREN];
    float nonLeafCost, pLeafCost;
};

#ifdef DEBUG
static bool print_critNets = FALSE;
static bool print_tp = FALSE;
static bool mytest = FALSE;
#endif DEBUG


/* ---------- Local Procedures ---------- */
static BearStatus   OptOrientOfChildren();

#ifdef DEBUG
static void         PrintTPInfo();
#endif DEBUG



/* 
 * -------------------------------------------------------------------------
 * General Information:
 *
 *      if nd_type != LEAF, then nd_orient = 0, 1, 2, 3 corresponding to
 *          GeoIdentity, Geo90, Geo180, and Geo270.
 *      if nd_type == LEAF, then nd_orient = 0, 1, ..., 7 corresponding to
 *          GeoIdentity, GeoRef45, and 4 possible reflections.
 *
 *      The leaf nodes are at level 0 and root is at the highest level. 
 *      The root node is at depth 0 and the leaf nodes at highest depth.
 *
 * -------------------------------------------------------------------------
 */

static BearStatus   EnumNonLeaf();
static BearStatus   GreedyOpt();
static BearStatus   EnumPseudoLeaf();

static float        areaWeight, wireWeight, critWeight;
static float        areaTol, aspectTol, sensTol;
static float        wireLowerBnd;



static BearStatus
EnumNonLeaf(nodePtr, shapeNr, nonLeafCost, tpPtr)
Ncl_Node      *nodePtr;
int           shapeNr;
float         nonLeafCost;
FloorplanInfo *tpPtr;
{
    if (nodePtr->nd_type != NON_LEAF) {
        Bear_PrintErr("EnumNonLeaf failed: nodePtr must be a non-leaf.");
        return (BEAR_ERROR);
    }

    if (nonLeafCost * sensTol < tpPtr->nonLeafCost) {
        (tpPtr->numUpdates)++;
        tpPtr->pattern = nodePtr->nd_pattern;
        tpPtr->label = nodePtr->nd_label;
        tpPtr->orient = nodePtr->nd_orient;
        tpPtr->shapeNr = shapeNr;
        tpPtr->nonLeafCost = nonLeafCost;

        if (ncl_TimingFlag == TRUE) {
            Ncl_LinkedCritNetInNode *critInNodePtr;

            for (critInNodePtr = nodePtr->nd_lnkCritNetList;
                    critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                    NPLNEXT(critInNodePtr)) {
                critInNodePtr->best_inter_l = critInNodePtr->inter_l;
            }
        }
    }

    return (BEAR_OK);
}



static BearStatus
EnumPseudoLeaf(nodePtr, shapeNr, partialCostAppr, partialCostExac, tpPtr)
Ncl_Node      *nodePtr;
int           shapeNr;
float         partialCostAppr, partialCostExac;
FloorplanInfo *tpPtr;
{
    Ncl_Node    *childNodePtr;
    float       neighbCost = 0, approxCost, exactCost;
    int         i, bestLeafO[MAX_CHILDREN], 
                numOfChildren = nodePtr->nd_numOfChildren;

    if (nodePtr->nd_type != P_LEAF) {
        Bear_PrintErr("EnumPseudoLeaf failed: nodePtr must be a pseudo-leaf.");
        return (BEAR_ERROR);
    }

    /* 
     * We initialize cells' nd_orient to no orientation. 
     */
    for (i = 0; i < numOfChildren; i++) {
        childNodePtr = nodePtr->nd_children[i];
        childNodePtr->nd_orient = childNodePtr->nd_nodeTrans.t_b*4;
    }

    if (ncl_TimingFlag == TRUE) {
        if (GreedyOpt(nodePtr, bestLeafO, &neighbCost) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
        approxCost = partialCostExac + neighbCost;
    } else {
        approxCost = partialCostAppr;
    }

    if (approxCost < tpPtr->nonLeafCost) {

        /*
         * We are using the nd_orient fields for the children that were
         * calculated during timing-cost minimization.  
         * if ncl_TimingFlag == FALSE, nd_orient are all zero due to prior
         * initialization.
         */
        if (GreedyOpt(nodePtr, bestLeafO, &neighbCost) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }

        exactCost = partialCostExac + neighbCost;

        if (exactCost < tpPtr->pLeafCost) {

            (tpPtr->numUpdates)++;
            tpPtr->pattern = nodePtr->nd_pattern;
            tpPtr->label = nodePtr->nd_label;
            tpPtr->orient = nodePtr->nd_orient;
            tpPtr->shapeNr = shapeNr;
            tpPtr->nonLeafCost = approxCost;
            tpPtr->pLeafCost = exactCost;

            for (i = 0; i < numOfChildren; i++) {
                tpPtr->leafO[i] = bestLeafO[i] + 
                    nodePtr->nd_children[i]->nd_nodeTrans.t_b*4;
            }

            /*
             * Final Optimization will overwrite best_inter_l .
             */
            if (ncl_TimingFlag == TRUE) {
                Ncl_LinkedCritNetInNode *critInNodePtr;

                for (critInNodePtr = nodePtr->nd_lnkCritNetList;
                        critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                        NPLNEXT(critInNodePtr)) {
                    critInNodePtr->best_inter_l = critInNodePtr->inter_l;
                }
            }

        }
    }

    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * NplEnumTPs() --
 *
 *      Chooses the best floorplanning solution for the nodePtr.
 *      Note that nodePtr->nd_status == PLACED means that nd_dim, 
 *      nd_nodeTrans, nd_tileTrans are correctly set.  This is a 
 *      pre-condition for any caller to this proc.
 *
 * Results:
 *      BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *      Assigns nd_pattern, nd_label, nd_orient to the nodePtr.
 *
 * ---------------------------------------------------------------------------- 
 */

BearStatus
NplEnumTPs(nodePtr)
Ncl_Node *nodePtr;
{
    register Ncl_ShapeData   *shapePtr;
    FloorplanInfo       bestTP;
    Ncl_ShapeSet        *nShapeSet, *iShapeSet, shapeSetStruct;
    double              levelAdj;
    float               curArea, curAspect, 
                        areaCost[2], aspectCost[2], wireCost[2],
                        cnCost[2], interCost[2], completeCost[2],
                        intraCost, partialCostExac, partialCostAppr, 
                        tmpArea, tmpAspect, 
                        desiredArea, desiredAspect;

    int                 i, shapeNr, pattern, labelID, *label, orient,
                        aspectTolIsOK, areaTolIsOK, foundSoln, 
                        desiredX, desiredY;

    int                 numOfChildren = nodePtr->nd_numOfChildren;

    if (nodePtr->nd_status == NOT_PLACED) {
        Bear_PrintErr("NplEnumTPs failed: nodePtr itself is not placed.");
        return (BEAR_ERROR);
    }
    if (nodePtr->nd_type == LEAF) {
        Bear_PrintErr("NplEnumTPs failed: nodePtr is a leaf.");
        return (BEAR_ERROR);
    }
    areaWeight = nplAreaWeight;
    wireWeight = nplWireWeight;
    critWeight = nplCritWeight;

    if (nodePtr->nd_depth == 0) {
        aspectTol = (float) (nplTolerance / 3.0);
    }
    else {
        aspectTol = (float) nplTolerance; 
    }
    areaTol = aspectTol / 2.0;
    sensTol = NPLSENS_TOT;

    levelAdj = ncl_ClusterTreeDepth - nodePtr->nd_depth;

    /* initialize shapeSetStruct */
    shapeSetStruct.size = 0;
    shapeSetStruct.shape = Ncl_InitShapes(ncl_MaxNumOfShapes);

    desiredX = nodePtr->nd_dim.p_x;
    desiredY = nodePtr->nd_dim.p_y;
    desiredAspect = ((float) desiredY) / ((float) desiredX);
    desiredArea = ((float) desiredX) * ((float) desiredY);

    if (desiredAspect <= 0) {
        Bear_PrintErr("NplEnumTPs failed: node aspect ratio is bad.");
        return (BEAR_ERROR);
    }

    if (nodePtr->nd_shapeSet == NIL (Ncl_ShapeSet)) {
        Bear_PrintErr("NplEnumTPs failed: shape function is NIL.");
        return (BEAR_ERROR);
    }

    if (nodePtr->nd_lengthSet == NIL (Ncl_LengthSet)) {
        Bear_PrintErr("NplEnumTPs failed: length function is NIL.");
        return (BEAR_ERROR);
    }

    /* HACK */
    wireLowerBnd = NPLMAX_FLOAT;
    for (i=0; i < nodePtr->nd_lengthSet->size; i++) {
        if (nodePtr->nd_lengthSet->length[i].l < wireLowerBnd) {
            /* 2.0 HEURISTIC factor */
            wireLowerBnd = 2.0 * nodePtr->nd_lengthSet->length[i].l;
        }
    }

    if (wireLowerBnd > NPLMAX_FLOAT - 1 || wireLowerBnd <= 0) {
        Bear_PrintErr("NplEnumTPs failed: length function is bad.");
        return (BEAR_ERROR);
    }

    bestTP.nonLeafCost = bestTP.pLeafCost = NPLMAX_FLOAT;
    bestTP.numUpdates = 0;
    aspectTolIsOK = areaTolIsOK = FALSE;

while (TRUE) {
    for (pattern = 0; pattern < NPL_MAX_PATTERNS; pattern++) {

        if (npl_FloorplanSizeArray[pattern] != numOfChildren) {
            continue;
        }

        for (labelID=0; labelID < npl_FactArray[numOfChildren-1]; labelID++) {
            label = npl_LableTable[labelID];

            nShapeSet = (*(nplShapeFuncArray[pattern])) (nodePtr, label, 0);
            iShapeSet = (numOfChildren == 1) ?
                    (Ncl_ShapeSet *) NULL :
                    (*(nplShapeFuncArray[pattern])) (nodePtr, label, 1);

            if (Ncl_MergeShapeSets(iShapeSet, nShapeSet, &shapeSetStruct) == 
                BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            if (nShapeSet != NULL) {
                FREE((char *) nShapeSet->shape);
                FREE((char *) nShapeSet);
            }
            if (iShapeSet != NULL) {
                FREE((char *) iShapeSet->shape);
                FREE((char *) iShapeSet);
            }

            for (shapeNr=0; shapeNr < shapeSetStruct.size; shapeNr++) {
                shapePtr = &(shapeSetStruct.shape[shapeNr]);

                if (NplGetIntraBlockL(nodePtr, shapePtr, &intraCost) == 
                    BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
                intraCost /= wireLowerBnd;

                if (ncl_TimingFlag) {
                    if (NplCalcCNIntraBlockL(nodePtr) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }
                }

                for (i=0; i < 2; i++) {
                    completeCost[i] = NPLMAX_FLOAT;
                }

                foundSoln = FALSE;
                for (i = 0; i < 2; i++) {
                    shapePtr->orient += 2*i;

                    if (NplAssignShapeToNode(shapePtr, nodePtr) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                     }

                    if (Npl_GenSpace(nodePtr, levelAdj) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }

                    if (nodePtr->nd_type == P_LEAF) {
                        if (NplAddPinSpace(nodePtr) == BEAR_ERROR) {
                            return (BEAR_ERROR);
                        }
                    }

                    if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }

                    curAspect = ((float) nodePtr->nd_dim.p_y) / 
                            ((float) nodePtr->nd_dim.p_x);
                    if (curAspect < (1 - aspectTol) * desiredAspect ||
                            curAspect > (1 + aspectTol) * desiredAspect) {
                        break;
                    }
                    if (! aspectTolIsOK) aspectTolIsOK = TRUE;

                    curArea = ((float) nodePtr->nd_dim.p_x) * 
                            ((float) nodePtr->nd_dim.p_y);
                    if (curArea > (areaTol + 1.0) * desiredArea) {
                        break;
                    }
                    if (! areaTolIsOK) {
                        areaTolIsOK = TRUE;
                    }

                    NplGeoPen(desiredX, desiredY, 
                            nodePtr->nd_dim.p_x, nodePtr->nd_dim.p_y, 
                            &(areaCost[i]), &(aspectCost[i]));
                    areaCost[i] /= desiredArea;
                    aspectCost[i] /= desiredArea;

                    if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }

                    if (NplGetInterBlockL(nodePtr, &(interCost[i])) == 
                        BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }
                    interCost[i] /= wireLowerBnd;

                    wireCost[i] = intraCost + interCost[i];

                    if (ncl_TimingFlag) {
                        if (NplCalcCNInterBlockL(nodePtr) == BEAR_ERROR) {
                            return (BEAR_ERROR);
                        }
                        if (NplGetCNCost(nodePtr, &(cnCost[i])) == BEAR_ERROR) {
                            return (BEAR_ERROR);
                        }
                    } 
                    else {
                        cnCost[i] = 0;
                    }

                    completeCost[i] = areaWeight*(areaCost[i] + aspectCost[i]) +
                            wireWeight * wireCost[i] + critWeight * cnCost[i];

                    foundSoln = TRUE;
                    if (nplSymmetric[pattern]) break;
                }  /* end for each i */

                if (! foundSoln) continue;

                if (! nplSymmetric[pattern]) {

                    if (completeCost[1] < completeCost[0]) {
                        completeCost[0] = completeCost[1];
                        areaCost[0] = areaCost[1];
                        aspectCost[0] = aspectCost[1];
                        wireCost[0] = wireCost[1];
                        cnCost[0] = cnCost[1];
                    } 
                    else {
                        shapePtr->orient -= 2;
                    }

                } /* end if nplSymmetric */ 


                if (nodePtr->nd_type == NON_LEAF) {
                    nodePtr->nd_orient = shapePtr->orient;
                    if (EnumNonLeaf(nodePtr, shapeNr, completeCost[0], 
                        &bestTP) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }
                } 
                else {
                    partialCostAppr = areaWeight*(areaCost[0] + aspectCost[0]) +
                            wireWeight*wireCost[0];
                    partialCostExac = areaWeight*(areaCost[0] + aspectCost[0]) +
                            wireWeight*intraCost;

                    if (NplAssignShapeToNode(shapePtr, nodePtr) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }

                    if (Npl_GenSpace(nodePtr, levelAdj) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }

                    if (nodePtr->nd_type == P_LEAF) {
                        if (NplAddPinSpace(nodePtr) == BEAR_ERROR) {
                            return (BEAR_ERROR);
                        }
                    }

                    if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }

                    if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }

                    if (EnumPseudoLeaf(nodePtr, shapeNr, partialCostAppr, 
                           partialCostExac, &bestTP) == BEAR_ERROR) {
                        return (BEAR_ERROR);
                    }

                }

#ifdef DEBUG
                if (print_tp == TRUE) {
                    PrintTPInfo(nodePtr, areaCost[0], aspectCost[0], 
                            wireCost[0], cnCost[0]);
                }
#endif DEBUG

            } /* end for each shape */

            /* FREE((char *) shapeSetStruct.shape); */

        } /* end for each label */

    } /* end for each pattern */

    if (bestTP.numUpdates > 0) {
        break;
    } 
    else {
        if (aspectTolIsOK == TRUE) {
            if (areaTol > 256.0) {
                Bear_PrintErr(
                "NplEnumTPs() failed: possibility of an infinite loop.");
                return (BEAR_ERROR);
            }

            areaTol *= 2.0;
            Bear_Print("NplEnumTPs(): Area tolerance low. Fix-up and go.");
            fflush(stdout);
        }
        else {
            if (aspectTol > 256.0) {
                Bear_PrintErr(
                "NplEnumTPs() failed: possibility of an infinite loop.");
                return (BEAR_ERROR);
            }

            aspectTol *= 2.0;
            areaTol *= 2.0;
            Bear_Print("NplEnumTPs(): Aspect tolerance low. Fix-up and go.");
            fflush(stdout);
        }
    }
}

    nShapeSet = 
        (*(nplShapeFuncArray[bestTP.pattern])) (nodePtr, bestTP.label, 0);
    iShapeSet = (numOfChildren == 1) ?
            (Ncl_ShapeSet *) NULL :
            (*(nplShapeFuncArray[bestTP.pattern])) (nodePtr, bestTP.label, 1);

    if (Ncl_MergeShapeSets(iShapeSet, nShapeSet, &shapeSetStruct) == 
        BEAR_ERROR) {
        return (BEAR_ERROR);
    }
    if (nShapeSet != NULL) {
        FREE((char *) nShapeSet->shape);
        FREE((char *) nShapeSet);
    }
    if (iShapeSet != NULL) {
        FREE((char *) iShapeSet->shape);
        FREE((char *) iShapeSet);
    }
    shapePtr = &(shapeSetStruct.shape[bestTP.shapeNr]);
    if (bestTP.pattern != shapePtr->pattern || 
            bestTP.orient%2 != shapePtr->orient) {
        Bear_PrintErr("NplEnumTPs(): shape is bad.");
        return (BEAR_ERROR);
    }

    if (nodePtr->nd_type == P_LEAF) {
        for (i = 0; i < numOfChildren; i++) {
            nodePtr->nd_children[i]->nd_orient = bestTP.leafO[i];
        }
    }

    shapePtr->orient = bestTP.orient;
    if (NplAssignShapeToNode(shapePtr, nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (Npl_GenSpace(nodePtr, levelAdj) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (nodePtr->nd_type == P_LEAF) {
        if (NplAddPinSpace(nodePtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }
    
    if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (nodePtr->nd_type == NON_LEAF) {
        if (NplRoomExtOfNode(nodePtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }

    if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    FREE((char *) shapeSetStruct.shape);    

    /*
    if (nodePtr->nd_type == NON_LEAF) {
        fprintf(stdout, "Best cost for node %d after %d updates was %.2f .\n",
        nodePtr->nd_id, bestTP.numUpdates, bestTP.nonLeafCost);
    } else {
        fprintf(stdout, "Best cost for node %d after %d updates was %.2f .\n",
        nodePtr->nd_id, bestTP.numUpdates, bestTP.pLeafCost);
    }
    fflush(stdout);
    */

    return (BEAR_OK);
}



/*
 * Does floorplanning by a two phase process.  That is, it uses the bottom-up
 * shape functions for area minimization and then minimizes the wire length
 * by swapping cells across the cuts subject to no or little change in area
 * and aspect ratio of bottym-up shape.
 */
BearStatus
NplEnumLabels(nodePtr)
Ncl_Node *nodePtr;
{
    register Ncl_ShapeData   *shapePtr;
    Ncl_Node        *childNodePtr;
    Ncl_ShapeSet    *shapeSet;
    FloorplanInfo   bestTP;
    float           aspDiff, minAspDiff = NPLMAX_FLOAT;
    int             i, labelID, shapeNr, bestShapeNr, bestLeafO[MAX_CHILDREN],
                    numOfChildren = nodePtr->nd_numOfChildren;
    double          levelAdj;
    float           curArea, curAspect, wireCost[2],
                    cnCost[2], interCost[2], completeCost[2],
                    intraCost, partialCostExac, partialCostAppr, 
                    desiredArea, desiredAspect;
    int             aspectTolIsOK, areaTolIsOK, foundSoln, orientSav,
                    desiredX, desiredY;

    if (nodePtr->nd_status == NOT_PLACED) {
        Bear_PrintErr("NplEnumLabels() failed : the given node is not placed.");
        return (BEAR_ERROR);
    }

    desiredAspect = (float) nodePtr->nd_dim.p_y / (float) nodePtr->nd_dim.p_x;
    if (desiredAspect <= 0) {
        Bear_PrintErr("NplEnumLabels failed: node aspect ratio is bad.");
        return (BEAR_ERROR);
    }

    shapeSet = nodePtr->nd_shapeSet;
    if (shapeSet == NIL(Ncl_ShapeSet)) {
        Bear_PrintErr("NplEnumLabels() failed: Bottom-up shapes do not exist.");
        return (BEAR_ERROR);
    }

    wireWeight = nplWireWeight;
    critWeight = nplCritWeight;

    if (nodePtr->nd_depth == 0) {
        aspectTol = (float) (nplTolerance / 3.0);
    }
    else {
        aspectTol = (float) nplTolerance;
    }
    areaTol = aspectTol * 2.0;
    sensTol = NPLSENS_TOT;

    levelAdj = ncl_ClusterTreeDepth - nodePtr->nd_depth;

    /* minAspDiff = 0.0 for all non-root nodes */
    for (shapeNr=0; shapeNr < shapeSet->size; shapeNr++) {
        shapePtr = &(shapeSet->shape[shapeNr]);

        curAspect = shapePtr->sd_y / shapePtr->sd_x;

        aspDiff = ABS(desiredAspect - curAspect);
        if (aspDiff < minAspDiff) {
            minAspDiff = aspDiff;
            bestShapeNr = shapeNr;
        }
    }

    shapePtr = &(shapeSet->shape[bestShapeNr]);
    desiredX = shapePtr->sd_x;
    desiredY = shapePtr->sd_y;
    desiredAspect = ((float) desiredY) / ((float) desiredX);
    desiredArea = ((float) desiredX) * ((float) desiredY);

    bestTP.nonLeafCost = bestTP.pLeafCost = NPLMAX_FLOAT;
    bestTP.numUpdates = 0;

    /* HACK */
    wireLowerBnd = NPLMAX_FLOAT;
    for (i=0; i < nodePtr->nd_lengthSet->size; i++) {
        if (nodePtr->nd_lengthSet->length[i].l < wireLowerBnd) {
            /* 2.0 HEURISTIC factor */
            wireLowerBnd = 2.0 * nodePtr->nd_lengthSet->length[i].l;
        }
    }

    if (wireLowerBnd > NPLMAX_FLOAT - 1 || wireLowerBnd <= 0) {
        Bear_PrintErr("NplEnumLabels failed: length function is bad.");
        return (BEAR_ERROR);
    }

    aspectTolIsOK = areaTolIsOK = FALSE;

    orientSav = shapePtr->orient;

    if (NplGetIntraBlockL(nodePtr, shapePtr, &intraCost) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }
    intraCost /= wireLowerBnd;

    if (ncl_TimingFlag) {
        if (NplCalcCNIntraBlockL(nodePtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }


while (TRUE) {
    for (labelID=0; labelID < npl_FactArray[numOfChildren-1]; labelID++) {

        shapePtr->label = npl_LableTable[labelID];

        for (i=0; i < 2; i++) {
            completeCost[i] = NPLMAX_FLOAT;
        }

        foundSoln = FALSE;
        for (i=0; i < 2; i++) {
            shapePtr->orient = orientSav + 2*i;

            if (NplAssignShapeToNode(shapePtr, nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
             }

            if (Npl_GenSpace(nodePtr, levelAdj) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (nodePtr->nd_type == P_LEAF) {
                if (NplAddPinSpace(nodePtr) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
            }

            if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            curAspect = ((float) nodePtr->nd_dim.p_y) / 
                    ((float) nodePtr->nd_dim.p_x);
            if (curAspect < (1 - aspectTol) * desiredAspect ||
                    curAspect > (1 + aspectTol) * desiredAspect) {
                break;
            }
            if (! aspectTolIsOK) aspectTolIsOK = TRUE;

            curArea = ((float) nodePtr->nd_dim.p_x) * 
                    ((float) nodePtr->nd_dim.p_y);
            if (curArea > (1 + areaTol) * desiredArea) {
                break;
            }
            if (! areaTolIsOK) areaTolIsOK = TRUE;

            if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (NplGetInterBlockL(nodePtr, &(interCost[i])) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            interCost[i] /= wireLowerBnd;

            wireCost[i] = intraCost + interCost[i];

            if (ncl_TimingFlag) {
                if (NplCalcCNInterBlockL(nodePtr) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
                if (NplGetCNCost(nodePtr, &(cnCost[i])) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
            } 
            else {
                cnCost[i] = 0;
            }

            completeCost[i] = wireWeight * wireCost[i] + 
                    critWeight * cnCost[i];

            foundSoln = TRUE;
            if (nplSymmetric[shapePtr->pattern]) break;
        }  /* end for each i */

        if (! foundSoln) continue;

        if (! nplSymmetric[shapePtr->pattern]) {
            if (completeCost[1] < completeCost[0]) {
                completeCost[0] = completeCost[1];
                wireCost[0] = wireCost[1];
                cnCost[0] = cnCost[1];
            } 
            else {
                shapePtr->orient -= 2;
            }
        } /* end if nplSymmetric */ 

        if (nodePtr->nd_type == NON_LEAF) {
            nodePtr->nd_orient = shapePtr->orient;
            if (EnumNonLeaf(nodePtr, bestShapeNr, completeCost[0], 
                &bestTP) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
        } 
        else {

            partialCostAppr = wireWeight*wireCost[0];
            partialCostExac = wireWeight*intraCost;

            if (NplAssignShapeToNode(shapePtr, nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (Npl_GenSpace(nodePtr, levelAdj) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (nodePtr->nd_type == P_LEAF) {
                if (NplAddPinSpace(nodePtr) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
            }

            if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (EnumPseudoLeaf(nodePtr, bestShapeNr, partialCostAppr, 
                   partialCostExac, &bestTP) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

        }

#ifdef DEBUG
        if (print_tp == TRUE) {
            PrintTPInfo(nodePtr, 0.0, 0.0, wireCost[0], cnCost[0]);
        }
#endif DEBUG

    } /* end for each label */

    if (bestTP.numUpdates > 0) {
        break;
    } 
    else {
        if (aspectTolIsOK == TRUE) {
            if (areaTol > 256.0) {
                Bear_PrintErr(
                    "NplEnumLabels() failed: possibility of an infinite loop.");
                return (BEAR_ERROR);
            }

            areaTol *= 2.0;
            Bear_Print("NplEnumLabels(): Area tolerance low. Fix-up and go.");
            fflush(stdout);
        }
        else {
            if (aspectTol > 256.0) {
                Bear_PrintErr(
                    "NplEnumLabels() failed: possibility of an infinite loop.");
                return (BEAR_ERROR);
            }

            aspectTol *= 2.0;
            areaTol *= 2.0;
            Bear_Print("NplEnumLabels(): Aspect tolerance low. Fix-up and go.");
            fflush(stdout);
        }
    }
}

    if (bestTP.pattern != shapePtr->pattern || 
            bestTP.orient%2 != orientSav) {
        Bear_PrintErr("NplEnumLabels(): shape is bad.");
        return (BEAR_ERROR);
    }

    if (nodePtr->nd_type == P_LEAF) {
        for (i = 0; i < numOfChildren; i++) {
            nodePtr->nd_children[i]->nd_orient = bestTP.leafO[i];
        }
    }

    shapePtr->label = bestTP.label;
    shapePtr->orient = bestTP.orient;
    if (NplAssignShapeToNode(shapePtr, nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (Npl_GenSpace(nodePtr, levelAdj) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (nodePtr->nd_type == P_LEAF) {
        if (NplAddPinSpace(nodePtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }
    
    if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    /*
    if (nodePtr->nd_type == NON_LEAF) {
        fprintf(stdout, "Best cost for node %d after %d updates was %.2f .\n",
        nodePtr->nd_id, bestTP.numUpdates, bestTP.nonLeafCost);
    } else {
        fprintf(stdout, "Best cost for node %d after %d updates was %.2f .\n",
        nodePtr->nd_id, bestTP.numUpdates, bestTP.pLeafCost);
    }
    fflush(stdout);
    */

    return (BEAR_OK);
}



/*
 * Uses bottom-up shapes; but does not minimize wire length during
 * the top-down shape propagation phase.
 */
BearStatus
NplUsePartialEnumTPs(nodePtr)
Ncl_Node *nodePtr;
{
    Ncl_Node        *childNodePtr;
    Ncl_ShapeSet    *shapeSet;
    Ncl_ShapeData   *shapePtr;
    float           neighbCost, curAspect, aspDiff;
    float           minAspDiff = NPLMAX_FLOAT, desiredAspect;
    int             i, shapeNr, bestShapeNr, bestLeafO[MAX_CHILDREN];
    double          levelAdj;

    if (nodePtr->nd_status == NOT_PLACED) {
    Bear_PrintErr("NplUsePartialEnumTPs() failed : the given node is not placed.");
        return (BEAR_ERROR);
    }

    /* HACK: 4.0 HEURISTIC factor */
    wireLowerBnd = 4.0 * (nodePtr->nd_dim.p_x + nodePtr->nd_dim.p_y);

    desiredAspect = (float) nodePtr->nd_dim.p_y / (float) nodePtr->nd_dim.p_x;
    if (desiredAspect <= 0) {
        Bear_PrintErr("NplUsePartialEnumTPs failed: node aspect ratio is bad.");
        return (BEAR_ERROR);
    }

    shapeSet = nodePtr->nd_shapeSet;
    if (shapeSet == NIL(Ncl_ShapeSet)) {
        Bear_PrintErr("NplUsePartialEnumTPs() failed: Bottom-up shapes do not exist.");
        return (BEAR_ERROR);
    }

    levelAdj = ncl_ClusterTreeDepth - nodePtr->nd_depth;

    for (shapeNr=0; shapeNr < shapeSet->size; shapeNr++) {
        shapePtr = &(shapeSet->shape[shapeNr]);

        curAspect = shapePtr->sd_y / shapePtr->sd_x;

        aspDiff = ABS(desiredAspect - curAspect);
        if (aspDiff < minAspDiff) {
            minAspDiff = aspDiff;
            bestShapeNr = shapeNr;
        }
    }

    shapePtr = &(shapeSet->shape[bestShapeNr]);

    if (! nplSymmetric[shapePtr->pattern]) {
        float intraCost, interCost[2], wireCost[2], cnCost[2], completeCost[2];
        
        if (NplGetIntraBlockL(nodePtr, shapePtr, &intraCost) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }

        if (ncl_TimingFlag) {
            if (NplCalcCNIntraBlockL(nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
        }

        for (i=0; i < 2; i++) {

            shapePtr->orient += 2*i;

            if (NplAssignShapeToNode(shapePtr, nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (Npl_GenSpace(nodePtr, levelAdj) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (nodePtr->nd_type == P_LEAF) {
                if (NplAddPinSpace(nodePtr) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
            }

            if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (NplGetInterBlockL(nodePtr, &interCost[i]) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            wireCost[i] = (intraCost + interCost[i]) / wireLowerBnd;

            if (ncl_TimingFlag) {
                if (NplCalcCNInterBlockL(nodePtr) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
                if (NplGetCNCost(nodePtr, &(cnCost[i])) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
            } else {
                cnCost[i] = 0;
            }
            completeCost[i] = wireWeight*wireCost[i] + critWeight*cnCost[i];
        }

        if (completeCost[1] >= completeCost[0]) {
            shapePtr->orient -= 2;
        }

    }

    if (NplAssignShapeToNode(shapePtr, nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (Npl_GenSpace(nodePtr, levelAdj) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (nodePtr->nd_type == P_LEAF) {
        if (NplAddPinSpace(nodePtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }
    
    if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (nodePtr->nd_type == P_LEAF) {
        for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
            childNodePtr = nodePtr->nd_children[i];
            childNodePtr->nd_orient = childNodePtr->nd_nodeTrans.t_b*4;
        }

        if (GreedyOpt(nodePtr, bestLeafO, &neighbCost) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }

        for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
            childNodePtr = nodePtr->nd_children[i];
            childNodePtr->nd_orient = bestLeafO[i] +
                    childNodePtr->nd_nodeTrans.t_b*4;
        }
    }

    return (BEAR_OK);
}



/*
 * GreedyOpt() --
 *
 *  nd_dim and nd_nodeTrans fields of the child nodes must have been set 
 *  prior to the call.  This optimization is order dependent.
 */

static BearStatus
GreedyOpt(nodePtr, leafO, neighbCostPtr)
Ncl_Node    *nodePtr;           /* IN  */
int         leafO[];            /* OUT */
float       *neighbCostPtr;     /* OUT */
{
    Ncl_Node    *childNodePtr;
    float       onePassCost, oneCombinCost, interCost, wireCost, cnCost;
    int         i, j, numOfChildren = nodePtr->nd_numOfChildren, k;
    bool        allFlexCellsFlag;

    if (nodePtr->nd_type != P_LEAF) {
        Bear_PrintErr("GreedyOpt() failed: nodePtr must be a pseudo-leaf.");
        return (BEAR_ERROR);
    }

    if (wireLowerBnd == 0) {
        Bear_PrintErr("GreedyOpt() failed: wireLowerBnd is zero.");
        return (BEAR_ERROR);
    }

    /* If all child cells are flexible, need not optimize orientations. */
    allFlexCellsFlag = TRUE;
    for (i = 0; i < numOfChildren; i++) {
        childNodePtr = nodePtr->nd_children[i];
        if (childNodePtr->nd_cellInstPtr->masterPtr->shapeFlag != 
            MAC_FLEX_SHAPE) {
            allFlexCellsFlag = FALSE;
            break;
        }
    }
    if (allFlexCellsFlag == TRUE) {
        for (i = 0; i < numOfChildren; i++) {
            leafO[i] = 0;
        }
        *neighbCostPtr = 0.0;

        return (BEAR_OK);
    }

    for (i = 0; i < numOfChildren; i++) {
        if (NplCalcTileDataOfLeafNode(nodePtr->nd_children[i]) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }

    *neighbCostPtr = NPLMAX_FLOAT;
    for (i = 0; i < numOfChildren; i++) {

        onePassCost = NPLMAX_FLOAT;

        childNodePtr = nodePtr->nd_children[i];

        for (j = 0; j < 4; j++) {
        
            childNodePtr->nd_orient = j + childNodePtr->nd_nodeTrans.t_b*4;

            if (NplCalcTileDataOfLeafNode(childNodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (NplGetInterLeafL(nodePtr, &interCost) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            wireCost = interCost / wireLowerBnd;

            if (ncl_TimingFlag == TRUE) {
                if (NplCalcCNInterLeafL(nodePtr) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
                if (NplGetCNCost(nodePtr, &cnCost) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
            } else {
                cnCost = 0;
            }

            oneCombinCost = wireWeight * wireCost + critWeight * cnCost;
            /*
            (void) printf("greedy orientations:");
            for (k = 0; k < numOfChildren; k++) {
                (void) printf(" %d",nodePtr->nd_children[k]->nd_orient);
            }
            (void) printf("\n   ; interCost = %.3f, oneCombinCost = %.3f,  onePassCost = %.3f.\n", 
                    interCost, oneCombinCost, onePassCost);
            */
            if (oneCombinCost < onePassCost) {
                leafO[i] = j;
                onePassCost = oneCombinCost;
                /*
                (void) printf("bestCost is %.3f.\n", onePassCost);
                */
            }

        } /* end for each leaf orientation */

        /* 
         * We keep the best orientation found for this child. 
         */
        childNodePtr->nd_orient = leafO[i] + childNodePtr->nd_nodeTrans.t_b*4;

        if (NplCalcTileDataOfLeafNode(childNodePtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }

        if (onePassCost < *neighbCostPtr) {
            *neighbCostPtr = onePassCost;
        }

    } /* end for each child */

    if (*neighbCostPtr == NPLMAX_FLOAT) {
        Bear_PrintErr("GreedyOpt() failed: did not optimize.");
        return (BEAR_ERROR);
    }

    if (*neighbCostPtr < 0.0) {
        Bear_PrintErr("GreedyOpt() failed: wire cost is negative!");
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}


static BearStatus
OptOrientOfChildren(nodePtr, bestO, childCount)
Ncl_Node    *nodePtr;
int         bestO[];
int         childCount;
{
    static      float bestCost;
    Ncl_Node    *childNodePtr;
    Transform   origTrans;
    float       currentCost, interCost, wireCost, cnCost, aspectVar, 
                aspectTolForRot = 0.15;
    int         i, j, upperOrient, lowerOrient, 
                numOfChildren = nodePtr->nd_numOfChildren;
    bool        reflectFlag = FALSE;

    if (childCount == numOfChildren) {
        return BEAR_OK;
    }
    if (childCount == 0) {
        /* First call initialization. */
        bestCost = NPLMAX_FLOAT;
    }
    childNodePtr = nodePtr->nd_children[childCount];
    origTrans = childNodePtr->nd_nodeTrans;
    aspectVar = (float) ABS((childNodePtr->nd_dim.p_x) - 
        (childNodePtr->nd_dim.p_y)) / 
        NPLMIN(childNodePtr->nd_dim.p_x, childNodePtr->nd_dim.p_y);
    if (ncl_TimingFlag == TRUE) {
        aspectTolForRot = 0.50;
    }
    if (aspectVar <= aspectTolForRot) {
        /* Aspect ratio is close to a square, try reflectiions */
        lowerOrient = 0;
        upperOrient = 8;
        reflectFlag = TRUE;
    }
    else {
        lowerOrient = childNodePtr->nd_nodeTrans.t_b*4;
        upperOrient = lowerOrient + 4;
        reflectFlag = FALSE;
    }
    for (j = lowerOrient; j < upperOrient; j++) {
        childNodePtr->nd_orient = j;

        if (reflectFlag) {
            if (j > 3) {
                if (origTrans.t_b == 0) {
                    GeoTransTrans(&GeoRef45Transform, &origTrans,
                          &childNodePtr->nd_nodeTrans);
                }
                else {
                    childNodePtr->nd_nodeTrans = origTrans;
                }
            }
            else {
                if (origTrans.t_b == 1) {
                    GeoTransTrans(&GeoRef45Transform, &origTrans,
                          &childNodePtr->nd_nodeTrans);
                }
                else {
                    childNodePtr->nd_nodeTrans = origTrans;
                }
            }
        }
        if (NplCalcTileDataOfLeafNode(childNodePtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }

        if (OptOrientOfChildren(nodePtr, bestO, childCount+1) == BEAR_ERROR) {
            return BEAR_ERROR;
        }
        if (childCount != numOfChildren - 1) continue;

        if (NplGetInterLeafL(nodePtr, &interCost) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
        wireCost = interCost / wireLowerBnd;
            
        if (ncl_TimingFlag == TRUE) {
            if (NplCalcCNInterLeafL(nodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            if (NplGetCNCost(nodePtr, &cnCost) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
        } else {
            cnCost = 0;
        }
            
        currentCost = wireWeight * wireCost + critWeight * cnCost;
        /*
        (void) printf("orientations:");
        for (i = 0; i < numOfChildren; i++) {
            (void) printf(" %d", nodePtr->nd_children[i]->nd_orient);
        }
        (void) printf("\n    cost = %.3f,  bestCost = %.3f.\n", 
                currentCost, bestCost);
        */
        if (currentCost < bestCost) {
            bestCost = currentCost;
            for (i = 0; i < numOfChildren; i++) {
                bestO[i] = nodePtr->nd_children[i]->nd_orient;
            }
            if (ncl_TimingFlag == TRUE) {
                Ncl_LinkedCritNetInNode *critInNodePtr;

                for (critInNodePtr = nodePtr->nd_lnkCritNetList;
                        critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                        NPLNEXT(critInNodePtr)) {
                    critInNodePtr->best_inter_l = critInNodePtr->inter_l;
                }
            }
            /*
            (void) printf("bestCost is %.3f.\n", bestCost);
            */
        }
    } /* end for each leaf orientation */
    childNodePtr->nd_nodeTrans = origTrans;

    return BEAR_OK;
}



/*
 * ----------------------------------------------------------------------------
 * NplFinalOpt() --
 *      Enumerate through all of the possible leaf orientations for the
 *      pseudo-leaf to find optimal cell orientation.
 *
 * Results:
 *      BEAR_OK is returned if no errors, otherwise BEAR_ERROR
 *
 * Side effects:
 *      Leaf cell orientations may be changed.
 *
 * ---------------------------------------------------------------------------- 
 */
BearStatus
NplFinalOpt(nodePtr)
Ncl_Node    *nodePtr;   /* IN  */
{
    Ncl_Node    **childArray, *childNodePtr;
    int         i, orient[MAX_CHILDREN];
    int         numOfChildren = nodePtr->nd_numOfChildren;
    bool        allFlexCellsFlag;
    BearStatus  bearStatus = BEAR_OK;
    double      t1;

    if (nodePtr->nd_type != P_LEAF) {
        Bear_PrintErr("NplFinalOpt() failed: nodePtr must be a pseudo-leaf.");
        return (BEAR_ERROR);
    }

    /* If all child cells are flexible, need not optimize orientations. */
    allFlexCellsFlag = TRUE;
    for (i = 0; i < numOfChildren; i++) {
        childNodePtr = nodePtr->nd_children[i];
        if (childNodePtr->nd_cellInstPtr->masterPtr->shapeFlag != MAC_FLEX_SHAPE) {
            allFlexCellsFlag = FALSE;
            break;
        }
    }
    if (allFlexCellsFlag == TRUE) {
        return (BEAR_OK);
    }

    if (wireLowerBnd == 0) {
        Bear_PrintErr("NplFinalOpt() failed: wireLowerBnd is zero.");
        return (BEAR_ERROR);
    }

    t1 = NplPtime();

    nplUseMSTCalcFlag = TRUE;

    childArray = nodePtr->nd_children;

#ifdef DEBUG
    if (mytest) {
        printf("Children:");
        for (i = 0; i < numOfChildren; i++) {
            (void) printf(" %s", childArray[i]->nd_cellInstPtr->name);
        }
        (void) printf("\nInitial orientation:");
        for (i = 0; i < numOfChildren; i++) {
            (void) printf(" %d", childArray[i]->nd_orient);
        }
        (void) printf("\n");
    }
#endif DEBUG

    for (i = 0; i < numOfChildren; i++) {
        childNodePtr = nodePtr->nd_children[i];
        childNodePtr->nd_orient = childNodePtr->nd_nodeTrans.t_b*4;
        if (NplCalcTileDataOfLeafNode(childNodePtr) == BEAR_ERROR) {
            bearStatus = BEAR_ERROR;
            goto finish;
        }
    }

    if (OptOrientOfChildren(nodePtr, orient, 0) == BEAR_ERROR) {
        bearStatus = BEAR_ERROR;
        goto finish;
    }
     
    childArray = nodePtr->nd_children;

#ifdef DEBUG
    if (mytest) {
        (void) printf("Final orientation:");
    }
#endif DEBUG
    /*
     * We need not optimize orientation of flexible child nodes.
     * Current code, however, does this. MP
     */
    for (i = 0; i < numOfChildren; i++) {
        childArray[i]->nd_orient = orient[i];
        if (orient[i] > 3) {
            childArray[i]->nd_nodeTrans = GeoRef45Transform;
        }
        else {
            childArray[i]->nd_nodeTrans = GeoIdentityTransform;
        }
#ifdef DEBUG
        if (mytest) {
            (void) printf(" %d",orient[i]);
        }
#endif DEBUG

        if (NplCalcTileDataOfLeafNode(childArray[i]) == BEAR_ERROR) {
            bearStatus = BEAR_ERROR;
            goto finish;
        }
    }
#ifdef DEBUG
    if (mytest) {
        (void) printf("\n");
    }
#endif DEBUG

    if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
        bearStatus = BEAR_ERROR;
        goto finish;
    }

    if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
        bearStatus = BEAR_ERROR;
        goto finish;
    }

finish:
    nplUseMSTCalcFlag = FALSE;

    (void) fprintf (stdout,
        "CPU-time (Final Orientation Optimization) = %g sec\n", (NplPtime() - t1));
    (void) fflush(stdout);

    return (bearStatus);
}



#ifdef DEBUG
static void 
PrintTPInfo(nodePtr, areaCost, aspectCost, wireCost, cnCost)
Ncl_Node    *nodePtr;
float       areaCost, aspectCost, wireCost, cnCost;
{
    int i;

    fprintf(stdout, "node id: %d, pattern: %d, label: ",
        nodePtr->nd_id, nodePtr->nd_pattern);
    for (i=0; i < nodePtr->nd_numOfChildren; i++) {
        (void) fprintf (stdout, "%d ", nodePtr->nd_label[i]);
    }
    fprintf(stdout, "orient: %d\n", nodePtr->nd_orient);

    fprintf(stdout, 
        "    areaCost: %.3f, aspectCost: %.3f, wireCost: %.3f, cnCost: %.3f\n", 
        areaCost, aspectCost, wireCost, cnCost);

    if (ncl_TimingFlag == TRUE) {
        Ncl_LinkedCritNetInNode  *critInNodePtr;

        for (critInNodePtr = nodePtr->nd_lnkCritNetList;
                critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                NPLNEXT(critInNodePtr)) {

          fprintf(stdout,
                "    crit net: %s, inter_l: %.2f, bu_l: %.2f, used_l: %.2f\n",
                critInNodePtr->critNetPtr->netPtr->name,
                critInNodePtr->inter_l, critInNodePtr->bu_l,
                critInNodePtr->critNetPtr->used_l);
        }
    }

    fflush(stdout);
}
#endif DEBUG

