
/* nplRouteEstimate.c --
 *
 *     Bernhard Eschermann / Massoud Pedram
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplRouteEstimate.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newPInt.h"
#include "nplArrays.h"


#define POS(x) ((x) > 0 ? (x) : 0)

#define dir(char)\
(((char) == 'N') ? NORTH :\
(((char) == 'S') ? SOUTH :\
(((char) == 'W') ? WEST : EAST)))

#define undir(number)\
(((number) == NORTH) ? 'N' :\
(((number) == SOUTH) ? 'S' :\
(((number) == WEST) ? 'W' : 'E')))

#define ctoi(char)\
(((char) <= '9') ? (char) - '0' :\
(((char) >= 'a') ? (char) - 'a' + 10 : dir(char)))

#define itoc(number)\
((number) < MAX_CHILDREN ? (number) + '0' : undir(number))

#define BNUM(roomNum) (1 + 3 * (roomNum))

#define indx(char) ((char) - 'a')

typedef int SHORT;

typedef struct _Bottleneck Bottleneck;
struct _Bottleneck {
    SHORT   block1, block2;
    double  adjust;
    SHORT   prob[MAX_CHILDREN][MAX_CHILDREN+4];
};


/* ---------- Local Procedures ---------- */
static BearStatus PinSpace();


/* ---------- Local Global Variables ---------- */
static SHORT probArray[] = {
    30, 25, 40, 40, 35, 35, 30, 50, 25, 60, 20, 100,
    10, 30, 25, 35, 15, 75, 40, 0, 0, 0, 0, 0
};
static int sizeOfProbArray = sizeof (probArray) / sizeof (SHORT);

static Bottleneck *spaceInfo[NPL_MAX_PATTERNS];
static double bneckAdjust[NPL_MAX_PATTERNS][MAX_CHILDREN][MAX_CHILDREN+4];


/*
 * -----------------------------------------------------------------------------
 *
 * Npl_ReadSpaceInfo() --
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Initializes the array spaceInfo with values
 *    from the string nplRoutingInfo.
 *
 * Comments:
 *
 *    Conventions for nplRoutingInfo
 *    =============================
 *
 *    The string contains NPL_MAX_PATTERNS sections beginning with the title
 *    > template #n
 *    If the template has r rooms, 3r+1 subsections, one for every bottleneck
 *    will follow. Each subsection begins with a line
 *    > mn k a
 *    where m and n are the numbers of the blocks that bound this bottleneck.
 *    m and n are in the range from 0 to MAX_CHILDREN or one of N, S. E, W.
 *    The sequence is such that m is smaller than n.
 *    k is the number of nonzero entries for the probability matrix that
 *    follow in the remaining lines of the bottleneck subsection. k is
 *    a number between 1 and 9 or a small letter starting from a (=10),
 *    b (=11), etc.  a is a number between 0 and 1 which specifies the
 *    adjust factor for the probability matrix that follows.
 *    The remaining lines of the subsection contain blocks of the form
 *    > mnc
 *    separated by blanks or newlines. m and n are in the form already
 *    described for the bottleneck specifiers, they stand for the blocks
 *    that are connected with a certain connectivity c(m,n). The probability
 *    that a connection between m and n passes through the current bottleneck
 *    is given by c, a small letter. It stands for a particular bottleneck/
 *    connection configuration, the probability that the connection passes
 *    through the bottleneck is the corresponding entry in the probArray.
 *    An entry of 100 corresponds to a probability of 1. All probabilities
 *    not explicitly specified are set to 0.
 *
 * -----------------------------------------------------------------------------
 */

void
Npl_ReadSpaceInfo()
{
    int        bottNum, block1, block2, line = 0, entries, row, col;
    int        i, j, k;
    char       *buffer;

    for (i = 0; i < NPL_MAX_PATTERNS; i++) {
        buffer = nplRoutingInfo[line++];

        if (*buffer != 't') {
            fputs("\n\n*** error in string nplRoutingInfo ***\n\n", stderr);
            exit(-1);
        }

        bottNum = BNUM(npl_FloorplanSizeArray[i]);
        MALLOC(Bottleneck *, spaceInfo[i], bottNum * sizeof (Bottleneck));

        for (j = 0; j < bottNum; j++) {
            buffer = nplRoutingInfo[line++];
            block1 = spaceInfo[i][j].block1 = ctoi(*buffer);
            buffer++;
            block2 = spaceInfo[i][j].block2 = ctoi(*buffer);
            buffer++;
            buffer++; /* skip blank */
            entries = ctoi(*buffer); /* a = 10, g = 16 ... */
            buffer++;
            sscanf(buffer, "%lf", &spaceInfo[i][j].adjust);
            bneckAdjust[i][block1][block2] = (spaceInfo[i][j].adjust);

            if (entries != 0) {
                buffer = nplRoutingInfo[line++];
            }

            for (row = 0; row < MAX_CHILDREN; row++) {
                for (col = 0; col < MAX_CHILDREN+4; col++) {
                    spaceInfo[i][j].prob[row][col] = 0;
                }
            }

            for (k = 0; k < entries; k++) {
                row = ctoi(*buffer);
                buffer++;
                col = ctoi(*buffer);
                buffer++;
#ifdef DEBUG
                if (indx(*buffer) >= sizeOfProbArray) {
      Bear_PrintFatalErr("Npl_ReadSpaceInfo() failed : overflowed probArray");
                }
#endif DEBUG
                spaceInfo[i][j].prob[row][col] = probArray[indx(*buffer)];
                if (*buffer == 'a' || *buffer == 'b' || *buffer == 'e' ||
                        *buffer == 'f' || *buffer == 'g' || *buffer == 's') {
                    spaceInfo[i][j].prob[row][col] *= -1;
                }
                /* The '-' means that this probability should only be used
                 * on the non leaf level, otherwise it is taken to be = 0 */
                buffer++;
                buffer++; /* skip blank */
            }
        }
    }
}



/*
 * -----------------------------------------------------------------------------
 *
 * Npl_GenSpace() --
 *    Generate the necessary routing space between
 *    a) the two block sides that are closest to each other or
 *    b) from the closest block side to an IO goal.
 *    Requires that the pattern, label and orient fields of nodePtr are set.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    The matrix nodePtr->nd_spacing is set to the estimated
 *    values for routing space.
 *    Note that nodePtr->nd_spacing[i][j] corresponds to the spacing 
 *    between block occupying room i and that occupying room j of 
 *    the nodePtr->nd_pattern.
 *
 * -----------------------------------------------------------------------------
 */

BearStatus
Npl_GenSpace(nodePtr, levelAdjust)
Ncl_Node    *nodePtr;
double      levelAdjust;
{
    int         intSpacing, extSpacing;
    int         bneck, row, col;
    SHORT       b1, b2;
    int         c1, c2;
    int         fpID, *rmAsg, *gAsg;
    double      routAdj;

#ifdef DEBUG
    if (nodePtr->nd_type == LEAF) {
        Bear_PrintErr("Npl_GenSpace failed: nodePtr cannot be a leaf.");
        return (BEAR_ERROR);
    }
#endif /* DEBUG */

    routAdj = NPLMAX(1.0, ncl_RouteAdjust);
    if (nodePtr->nd_type == P_LEAF) {
        intSpacing = 4 * ncl_WireWidth * routAdj;
        extSpacing = 2 * ncl_WireWidth * routAdj;
    }
    else {
        intSpacing = extSpacing = ncl_WireWidth * routAdj;
    }

    if (ncl_ChannelFreeFlag == TRUE || ncl_QuickRouteEstFlag == TRUE) {
        for (row = 0; row < MAX_CHILDREN; row++) {
            for (col = row + 1; col < MAX_CHILDREN; col++) {
                if (col >= nodePtr->nd_numOfChildren) {
                    nodePtr->nd_spacing[row][col] = 0;
                } else {
                    nodePtr->nd_spacing[row][col] = intSpacing;
                }
            }
            for (col = MAX_CHILDREN; col < MAX_CHILDREN + 4; col++) {
                nodePtr->nd_spacing[row][col] = extSpacing;
            }
        }

        return (BEAR_OK);
    }

    fpID = nodePtr->nd_pattern;
    rmAsg = nodePtr->nd_label;
    gAsg = nplGoalAssignTable[nodePtr->nd_orient];

    /* 
     * HACK!  OVERKILL!
     * If we do not initialize everything to zero, then nd_spacing[0][3]
     * and nd_spacing[1][2] will not get set for template 4.  
     * This problem arises due to the fact that nplRoutingInfo does not 
     * have entries for these two bottlenecks.  However, FDim4 function
     * uses these entries. MP
     * if (fpID == 4) {
     * }
     */
    for (row = 0; row < MAX_CHILDREN; row++) {
        for (col = row + 1; col < MAX_CHILDREN + 4; col++) {
            nodePtr->nd_spacing[row][col] = 0;
        }
    }

    for (bneck = 0; bneck < BNUM(npl_FloorplanSizeArray[fpID]); bneck++) {

        b1 = spaceInfo[fpID][bneck].block1;
        b2 = spaceInfo[fpID][bneck].block2;

        /* 
         * The factor 10000 comes from the fact that probabilities and
         * connectivities are both multiplied by 100 to make them integers 
         */
        if (b2 < MAX_CHILDREN) {

            if (nodePtr->nd_type == P_LEAF) {
                /* 
                 * A compensation for the routing space around the pins 
                 * is necessary. The total number of pins is divided by 
                 * 4 to get the average number of pins per side, then 
                 * the # of tracks is multiplied with 10000 (see above) 
                 * and the resulting # of tracks is divided by a heuristic 
                 * factor of 4, because not for every pin on a side one 
                 * routing track is necessary. 
                 */
                nodePtr->nd_spacing[b1][b2] =
                    (nodePtr->nd_children[rmAsg[b1]]->nd_numOfPins +
                    nodePtr->nd_children[rmAsg[b2]]->nd_numOfPins) * 625.0;
            } else {
                /* 
                 * We leave 2 tracks free in any case, see above for 10000.
                 */
                nodePtr->nd_spacing[b1][b2] = 20000;
            }

        } else {
            nodePtr->nd_spacing[b1][b2] = 0;
        }

        for (row = 0; row < MAX_CHILDREN; row++) {
            c1 = (row < MAX_CHILDREN) ? rmAsg[row] : gAsg[row-MAX_CHILDREN];

            for (col = row + 1; col < MAX_CHILDREN+4; col++) {

                c2 = (col < MAX_CHILDREN) ? rmAsg[col] :
                    gAsg[col-MAX_CHILDREN];

                if (nodePtr->nd_type == P_LEAF) {
                    nodePtr->nd_spacing[b1][b2] +=
                        nodePtr->nd_connectivity[NPLMIN(c1, c2)][NPLMAX(c1, c2)] *
                        (int) POS(spaceInfo[fpID][bneck].prob[row][col]);
                } else {
                    nodePtr->nd_spacing[b1][b2] +=
                        nodePtr->nd_connectivity[NPLMIN(c1, c2)][NPLMAX(c1, c2)] *
                        (int) ABS(spaceInfo[fpID][bneck].prob[row][col]);
                }
            }
        }

        /* 
         * 10000 comes from the fact that the probabilities are given in %
         * (factor 100) and connectivities are multiplied by 100 in the
         * clustering to transform them in integers (second factor 100) 
         */
        nodePtr->nd_spacing[b1][b2] = (int) ( nodePtr->nd_spacing[b1][b2] * 
                spaceInfo[fpID][bneck].adjust * ncl_WireWidth / 
                (10000.0 * levelAdjust) );
    }

    for (row = 0; row < MAX_CHILDREN; row++) {
        for (col = row + 1; col < MAX_CHILDREN; col++) {
            if (col >= nodePtr->nd_numOfChildren) {
                nodePtr->nd_spacing[row][col] = 0;
            } else {
                nodePtr->nd_spacing[row][col] = 
                        NPLMAX(nodePtr->nd_spacing[row][col], intSpacing);
            }
        }
        for (col = MAX_CHILDREN; col < MAX_CHILDREN + 4; col++) {
            nodePtr->nd_spacing[row][col] = 
                    NPLMAX(nodePtr->nd_spacing[row][col], extSpacing);
        }
    }

    return (BEAR_OK);
}



/*
 * -----------------------------------------------------------------------------
 *
 * NplAddPinSpace() --
 *    Add the necessary routing space to get from the pin location to
 *    the side of the block that is closest to the goal of the connection
 *    (only on the lowest level of the hierarchy).
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    The values in the matrix nodePtr->nd_spacing are increased according
 *    to the pins and their connections on the sides of the cells.
 *    If the child cell has flexible shape, use totalPinCnt to estimate wiring.
 *
 * -----------------------------------------------------------------------------
 */

BearStatus
NplAddPinSpace(nodePtr)
Ncl_Node *nodePtr;
{
    
    Ncl_Node    *childNodePtr;
    int         i, j, numOfChildren = nodePtr->nd_numOfChildren;
    float       pinSpacing;

    if (ncl_ChannelFreeFlag == TRUE) {
        return (BEAR_OK);
    }

#ifdef DEBUG
    if (nodePtr->nd_type != P_LEAF) {
        Bear_PrintErr("NplAddPinSpace failed: nodePtr is not a pLeaf.");
        return (BEAR_ERROR);
    }
#endif /* DEBUG */

    if (nodePtr->nd_status != PLACED) {
        Bear_PrintErr("NplAddPinSpace failed: nodePtr is not placed.");
        return (BEAR_ERROR);
    }

    for (i = 0; i < numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[nodePtr->nd_label[i]];

        if (! ncl_QuickRouteEstFlag) {
            if (childNodePtr->nd_cellInstPtr->masterPtr->shapeFlag == MAC_FLEX_SHAPE) {
                /* Hueristic factor 1.75 */
                pinSpacing = (childNodePtr->nd_totalPinCnt * ncl_WireWidth *
                        ncl_RouteAdjust) / (4 * 1.75);

                for (j = 0; j < i; j++) {
                    nodePtr->nd_spacing[j][i] += ((int) pinSpacing);
                }
                for (j = i + 1; j < MAX_CHILDREN + 4; j++) {
                    if (j >= numOfChildren && j < MAX_CHILDREN) continue;
                    nodePtr->nd_spacing[i][j] += ((int) pinSpacing);
                }
            }
            else {
                if (PinSpace(childNodePtr, i) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }

                for (j = 0; j < i; j++) {
                    nodePtr->nd_spacing[j][i] += 
                        ((int) (childNodePtr->nd_spacing[i][j] * ncl_RouteAdjust));
                }
                for (j = i + 1; j < MAX_CHILDREN + 4; j++) {
                    if (j >= numOfChildren && j < MAX_CHILDREN) continue;
                    nodePtr->nd_spacing[i][j] += 
                        ((int) (childNodePtr->nd_spacing[i][j] * ncl_RouteAdjust));
                }
            }
        }
        else {
            pinSpacing = (childNodePtr->nd_totalPinCnt * ncl_WireWidth *
                    ncl_RouteAdjust) / (4 * 1.325);
            for (j = 0; j < i; j++) {
                nodePtr->nd_spacing[j][i] += ((int) pinSpacing);
            }
            for (j = i + 1; j < MAX_CHILDREN + 4; j++) {
                if (j >= numOfChildren && j < MAX_CHILDREN) continue;
                nodePtr->nd_spacing[i][j] += ((int) pinSpacing);
            }
        }
    }

    return (BEAR_OK);
}



/*
 * -----------------------------------------------------------------------------
 *
 * PinSpace() --
 *    Generate the necessary routing space to get from the pin location to
 *    the side of the block that is closest to the goal of the connection
 *    (only on the lowest level of the hierarchy).
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    The values in the matrix nodePtr->nd_spacing are set according
 *    to the pins and their connections on the sides of the cell.
 *
 * -----------------------------------------------------------------------------
 */

static BearStatus
PinSpace(nodePtr, room)
Ncl_Node    *nodePtr;
int         room;
{
    Ncl_Node    *parentPtr;
    double      spacing;
    int         fpID, *rmAsg, *gAsg, dirAsg;
    int         j, k, l;
    int         direc, goal, direc2;
    char        *ptr, *ptr2;

#ifdef DEBUG
    if (nodePtr->nd_type != LEAF) {
        Bear_PrintErr("PinSpace failed: nodePtr is not a leaf.");
        return (BEAR_ERROR);
    }
#endif /* DEBUG */

    /* initialize ALL spacing fields to 0 */
    for (l = 0; l < MAX_CHILDREN+4; l++) {
        nodePtr->nd_spacing[room][l] = 0;
    }

    parentPtr = nodePtr->nd_parentPtr;
    fpID = parentPtr->nd_pattern;
    rmAsg = parentPtr->nd_label;
    gAsg = nplGoalAssignTable[parentPtr->nd_orient];
    dirAsg = nodePtr->nd_orient;

    for (j = 0; j < 4; j++) {   
        /* for all pin directions */
        direc = nplDirectionTable[dirAsg][gAsg[j] - MAX_CHILDREN] - MAX_CHILDREN;

        for (ptr = nplNeighb[fpID][room][j]; *ptr != '\0'; ptr++) {   
            /* for all adjacent cells/boundaries in this direction */
            l = ctoi(*ptr);
            spacing = 0.0;

            for (k = 0; k < MAX_CHILDREN+4; k++) {
                /* for all possible conns with pin in j direction */
                goal = (k < MAX_CHILDREN) ? rmAsg[k] : gAsg[k-MAX_CHILDREN];

                if (index(nplNeighb[fpID][room][j], itoc(k))) {
                    spacing += nodePtr->nd_connectivity[direc][goal] / 3;
                } else {
                    spacing += nodePtr->nd_connectivity[direc][goal];
                }
            }

            spacing += nodePtr->nd_connectivity[(direc+1)%4][rmAsg[room]] / 4;
            spacing += nodePtr->nd_connectivity[(direc+3)%4][rmAsg[room]] / 4;

            for (ptr2 = nplInDir[fpID][room][j]; *ptr2 != '\0'; ptr2++) {   
                /* for all conns to cells in j direction with pins on
                 * neighboring side of block. 
                 */
                k = ctoi(*ptr2);
                goal = (k < MAX_CHILDREN) ? rmAsg[k] : gAsg[k-MAX_CHILDREN];

                spacing += (nodePtr->nd_connectivity[(direc+1)%4][goal] +
                        nodePtr->nd_connectivity[(direc+3)%4][goal]) / 4;
            }

            for (ptr2 = nplInDir[fpID][room][(j+1)%4]; *ptr2 != '\0'; ptr2++) {
                /* The next two loops are for all the connections from a
                 * neighboring side of the cell to the opposite neighboring
                 * side / direction. 
                 */
                k = ctoi(*ptr2);
                goal = (k < MAX_CHILDREN) ? rmAsg[k] : gAsg[k-MAX_CHILDREN];
                direc2 = nplDirectionTable[dirAsg][gAsg[(j+3)%4] - MAX_CHILDREN]
                             - MAX_CHILDREN;

                spacing += nodePtr->nd_connectivity[direc2][goal] / 2;
            }

            for (ptr2 = nplInDir[fpID][room][(j+3)%4]; *ptr2 != '\0'; ptr2++) {
                k = ctoi(*ptr2);
                goal = (k < MAX_CHILDREN) ? rmAsg[k] : gAsg[k-MAX_CHILDREN];
                direc2 = nplDirectionTable[dirAsg][gAsg[(j+1)%4] - MAX_CHILDREN]
                             - MAX_CHILDREN;

                spacing += nodePtr->nd_connectivity[direc2][goal] / 2;
            }

            /* HEURISTIC: adjustment with factor 2.0 */
            nodePtr->nd_spacing[room][l] = 
                    (int) (spacing * ncl_WireWidth / 200.0 * 
                    bneckAdjust[fpID][NPLMIN(room, l)][NPLMAX(room, l)]);
            /* 
             * for a new j, l is guaranteed to be different from previous
             * l's in the loop, therefore no '+=' is necessary (rationale:
             * a cell can't be a neighbor of another cell in more than one
             * direction since the cells are rectangular and not L-shaped) 
             */

        }

    }

    return (BEAR_OK);
}


