
/*
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/matrix.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif
*/

/*
 *  matrix.c
 *  ********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for obtaining matrix values for FIX and WGT.
 *
 *  functions contained in this file:
 *	calcFIX()
 *	calcWGT()
 *	traceRoute()
 *	calcMatrix()
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "initGlb.h"



/*
 *  calcFIX()
 *
 *  DESCRIPTION
 *  -----------
 *	Estimate initial demand for region usage by shortest routes 
 *	for each net.
 *	Also sum up the fixed usage for each region by critical nets.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	netGr		: current net group
 *	pass_times[i]	: # of times the given net passes region i
 *				pass-thru	: 1
 *				termination	: 0.5
 *				T cross		: 1.5
 *				+ cross		: 2.0
 *
 *  Local variables:
 *  ---------------
 *	alphaPitch	: pitch per terminating net in OCT units
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	parm		: parameter struct
 *	statis		: statistics struct
 *
 *
 */


calcFIX(netGr, pass_times)

GNetGrDef	*netGr;
float		pass_times[G_MAXREGION];

{


int		i;
int		alphaPitch;
int		regID, regType;
int		numNet;
int		numCritNet;



    /* alpha is used for a net terminating in this region */
    alphaPitch = (int)((float)G_UNITPITCH * parm.alpha);

    numNet = netGr->NoNet;
    numCritNet = netGr->NoCritNet;

    for (i = 0; i < G_MAXREGION; ++i) {
	if (!GID2Reg[i]) continue;
	regType = GID2Reg[i]->type;
	regID = i + 1;
	if (pass_times[i] == 0.0) {
	    continue;
	} else if (pass_times[i] == 0.5) {
	    if ((regType == G_HOR_BN) || (regType == G_VER_BN)) {
		GID2Reg[i]->estUsage += alphaPitch * numNet;
		FIX[regType][regID - 1] += alphaPitch * numCritNet;
	    } else if (regType == G_OVERCELL) {
		FIX[G_OVERCELL][regID - 1] += alphaPitch * numCritNet;
	    }
	} else {
	    if ((regType == G_HOR_BN) || (regType == G_VER_BN)) {
		GID2Reg[i]->estUsage += G_UNITPITCH * numNet;
		FIX[regType][regID - 1] += G_UNITPITCH * numCritNet;
	    } else if (regType == G_OVERCELL) {
		FIX[G_OVERCELL][regID - 1] += G_UNITPITCH * numCritNet;
	    }
	}
    }
}


/*
 *  calcWGT()
 *
 *  DESCRIPTION
 *  -----------
 *	Calculate WGT matrix
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	pass_times[i]	: # of times the given net passes region i
 *				pass-thru	: 1
 *				termination	: 0.5
 *				T cross		: 1.5
 *				+ cross		: 2.0
 *	route		: route to be traced
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	parm		: struct for parameter
 *
 */

calcWGT(pass_times, routeNo)

float	pass_times[G_MAXREGION];
int	routeNo;

{

int	i;
int	regID;
int	regType;

	/* Get a weight matrix */
	/*---------------------*/
	for (i = 0; i < G_MAXREGION; ++i) {
	    if (!GID2Reg[i]) continue;
	    regType = GID2Reg[i]->type;
	    regID = i + 1;
	    if (pass_times[i] == 0.0) {
		continue;
	    } else if (pass_times[i] == 0.5) {
		if (regType != G_DOM_XSEC) {
		    WGT[regID - 1][routeNo] += parm.alpha;
		}
	    } else {
		if (regType != G_DOM_XSEC) {
		    WGT[regID - 1][routeNo] += 1.0;
		}
	    }
	}

}


/*
 *  traceRoute()
 *
 *  DESCRIPTION
 *  -----------
 *	Calculate usage of routes for each region
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	route		: route to be traced
 *	pass_times[i]	: # of times the given net passes region i
 *				pass-thru	: 1
 *				termination	: 0.5
 *				T cross		: 1.5
 *				+ cross		: 2.0
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */

traceRoute(route, pass_times)

routeType	*route;
float		pass_times[G_MAXREGION];

{

GRegionDef	*nextRegion,*getNextRegion();
GRegionDef	*preRegion;
pathType 	*nextPath;
nodeGroupType	*nextNode;
int		i;
int		regID, regType;


	for (i = 0; i < G_MAXREGION; ++i) {
	    pass_times[i] = 0.0;
	}

	for (nextPath = route->path;
		nextPath != NULL; nextPath = nextPath->next) {

	    for (nextNode = nextPath->node, preRegion = NULL;
			(nextRegion = getNextRegion(nextNode)) != NULL;
			nextNode = NULL, preRegion = nextRegion) {

	    /* Through edge is counted  by 1/2 when go into here 
	     * and 1/2 when go out, which amounts to 1.0 in all */
	    /* Add weight for going out of the preceding region */

		if (preRegion != NULL) {
		    regType = preRegion->type;
		    regID = preRegion->id;
		    if (regType != G_DOM_XSEC) {
			pass_times[regID - 1] += 0.5;
		    }
		}

	    /* Add weight for coming into the current region */

		if (nextRegion != GID2Reg[nextPath->node->id - 1]) {
		    regType = nextRegion->type;
		    regID = nextRegion->id;
		    if (regType != G_DOM_XSEC) {
			pass_times[regID - 1] += 0.5;
		    }
		}
	    }
	}
}



/*
 *  calcMatrix()
 *
 *  DESCRIPTION
 *  -----------
 *	Calculate the values of matrices for the given net group
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	netGr		: current net group
 *	netGrCount	: sequential number for the current net group
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	parm		: struct for parameter
 *
 */

calcMatrix(netGr)

GNetGrDef	*netGr;

{

int	i;
float	pass_times[G_MAXREGION];


    for (i = 0; i <= parm.M - 1; ++i) {

	if (netGr->candRoute[i] == NULL) {
	    return;
	}

	traceRoute(netGr->candRoute[i], pass_times);

	if (i == 0) {
	    calcFIX(netGr, pass_times);
	    statis.init_netlength += 
			netGr->candRoute[0]->length * netGr->NoNet /
			OCT_PER_MICRON;
	}
	if ((askTrivialRoute(netGr) == YES) 
		|| (netGr->NoCritNet == netGr->NoNet)) {
	    continue;
	} else {
	    calcWGT(pass_times, (netGr->id - 1) * parm.M + i);
	}
    }

}


