#include "port.h"
#include "utility.h"
#include "errtrap.h"
#include "oct.h"
#include "tap.h"
#include "oh.h"
#include "wolfe.h"
#include "vov.h"

static void options_error(s, i)
char *s;
int i;
{
    errRaise("wolfe", OCT_ERROR,
	     "Error parsing rules file: line %d - cannot parse '%s'", i, s);
}


int read_rules_file(rules_file,facet)
char *rules_file;
octObject *facet;
{
    tapLayerListElement layer_list[2];
    octObject tech_facet, vert_layer, horiz_layer;
    octObject dummy_facet, prop, contact;
    octStatus stat;
    char lname[1024], cell[1024], view[1024], token[1024];
    char filename[1024], side[1024];
    int width, status, line_num = 0;
    struct octBox bbox;
    double temp, temp1;
    FILE *fp;

    tech_facet.type = OCT_FACET;
    vert_layer.type = OCT_LAYER;
    horiz_layer.type = OCT_LAYER;

    /* Get the technology prop from the cell - default scmos */
    if (ohGetByPropName(facet,&prop,"TECHNOLOGY") != OCT_OK){
	(void) printf("WARNING: No TECHNOLOGY property on facet, Using scmos \n");
	tapOpenTechFacet("scmos","symbolic",&tech_facet,"r");
	OH_ASSERT(ohCreateOrModifyPropStr(facet,&prop,"TECHNOLOGY","scmos"));
    } else {
	tapOpenTechFacet(prop.contents.prop.value.string, "symbolic", &tech_facet, "r");
    }

    /* get the units/lambda from tap */
    if (tapGetProp(facet, TAP_LAMBDA)) {
        units_per_lambda = TAP_LAMBDA->value.integer;
    }

    (void) strcpy(filename, rules_file);
    if ((fp = fopen(filename, "r")) == NULL) {
	(void) sprintf(filename, "%s/%s",
				 util_file_expand(tapGetDefaultRoot()),
				 rules_file);
	if ((fp = fopen(filename, "r")) == NULL) {
	    (void) fprintf(stderr, "Can't find rules file '%s'\n", filename);
	    return 1;
	}
    }
    VOVinput( VOV_UNIX_FILE, util_logical_compress( filename, "OCTTOOLS" ) );

    if (debug > 0) {
	(void) fprintf(stderr, "Using rules file: %s\n", filename);
    } 

loop:
    status = fscanf(fp, "%s", token);
    if (status == 1) {
	line_num++;
	if (strcmp(token, "units_per_lambda") == 0) {
	    if (fscanf(fp, " %d\n", &units_per_lambda) != 1) {
		options_error(token, line_num);
	    }
	} else if (strcmp(token, "feedthru") == 0) {
	    if (fscanf(fp, " %s %s\n", cell, view) != 2) {
		options_error(token, line_num);
	    } else {
		feedthru_cell = util_strsav(cell);
		feedthru_view = util_strsav(view);
	    }
	} else if (strcmp(token, "v_layer") == 0) {
	    if (fscanf(fp, "%s \n", lname) != 1) {
		options_error(token, line_num);
	    } else {
		vlayer_name = util_strsav(lname);
		OH_ASSERT(ohGetByLayerName(&tech_facet,&vert_layer,vlayer_name));
		vlayer_width = tapGetMinWidth(&vert_layer) ;
		if( (vlayer_width == TAP_NO_RULE) || (vlayer_width == TAP_RULE_ERROR)){
		    (void) printf("no min width rule for %s\n", vlayer_name);
		    VOVend(1);
		}
		vlayer_pitch = tapGetMinSpacing(&vert_layer,&vert_layer) ;
		if( (vlayer_pitch == TAP_NO_RULE) || (vlayer_pitch == TAP_RULE_ERROR)){
		    (void) printf("no min spacing rule for %s\n", vlayer_name);
		    VOVend(1);
		}
	    }

	} else if (strcmp(token, "power_position") == 0) {
	    if (fscanf(fp, "%s \n", side) != 1) {
		options_error(token, line_num);
	    } else {
		if ( strcmp(side,"LEFT") == 0){
			power_position = LEFT;
		}
		else if ( strcmp(side,"RIGHT") == 0){
			power_position = RIGHT;
		}
		else if ( strcmp(side,"BOTH") == 0){
			power_position = BOTH;
		}
		else{
		(void) printf("Unknown side for Power Terminal - Using LEFT \n");
		}
	    }
	} else if (strcmp(token, "v_net_weight") == 0) {
	    if (fscanf(fp, "%lf \n", &temp) != 1) {
		options_error(token, line_num);
	    } else v_net_weight = temp;
	} else if (strcmp(token, "h_net_weight") == 0) {
	    if (fscanf(fp, "%lf \n", &temp) != 1) {
		options_error(token, line_num);
	    } else h_net_weight = temp;
	} else if (strcmp(token, "rowSep") == 0) {
	    if (fscanf(fp, "%lf \n", &temp) != 1) {
		options_error(token, line_num);
	    } else rowSep = temp;
	} else if (strcmp(token, "power_width") == 0) {
	    if (fscanf(fp, "%d \n", &width) != 1) {
		options_error(token, line_num);
	    } else power_width = width;
	} else if (strcmp(token, "fast") == 0) {
	    fast_mode = 1;
	} else if (strcmp (token, "num_rows") == 0) {
	    if (fscanf (fp, "%d \n", &num_rows) != 1) {
                options_error (token);
	    }
	} else if (strcmp (token, "implicit_range") == 0) {
	    if (fscanf (fp, "%lf \n", &implicit_feed_thru_range) != 1) {
	        options_error (token);
            }
	} else if (strcmp (token, "add_implicit_feeds") == 0) {
	    Add_implicit = 0;
        } else if (strcmp (token, "do_global_route") == 0) {
	    do_global_route = 0;
	} else if (strcmp(token, "h_layer") == 0) {
	    if (fscanf(fp, "%s \n", lname) != 1) {
		options_error(token, line_num);
	    } else {
		hlayer_name = util_strsav(lname);
		OH_ASSERT(ohGetByLayerName(&tech_facet,&horiz_layer,hlayer_name));
		hlayer_width = tapGetMinWidth(&horiz_layer) ;
		if( (hlayer_width == TAP_NO_RULE) || (hlayer_width == TAP_RULE_ERROR)){
			(void) printf("no min width rule for %s\n", hlayer_name);
			VOVend(1);
		}
		hlayer_pitch = tapGetMinSpacing(&horiz_layer,&horiz_layer) ;
		if( (hlayer_pitch == TAP_NO_RULE) || (hlayer_pitch == TAP_RULE_ERROR)){
			(void) printf("no min spacing rule for %s\n", hlayer_name);
			VOVend(1);
		}
	    }
	} else if (strcmp(token, "rowSep") == 0) {
	    if (fscanf(fp, "%lf", &rowSep) != 1) {
		options_error(token, line_num);
	    }

#ifdef TW3_3
	} else if (strcmp(token, "indent") == 0) {
	    if (fscanf(fp, "%lf", &indent) != 1) {
		options_error(token, line_num);
	    }
#endif
	} else if (strcmp(token, "do.global.route.cell.swaps") == 0) {
	    if (fscanf(fp, "%lf", &temp) != 1) {
		options_error(token, line_num);
	    }
#ifdef TW3_3
	    /* If set on command line, do not reset from the rules file */
	    if (do_global_route_cell_swaps == 0) {
		do_global_route_cell_swaps = temp;
	    }
#endif

	} else {
	    options_error(token, line_num);
	}
    } else if (status != EOF) {
	options_error("keyword", line_num);
    } else if ( status== EOF) {
	/* Read the data about the connectors and derive the variables */

	if ( line_num == 0) {
	    (void) printf("ERROR: Empty rules file %s\n", rules_file);
	    VOVend(-1);
	}
        if (units_per_lambda == 0) {
            (void) printf("ERROR: units_per_lambda is zero\n");
            VOVend(-1);
        }
	vlayer_width /= units_per_lambda;
	vlayer_pitch /= units_per_lambda;
	hlayer_pitch /= units_per_lambda;
	hlayer_width /= units_per_lambda;

	layer_list[0].layer = horiz_layer;
	layer_list[0].width = 0;
	layer_list[0].direction = TAP_HORIZONTAL;
	layer_list[1].layer = vert_layer;
	layer_list[1].width = 0;
	layer_list[1].direction = TAP_VERTICAL;
	contact.type = OCT_INSTANCE;
	if( (stat = tapGetConnector(2,layer_list,&contact)) != OCT_OK){
		(void)fprintf(stdout,"ERROR: %s-%s contact of size %dx%d not found\n", hlayer_name, vlayer_name, hlayer_width, vlayer_width);
		OH_ASSERT(stat);
		VOVend (1);
	}

	hv_contact_name = util_strsav(contact.contents.instance.master);
	hv_contact_view = util_strsav(contact.contents.instance.view);

	if (ohOpenFacet(&dummy_facet,hv_contact_name,
		hv_contact_view,"contents","r") < OCT_OK){
	    (void) printf("Opening the master of v_contact\n");
	    octError(NIL(char));
	    VOVend(1);
	}
	VOVinputFacet( &dummy_facet );

	OH_ASSERT(octBB(&dummy_facet,&bbox));
	temp1 = MAX(vlayer_pitch, hlayer_pitch);
	hv_contact_height = (bbox.upperRight.y - bbox.lowerLeft.y)/units_per_lambda;
	hv_contact_width = (bbox.upperRight.y - bbox.lowerLeft.y)/units_per_lambda;
	vlayer_pitch = temp1 + hv_contact_height;
	hlayer_pitch = temp1 + hv_contact_width;
	OH_ASSERT( octCloseFacet(&dummy_facet));

	layer_list[0].layer = vert_layer;
	if( (stat = tapGetConnector(2,layer_list,&contact)) != OCT_OK){
		(void)fprintf(stdout,"ERROR: %s-%s contact of size %dx%d not found\n", vlayer_name, vlayer_name, vlayer_width, vlayer_width);
		OH_ASSERT(stat);
		VOVend (1);
	} else{
		v_contact_name = util_strsav(contact.contents.instance.master);
		v_contact_view = util_strsav(contact.contents.instance.view);
	}

	layer_list[1].layer = horiz_layer;
	layer_list[0].layer = horiz_layer;
	if( (stat = tapGetConnector(2,layer_list,&contact)) != OCT_OK){
		(void)fprintf(stdout,"ERROR: %s-%s contact of size %dx%d not found\n", hlayer_name, hlayer_name, hlayer_width, hlayer_width);
		OH_ASSERT(stat);
		VOVend (1);
	} else{
		h_contact_name = util_strsav(contact.contents.instance.master);
		h_contact_view = util_strsav(contact.contents.instance.view);
	}

	if( debug > 2){
		(void) printf("layer %s Pitch %d Width %d \n",hlayer_name,
			hlayer_pitch, hlayer_width);
		(void) printf("layer %s Pitch %d Width %d \n",vlayer_name,
			vlayer_pitch, vlayer_width);
	}

	return 0;
    }
    goto loop;
}
