/* sccsid = @(#)read_pins.c	1.7 wolfe (ITD) 7/30/90   */
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "symlib.h"
#include "wolfe.h"

static int create_formal_pin();
static void find_feedthru_pin();
static void find_pin();
static void sort_pins();
static void dump_pins();
static void move_top_bottom_pins();
static void find_match_pin();
static void redistribute_pins();

static int sparc_constraint[4] = {SYM_PUT_CONST_TOP, SYM_PUT_CONST_BOTTOM, SYM_PUT_CONST_LEFT, SYM_PUT_CONST_RIGHT};


/*
 *  Read the TimberWolf pin file.  This file describes each pin on 
 *  each channel edge.  The format is:
 *
 *	1) net name
 *	2) group number
 *	3) instance name
 *	4) terminal name
 *	5) pin x location (ignored)
 *	6) pin y location (ignored)
 *	7) channel number (1 ... nrows+1; -1 for left side, -2 for right side)
 *	8) location within channel (-1=BOTTOM, 1=TOP, -2=LEFT, 2=RIGHT)
 *	9) stranger (backward compatibility, must be zero)
 *
 *  An array channel[channel_number][channel_location] is created which
 *  is a linked list of all of the pins along this channel edge.  
 *
 *  We trust that this information is complete and correct !
 */


int read_pin_file(prefix, facet)
char *prefix;			/* IN: prefix for the pin filename */
octObject *facet;		/* IN: the facet */
{
    struct octBox bb;
    octObject conn_bag, inst, aterm, net;
    pin_t *newpin;
    FILE *fp;
    char filename[1024], net_name[1024], cell_name[1024], term_name[1024];
    int i, j, xl, yl, channel_num, channel_loc, group, stranger, linenum;

    linenum = 0;

    /* Open the TimberWolf pin file */
    (void) sprintf(filename, "%s.pin", prefix);
    if ((fp = fopen(filename, "r")) == NULL) {
	perror(filename);
	return 1;
    }

    /* pre-allocate the channel information table -- GLOBAL TABLE */
    channel = ALLOC(pin_t **, num_rows + 4);
    channel = channel + 2;		/* allow negative subscripts */
    for(i = -2; i <= num_rows+1; i++) {
			channel[i] = ALLOC(pin_t *, 4);
			for(j = 0; j < 4; j++) {
				channel[i][j] = NIL(pin_t);
			}
    }


    /* Read TimberWolf .pin file */
    for(;;) {
	linenum++;
	i = fscanf(fp, "%s %d %s %s %d %d %d %d %d\n", 
		net_name, &group, cell_name, term_name, 
		&xl, &yl, &channel_num, &channel_loc, &stranger);
	if (i == EOF) {
	    break;		/* natural termination of the loop */
	} else if (i != 9) {
	    (void) fprintf(stderr, "\"%s\", line %d: bad count from scanf\n",
		filename, linenum);
	    return 1;
	}
	    
	/* Check the channel number field */
	if (channel_num == -1 || channel_num == -2) continue;  /* skip */
	if (channel_num < 1 || channel_num > num_rows+1) {
	    (void) fprintf(stderr, 
		"\"%s\", line %d: Channel number '%d' out of range\n",
		filename, linenum, channel_num);
	    return 1;
	}

	/* Check the pin location field */
	if (channel_loc == 1) {
	    channel_loc = TOP;
	} else if (channel_loc == -1) {
	    channel_loc = BOTTOM;
	} else if (channel_loc == 2) {
	    channel_loc = RIGHT;
	} else if (channel_loc == -2) {
	    channel_loc = LEFT;
	} else {
	    (void) fprintf(stderr, 
		"\"%s\", line %d: Channel location '%d' out of range\n",
		filename, linenum, channel_loc);
	    return 1;
	}

	    if (stranger != 0) {
		(void) fprintf(stderr, 
		    "\"%s\", line %d: Stranger pin field must be zero\n",
		    filename, linenum);
	    }

	/*
	 *  Look for the pin, get its instance, terminal, and net
	 */
	OH_ASSERT(ohGetOrCreateBag(facet,&conn_bag,"CONNECTORS"));
	if (strcmp(cell_name, "WOLFE-PAD") == 0) {
	    if (! create_formal_pin(facet, channel_loc, net_name, &inst,&aterm, &net,&conn_bag)) {
		(void) fprintf(stderr, 
			       "\"%s\", line %d: cannot find formal terminal on net %s\n", 
			       filename, linenum, net_name);
		continue;
	    }
	} else if (strcmp(cell_name, "PSEUDO_CELL") == 0) {
	    if (! create_formal_pin(facet, channel_loc, net_name, &inst, &aterm, &net,&conn_bag)) {
		(void) fprintf(stderr, 
			       "\"%s\", line %d: cannot find formal terminal on net %s\n", 
			       filename, linenum, net_name);
		continue;
	    }
	} else if ((strncmp(cell_name, "twfeed", 6) == 0) && !complex_feeds) {
	    find_feedthru_pin(facet, channel_loc, cell_name, net_name, 
				&inst, &aterm, &net);
	} else {
	    find_pin(facet, channel_num,channel_loc,cell_name, 
			term_name, net_name, &inst, &aterm, &net, xl);

	}

	/*
	 *  Allocate the pin structure, and link into the channel edges
	 */
	newpin = ALLOC(pin_t, 1);
	newpin->next = channel[channel_num][channel_loc];
	channel[channel_num][channel_loc] = newpin;
	newpin->group = group;
	newpin->instId = inst.objectId;
	newpin->termId = aterm.objectId;
	newpin->netId = net.objectId;

	/*
	 *  Finally, insert the terminal width and position info
	 */
	OH_ASSERT(ohTerminalBB(&aterm, &bb));
	newpin->width = bb.upperRight.x - bb.lowerLeft.x;
	if((channel_loc == LEFT) || (channel_loc == RIGHT))
	  newpin->position = bb.lowerLeft.y;
	else  newpin->position = bb.lowerLeft.x;
    }

    move_top_bottom_pins();
    sort_pins();
    if (debug > 2) dump_pins();
    return 0;
}

/*
 *  create_formal_pin -- The current 'pin' is a PSEUDO_CELL (i.e., a left
 *  or right pad) or a WOLFE-PAD (i.e., a top or bottom pad).  Find the
 *  instance, actual terminal, and net given just the net name.
 */

static int create_formal_pin(facet, loc, net_name, inst, aterm, net,conn_bag)
    octObject *facet;
    int loc;			/* Probably the side. */
    char *net_name;
    octObject *inst, *aterm, *net;
    octObject *conn_bag;
{
    octGenerator gen;
    octStatus s;
    int MAGIC = -12345;		/* UGLY!! */
    static int term_no = 0;
    char inst_name[16];
    float frac;
    int pos, is_signal, connector_present=0;
    octObject fterm, power_conn, layer, vertlayer;
    octObject fterm_edge_prop, myfterm, myfterm_edge_prop;
    octGenerator fterm_gen;


    /* First get the net by name */
    OH_ASSERT_DESCR(ohGetByNetName(facet, net, net_name),
		    "attempting to find a net by name");

    /* Look at all formal terminals on this net */
    OH_ASSERT(octInitGenContents(net, OCT_TERM_MASK, &gen));
    while ((s = octGenerate(&gen, &fterm)) == OCT_OK) {
	if (!octIdIsNull(fterm.contents.term.instanceId)) continue; /* Skip Actual */

	/*
	 * We need to special case the SUPPLY and GROUND net....
	 * If the net was present in the file ***.cel it means that
	 * a signal terminal of a cell was connected to a power net.
	 * We must make sure that the corresponding formal terminal
	 * is a connector for the power rail, irrespective of the
	 * fact that the "extend-terms" flag is set or not .....
	 */
	is_signal = get_terminal_position(&fterm, &pos, &frac);
	if (pos != loc) {
	    /* fprintf( stderr, "%s on side %d not %d\n", ohFormatName( &fterm), pos, loc ); */
	    continue; /* not on correct side */
	}

	/* Get the appropriate connector if required */
	if (!is_signal){
	    OH_ASSERT(ohGetOrCreateLayer(facet, &layer, hlayer_name));
	    OH_ASSERT(ohGetOrCreateLayer(facet, &vertlayer, vlayer_name));
	    connector_present = 1;
	    if (get_connector(vertlayer,layer,power_width*units_per_lambda,
			      hlayer_width*units_per_lambda ,&power_conn) != OCT_OK){
		/* Warn that appropriate connector is absent */
		(void)fprintf(stdout,"WARNING: Connector of size %dx%d not available in the technology.\n",
			      power_width, hlayer_width);
		(void)fprintf(stdout,"\tNet %s needs to be connected manually\n", 
			      net->contents.net.name);
		connector_present = 0;
	    }
	}

	/* 
	 * This net might have multiple formals; hack to see if
	 * we have connected this particular formal before.
	 */
	if (octGenFirstContent(&fterm, OCT_TERM_MASK, aterm) != OCT_OK) {
	    /*
	     * No actual terminal.
	     * Create an instance of a connector or contact for formal term
	     */
	    (void) sprintf(inst_name,"T%d",++term_no);
	    if (!is_signal){
		if (connector_present){
		    OH_ASSERT(ohCreateInstance(facet,inst,inst_name,
					       power_conn.contents.instance.master,
					       power_conn.contents.instance.view ,MAGIC,0,
					       OCT_NO_TRANSFORM));
		} else {
		    OH_ASSERT(ohCreateInstance(facet,inst,inst_name,
					       h_contact_name,h_contact_view,MAGIC,0,
					       OCT_NO_TRANSFORM));
		}
	    } else {
		/* This is a signal terminal. */
		if ( extend_terms ){
		    OH_ASSERT(ohCreateInstance(facet,inst,inst_name,
					       hv_contact_name,hv_contact_view,MAGIC,0,
					       OCT_NO_TRANSFORM));
		} else if ( (pos == LEFT) || (pos == RIGHT)){
		    OH_ASSERT(ohCreateInstance(facet,inst,inst_name,
					       h_contact_name,h_contact_view,MAGIC,0,
					       OCT_NO_TRANSFORM));
		} else if ( (pos == TOP) || (pos == BOTTOM)){
		    OH_ASSERT(ohCreateInstance(facet,inst,inst_name,
					       v_contact_name,v_contact_view,MAGIC,0,
					       OCT_NO_TRANSFORM));
		}
		/* Add a constraint for the compaction */
		if (symSideConstraint(inst,sparc_constraint[pos], 0, 0) == SYM_ERROR){
		    (void)fprintf(stderr,"Unable to write sparcs constraint\n");
		}
	    }
	    OH_ASSERT(octAttach(conn_bag,inst));
	    OH_ASSERT(octGenFirstContent(inst,OCT_TERM_MASK,aterm));
	    OH_ASSERT(octAttach(&fterm,aterm));
	    /*
	     * The following code added by Barry Boes to attach all formal terminals
	     * on the same side and on the same net as the terminal instance just
	     * creeated to that instance.
	     */
	    if (ohGetByPropName (&fterm, &fterm_edge_prop,"TERM_EDGE") == OCT_OK)  {
		/* Look at all formal terminals on this net */
		OH_ASSERT(octInitGenContents(net, OCT_TERM_MASK, &fterm_gen));
		while (octGenerate(&fterm_gen, &myfterm) == OCT_OK) {
		    if (!octIdIsNull(myfterm.contents.term.instanceId)) continue; /* Skip */
		    if (octIdsEqual(fterm.objectId, myfterm.objectId)) continue; /* Same term, skip. */
		    if (ohGetByPropName(&myfterm,&myfterm_edge_prop,"TERM_EDGE")!=OCT_OK) continue;
		    if (strcmp(fterm_edge_prop.contents.prop.value.string,
			       myfterm_edge_prop.contents.prop.value.string) == 0) {
			octAttach(&myfterm, aterm);
		    }
		}
	    }
	    octFreeGenerator( &gen );
	    return 1;
	} else {
	    /*
	     * If the implemention is already there we MIGHT WANT TO
	     * check if it is correct ... Since wolfe will remove all
	     * implementations of formal terminals initially, right now
	     * this is not implemented
	     */
	    OH_ASSERT(ohGetById(inst, aterm->contents.term.instanceId));
	    if( inst->contents.instance.transform.translation.x != MAGIC){
		inst->contents.instance.transform.translation.x = MAGIC;
		OH_ASSERT(octModify(inst));
	    }
	    
	    octFreeGenerator( &gen );
	    return 1;
	}
    }
    OH_ASSERT(s == OCT_GEN_DONE ? OCT_OK : s);
 
    /*
     *  not quite sure what causes this, but try again, but
     * be less picky about correct position.  TW sometimes
     * decides to move formals !!
     */
    fprintf(stderr, "Using useless code\n" );

    /* Look at all formal terminals on this net */
    OH_ASSERT(octInitGenContents(net, OCT_TERM_MASK, &gen));
    while ((s = octGenerate(&gen, &fterm)) == OCT_OK) {
	if ( ! octIdIsNull(fterm.contents.term.instanceId))  continue; /* Skip actual terms. */

	/* This net might have multiple formals; hack to see if
	 *  we have connected this particular formal before
	 */
	if (octGenFirstContent(&fterm, OCT_TERM_MASK, aterm) != OCT_OK){
	    /* Create an instance of a connector or contact */
	    (void)strcpy(inst_name,"");
	    (void) sprintf(inst_name,"T%d",++term_no);
	    if ( extend_terms){
		OH_ASSERT(ohCreateInstance(facet,inst,inst_name,
					   hv_contact_name,hv_contact_view,MAGIC,0,
					   OCT_NO_TRANSFORM));
		OH_ASSERT(octAttach(conn_bag,inst));
		OH_ASSERT(octGenFirstContent(inst,OCT_TERM_MASK,aterm));
		OH_ASSERT(octAttach(&fterm,aterm));
		octFreeGenerator( &gen );
		return 1;
	    } else if ( (pos == LEFT) || (pos == RIGHT)){
		OH_ASSERT(ohCreateInstance(facet,inst,inst_name,
					   h_contact_name,h_contact_view,MAGIC,0,
					   OCT_NO_TRANSFORM));
		OH_ASSERT(octAttach(conn_bag,inst));
		OH_ASSERT(octGenFirstContent(inst,OCT_TERM_MASK,aterm));
		OH_ASSERT(octAttach(&fterm,aterm));
		octFreeGenerator( &gen );
		return 1;
	    } else if ( (pos == TOP) || (pos == BOTTOM)) {
		OH_ASSERT(ohCreateInstance(facet,inst,inst_name,
					   v_contact_name,v_contact_view,MAGIC,0,
					   OCT_NO_TRANSFORM));
		OH_ASSERT(octAttach(conn_bag,inst));
		OH_ASSERT(octGenFirstContent(inst,OCT_TERM_MASK,aterm));
		OH_ASSERT(octAttach(&fterm,aterm));
		octFreeGenerator( &gen );
		return 1;
	    }
	} else {
	    if( inst->contents.instance.transform.translation.x != MAGIC){
		inst->contents.instance.transform.translation.x = MAGIC;
		OH_ASSERT(octModify(inst));
		octFreeGenerator( &gen );
		return 1;
	    }
	}
    }
    OH_ASSERT(s == OCT_GEN_DONE ? OCT_OK : s);

    return 0;			/* Formal terminal not found  */
}

/*
 *  find_feedthru_pin -- The current 'pin' is a pin on a feed-thru instance.
 *  We have to guess on the pin name based on the channel edge and the
 *  transformation on the feedthru instance (should probably always be
 *  OCT_NO_TRANSFORM anyways, but we might as well check anyways).
 */
static void find_feedthru_pin(facet, loc, cell_name, net_name, inst, aterm, net)
octObject *facet;
int loc;
char *cell_name, *net_name;
octObject *inst, *aterm, *net;
{
    octStatus s;
    octTransformType type;

    /* Get the instance by name */
    OH_ASSERT_DESCR(ohGetByInstName(facet, inst, cell_name),
	"attempting to find feedthru instance by name");

    /*  Pin on a feedthru --  map name into TOP or BOTTOM */
    type = inst->contents.instance.transform.transformType;
    if (loc == TOP && type == OCT_NO_TRANSFORM) {
	OH_ASSERT_DESCR(ohGetByTermName(inst, aterm, "BOTTOM"),
	    "attempting to find BOTTOM terminal on feedthru");
    } else if (loc == BOTTOM && type == OCT_MIRROR_Y) {
	OH_ASSERT_DESCR(ohGetByTermName(inst, aterm, "BOTTOM"),
	    "attempting to find BOTTOM terminal on feedthru");
    } else if (loc == TOP && type == OCT_MIRROR_Y) {
	OH_ASSERT_DESCR(ohGetByTermName(inst, aterm, "TOP"),
	    "attempting to find TOP terminal on feedthru");
    } else if (loc == BOTTOM && type == OCT_NO_TRANSFORM) {
	OH_ASSERT_DESCR(ohGetByTermName(inst, aterm, "TOP"),
	    "attempting to find TOP terminal on feedthru");
    } else {
	(void) fprintf(stderr, "Problem finding pin on a feedthru\n");
	exit(1);
    }

    OH_ASSERT_DESCR(ohGetByNetName(facet, net, net_name),
	"attempting to find net by name");
    s = octAttachOnce(net, aterm);
    if (s != OCT_ALREADY_ATTACHED) OH_ASSERT(s);
}

/*
 *  find_pin -- normal case, the current 'pin' is on a normal cell;
 *  Just get the instance, actual terminal and nets by name
 */
static void
find_pin(facet, num, loc,cell_name, term_name, net_name, inst, aterm, net,xl)
octObject *facet;
char *cell_name, *term_name, *net_name;
octObject *inst, *aterm, *net;
int num, loc, xl;		/* For the feedthru info*/
{
    int find_corresp_feed();
    octStatus s;
    octObject new_inst, conn_bag;
    char new_name[1024];

    /* Unhack the pin name */
    if (strncmp(term_name, "WOLFE-TOP-", 10) == 0) {
	term_name += 10;
    } else if (strncmp(term_name, "WOLFE-BOT-", 10) == 0) {
	term_name += 10;
    }
    OH_ASSERT_DESCR(ohGetByInstName(facet, inst, cell_name),
	"attempting to find instance by name");
    OH_ASSERT_DESCR(ohGetByNetName(facet, net, net_name),
	"attempting to find net by name");

    if ( strncmp(term_name,"FEED",4) != 0){
	OH_ASSERT_DESCR(ohGetByTermName(inst, aterm, term_name),
	    "attempting to find terminal by name");

	/*
	 * The terminal may not be on the net because it
	 * may be JOINEDTERMS
	 */
	s = octAttachOnce(net, aterm);
	if (s != OCT_ALREADY_ATTACHED) OH_ASSERT(s);
    }
    else{
	if (find_corresp_feed(num, loc, aterm, net->objectId, inst->objectId, xl)){
	/* 
	 * Create a terminal for the feed_thru and 
	 * attach to CONNECTORS bag
	 */
	     (void)sprintf(new_name,"%s_%s",cell_name,term_name);
	     OH_ASSERT(ohCreateInstance(facet,&new_inst,new_name,
		 v_contact_name, v_contact_view, ((xl*units_per_lambda)
		 - min_xposition), -12345, OCT_NO_TRANSFORM));
	     OH_ASSERT(octGenFirstContent(&new_inst,OCT_TERM_MASK,aterm));
	     OH_ASSERT(ohGetByBagName(facet,&conn_bag,"CONNECTORS"));
	     OH_ASSERT(octAttach(&conn_bag,&new_inst) );
	}
    }
}

int
find_corresp_feed(channel_no, location, aterm, netId, cellId, xl)
octObject *aterm;
octId netId, cellId;
int channel_no, location;
int xl;
{
  pin_t *pin;
  int i,j;
  octObject inst;
  
  if ( location == TOP){
    j = BOTTOM;
    i = channel_no + 1;
  }
  else if (location == BOTTOM){
    j = TOP;
    i = channel_no - 1;
  }
  else{
    (void) printf("Unknown location for implicit feedthru \n");
    exit(1);
  }
  for( pin = channel[i][j]; pin != NIL(pin_t); pin = pin->next){
    if (octIdsEqual(pin->netId, netId) && octIdsEqual(pin->instId, cellId)) {
      /* Feedthru terminal implementation found */
      OH_ASSERT (ohGetById(aterm,pin->termId));
      OH_ASSERT (octGenFirstContainer (aterm, OCT_INSTANCE_MASK,
                                       &inst));
      if ((xl * units_per_lambda - min_xposition) ==
          inst.contents.instance.transform.translation.x)
        return 0;
    }
  }
  return 1;
}

/*
 *  Use the 'generic' linked-list package to sort the pins
 */

#define SORT1 	sort_pins_list
#define SORT 	sort_pins_list_helper
#define TYPE 	pin_t
#define NEXT 	next
#include "lsort.h"

static int pin_compare(pi, pj)
pin_t *pi, *pj;
{
    return pi->position - pj->position;
}


static void sort_pins()
{
    int i,j;
    /* Sort the pins by x-position */
    for(i = -2; i <= num_rows + 1; i++) {
	for(j = 0; j < 4; j++) {
	    channel[i][j] = sort_pins_list(channel[i][j], pin_compare);
	}
    }
}


static void dump_pins()
{
    octObject term, inst, net;
    int i, j;
    pin_t *p;

    (void) printf("\n\nPIN TABLE\n");
    for(i = 1; i <= num_rows+1; i++) {
	for(j = 0; j < 4; j++) {
	    (void) printf("Channel %d %s\n", i, print_location[j]);
	    for(p = channel[i][j]; p != NIL(pin_t); p = p->next) {
		OH_ASSERT(ohGetById(&term, p->termId));
		OH_ASSERT(ohGetById(&inst, p->instId));
		OH_ASSERT(ohGetById(&net, p->netId));
		(void) printf("  pos=%5d width=%2d group=%3d inst=%s term=%s net=%s\n",
			p->position, p->width, p->group, 
			inst.contents.instance.name, term.contents.term.name,
			net.contents.net.name);
	    }
	}
    }
}

int discard_pin_table()
{
    int i, j;
    pin_t *p, *pnext;

    for(i = -2; i <= num_rows+1; i++) {
	for(j = 0; j < 4; j++) {
	    for(p = channel[i][j]; p != NIL(pin_t); p = pnext) {
		pnext = p->next;
		FREE(p);
	    }
	}
    }
	return 0;
}


static void find_match_pin(p, edge)
pin_t *p, *edge;
{
    pin_t *p1;
    static int last_position = 40;

    for(p1 = edge; p1 != NIL(pin_t); p1 = p1->next) {
	if (p->group == p1->group) {
	    p->position = p1->position;
	    return;
	}
    } 

    {
	/* Generate nice error message */
	octObject aterm, fterm, net;

	OH_ASSERT(ohGetById(&aterm, p->termId));
	OH_ASSERT(octGenFirstContainer(&aterm, OCT_TERM_MASK, &fterm));
	OH_ASSERT(octGenFirstContainer(&fterm, OCT_NET_MASK, &net ));
	(void) fprintf(stderr, 
		       "I am having some (unspecified) problem routing to\n\tformal terminal %s\n\ton net %s\n", 
		       fterm.contents.term.name, net.contents.net.name);
    }
    p->position = last_position;
    last_position += vlayer_pitch * units_per_lambda;
}


static void move_top_bottom_pins()
{
    pin_t *p;

    for(p = channel[num_rows+1][TOP]; p != NIL(pin_t); p = p->next) {
	find_match_pin(p, channel[num_rows+1][BOTTOM]);
    }
    for(p = channel[1][BOTTOM]; p != NIL(pin_t); p = p->next) {
	find_match_pin(p, channel[1][TOP]);
    }

    /*
     * Make another pass and separate pins . Necessary if
     * there is more than 1 pin belonging to a group. The above
     * routine assigns them the same position with the result
     * that some of them are dropped during the gridding 
     */
    if (channel[num_rows+1][TOP] != NIL(pin_t)){
	channel[num_rows+1][TOP] = sort_pins_list(channel[num_rows+1][TOP], pin_compare);
	redistribute_pins(channel[num_rows+1][TOP], num_rows+1);
    }
    if (channel[1][BOTTOM] != NIL(pin_t)){
	channel[1][BOTTOM] = sort_pins_list(channel[1][BOTTOM], pin_compare);
	redistribute_pins(channel[1][BOTTOM], 1);
    }
}


static void
redistribute_pins(list, num)
pin_t *list;
int num;
{
    pin_t **dup_list;
    int *pos_list, curpos, vlp;
    int num_dup, num_pos, dup_list_len, pos_list_len, i;
    pin_t *p;

    dup_list_len = pos_list_len = 10;
    num_pos = num_dup = 0;
    dup_list = ALLOC( pin_t *, dup_list_len);
    pos_list = ALLOC( int , pos_list_len);
    vlp = vlayer_pitch * units_per_lambda;

    for ( p = list; p != NIL(pin_t); p = p->next){
	if ( p->next != NIL(pin_t)){
	    if(( p->next->position - p->position) < vlp){
   /* Add to list of the duplicate pins */
		num_dup++;
		if ( num_dup > dup_list_len){
		    dup_list_len += 10;
		    dup_list = REALLOC( pin_t *, dup_list, dup_list_len);
		}
		dup_list[num_dup-1] = p->next;
	    } else {
   /* Add to list of possible positions */
		curpos = p->position;
		while ( (curpos + 2 * vlp) < p->next->position){
		    curpos += vlp;
		    num_pos++;
		    if ( num_pos > pos_list_len){
			pos_list_len += 10;
			pos_list = REALLOC( int, pos_list, pos_list_len);
		    }
		    pos_list[num_pos-1] = curpos;
		}
	    }
	}
    }

    /*
     * For all the pins stored in the dup_list( actually
     * an array) put a value for the position.
     */
    if ( num_dup <= num_pos) {
	for ( i = 0; i < num_dup; i++){
	    (dup_list[i])->position = pos_list[i];
	}
    } else {
	(void) printf("WARNING: Unable to redistribute pins in channel %d \n", num);
	for ( i = 0; i < num_pos; i++){
	    (dup_list[i])->position = pos_list[i];
	}
    }

    FREE( dup_list);
    FREE( pos_list);
}
