;;;;(
;;;;    The Official LightLisp Test and Validation Suite
;;;;    

(defun rcs (symbol)
  "use the symbol as a key to what to pick out of rcs-list,
  and then remove the leading and trailing $ signs"
  (let* ((rcs-list '((Revision . "$Revision: 1.1.1.1 $")
		     (Date . "$Date: 2002/12/31 15:50:57 $")
		     (Author . "$Author: saj $")
		     (Source . "$Source: /home/cvs/octtools/src/LightLisp/src/lightlisp/validation/comprehensive.ll,v $")))
	 (string (cdr (assoc symbol rcs-list))))
	(subseq string 1 (1- (length string)))))

(defvar validations nil
  "an a-list stating what validations have been done.
  Values for the a-list value slot are
  'not-validated	- indicating that the test was not run
  'success		- indicating a successful test
  'fail			- indicating an unsuccessful test
  'load-failure		- indicating a failure to load the test")

(defun validation-symbolp (symbol)
  "see if the symbol is known to the validation system"
  (assoc symbol validations))

(defun not-validatedp (a-item)
  "take the a-item in validations - if it was not-validated"
  (eq (cdr a-item) 'not-validated))

(defun validation-for (symbol)
  "stick the symbol in the a-list as not-validated"
  (if (not (assoc symbol validations))
      (setq validations (acons symbol 'not-validated validations))))

(defun successp (a-item)
  "take the a-item in validations - if it was successful"
  (eq (cdr a-item) 'success))

(defun invalidated (symbol)
  "note the symbol in the a-list as unsuccessfully"
  (rplacd (assoc symbol validations) 'fail))

(defun validated (symbol)
  "note the symbol in the a-list as successfully validated"
  (rplacd (assoc symbol validations) 'success))

(defun bad-loadp (a-item)
  "take the a-item in validations - if it was a bad load"
  (eq (cdr a-item) 'load-failure))

(defun bad-load (symbol)
  "note the symbol in the a-list as badly-loaded"
  (rplacd (assoc symbol validations) 'load-failure))

(defun select (predp list)
  "select those in the list that are match predp"
  (cond ((null list) nil)
	((funcall predp (car list))
	 (cons (caar list) (select predp (cdr list))))
	(t
	 (select predp (cdr list)))))

(defmacro validation (symbol &rest body)
  "set up a progn that will print out some stuff,
  protect against (throw 'lightlisp ... ) and 
  if all goes well, declare the symbol as validated"
  `(progn
    (format t "~%~%~%~%~
	    --------------------~%~%    ~
	    The ~S Test~%~
	    ~%--------------------~%"
	    ',symbol)
    (cond (just-examinep
	   t)
	  ((catch 'lightlisp
		  (catch 'lightlisp
			 ,@body))
	   (validated ',symbol))
	  (t
	   (invalidated ',symbol)))))

(defmacro expect (result code)
  "write up some code to print out some stuff about result and generated-by"
  (let ((codename (gensym)) (execname (gensym)) (resultname (gensym)))
       `(let ((,codename ',code)
	      (,execname ,code)
	      (,resultname ,result))
	     (princ "Form: ") (pprint codename) (terpri)
	     (format t "Result: ~S~%Expected: ~S (~A)~%"
		     execname resultname (equalp execname resultname)))))

(defmacro error-recovery (&body body)
  "when LightLisp is bound, try the error-recovery bit"
  `(when LightLisp
	 (catch 'lightlisp
		,@body)))

(defun test (list &optional (examine-onlyp nil))
  "Foreach symbol in the list, load the file ``comprehensive/<symbol>.ll''
  and note failures.
  If just-examinep is non-nil, then the validation macro
  will just load, but not execute - allows a test-the-tester feature"
  (dolist (symbol list)
	  (let ((filename (concatenate 'string "validation/comprehensive/"
				       (symbol-name symbol) ".ll")))
	       (validation-for symbol)
	       (setq just-examinep examine-onlyp)
	       (if (not (load filename))
		   (bad-load symbol)))))

(defun successes ()
  "return the successful tests"
  (select #'successp validations))

(defun failures ()
  "return the not-successful tests"
  (select #'(lambda (v) (not (successp v))) validations))

;;;
;;;    Oblist management
;;;
(let ((ignore '(&body &optional &rest :direction :output :write)))
  (flet ((ignore-symbolp (symbol)
			 "see if the symbol should be ignored by this validation suite"
			 (member symbol ignore)))
	(defun untested-symbols ()
	       "return a list of untested symbols"
	       (let ((ret nil))
		    (dolist (symbol (oblist) ret)
			    (if (and (not (ignore-symbolp symbol))
				     (null (assoc symbol validations)))
				(setq ret (cons symbol ret))))))))

;;;
;;;    Useful functions
;;;
(defmacro while (test &body body)
  "while test is true evaluate body"
  `(do ()
      (,test)
      ,@body))

;;;
;;;    The Validation Parameters
;;;
(defparameter lightlisp-oblist nil
  "the functions to validate")

(defparameter LightLisp nil
  "if non-nil, then certain LightLisp-specific error recovery
  mechanisms will be tested.  If nil, then this can be run under CommonLisp")

(defparameter just-examinep nil
  "if nil, then run the whole test
  if non-nil, then just load the tester, but don't execute it")

;;;
;;;    The Validation itself ...
;;;
(format t "~
The LightLisp Test and Validation Suite~%~
    Wendell C. Baker~%~
~%~
    ~A - ~A~%~
" (RCS 'Revision) (RCS 'Date))

(test lightlisp-oblist)

(format t "~
The following tests succeeded: ~S~%~
The following tests failed:    ~S~%~
The following were untested:   ~S~%~
" (successes) (failures) (untested-symbols))
;)
