;;;;(
;;;;    The Destructuring Test
;;;;
;;;;    Guy Steele's CommonLisp page 146 ...  is the inspiration
;;;;       for this little suite of testers
;;;;

;;    in order to test (not in Steele)
(defmacro print-eval (&body forms)
  "make a progn that will print out the execution of each form"
  (cons 'progn
	(let ((code nil) tmp)
	     (dolist (f forms code)
		     (setq tmp `(format t
					"--------------------~
					~%form: ~S~
					~%    expansion: ~S~
					~%    result: ~S~
					~%--------------------~%~%"
					',f (macroexpand ',f) ,f))
		     ; get it in the right order (no matter how expensive)
		     (setq code (append code (cons tmp nil)))))))

;;;
;;;    The dolist test
;;;
(defparameter the-list '(apples oranges bananas cherries guavas avocados)
  "a list for use in use-dolist")

(defun use-dolist (list)
  "a use of dolist"
  (format t "The members of the list ~%    ~S~%are:~%" list)
  (let ((i 0))
       (dolist (m list)
	       (setq i (1+ i))
	       (format t "    member ~D is ~S~%" i m)))
  (princ "done") (terpri))

;;
;;    Before redefinition
;;
(use-dolist the-list)

;; cannot use defmacro to redefine a special form
;; but what you can do is unbind the special form first!
(fmakunbound 'dolist)
(defmacro dolist ((var listform &optional resultform) &rest body)
  "a replacement dolist"
  (let ((listvar (gensym))
	(genvar (gensym))
	(genret (gensym)))
       `(let (,genret)
	     (princ "The New Dolist!") (terpri)
	     (setq ,genret (do* ((,genvar ,listform (cdr ,genvar))
				 (,var (car ,genvar) (car ,genvar)))
				((null ,genvar) ,resultform)
				,@body))
	     (princ "The New Dolist Is for YOU!") (terpri)
	     ,genret)))

;;
;;    After redefinition, but before reredefinition
;;
(use-dolist the-list)

(macrolet ((dolist ((var listform &optional resultform) &rest body)
		   (let ((listfar (gensym))
			 (ivar (gensym))
			 (var (gensym)))
			`(progn
			  (princ "The Secret Do!")
			  (terpri)
			  (let ((,listvar ,listform) ,var)
			       (dotimes (,ivar ,length ,resultform)
					(setq ,var (nth ,ivar ,listvar))
					,@body))
			  (princ "The Secret Do Is for YOU (only)!")
			  (terpri)))))
  (use-dolist the-list))


;;;
;;;    Moving right along to the arithmetic-if (pageh 147)
;;;
(defmacro arithmetic-if (test neg-form zero-form pos-form)
  (let ((var (gensym)))
       `(let ((,var ,test))
	     (cond ((< ,var 0) ,neg-form)
		   ((= ,var 0) ,zero-form)
		   (t ,pos-form)))))

(print-eval
  (arithmetic-if (- 5 6) 'neg 'zero 'pos)
  (arithmetic-if (- 5 5) 'neg 'zero 'pos)
  (arithmetic-if (- 5 4) 'neg 'zero 'pos))

;;;
;;;    And again, moving right along to the infamous halibut macro
;;;
(defmacro halibut ((mouth eye1 eye2) ((fin1 length1) (fin2 length2)) tail)
  "cons up some code to print out the value of each variable"
  (cons 'progn
	(let* ((vars '(mouth eye1 eye2 fin1 length1 fin2 length2 tail))
	       (vals (list mouth eye1 eye2 fin1 length1 fin2 length2 tail))
	       (code nil)
	       tmp)
	      (dolist (v (pairlis vars vals) code)
		      (setq tmp `(format t "~S	~S~%" ',(car v) ',(cdr v)))
		      (setq code (cons tmp code))))))

(halibut (m (car eyes) (cdr eyes))
  ((f1 (count-scales f1)) (f2 (count-scales f2)))
  my-favorite-tail)

(catch 'lightlisp
  (princ "Expect an error...")
  (terpri)
  (princ "    not enough required arguments (fin1 length1)")
  (terpri)
  (halibut (m (car eyes) (cdr eyes))
	   ((f1) (f2 (count-scales f2)))
	   my-favorite-tail))

(catch 'lightlisp
  (princ "Expect an error...")
  (terpri)
  (princ "    formals and actuals must both be lists (mouth eye1 eye2)")
  (terpri)
  (halibut my-favorite-head
	   ((f1 (count-scales f1)) (f2 (count-scales f2)))
	   my-favorite-tail))

;;;
;;;    Test for invalid destructuring
;;;
(catch 'lightlisp
  (princ "Expect an error...")
  (terpri)
  (princ "    badly formed &optional spec")
  (terpri)
  (defmacro loser (x &optional (a b &rest c) &rest z)
	    `(progn
	      (princ "This should never execute (a loser)")
	      (terpri)))
  (loser 45))

(defmacro loser (x &optional ((a b &rest c)) &rest z)
  `(progn
    (princ "This should execute for some forms (a kindof-loser)")
    (terpri)
    (format t "x - ~S~%" ',x)
    (format t "a - ~S~%" ',a)
    (format t "b - ~S~%" ',b)
    (format t "c - ~S~%" ',c)
    (format t "z - ~S~%" ',z)
    (terpri)))

(loser (car pool) (1 2 3 4 5) 5 6 7)
(loser (car pool) (1 2 3 4 5))
(loser (car pool) (1 2))
(catch 'lightlisp
  (princ "Expect an error...")
  (terpri)
  (princ "    not enough required arguments (a b &rest c)")
  (terpri)
  (loser (car pool)))

(defmacro non-loser (x &optional ((a b &rest c) '(nil nil)) &rest z)
  "this should execute"
  (format t "A Non-Loser that should execute~%")
  (cons 'progn
	(let* ((vars '(x a b c z))
	       (vals (list x a b c z))
	       (code nil)
	       tmp)
	      (dolist (v (pairlis vars vals) code)
		      (setq tmp `(format t "~S	~S~%" ',(car v) ',(cdr v)))
		      (setq code (cons tmp code))))))

(print-eval
  (non-loser 1)
  (non-loser 1 (2 3))
  (non-loser 1 (2 3 3.1 3.2))
  (non-loser 1 (2 3) 4 5 6 7 8))

(defmacro non-loser1 (x &optional ((a b &rest c) '(nil nil)) &rest z)
  "this should execute also"
  (format t "A Nother-Non-Loser that should execute~%")
  (cons 'progn
	(let* ((vars '(x a b c z))
	       (vals (list x a b c z))
	       (code nil)
	       tmp)
	      (dolist (v (pairlis vars vals) code)
		      (setq tmp `(format t "~S	~S~%" ',(car v) ',(cdr v)))
		      (setq code (cons tmp code))))))

(print-eval
  (non-loser1 1)
  (non-loser1 1 (2 3))
  (non-loser1 1 (2 3 3.1 3.2))
  (non-loser1 1 (2 3) 4 5 6 7 8))

(setq pool '(a b c d e))
(setq x -8)

(non-loser (car pool) (+ x 1))
(catch 'lightlisp
  (princ "Expect an error...")
  (terpri)
  (non-loser1 (car pool) (+ x 1)))

;)
