;;;;(
;;;;    Regression testing for LightLisp
;;;;

(defparameter catch-tags '(lightlisp)
  "a list of the catch tags that are important
  for this regression tester")

(let ((last-regression '(0 . 0))
      (regression-sequence nil))
  (macrolet ((major-of (major-minor)
		       `(car ,major-minor))
	     (minor-of (major-minor)
		       `(cdr ,major-minor)))
	    (labels ((make-catches (body &optional (catch-tags catch-tags))
				   `(catch ',(car catch-tags)
					   ,@(if (null (cdr catch-tags))
						 body
						 (list (make-catches body (cdr catch-tags)))))))
		    (defmacro regression (major minor description &body test-forms)
			      "major and minor should be integers
			      test-forms may be any form at all"
			      (let ((this-regression (cons major minor))
				    (codelist nil)
				    (magic-key (gensym))
				    (form (gensym))
				    (result (gensym))
				    (catch-result (gensym))
				    tmp)
				   ;; insure that all regression tests are in the correct order
				   (when (or (< (major-of this-regression) (major-of last-regression))
					     (and (= (major-of this-regression) (major-of last-regression))
						  (<= (minor-of this-regression) (minor-of last-regression))))
					 (format t "Regression Error~%    ~
						 regression ~S is not increasing in the regression sequence~%    ~
						 last regression: ~S~%regressions: ~S~%"
						 this-regression last-regression regression-sequence)
					 (throw 'regression nil))
				   ;; update the bookkeeping
				   (setq last-regression this-regression)
				   (setq regression-sequence (cons last-regression regression-sequence))
				   ;; generate the code
				   (dolist (tf test-forms)
					   (setq tmp (let* ((catch-body `((setq ,result ,tf)
									  ',magic-key))
							    (catch-form (make-catches catch-body)))
							   `(let ((,form ',tf)
								  ,result
								  ,catch-result)
								 (format t "Form:~%    ") (pprint ,form) (terpri)
								 (setq ,catch-result
								       ,catch-form)
								 (format t "Result:~%    ") (pprint ,result) (terpri)
								 (unless (eq ',magic-key ,catch-result)
									 (princ "Warning: a throw occurred (the result is from the throw)")
									 (terpri))
								 (terpri)
								 (terpri))))
					   ;; use nconc to append it to the list
					   (setq codelist (cons tmp codelist)))
				   ;; reverse the codelist
				   (labels ((reverse (list)	; not flet
						     (cond ((null list) nil)
							   ((null (cdr list)) list)
							   (t (nconc (reverse (cdr list)) (cons (car list) nil))))))
					   (setq codelist (reverse codelist)))
				   ;; return the progn
				   `(progn
				     (format t "-----~%Regression Test ~S.~S~%~%Description:~%~A~%~%"
					     ,(major-of this-regression)
					     ,(minor-of this-regression)
					     ,description)
				     ,@codelist
				     (format t "-----~%~%~%~%")))))))

(dotimes (i 2) (terpri))
(princ lisp-implementation-version)
(dotimes (i 3) (terpri))

;;;
;;;    Release 1 --------------------
;;;
(regression 1 0
  "trivial - it ran")

(regression 1 1
  "insure that result-type in concatenate, coerce, and map is always a symbol.
  LLApplyC correctly causes an error on too-few-args."
  (concatenate "yes" "no")			; result-type is a string
  (coerce '("y" "e" "s") "no")			; result-type is a string
  (map "yes" #'(lambda (x) x) '("y" "e" "s"))	; result-type is a string
  (cons 'a)	; too few args
  (progn
   (defun fig (r &optional s)
	  (cons r s))
   (fig 1)
   (fig)))	; too few args 

(regression 1 2
  "be able to coerce integers and reals into strings"
  (coerce 4 'string)		; coerce an integer to a string
  (coerce 4.4 'string))		; coerce a real to a string

(regression 1 3
  "(format nil ... ) now returns the string instead of printing on stderr"
  (format nil "this is a string ~S" "yes a string"))	; return the string

(regression 1 4
  "(listp nil) now returns t
  added LLNot (untestable from lisp)
  added LLListp (untestable from lisp)"
  (listp nil))	; return t

(regression 1 5
  "fixed a bug in let/let*/flet/labels/macrolet
  (all were installing in the functionSlot)"
  ;; this is implemented as a major part of two-slot validation
  ; first with let
  (let ((re-plast '(f g h i j)))
       (defun re-plast (&optional (desc "executed"))
	      (format t "re-plast - ~A~%" desc)
	      (format t "re-plast is ~S~%" re-plast))
       (re-plast "locally"))
  
  (re-plast)
  
  (let ((be-plast '(f g h i j)))
       (flet ((be-plast (&optional (desc "executed"))
			(format t "be-plast is ~S~%" be-plast)))
	     (be-plast)))
  
  ; then with let*
  (let* ((ye-plast '(f g h i j)))
	(defun ye-plast (&optional (desc "executed"))
	       (format t "ye-plast - ~A~%" desc)
	       (format t "ye-plast is ~S~%" ye-plast))
	(ye-plast "locally"))
  
  (ye-plast)
  
  (let* ((me-plast '(f g h i j)))
	(flet ((me-plast (&optional (desc "executed"))
			 (format t "me-plast - ~A~%" desc)
			 (format t "me-plast is ~S~%" me-plast)))
	      (me-plast)))
  
  ; then with flet
  (let* ((pe-plast '(f g h i j)))
	(labels ((pe-plast (&optional foodle)
			   (format t "this should not be seen~%")
			   'error))
		(flet ((pe-plast (&optional (desc "executed"))
				 (format t "pe-plast - ~A~%" desc)
				 (format t "pe-plast is ~S~%" pe-plast)))
		      (pe-plast))))
  
  ; then with labels
  (let* ((que-plast '(f g h i j)))
	(flet ((que-plast (&optional foodle)
			  (format t "this should not be seen~%")
			  'error))
	      (labels ((que-plast (&optional (desc "executed"))
				  (format t "que-plast - ~A~%" desc)
				  (format t "que-plast is ~S~%" que-plast)))
		      (que-plast))))
  
  ; then with macrolet
  (let* ((we-plast '(f g h i j)))
	(flet ((we-plast (&optional foodle)
			 (format t "this should not be seen~%")
			 'error))
	      (macrolet ((we-plast (&optional (desc "executed"))
				   `(progn
				     (format t "we-plast - ~A~%" ,desc)
				     (format t "we-plast is ~S~%" we-plast))))
			(we-plast)))))

(regression 1 6
  "fixed a bug in flet (was calling LLMacroClosure)"
  (flet ((list (&rest args)
	       (map 'list #'(lambda (x) (if (consp x) (car x))) args)))
	(list '((flowchart posteriori bypass mensuration)
		(sequestration testate exaltation acuity)
		(hayward blacken knot namesake)
		(anticipatory primitive sewerage intemperance)
		(l Neapolitan edition profession)
		(graduate barre circumferential dialectic)
		(pica nepenthe eigenvector promptitude)))))

(regression 1 7
  "~S in LLFormatC now uses LL_prin1 (as suitable for (read ... ))
         (untestable from lisp)
   ~S in LLFormatS now uses LL_prin1 (as suitable for (read ... ))"
  (let ((all-types '(1003 1.3093e-4 "yes\nand\nno" 'apple :keyword)))
       (format t "Original: ~S~%Format: ~A~%"
	       all-types
	       (format nil "~S" all-types))))

(regression 1 8
  "The reader now barfs on  { match this ... ( } ')
  (not testable from lisp)
  (dolist (v nil) ... ) now works"
  (dolist (v nil)
	  (format t "This should not print~%")))

(regression 1 9
  "(defun foodle (r) r)  (funcall #'foodle 's) now works"
  (progn
   (defun foodle (r)
	  r)
   (funcall #'foodle 's)))

(regression 1 10
  ",@nil should not substitute anything (not barf)"
  (let ((a '(first-piece second-piece)) (b nil))
       `'(,b ,@b ,@a ,@b ,b)))

(regression 1 11
  "rescan macro-closure cells correctly across gc
       (not directly testable from a script b/c gc must intervene)"
  (progn
   (defmacro a (b c)
	     `(list ,b ,c))
   ;; must gc here
   (macroexpand '(a 9 10))))

(regression 1 12
  "catch returns the return value of the last member executed"
  (let ((rat 'scum))
       (catch 'here
	      (setq rat 'bat))
       rat)
  (catch 'foodle
	 (let ((noodle 2))
	      (dotimes (i 10)
		       (+ noodle 222 333))
	      (throw 'foodle noodle)
	      (setq noodle 'this-should-not-happen))))

(regression 1 13
  "equal on lists should not infinite-recur"
  (equal '(a b) '(a b c))
  (equal '(a b c) '(a b c)))

(regression 1 14
  "defvar, and defconstant would only setq on symbols that were already defined"
  (progn
   (setq fiddle 1)
   (defvar fiddle 2)
   fiddle)
  (progn
   (setq faddle 3)
   (defconstant faddle 4)
   faddle)
  (progn
   (format t "poodle is ~A~%" (if (boundp 'poodle) "already defined" "undefined"))
   (defvar poodle 5)
   poodle)
  (progn
   (format t "piddle is ~A~%" (if (boundp 'piddle) "already defined" "undefined"))
   (defconstant piddle 6)
   piddle))

(regression 1 15
  "labels defines a recursive and mutually recursive scope
  flet on the other hand does not"
  (labels ((ones (list)
		 (if (null list)
		     nil
		     (cons 1 (cdr list)))))
	  (apply #'+ (ones '(a b c d e f g h i j))))
  (labels ((5+ (number)
	       (1+ (4+ number)))
	   (4+ (number)
	       (1+ (3+ number)))
	   (3+ (number)
	       (1+ (2+ number)))
	   (2+ (number)
	       (1+ (1+ number)))
	   (1+ (number)
	       (+ 1 number)))
	  (+ (5+ 5) (5+ 5)))
  (labels ((ones (list)
		 (if (null list)
		     nil
		     (cons 1 (ones (cdr list))))))
	  (flet ((ones (list)
		       (if (null list)
			   nil
			   (cons 100 (ones (cdr list))))))
		(apply #'+ (ones '(a b c d e f g h i j))))))

(regression 1 16
  "continuing with 1.2 - coerce works as advertized for strings and lists"
  (coerce 4 'string)		; coerce an integer to a string
  (coerce 4.4 'string)		; coerce a real to a string
  (coerce "foodle" 'string)	; coerce a string to a string
  (coerce '("pig" "pag" "pug" "pog" "pid" "pod") 'string)	; coerce a list of strings to a string
  (coerce "unboundoid" 'list)	; coerce a string to a list
  (coerce '(a 2 3.3 #'cons) 'list))	; coerce a list to a list

(regression 1 17
  "concatenate works as advertized for lists - it copies all members
  some error messages in append were changed (when members are not lists)"
  (concatenate 'string "apple" " " "banana" "cherry" " - " " yes")
  (concatenate 'list '(:apple :banana :cherry) '(:date :eclair) nil nil '(:fruit :guava))
  (concatenate 'list '(a b c . d) '(e f g h) nil '(e . j))
  (concatenate 'list 1 nil)
  (append 1 nil))

(regression 1 18
  "that do, and do* can have (end-test . result) where result can be a list (a progn)"
  (do ((i 1 (1+ i)))
      ((= i 4)
       (setq i (1+ i))	; four of them
       (setq i (1+ i))
       (setq i (1+ i))
       (setq i (1+ i)) i))
  (do* ((i 1 (1+ i)))
      ((= i 4)
       (setq i (1+ i))	; four of them
       (setq i (1+ i))
       (setq i (1+ i))
       (setq i (1+ i)) i))
  (do ((i 1 (1+ i)))
      ((= i 4) (+ i 4)))
  (do* ((i 1 (1+ i)))
      ((= i 4) (+ i 4))))

(regression 1 19
  "pprint of a string handles quotes correctly, ~S in format too"
  (pprint "a string")
  (pprint "\"a quoted string\"")
  (format t "~S" "a string")
  (format t "~S" "\"a quoted string\"")
  (pprint (format nil "~S" "a string"))
  (pprint (format nil "~S" "\"a quoted string\"")))

(regression 1 20
  "printing output-only streams works correctly"
  (prin1 *standard-output*)
  (prin1 *error-output*)
  (format t "~S" *standard-output*)
  (format t "~S" *error-output*))

(regression 1 21
  "added streamp input-stream-p and output-stream-p"
  (streamp *standard-output*)
  (streamp *error-output*)
  (input-stream-p *standard-output*)
  (input-stream-p *error-output*)
  (output-stream-p *standard-output*)
  (output-stream-p *error-output*))

(regression 1 22
  "*error-output* now is bound to LLErrorStream (not LLStandardStream)"
  *standard-output*
  *error-output*)

(regression 1 23
  "result-types may be symbols or types for concatenate, coerce, map"
  (coerce 4 'string)	
  (coerce 4 (type-of "string"))
  (concatenate 'string "apple" " " "banana" "cherry" " - " " yes")
  (concatenate (type-of "string") "apple" " " "banana" "cherry" " - " " yes")
  (map 'list #'(lambda (x) (cons 'touched x))
       '(:apple :banana :cherry) '(:date :eclair) nil nil '(:fruit :guava))
  (map (type-of "string") #'(lambda (x) (cons 'touched x))
       '(:apple :banana :cherry) '(:date :eclair) nil nil '(:fruit :guava)))

(regression 1 24
  "use LLPprint for LLShell (untestable from here)")

(regression 1 25
  "added the time function"
  (time (+ 1 1))
  (time (dotimes (i 10000)
		 (+ 1 1))))

(regression 1 26
  "coerce will look for both sequence keys and type keys now"
  (coerce '("a" "b" "c") 'string))
  (coerce '("a" "b" "c") 'cons)
  (coerce '("a" "b" "c") 'list))

(regression 1 27
  "the caar ... system now reads from the inside out not vice versa"
  ;;
  (caar '((yes no-1) no-2 no-3))
  (cadr '((no-1 no-2) yes no-3))
  (cdar '((no-1 yes-list ... ) no-2 no-3))
  (cddr '((no-1 no-2) no-3 yes-list ... ))
  ;;
  (caaar '(((yes no-1) no-2) no-3))
  (caadr '(no-1 ((yes no-2) no-3) no-4))
  (cadar '((no-1 (yes no-2) no-3) no-4))
  (caddr '(no-1 no-2 yes no-3))
  (cdaar '(((no-1 yes-list ... ) no-2) no-3))
  (cdadr '(no-1 (no-2 yes-list ... ) no-3))
  (cddar '((no-1 no-2 yes-list ... ) no-3))
  (cdddr '(no-1 no-2 no-3 yes-list ... ))
  ;;
  (caaaar '((((yes no-1) no-2) no-3) no-4) no-5)
  (caaadr '(no-1 (((yes no-2) no-3) no-4) no-5))
  (caadar '((no-1 ((yes no-2) no-3) no-4) no-5))
  (caaddr '(no-1 no-2 ((yes no-3) no-4) no-5))
  (cadaar '(((no-1 (yes no-2) no-3) no-4) no-5))
  (cadadr '(no-1 (no-2 yes no-3) no-4))
  (caddar '((no-1 no-2 yes no-3) no-4))
  (cadddr '(no-1 no-2 no-3 yes no-4))
  (cdaaar '((((no-1 yes-list ... ) no-2) no-3) no-4))
  (cdaadr '(no-1 ((no-2 yes-list ... ) no-3) no-4))
  (cdadar '((no-1 (no-2 yes-list ... ) no-3) no-4))
  (cdaddr '(no-1 no-2 (no-3 (yes-list ... ) no-4) no-5))
  (cddaar '(((no-1 no-2 yes-list ... ) no-3) no-4))
  (cddadr '(no-1 (no-2 no-3 yes-list ... ) no-4))
  (cdddar '((no-1 no-2 no-3 yes-list ... ) no-4))
  (cddddr '(no-1 no-2 no-3 no-4 yes-list ... )))

(regression 1 28
  "call LLTerpri after LLPprint in LLShell and LLSubshell - can't demo from here")

(regression 1 29
  "Fixed a bug in LLCatchBegin that caused unwind-protect
  (or any other promise-based routine) to not get their promises evaluated."
  (unwind-protect (princ "protected")
		  (prin1 " cleanup1")
		  (prin1 " cleanup2")
		  (terpri))
  (progn
   (unwind-protect (setq x '(this should never happen))
		   (setq x 'this-should-happen))
   x))

(regression 1 30
  "added the LL_type_of macro so that LL_nil could get its type
  correctly.  All code should be written as LL_type_of(obj), not
  obj->type, for LL_nil may be implemented as 0x0.
  This should make nil work for many routines where it would
  otherwise coredump"
  (coerce nil 'list)
  (coerce nil 'cons)
  (concatenate 'list nil nil nil nil nil nil)
  (copy-seq nil)
  (elt nil 0)
  (eql nil nil)
  (eql nil 1)
  (equal nil nil)
  (equal 3 nil)
  (length nil)
  (length '(a b c))
  (prin1 nil)
  (prin1 '(a))
  (princ nil)
  (princ '(b))
  (print nil)
  (print '(b))
  (pprint nil)
  (pprint '(b))
  (subseq nil 0 0)
  (type-of nil)
  (map 'list #'identity nil))

(regression 1 31
  "insure that &optional init-forms are evaluated in the
  environment of the lambda"
  (let ((outer-value 338821))
       (flet ((calculate (list &optional (length (length list)) (random outer-value))
			 (format t "Calculate:~%    list - ~S~%    length - ~S~%    random - ~S~%"
				 list length random)))
	     (calculate '(a b c d e))
	     (calculate nil)
	     (calculate '(r s t u v) 90000)
	     (calculate nil 100)
	     (calculate '(m n o p) 100100 300300)
	     (calculate 'nil 777888 12345))))

(regression 1 32
  "random now checks correctly for positivity of its first argument"
  (random 45)
  (random -45))

(regression 1 33
  "fixed a problem with div - if the two were divisible,
  the contagion prevented any division from being performed"
  (/ 4 4)
  (/ 4 2)
  (/ 2 2)
  (/ 2 3)
  (/ 3 3))

(regression 1 34
  "(random <integer>) calls LLMakeInteger, not LLMakeReal resulting
  in a correct result"
  (defmacro type-value (exp)
	    `(list ,exp (type-of ,exp)))
  (type-value (random 1000))
  (type-value (random 1000.0))
  (type-value (random 10)))

(regression 1 35
  "if there were no stepforms in a do, or do*, then the initial binding
  was not properly being carried across iterations of the loop"
  (do* ((end nil)
	(b 0))
       (end b)
       (princ b)
       (princ "-")
       (setq b (1+ b))
       (if (= b 10)
	   (setq end t)))
  (do ((end nil)
       (b 0))
      (end b)
      (princ b)
      (princ "-")
      (setq b (1+ b))
      (if (= b 10)
	  (setq end t))))

(regression 1 36
  "apply was producing the wrong argument set for one-arg functions"
  (apply #'> '(5 4 3 2 1))
  (apply #'< '(5 4 3 2 1))
  (apply #'(lambda (rat) (list rat rat rat rat)) 'regular nil)
  (apply #'(lambda (rat) (list rat rat rat rat)) '(regular))
  (apply #'(lambda (rat) (list* rat rat rat rat)) 'dotted nil)
  (apply #'(lambda (rat) (list* rat rat rat rat)) '(dotted)))

(regression 1 37
  "format for ~newline ignores all whitespace after the newline"
  (format t "~
	  And things - ~
		And other things~%")
  (format t "Only one newline~%~&~&~&~&~&~&~&"))

(regression 1 38
  "fixed a bug in LLPrintFullStats (was using ~S instead of ~A)
  not entirely demonstrable here unless we over-cons"
  (time (dotimes (i 100000) (cons 1 1))))

(regression 1 39
  "uses a more rational flonum printing strategy %lf %le %lg"
  (dolist (f `(.1 .222  10.3034 8.309 21.8e-4 .4e8 ,pi ,e 338e34))
	  (format t "A f-float ~F~%A e-float ~E~%A g-float ~G~%" f f f)))

(regression 1 40
  "added read and read-from-string which means LLReadValue changed
  in the number of arguments that it takes"
  (read-from-string "orange")
  (read-from-string "(apple banana cherry)")
  (read-from-string "(defun foodle (poodle) (cons 'kit (cons 'kaboodle poodle)))")
  (eval (read-from-string "(defun foodle (poodle) (cons 'kit (cons 'kaboodle poodle)))"))
  (poodle 45)
  (read-from-string "" nil nil)
  (read-from-string "" nil 5)
  (read-from-string "" nil 'eof)
  (read-from-string "" t 'eof)
  ;;
  (setq file (open "/tmp/fopple" :direction :output))
  (princ "orange~%(apple banana cherry) ~
	 (defun noodle (spoon) (cons 'pesto (cons 'tomato spoon)))~%~
	 (defun noodle (spoon) (cons 'pesto (cons 'tomato spoon)))" file)
  (close file)
  (setq fig (open "/tmp/fopple" :direction :input))
  (read fig)
  (read fig)
  (read fig)
  (eval (read fig))
  (noodle 45)
  ; at eof now
  (read fig nil nil)
  (read fig nil 5)
  (read fig nil 'eof)
  (read fig t 'eof)
  (close fig))

(regression 1 41
  "print the line number in LLSource and LLSourceString before quitting"
  (read-from-string "(this is\nthe\beginning\nof\nthe\nlist - no end!"	; close-paren missing )
		    nil nil)
  (let ((file (open "/tmp/yipple" :direction :output)))
       (princ  "(this is\nthe\beginning\nof\nthe\nlist - no end!"  ; close-paren missing )
	       file)
       (close file)
       (load "/tmp/yipple")))

(regression 1 42
  "do with no initializers was illegal"
  (do ((a 1))
      ((null a) (terpri))
      (if (= a 10)
	  (setq a nil)
	  (setq a (1+ a)))
      (princ a) (princ "-"))
  (let ((a 1))
       (do ()
	   ((null a) (terpri))
	   (if (= a 10)
	       (setq a nil)
	       (setq a (1+ a)))
	   (princ a) (princ "-"))))

(regression 1 43
  "subshell now prints the evaluated value, not the value that was read
  (have to take my word for this one (can't go interactive))")

(regression 1 44
  "do initializer with var only (and no initializer) now works"
  (do ((apple 1)
       (banana))
      ((null apple) (terpri))
      (setq banana (* 100 apple))
      (if (= apple 10)
	  (setq apple nil)
	  (setq apple (1+ apple)))
      (princ apple) (princ "/") 
      (princ banana) (princ "-")))

(regression 1 45
  "if any of the c[ad]{2,3}r combinations resulted in a atom they failed"
  (caar '((CELLTYPE "LEAF")))
  (cadr '(CELLTYPE "LEAF"))
  (cdar '((CELLTYPE "LEAF")))
  (cddr '(CELLTYPE "LEAF" rest ... )))

(regression 1 46
  "don't print environment for backquote/comma evaluation failure"
  `(apples ,'oranges)
  `(apples ,oranges))

(regression 1 47
  "it is admissible to close a closed stream"
  (let ((file (open "/tmp/flop" :direction :output)))
       (close file)
       (close file)
       (close file)))

(regression 1 48
  "type-of returns the print-name of the object's type"
  (type-of nil)
  (type-of '(a))
  (type-of 4)
  (type-of 4.5)
  (type-of 'a4)
  (type-of (type-of 4))
  (type-of #'car))

(regression 1 49
  "defconstant wasn't even calling LLSetValue"
  (defconstant bobby 3)
  bobby
  (defconstant fern (+ 1 3))
  fern)

(regression 1 50
  "insure destructuring works for &optional destructurings too"
  (defmacro mush (x &optional ((a b &rest c) '(nil nil)) &rest z)
	    "this should execute"
	    (cons 'progn
		  (let* ((vars '(x a b c z))
			 (vals (list x a b c z))
			 (code nil)
			 tmp)
			(dolist (v (pairlis vars vals) code)
				(setq tmp `(format t "~S   ~S~%" ',(car v) ,(cdr v)))
				(setq code (cons tmp code))))))
  (mush 1)
  (mush 1 (2 3))
  (mush 1 (2 3 4))
  (mush 1 (2 3 4 5))
  (mush 1 (2 3 4 5 6))
  (mush 1 (2 3 4 5 6) 7 8))

(regression 1 51
  "if progn is given a dotted list (or an atom) as its arg, then nil"
  (do* ((end nil)
	(b 0))
       (end . b)
       (princ b)
       (princ "-")
       (setq b (1+ b))
       (if (= b 10)
	   (setq end t)))
  (do ((end nil)
       (b 0))
      (end . b)
      (princ b)
      (princ "-")
      (setq b (1+ b))
      (if (= b 10)
	  (setq end t)))
  (cond (t . 4))
  (cond (nil 45)
	(t . 4)))

(regression 1 52
  "format ~& prints out the &"
  (format t "Only one newline~%~&~&~&~&~&~&~&"))

(regression 1 53
  "LLShell now interprets unknown args as filenames to source
  but that is hardly testable from here")

(regression 1 54
  "gc now drops integers that are unused, but that is hardly
  demonstrable from here.")

(regression 1 55
  "make LLErrorF flush LLStandardStream before continuing
  but this is barely demonstrable from here")

(regression 1 56
  "format now returns nil or a string at the correct times"
  (format t "Format to standard-output~%")
  (format *error-output* "Format to error-output~%")
  (format nil "Format to a string~%"))

(regression 1 57
  "pprint now knows about 80 char lines"
  ;; typically the following would look icky without linebreaks
  (do* ((end nil)
	(b 0))
       (end . b)
       (princ b)
       (princ "-")
       (setq b (1+ b))
       (if (= b 10)
	   (setq end t)))
  (do ((end nil)
       (b 0))
      (end . b)
      (princ b)
      (princ "-")
      (setq b (1+ b))
      (if (= b 10)
	  (setq end t)))
  (cond (t . 4))
  (cond (nil 45)
	(t . 4)))

(regression 1 58
  "#' now works on lexically bound functions"
  (flet ((stringify (thing)
		    (cond ((symbolp thing)
			   (symbol-name thing))
			  ((numberp thing)
			   (coerce thing 'string))
			  ((stringp thing)
			   thing)
			  (t
			   (format nil "~S" thing)))))
	(dolist (t `(a 11 1.11 "yes" ,*standard-output*))
		(apply #'stringify t))))

(regression 1 59
  "prin1 princ print all understand about , ,@ and ` readmacros"
  (let ((objects
	 (list 1 1.1
	       "a string"
	       *standard-output*
	       #'car
	       nil
	       '(a b c d e f g h i j k l m n o p q r s t u v w x y z
		   A B C D E F G H I J K L M N O P Q R S T U V W X Y Z)
	       '(apple (papple bananana yipple papple fopple ) cherry)
	       '(halibut (m (car eyes) (cdr eyes))
			 ((f1 (count-scales f1)) (f2 (count-scales f2)))
			 my-favorite-tail)
	       '(A list `(with some backquoted ,stuff and  ,o ther stuff))
	       '(catch 'lightlisp
		       (princ "Expect an error...")
		       (terpri)
		       (halibut my-favorite-head
				((f1 (count-scales f1)) (f2 (count-scales f2)))
				my-favorite-tail)))))
       (dolist (printer (list #'prin1 #'princ #'print #'pprint))
	       (format t "Printer: ~S~%------------------------------~%~%" printer)
	       (dolist (atom objects)
		       (funcall printer atom))
	       (terpri)
	       (terpri)
	       (terpri))))

(regression 1 60
  "macroexpand works until it sees an unbound or function symbol"
  (defmacro d (type class name)
	    "return this neato thing
	    `(define ,type ,class ,name))
  (macroexpand '(d apple banana cherry))
  (defmacro define (type class name)
	    "return a better neato thing"
	    `((type . ,type) (class . ,class) (name . ,name)))
  (macroexpand '(d apple banana cherry)))

(regression 1 61
  "fixed a bug in prin1, princ and print with respect to
  printing forms with ' , ` ,@ in them"
  (setq terrible '(a list (a gist) for this is it
		     `((type . ,type) (class . ,class) (name . ,name))
		     `((type . ,@type-list) (class . ,@class-list) (name . ,@name-list))
		     (#'car #'boat #'train)
		     'car 'boat 'train))
  (prin1 terrible)
  (princ terrible)
  (print terrible)
  (pprint terrible))

(regression 1 62
  "backquotes and commas in dotted lists now print correctly
  this is just a bugfix on top of 1.61 really"
  (setq terrible '(a list (a gist) for this is it
		     `((type . ,type) (class . ,class) (name . ,name))
		     `((type . ,@type-list) (class . ,@class-list) (name . ,@name-list))
		     (#'car #'boat #'train)
		     'car 'boat 'train))
  (prin1 terrible)
  (princ terrible)
  (print terrible)
  (pprint terrible))

(regression 1 63
  "do and do* now use a single environment over their life
  this makes the stepforms work correctly as well as
  lexical closures made inside the do work correctly"
  (let ((funcs (do ((i 0 (1+ i))
		    (ret nil))
		   ((= i 10) ret)
		   (setq ret (cons #'(lambda () i) ret)))))
       (dolist (f funcs)
	       (format t "~d~%" (funcall f)))
       (apply #'equal (map 'list #'funcall funcs))))

(regression 1 64
  "defun and defmacro would barf if the functionSlot was
  previously unbound, but the valueSlot was bound"
  (progn
   (makunbound 'shipwhicky-foodle-fowpoidal-and-things)
   (fmakunbound 'shipwhicky-foodle-fowpoidal-and-things)
   (defvar shipwhicky-foodle-fowpoidal-and-things 49
	   "nobody will be using a name like this"))
  (defun shipwhicky-foodle-fowpoidal-and-things ()
	 "nobody will be using a name like this"
	 49))

(regression 1 65
  "oddp was returning evenp"
  (map 'list #'oddp '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'evenp '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'zerop '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'minusp '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'plusp '(-4 -3 -2 -1 0 1 2 3 4)))

(regression 1 66
  "re-fixed the goddamned pprint; was printing the car-slot only for dotted pairs"
  (pprint '(a . b))
  (pprint '((a . b)
	    (c . d)
	    (e . f)
	    (g . h)))
  (pprint '(a b c d e f g . h))
  (pprint '(a (b c . d) e f (g h i (j k . l)) . m)))

(regression 1 67
  "yet another attempt at getting pprint to do the right thing
  attempt to insert line-breaks before recurring"
  (pprint
   '(regression 1 2
		"OLObjectReform now does a octGetById in order to refresh the fields.
		OLObjectReform and OLObjectForm now behave correctly with respect
		to indexing into the octObject field tables"
		(let* ((facet (open-facet-by-name "/tmp/flop" "symbolic" "contents" "" :readwrite-append))
		       (prop (create facet `(object ,OCT-PROP
						    ((name . "SAM-I-AM")
						     (value . "GREEN-EGGS"))))))
		      (object-of facet)))))

(regression 1 68
  "fixed the error printfs in format to be more informative"
  (format 3 "this won't work~%")
  (format '(a pple banana cherry) "neither will this")
  (format *standard-output* "This will work~%"))

(regression 1 69
  "added LLSourceFile (not directly testable from here)")

(regression 1 70
  "subshell now flushes LLStandardStream and LLErrorStream before starting up
  (this is not directly demonstrable from here)")


(regression 1 71
  "added LLProbeValueP (not testable from here)")

(regression 1 72
  "strings and symbols were clashing (case sensitive vs sensitive)
  added a string table and LLInternString - moved LLGenerateString
  to LLGenerateSymbol")

(regression 1 73
  "added LLSetCaseSensitivity (not directly testable from here)")

(regression 1 74
  "open :direction :output now uses "w" and fopen
  also, LLApplyC was ignoring extra args to fixed-argument functions"
  (let ((filename "/tmp/fink-dink-pink"))
       (let ((s (open filename :direction :output)))
	    (format s "\"Four Score and Seven Years Ago\")
	    (close s))
       (let ((s (open filename :direction :input)))
	    (pprint (read s nil :eof-unexpected nil))
	    (pprint (read s nil :eof-expected nil))
	    (close s))
       ; truncate it
       (let ((s (open filename :direction :output)))
	    (close s))
       (let ((s (open filename :direction :input)))
	    (pprint (read s nil :eof-expected nil))))
  (cons 1 2)
  (cons 1)
  (cons 1 2 3))

(regression 1 76
  "free of bogus pointer (LLNilCell->sequence) in LL_Mach_Uninitialize
  (not testable from here)")

(regression 1 77
  "use &end instead of end in the LLGetFullStats memory calculation
  (not demonstrable here)")

(regression 1 78
  "let, let*, flet, labels and macrolet can have a nil initializer"
  (let ()
	'ok)
  (let* ()
	'ok)
  (flet ()
	'ok)
  (labels ()
	  'ok)
  (macrolet ()
	    'ok))

(regression 1 79
  "mod and rem were bogus"
  (mod 13 4)
  (mod -13 4)
  (mod 13 -4)
  (mod -13 -4)
  (rem 13 4)
  (rem -13 4)
  (rem 13 -4)
  (rem -13 -4)
  (map 'list #'(lambda (i) (mod i 3)) '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'(lambda (i) (mod i -3)) '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'(lambda (i) (rem i 3)) '(-4 -3 -2 -1 0 1 2 3 4)))

(regression 1 80
  "copy-list and copy-tree had problems with dotted lists"
  (copy-list '(a b))
  (copy-list '(a . b))
  (copy-list '(a b c d e f))
  (copy-list '(a b c d e . f))
  (copy-tree '(a b))
  (copy-tree '(a . b))
  (copy-tree '(a b c d e f))
  (copy-tree '(a b c d e . f))
  (copy-tree '((a b) (c d)))
  (copy-tree '((a . b) (c . d)))
  (copy-tree '((a b c d e f) (a b c d e f)))
  (copy-tree '((a b c d e . f) (a b c d e . f))))

(regression 1 81
  "equal vs equalp should be case-sensitive vs case-insensitive"
  (progn
   (setq small "frobbery")
   (setq big "FROBBERY"))
  (equal small small)
  (equal big big)
  (equal small big)
  (equal (copy-seq small) (copy-seq big))
  (equalp small small)
  (equalp big big)
  (equalp small big)
  (equal (copy-seq small) (copy-seq big)))

(regression 1 82
  "map with a nil result-type coredumped"
  (map 'list (symbol-function '+) '(1 2 3) '(2 4 6))
  (map nil (symbol-function '+) '(1 2 3) '(2 4 6)))

(regression 1 83
  "insure that mod works for negatives
  insure that mod and rem work for both reals and integers alike"
  (mod 13 4)
  (mod -13 4)
  (mod 13 -4)
  (mod -13 -4)
  (mod 13.4 1)
  (mod -13.4 1)
  (mod 13.4 -1)
  (mod -13.4 -1)
  (rem 13 4)
  (rem -13 4)
  (rem 13 -4)
  (rem -13 -4)
  (rem 13.4 1)
  (rem -13.4 1)
  (rem 13.4 -1)
  (rem -13.4 -1)
  (map 'list #'(lambda (i) (mod i 3)) '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'(lambda (i) (mod i -3)) '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'(lambda (i) (rem i 3)) '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'(lambda (i) (rem i -3)) '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'(lambda (i) (mod i 3.0)) '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'(lambda (i) (mod i -3.0)) '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'(lambda (i) (rem i 3.0)) '(-4 -3 -2 -1 0 1 2 3 4))
  (map 'list #'(lambda (i) (rem i -3.0)) '(-4 -3 -2 -1 0 1 2 3 4)))

(regression 1 84
  "LLStandardStream and LLErrorStream don't have their input-sides
  named <input> any more"
  *standard-output*
  *error-output*)

(regression 1 85
  "copy-tree and copy-list had bugs with dotted-lists"
  (copy-tree 'a)
  (copy-tree '(a . b))
  (copy-tree '((a . b) (c . d)))
  (copy-tree '((a b c . d) (e f g . h)))
  (copy-list '(a . b))
  (copy-list '((a . b) (c . d)))
  (copy-list '((a b c . d) (e f g . h))))

(regression 1 86
  "first, second ... tenth if given a dotted list or non-cons
  would seg-fault (the distinction between ~S and ~A)"
  (first 'a)
  (first '(a . b))
  (second 'a)
  (second '(a . b))
  (third 'a)
  (third '(a . b))
  (fourth 'a)
  (fourth '(a . b))
  (fifth 'a)
  (fifth '(a . b))
  (sixth 'a)
  (sixth '(a . b))
  (seventh 'a)
  (seventh '(a . b))
  (eighth 'a)
  (eighth '(a . b))
  (ninth 'a)
  (ninth '(a . b))
  (tenth 'a)
  (tenth '(a . b)))

(regression 1 87
  "treat EOF as a token terminator"
  (read-from-string "1")
  (read-from-string "a"))

(regression 1 88
  "use EOF as a token terminator"
  (mod 0 1)
  (mod 0 1.0)
  (rem 0 1)
  (rem 0 1.0))

;)
