#if ! defined(LightLispMachine_h)
#define LightLispMachine_h

/*
 *    The (virtual) LightLisp machine
 *
 *    This is only really designed for use by knowledgeable types.
 *    The gotcha is that you cannot hold pointers into the heap
 *    across garbage collections.   Because this is tempting to do,
 *    only the wary should access the heap.
 */

typedef String LLstring;	/* a == b <=> strequal(a, b) */

extern void LLCaseSensitivity();	/* must be called BEFORE LLInitialize() */

extern LLstring LLInternSymbol();
extern LLstring LLGenerateSymbol();

extern String LLInternString();

/*
 *    LLobject *
 *    coerce(sequence, result_type)
 *    LLobject *sequence;
 *    LLobject *result_type;
 *
 *    LLobject *
 *    concatenate(result_type, sequences)
 *    LLobject *result_type;
 *    LLobject *sequences;		- &rest
 *
 *    LLobject *
 *    copy_seq(sequence)
 *    LLobject *sequence;
 *
 *    LLobject *
 *    elt(sequence, index)
 *    LLobject *sequence;
 *    LLobject *index;
 *
 *    LLobject *
 *    length(sequence)
 *    LLobject *sequence;
 *
 *    LLobject *
 *    subseq(sequence, start, end)
 *    LLobject *sequence;
 *    LLobject *start;
 *    LLobject *end;			- &optional - may be nil
 */
typedef struct LLobject LLobject;
typedef struct LLobjectType LLobjectType;
typedef struct LLsequence LLsequence;
struct LLsequence {
    LLstring sname;		/* name of the sequence */
    LLobjectType *type;
    LLobject *(*coerce)();
    LLobject *(*concatenate)();
    LLobject *(*copy_seq)();
    LLobject *(*elt)();
    LLobject *(*length)();
    LLobject *(*subseq)();
};

/*
 *    void
 *    prin1(object, stream)
 *    LLobject *object;
 *    STREAM *stream;
 *
 *    void
 *    princ(object, stream)
 *    LLobject *object;
 *    STREAM *stream;
 *
 *    void
 *    print(object, stream)
 *    LLobject *object;
 *    STREAM *stream;
 *
 *    void
 *    pprint(object, stream)
 *    LLobject *object;
 *    STREAM *stream;
 *
 *    LLobject *
 *    scan(old)
 *    LLobject *old;
 *
 *    void
 *    rescan(new, relocate)
 *    LLobject *new;
 *    LLobject *(relocate)();
 *
 *        LLobject *
 *        relocate(old)
 *        LLobject *old;
 *
 *    If you add a new type, stick the variable in config.c
 *    export it via LightLispMachine, add a printer in
 *    pr_functions.c and a garbage collector in gc_functions.c
 */
/* typedef struct LLobjectType LLobjectType; with LLsequence */
struct LLobjectType {
    LLstring tname;		/* type name (must be unique across all types) */
    LLstring pname;		/* print name (need not be unique) */
    LLobject *object;		/* the object representing this */
    int flags[4];		/* random bits for you play with */
    int count;			/* number of this type that are available */
    void (*prin1[3])();
    void (*princ[3])();
    void (*print[3])();
    void (*pprint[3])();
    struct {
	LLobject *(*scan)();	/* fault over into the new half */
	void (*rescan)();	/* scan the members for pointers */
    } garbage_collect;
    LLsequence *sequence;	/* possibly NULL */
};

#define LL_prin1(obj, i, stream)	(*LL_type_of(obj)->prin1[i])(obj, stream)
#define LL_princ(obj, i, stream)	(*LL_type_of(obj)->princ[i])(obj, stream)
#define LL_print(obj, i, stream)	(*LL_type_of(obj)->print[i])(obj, stream)
#define LL_pprint(obj, i, stream)	(*LL_type_of(obj)->pprint[i])(obj, stream)

/* 
 *    If you change this, you must change the function
 *    dispatcher in eval.c so that it can know to stuff
 *    more args on the stack.
 */
#define LL_MAXARGS	20	/* (50 is the CommonLisp minimum) */

extern LLobjectType *LLNilCell;
extern LLobjectType *LLIntegerCell;
extern LLobjectType *LLRealCell;
extern LLobjectType *LLStringCell;
extern LLobjectType *LLSymbolCell;
extern LLobjectType *LLTypeCell;
extern LLobjectType *LLMacroClosureCell;
extern LLobjectType *LLFunctionClosureCell;
extern LLobjectType *LLConsCell;
extern LLobjectType *LLFunctionCells[LL_MAXARGS][LL_MAXARGS][2];/* not all filled */
extern LLobjectType *LLMacroCell;
extern LLobjectType *LLSpecialFormCell;
extern LLobjectType *LLStreamCell;
extern LLobjectType *LLDeadCell;	/* no value at all (e.g. a closed stream) */

/* typedef struct LLobject LLobject; above LLsequence */
struct LLobject {
    LLobjectType *type;
    union {
	int32 a_integer;		/* an integer */
	double a_real;
	String a_string;
	struct {
	    LLstring pname;
	    LLobject *value;
	} a_symbol;
	LLobjectType *a_type;
	struct {
	    LLobject *environment;
	    LLobject *code;
	} a_closure;
	struct {
	    LLobject *car;
	    LLobject *cdr;
	} a_cons;
	struct {
	    LLstring pname;
	    LLobject *(*function)();	/* see below for function args */
	} a_function;
	struct {	/* wrapped in a structure so you can structure-assign it */
	    FitAny value[2];		/* ya get 2 fitany quantities to play with */
	} a_userDefined;
	struct {
	    LLstring pname;
	} a_distinguished;
	STREAM *a_stream;
	struct {
	    LLobject *broken_heart;
	} a_garbage;
    } value;
};

/*
 *    Primitive creators
 */
extern LLobject *LLMakeObject(); /* the internal allocator */
extern LLobjectType *LLMakeObjectType();
extern void LLMakeASequence();

extern LLobjectType *LLLookupType();
extern LLsequence *LLLookupSequence();

extern LLobject *LLMakeDistinguished();
extern LLobject *LLMakeFunction();
extern LLobject *LLMakeInteger();
extern LLobject *LLMakeMacro();
extern LLobject *LLMakeReal();
extern LLobject *LLMakeSpecialForm();
extern LLobject *LLMakeStream();
extern LLobject *LLMakeString();
extern LLobject *LLMakeSymbol();
extern LLobject *LLMakeType();
extern LLobject *LLMakeUninternedSymbol();
extern void LLUnmakeStream();

/* 
 *    LLobject *
 *    evaluator(sexp, environment)
 *    LLobject *sexp;
 *    LLobject *environment;
 *
 *    LLEvaluator fits
 */
extern LLobject *LLEvaluator();
extern LLobject *LLExtendEnvironment();
extern LLobject *LLFunctionClosure();
extern LLobject *LLGetValue();
extern LLobject *LLMacroClosure();
extern LLobject *LLReadValue();
extern LLobject *LLSetValue();

extern Boolean LLProbeValueP();
extern Boolean LLInternedSymbolP();	

extern int LLGetLength();		/* returns C int count */
extern LLobject *LLGetNth();		/* takes a C int argument */
extern LLobject *LLGetNthcdr();		/* takes a C int argument */

/*
 *    Now the ``machine''
 *
 *    If cleanup is to be evaled in a lexical environment,
 *    then put the value to be evaluated in a closure
 */
typedef struct LLcatcher LLcatcher;
struct LLcatcher {
    LLobject *tag;		/* compared via eq */
    LLobject *cleanup;		/* evaled on cleanup */
    jmp_buf state;
};

/*
 *    void
 *    scan(userData, relocate)
 *    RefAny userData;
 *    LLobject *(*relocate)();
 *
 *        LLobject *
 *        relocate(old)
 *        LLobject *old;
 */
typedef struct LLxregister LLxregister;
struct LLxregister {
    void (*scan)();
    RefAny userData;
};

/*
 *    If you add to this, you have to initialize the ``register''
 *    in init.c, and change scan_registers in garbage_collect.c
 */
typedef struct LLmachine LLmachine;
struct LLmachine {
    jmp_buf top_level;			/* top level catcher (not always valid) */
    DYNStkgeneralStack catchers;	/* of LLcatcher (for (escape) and errors) */
    LLobject *(*exit)();		/* called from (exit) */
    st_table *types;			/* of (LLstring, LLobjectType *) */
    st_table *sequences;		/* of (LLstring, LLsequence *) */
    st_table *oblist;			/* of (LLstring, LLobject *) b/c an a-list is too slow */
    st_table *handles;			/* of (LLstring, LLobject *) handles into the heap */
    st_table *integers;			/* of (int, LLobject *) b/c a copy of each is too much */
    st_table *streams;			/* of (STREAM *, LLobject *) b/c a copy of each is too much */
    DYNgeneralArray externals;		/* of LLxregister */
    LLobject *readtable[128];		/* standard macros */
    LLobject *dispatchreadtable[128];	/* # macros */
    LLobject *alwayscatchable;		/* magic alwayscatchable value */
    LLobject *atsign;			/* the ,@ special form */
    LLobject *backquote;		/* the ` special form */
    LLobject *comma;			/* the , special form */
    LLobject *environment;		/* the top-level environment (nil) */
    LLobject *function;			/* the #' special form (function) */
    LLobject *eofvalue;			/* the magic eof value */
    LLobject *lightlisp;		/* to throw to */
    LLobject *lambda;			/* in order to make (lambda () ... ) */
    LLobject *nil;
    LLobject *quote;			/* the ' special form (quote) */
    LLobject *t;
    LLobject *unbound;			/* the magic unbound value */
    LLobject *uncatchable;		/* the magic uncatchable value */
};

extern LLmachine LLMachine;

/*
 *    The internal macro version of things
 *
 *    They assume that you know what you're doing
 */
#define LL_alwayscatchable	(LLMachine.alwayscatchable)
#define LL_atsign		(LLMachine.atsign)
#define LL_backquote		(LLMachine.backquote)
#define LL_comma		(LLMachine.comma)
#define LL_function		(LLMachine.function)
#define LL_environment		(LLMachine.environment)
#define LL_eofvalue		(LLMachine.eofvalue)
#define LL_lightlisp		(LLMachine.lightlisp)
#define LL_lambda		(LLMachine.lambda)
#define LL_nil			(LLMachine.nil)
#define LL_quote		(LLMachine.quote)
#define LL_t			(LLMachine.t)
#define LL_unbound		(LLMachine.unbound)
#define LL_uncatchable		(LLMachine.uncatchable)

/*
 *    The catch primitives
 *
 *    catch-begin is
 *        push LLcatcher
 *        if nil cleanup-code fine
 *           else `#'(lambda () ,@cleanupcode)
 *
 *    Because setjmp only works in a macro, these have to be macros (damnit)
 */
extern LLcatcher LLPreCatcher;
extern LLobject *LLCaughtValue;
#define LLCatchBegin(tagvalue, cleanupcode, environment)	\
	    (LLPreCatcher.tag = (tagvalue), \
	     LLPreCatcher.cleanup = LL_null(cleanupcode) ? (cleanupcode): \
		    LLCons(LLFunctionClosure(LLCons(LL_nil, cleanupcode), environment), LL_nil), \
	     DYNStkpPush(&LLMachine.catchers, &LLPreCatcher, LLcatcher))

/*
 *    if (LLTopLevelCatch()) {
 *        - things failed -
 *    }
 *
 *    LLobject *value;
 *
 *    if (LLTopLevelCatchValue(&value)) {
 *        - things failed -
 *    }
 *
 *    LLobject *tag, *cleanup, *environment;
 *
 *    if (LLCatchVoid(tag, cleanup, environment)) {
 *        - things failed -
 *    }
 *
 *    LLobject *tag, *cleanup, *environment, *value;
 *
 *    if (LLCatchValue(tag, cleanup, environment, &value)) {
 *        - things failed -
 *    }
 */
#define LLTopLevelCatchVoid()	\
	    ((Boolean) setjmp(LLMachine.top_level))
#define LLCatchVoid(tagvalue, cleanupcode, environment)	\
	    (LLCatchBegin(tagvalue, cleanupcode, environment), \
	     (Boolean) setjmp(DYNStkpTop(&LLMachine.catchers, LLcatcher)->state))
#define LLTopLevelCatchValue(p_val)		\
	    (LLCaughtValue = (LLobject *) setjmp(LLMachine.top_level), \
	     *(p_val) = LL_car(LLCaughtValue), \
	     ! LL_null(LLCaughtValue))
#define LLCatchValue(tagvalue, cleanupcode, environment, p_val)	\
	    (LLCatchBegin(tagvalue, cleanupcode, environment), \
	     LLCaughtValue = (LLobject *) \
		    setjmp(DYNStkpTop(&LLMachine.catchers, LLcatcher)->state), \
	     *(p_val) = LL_car(LLCaughtValue), \
	     ! LL_null(LLCaughtValue))
#define LLEndCatch()		/* eval AFTER pop */ \
	    (LLEvaluator(DYNStkpPop(&LLMachine.catchers, LLcatcher)->cleanup, LL_environment))

/*
 *    LLobject *
 *    function( ... argument ... )
 *    LLobject *argument0;
 *    LLobject *argument1;
 *    ...
 *
 *        returns the result
 *
 *    LLobject *
 *    expander(arguments, environment)
 *    LLobject *arguments;	- a list of the (unevaluated) arguments
 *    LLobject *environment;
 *
 *        returns the result
 *
 *    LLobject *
 *    C_call_special_form(arguments, environment)
 *    LLobject *arguments;	- a list of the (unevaluated) arguments
 *    LLobject *environment;
 *
 *        returns the result
 *
 *    LLobject *
 *    special_form(arguments, environment)
 *    LLobject **arguments;	- a list of the (unevaluated) arguments
 *    LLobject **environment;
 *
 *        returns {t, nil} on {tail-recursive, non-tail-recursive}
 *        if nil, then *arguments is the return value
 *        if t, then *arguments must be evaluated in *environment
 */
#define LL_NON_FUNCTION		-1		/* ->flags[0] */
#define LL_FUNCTION		0		/* ->flags[0] */
#define LL_MACRO		1		/* ->flags[0] */
#define LL_SPECIAL_FORM		2		/* ->flags[0] */

#define LL_function_call(o)	(*(o)->value.a_function.function)/* ( args ... ) */
#define LL_function_nrequired(o)	((o)->type->flags[1])
#define LL_function_noptional(o)	((o)->type->flags[2])
#define LL_function_restp(o)		((o)->type->flags[3]) /* 0 or 1 */

#define LL_function_ntotal(o)		(LL_function_nrequired(o)+LL_function_noptional(o)+(LL_function_restp(o) ? 1: 0))

/*
 *    Examine flags[0] b/c it is too hard to figure
 *    out if something is in LLFunctionCells[][][] or not.
 */
#define LL_functionp(o)		((o) != LL_nil && (o)->type->flags[0] == LL_FUNCTION)
#define LL_macrop(o)		((o) != LL_nil && (o)->type == LLMacroCell)
#define LL_special_formp(o)	((o) != LL_nil && (o)->type == LLSpecialFormCell)

/*
 *    Dealing with closures
 */
#define LL_function_closurep(o)		((o) != LL_nil && (o)->type == LLFunctionClosureCell)
#define LL_macro_closurep(o)		((o) != LL_nil && (o)->type == LLMacroClosureCell)
#define LL_closurep(o)			((o) != LL_nil && \
					 ((o)->type == LLFunctionClosureCell || \
					 (o)->type == LLMacroClosureCell))

#define LL_code(o)		((o)->value.a_closure.code)
#define LL_env(o)		((o)->value.a_closure.environment)

/*
 *    Commonly used macros
 */

#define LL_atsignp(o)		(LL_consp(o) && LL_eq(LL_first(o), LL_atsign))
#define LL_atom(o)		( ! LL_consp(o) )
#define LL_car(o)		((o) == LL_nil ? LL_nil: (o)->value.a_cons.car)
#define LL_cdr(o)		((o) == LL_nil ? LL_nil: (o)->value.a_cons.cdr)
#define LL_commap(o)		(LL_consp(o) && LL_eq(LL_first(o), LL_comma))
#define LL_consp(o)		((o) != LL_nil && (o)->type == LLConsCell)
#define LL_endp(o)		((o) == LL_nil)
#define LL_eq(o1, o2)		((o1) == (o2))
#define LL_integerp(o)		((o) != LL_nil && (o)->type == LLIntegerCell)
#define LL_lambdap(o)		(LL_consp(o) && LL_first(o) == LL_lambda)
#define LL_listp(o)		((o) == LL_nil || (o)->type == LLConsCell)
#define LL_not(o)		LL_null(o)
#define LL_null(o)		((o) == LL_nil)
#define LL_numberp(o)		(LL_integerp(o) || LL_realp(o))
#define LL_pname(o)		(LL_functionp(o) ? \
					(o)->value.a_function.pname: \
				 LL_symbolp(o) ? \
					(o)->value.a_symbol.pname: \
					(o)->value.a_distinguished.pname)
#define LL_realp(o)		((o) != LL_nil && (o)->type == LLRealCell)
#define LL_rplaca(o, v)		((o)->value.a_cons.car = (v))
#define LL_rplacd(o, v)		((o)->value.a_cons.cdr = (v))
#define LL_sequencep(o)		(((o) == LL_nil ? LLNilCell: (o)->type)->sequence != NULL)
#define LL_streamp(o)		((o) != LL_nil && (o)->type == LLStreamCell)
#define LL_stringp(o)		((o) != LL_nil && (o)->type == LLStringCell)
#define LL_symbolp(o)		((o) != LL_nil && (o)->type == LLSymbolCell)
#define LL_typep(o, tCell)	(((o) == LL_nil ? LLNilCell: (o)->type) == (tCell))

#define LL_first(o)		(LL_car(o))
#define LL_second(o)		(LL_car(LL_cdr(o)))
#define LL_third(o)		(LL_car(LL_cdr(LL_cdr(o))))
#define LL_fourth(o)		LLGetNth(3, o)
#define LL_fifth(o)		LLGetNth(4, o)
#define LL_sixth(o)		LLGetNth(5, o)
#define LL_seventh(o)		LLGetNth(6, o)
#define LL_eighth(o)		LLGetNth(7, o)
#define LL_ninth(o)		LLGetNth(8, o)
#define LL_tenth(o)		LLGetNth(9, o)

#define LL_rest(o)		(LL_cdr(o))

/* returning a (LLobjecType *) */
#define LL_type_of(o)		((o) == LL_nil ? LLNilCell: (o)->type)

/*
 *    Accessing the various value cells of an environment
 *
 *    Value of a list of frames consisting of frames
 *
 *    environment = (value-alist . function-alist)
 *    alist item = (symbol . value)
 */
typedef enum LLslot LLslot;		/* pegs and slots */
enum LLslot {
    LL_valueSlot,
    LL_functionSlot,
};

/*
 *    for those pesky a-lists in the lexical environment
 */
#define LL_value_slot(e)		LL_car(e)
#define LL_function_slot(e)		LL_cdr(e)
#define LL_slot(e, slot)		((slot) == LL_valueSlot ? \
						LL_value_slot(e): \
						LL_function_slot(e))

/*
 *   for functions at the top level
 */
#define LL_symbol_value(sym)		LL_car((sym)->value.a_symbol.value)
#define LL_symbol_function(sym)		LL_cdr((sym)->value.a_symbol.value)
#define LL_symbol_slot(sym, slot)	((slot) == LL_valueSlot ? \
					 LL_symbol_value(sym): \
					 LL_symbol_function(sym))
#define LL_unboundp(sym, slot)		LL_eq(LL_symbol_slot(sym, slot), LL_unbound)

#define LL_set_symbol_value(sym, v)		LL_rplaca((sym)->value.a_symbol.value, v)
#define LL_set_symbol_function(sym, v)		LL_rplacd((sym)->value.a_symbol.value, v)
#define LL_set_symbol_slot(sym, slot, v)	((slot) == LL_valueSlot ? \
						 LL_set_symbol_value(sym, v): \
						 LL_set_symbol_function(sym, v))

#define LL_slot_name(slot)		((slot) == LL_valueSlot ? "value": "function")

/*
 *    Neat functions to make it all useable (in LightLisp)
 */
extern LLobject *LL1Minus();		/* function */
extern LLobject *LL1Plus();		/* function */
extern LLobject *LLAbs();		/* function */
extern LLobject *LLAcons();		/* function */
extern LLobject *LLAcos();		/* function */
extern LLobject *LLAcosh();		/* function */
extern LLobject *LLAnd();		/* C-call special form (macro)*/
extern LLobject *LLAnd_tr();            /* special form (macro) */
extern LLobject *LLAppend();		/* function */
extern LLobject *LLApply();		/* function */
extern LLobject *LLAsin();		/* function */
extern LLobject *LLAsinh();		/* function */
extern LLobject *LLAssoc();		/* function */
extern LLobject *LLAtan();		/* function */
extern LLobject *LLAtanh();		/* function */
extern LLobject *LLAtom();		/* function */
extern LLobject *LLBackquote();		/* C-call special form */
extern LLobject *LLBoundp();		/* function */
extern LLobject *LLCaaaar();		/* function */
extern LLobject *LLCaaadr();		/* function */
extern LLobject *LLCaaar();		/* function */
extern LLobject *LLCaadar();		/* function */
extern LLobject *LLCaaddr();		/* function */
extern LLobject *LLCaadr();		/* function */
extern LLobject *LLCaar();		/* function */
extern LLobject *LLCadaar();		/* function */
extern LLobject *LLCadadr();		/* function */
extern LLobject *LLCadar();		/* function */
extern LLobject *LLCaddar();		/* function */
extern LLobject *LLCadddr();		/* function */
extern LLobject *LLCaddr();		/* function */
extern LLobject *LLCadr();		/* function */
extern LLobject *LLCar();		/* function */
extern LLobject *LLCase();		/* C-call special form (macro) */
extern LLobject *LLCase_tr();           /* special form (macro) */
extern LLobject *LLCatch();		/* C-call special form */
extern LLobject *LLCatch_tr();          /* special form */
extern LLobject *LLCdaaar();		/* function */
extern LLobject *LLCdaadr();		/* function */
extern LLobject *LLCdaar();		/* function */
extern LLobject *LLCdadar();		/* function */
extern LLobject *LLCdaddr();		/* function */
extern LLobject *LLCdadr();		/* function */
extern LLobject *LLCdar();		/* function */
extern LLobject *LLCddaar();		/* function */
extern LLobject *LLCddadr();		/* function */
extern LLobject *LLCddar();		/* function */
extern LLobject *LLCdddar();		/* function */
extern LLobject *LLCddddr();		/* function */
extern LLobject *LLCdddr();		/* function */
extern LLobject *LLCddr();		/* function */
extern LLobject *LLCdr();		/* function */
extern LLobject *LLCeiling();		/* function */
extern LLobject *LLClearoutput();	/* function */
extern LLobject *LLClose();		/* function */
extern LLobject *LLCoerce();		/* function */
extern LLobject *LLConcatenate();	/* function */
extern LLobject *LLCond();		/* C-call special form (macro) */
extern LLobject *LLCond_tr();           /* special form (macro) */
extern LLobject *LLCons();		/* function */
extern LLobject *LLConsp();		/* function */
extern LLobject *LLCopylist();		/* function */
extern LLobject *LLCopyseq();		/* function */
extern LLobject *LLCopytree();		/* function */
extern LLobject *LLCos();		/* function */
extern LLobject *LLCosh();		/* function */
extern LLobject *LLDefconstant();	/* C-call special form (macro) */
extern LLobject *LLDefconstant_tr();	/* special form (macro) */
extern LLobject *LLDefmacro();		/* C-call special form (macro) */
extern LLobject *LLDefmacro_tr();       /* special form (macro) */
extern LLobject *LLDefparameter();	/* C-call special form (macro) */
extern LLobject *LLDefparameter_tr();	/* special form (macro) */
extern LLobject *LLDefun();		/* C-call special form (macro) */
extern LLobject *LLDefun_tr();          /* special form (macro) */
extern LLobject *LLDefvar();		/* C-call special form (macro) */
extern LLobject *LLDefvar_tr();		/* special form (macro) */
extern LLobject *LLDiv();		/* function */
extern LLobject *LLDo();		/* C-call special form (macro) */
extern LLobject *LLDo_tr();             /* special form (macro) */
extern LLobject *LLDolist();		/* C-call special form (macro) */
extern LLobject *LLDolist_tr();         /* special form (macro) */
extern LLobject *LLDostar();		/* C-call special form (macro) */
extern LLobject *LLDostar_tr();         /* special form (macro) */
extern LLobject *LLDotimes();		/* C-call special form (macro) */
extern LLobject *LLDotimes_tr();        /* special form (macro) */
extern LLobject *LLEighth();		/* function */
extern LLobject *LLElt();		/* function */
extern LLobject *LLEq();		/* function */
extern LLobject *LLEql();		/* function */
extern LLobject *LLEqual();		/* function */
extern LLobject *LLEqualp();		/* function */
extern LLobject *LLEval();		/* function */
extern LLobject *LLEvenp();		/* function */
extern LLobject *LLExit();		/* function */
extern LLobject *LLExp();		/* function */
extern LLobject *LLExpt();		/* function */
extern LLobject *LLFboundp();		/* function */
extern LLobject *LLFceiling();		/* function */
extern LLobject *LLFfloor();		/* function */
extern LLobject *LLFifth();		/* function */
extern LLobject *LLFinishoutput();	/* function */
extern LLobject *LLFirst();		/* function */
extern LLobject *LLFlet();		/* C-call special form */
extern LLobject *LLFlet_tr();		/* special form */
extern LLobject *LLFloatp();		/* function */
extern LLobject *LLFloor();		/* function */
extern LLobject *LLFmakunbound();	/* function */
extern LLobject *LLForceoutput();	/* function */
extern LLobject *LLFormat();		/* function */
extern LLobject *LLFourth();		/* function */
extern LLobject *LLFreshline();		/* function */
extern LLobject *LLFround();		/* function */
extern LLobject *LLFtruncate();		/* function */
extern LLobject *LLFuncall();		/* function */
extern LLobject *LLFunction();		/* C-call special form */
extern LLobject *LLFunction_tr();       /* special form */
extern LLobject *LLFunctionp();		/* function */
extern LLobject *LLGe();		/* function */
extern LLobject *LLGensym();		/* function */
extern LLobject *LLGt();		/* function */
extern LLobject *LLIdentity();		/* function */
extern LLobject *LLIf();		/* C-call special form */
extern LLobject *LLIf_tr();             /* special form */
extern LLobject *LLIntegerp();		/* function */
extern LLobject *LLInputstreamp();	/* function */
extern LLobject *LLIsqrt();		/* function */
extern LLobject *LLLabels();		/* C-call special form */
extern LLobject *LLLabels_tr();		/* special form */
extern LLobject *LLLast();		/* function */
extern LLobject *LLLe();		/* function */
extern LLobject *LLLength();		/* function */
extern LLobject *LLLet();		/* C-call special form */
extern LLobject *LLLet_tr();            /* special form */
extern LLobject *LLLetstar();		/* C-call special form */
extern LLobject *LLLetstar_tr();        /* special form */
extern LLobject *LLList();		/* function */
extern LLobject *LLListlength();	/* function */
extern LLobject *LLListp();		/* function */
extern LLobject *LLListstar();		/* function */
extern LLobject *LLLoad();		/* function */
extern LLobject *LLLog();		/* function */
extern LLobject *LLLt();		/* function */
extern LLobject *LLMacrofunction();	/* function */
extern LLobject *LLMacroexpand();	/* function */
extern LLobject *LLMacroexpand1();	/* function */
extern LLobject *LLMacrolet();		/* C-call special form */
extern LLobject *LLMacrolet_tr();	/* special form */
extern LLobject *LLMakesymbol();	/* function */
extern LLobject *LLMakunbound();	/* function */
extern LLobject *LLMap();		/* function */
extern LLobject *LLMax();		/* function */
extern LLobject *LLMember();		/* function */
extern LLobject *LLMin();		/* function */
extern LLobject *LLMinus();		/* function */
extern LLobject *LLMinusp();		/* function */
extern LLobject *LLMod();		/* function */
extern LLobject *LLNconc();		/* function */
extern LLobject *LLNinth();		/* function */
extern LLobject *LLNot();		/* function */
extern LLobject *LLNth();		/* function */
extern LLobject *LLNthcdr();		/* function */
extern LLobject *LLNull();		/* function */
extern LLobject *LLNumberp();		/* function */
extern LLobject *LLNumeq();		/* function */
extern LLobject *LLNumnoteq();		/* function */
extern LLobject *LLOblist();		/* function */
extern LLobject *LLOddp();		/* function */
extern LLobject *LLOpen();		/* function */
extern LLobject *LLOr();		/* C-call special form (macro) */
extern LLobject *LLOr_tr();             /* special form (macro) */
extern LLobject *LLOutputstreamp();	/* function */
extern LLobject *LLPairlis();		/* function */
extern LLobject *LLPlus();		/* function */
extern LLobject *LLPlusp();		/* function */
extern LLobject *LLPprint();		/* function */
extern LLobject *LLPrin1();		/* function */
extern LLobject *LLPrinc();		/* function */
extern LLobject *LLPrint();		/* function */
extern LLobject *LLProg1();		/* C-call special form (macro) */
extern LLobject *LLProg1_tr();          /* special form (macro) */
extern LLobject *LLProg2();		/* C-call special form (macro) */
extern LLobject *LLProg2_tr();          /* special form (macro) */
extern LLobject *LLProgn();		/* C-call special form */
extern LLobject *LLProgn_tr();          /* special form */
extern LLobject *LLQuote();		/* C-call special form */
extern LLobject *LLQuote_tr();          /* special form */
extern LLobject *LLRandom();		/* function */
extern LLobject *LLRead();		/* function */
extern LLobject *LLReadfromstring();	/* function */
extern LLobject *LLRem();		/* function */
extern LLobject *LLRest();		/* function */
extern LLobject *LLRound();		/* function */
extern LLobject *LLRplaca();		/* function */
extern LLobject *LLRplacd();		/* function */
extern LLobject *LLSecond();		/* function */
extern LLobject *LLSet();		/* function */
extern LLobject *LLSetq();		/* C-call special form */
extern LLobject *LLSetq_tr();		/* special form */
extern LLobject *LLSeventh();		/* function */
extern LLobject *LLSignum();		/* function */
extern LLobject *LLSin();		/* function */
extern LLobject *LLSinh();		/* function */
extern LLobject *LLSixth();		/* function */
extern LLobject *LLSqrt();		/* function */
extern LLobject *LLSpecialformp();	/* function */
extern LLobject *LLStreamp();		/* function */
extern LLobject *LLStringp();		/* function */
extern LLobject *LLSubseq();		/* function */
extern LLobject *LLSubshell();		/* C-call special form */
extern LLobject *LLSubshell_tr();	/* special form */
extern LLobject *LLSymbolfunction();	/* function */
extern LLobject *LLSymbolname();	/* function */
extern LLobject *LLSymbolvalue();	/* function */
extern LLobject *LLSymbolp();		/* function */
extern LLobject *LLTan();		/* function */
extern LLobject *LLTanh();		/* function */
extern LLobject *LLTenth();		/* function */
extern LLobject *LLTerpri();		/* function */
extern LLobject *LLThird();		/* function */
extern LLobject *LLThrow();		/* function */
extern LLobject *LLTime();		/* C-call special form (macro) */
extern LLobject *LLTime_tr();		/* C-special form */
extern LLobject *LLTimes();		/* function */
extern LLobject *LLTruncate();		/* function */
extern LLobject *LLTypeof();		/* function */
extern LLobject *LLUnless();		/* C-call special form (macro) */
extern LLobject *LLUnless_tr();         /* special form (macro) */
extern LLobject *LLUnwindprotect();	/* C-call special form */
extern LLobject *LLUnwindprotect_tr();  /* special form */
extern LLobject *LLWhen();		/* C-call special form (macro) */
extern LLobject *LLWhen_tr();           /* special form (macro) */
extern LLobject *LLZerop();		/* function */

/*
 *    Printing
 */
extern void LLErrorF();                 /* C variables are the args */
extern void LLPrintF();                 /* C variables are the args */
extern STREAM *LLErrorStream;
extern STREAM *LLStandardStream;

extern void LLFormatC();                /* C variables are the args (avec un stream)*/
extern void LLFormatS();                /* s-expressions are the args (avec un stream) */


/*
 *    Workhorse routines for the handles
 */
typedef int LLhandle;

extern LLhandle LLMakeHandle();
extern void LLKillHandle();		/* yes handles don't garbage-collect */

extern LLobject*LLSetHandle();		/* returns the value set */
extern LLobject *LLEsetHandle();	/* returns the environment set */
extern LLobject *LLRefHandle();
extern LLobject *LLErefHandle();

/*
 *    Garbage collection - (for now)
 *
 *    The user must know to call garbage-collect
 *    when the system is in a quiescent state.
 */
extern void LLGarbageCollect();

typedef struct LLstats LLstats;
struct LLstats {
    int totalHeapBytes;                 /* number of bytes in the heap (only) */
    int totalCells;                     /* total cells allocated */
    int cellsUsed;                      /* cells used up so far */
    int cellsFree;                      /* cells free */
    int maximumCellsAllowed;            /* maximumCellsAllowed == 0 implies unlimited */
};
extern void LLGetStats();
extern void LLPrintStats();

typedef struct LLcellsByType LLcellsByType;
struct LLcellsByType {
    LLstring pname;
    int count;
};

typedef struct LLfullStats LLfullStats;
struct LLfullStats {
    int totalHeapBytes;			/* number of bytes in the heap (only) */
    int totalCells;			/* total cells allocated */
    int cellsUsed;			/* cells used up so far */
    int cellsFree;			/* cells free */
    int maximumCellsAllowed;		/* maximumCellsAllowed == 0 implies unlimited */
    DYNgeneralArray cellsByType;	/* of LLcellsByType */
    int cellBytes;			/* bytes used for cells */
    int stringBytes;			/* bytes used for strings */
    int totalBytes;			/* number of bytes in heap + cells + string + symbol */
    int totalCoreBytes;			/* number of bytes in in core */
    int garbageCollections;		/* number of garbage collections */
    int garbageCollectTime;		/* milliseconds in gc */
    int totalTime;			/* milliseconds since LLInitialize */
};
extern void LLGetFullStats();
extern void LLFreeFullStats();
extern void LLPrintFullStats();

#endif /* LightLispMachine_h */
