#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

static LLobject *eval_args();
static LLobject *function_call();

/*
 *    LLLambdaArgs - handle the binding of the lambda arguments
 *
 *    LLobject *
 *    evaluator(arg, environment)
 *    LLobject *arg;
 *    LLobject *environment;
 *
 *    ( { var }*
 *      [ &optional { var | (var [initform [svar]]) }* ]
 *      [ &rest var ]
 *
 *    return the resulting environment
 */
LLobject *
LLLambdaArgs(formals, formals_env, actuals, actuals_env, evaluator, destructuringP)
LLobject *formals;
LLobject *formals_env;
LLobject *actuals;
LLobject *actuals_env;
LLobject *(*evaluator)();	/* may be NULL */
Boolean destructuringP;
{
	Boolean initformP;
	LLobject *f, *a, *formal, *actual, *lambda_env, *next,
		*sexp_formal, *suppliedp_formal, *tmp, *last, *rest;
	LLobject *f2;	/* goddamnit, this is only for pcc */

	if ( ! LL_listp(formals) || ! LL_listp(actuals) ) {
	    LLErrorF("\
Badly formed argument lambda list\n\
\tformals and actuals for a lambda-list must both be lists\n\
\tformals: ~S\n\
\tactuals: ~S\n\
", formals, actuals);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	lambda_env = formals_env;

	/*
	 *    { var }*
	 */
	for (f=formals, a=actuals; LL_consp(f); f=LL_cdr(f), a=LL_cdr(a)) {
	    formal = LL_car(f);

	    if (LL_symbolp(formal) && LL_pname(formal)[0] == '&') {
		/*
		 *    &optional, &rest or such 
		 */
		break;
	    }

	    if ( ! LL_consp(a) ) {
		LLErrorF("\
Badly formed argument lambda list\n\
\tnot enough required arguments for lambda list\n\
\tformals: ~S\n\
\tactuals: ~S\n\
", formals, actuals);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    actual = LL_car(a);

	    if (LL_consp(formal)) {
		if ( ! destructuringP ) {
		    LLErrorF("\
Badly formed formal list for lambda list\n\
\tdestructuring s-exp are not allowed at this point\n\
\toffender: ~S\n\
\tformals:  ~S\n\
\tactuals:  ~S\n\
", formal, formals, actuals);
		    LLThrow(LL_lightlisp, LL_nil);
		}
		lambda_env = LLLambdaArgs(formal, lambda_env,
				actual, actuals_env,
				evaluator, destructuringP);
		continue;
	    } else if ( ! LL_symbolp(formal) ) {
		LLErrorF("\
Badly formed formal list for lambda list\n\
\texpecting symbol or sexp only\n\
\toffender: ~S\n\
\tformals:  ~S\n\
\tactuals:  ~S\n\
", formal, formals, actuals);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    /* ExtendEnvironment will check for (symbolp formal) */
	    if (evaluator != NULL)
		actual = (*evaluator)(actual, actuals_env);
	    lambda_env = LLExtendEnvironment(formal, LL_valueSlot, actual, lambda_env);
	}

	if (LL_symbolp(LL_car(f)) && LL_pname(LL_car(f)) == LLAndOptional) {
	    /*
	     *   [ &optional { var | (var [initform [svar]]) }* ]
	     */
	    for (f=LL_cdr(f) /* inherit a */ ; LL_consp(f); f=LL_cdr(f)) {
		formal = LL_car(f);

		if (LL_symbolp(formal) && LL_pname(formal)[0] == '&') {
		    /*
		     *    &rest, &body or such 
		     */
		    break;
		}

		if (LL_consp(a)) {
		    initformP = FALSE;
		    actual = LL_car(a);
		    a = LL_cdr(a);
		} else {
		    initformP = TRUE;
		    actual = LL_nil;
		}

		suppliedp_formal = LL_nil;
		if (LL_consp(formal)) {
		    /*
		     *    (var [ initform [ suppliedp ] ])
		     */
		    sexp_formal = formal;

		    formal = LL_first(sexp_formal);
		    if (initformP)
			actual = LL_second(sexp_formal);
		    if ( ! LL_null(LL_cdr(LL_cdr(sexp_formal))) ) {
			/*
			 *    (var initform suppliedp)
			 */
			suppliedp_formal = LL_third(sexp_formal);
			if ( ! LL_symbolp(suppliedp_formal) ) {
			    LLErrorF("\
Badly formed formal list for lambda list\n\
\texpecting symbol for supplied-p after &optional\n\
\tformal:  ~S\n\
\tformals: ~S\n\
\tactuals: ~S\n\
", sexp_formal, formals, actuals);
			    LLThrow(LL_lightlisp, LL_nil);
			}
		    }
		}

		/*
		 *    If formal is a cons, then more destructuring
		 *    If it is not a cons then it must be a symbol.
		 *
		 *    Actual contains a value that needs to be evaluated or not.
		 */

		if (LL_consp(formal)) {
		    /*
		     *    More destructuring
		     * 
		     *    (( ... ) [ initform [ suppliedp ] ])
		     *    AND enough args to cover it (e.g. actual is value)
		     */
		    if ( ! destructuringP ) {
			LLErrorF("\
Badly formed formal list for lambda list\n\
\tdestructuring s-exp are not allowed at this point\n\
\toffender: ~S\n\
\tformals: ~S\n\
\tactuals: ~S\n\
", formal, formals, actuals);
			LLThrow(LL_lightlisp, LL_nil);
		    }
		} else if ( ! LL_symbolp(formal) ) {
		    LLErrorF("\
Badly formed formal list for lambda list\n\
\texpecting symbol or sexp for &optional\n\
\toffender: ~S\n\
\tformals:  ~S\n\
\tactuals:  ~S\n\
", formal, formals, actuals);
		    LLThrow(LL_lightlisp, LL_nil);
		}

		/* ExtendEnvironment will check for (symbolp formal) */
		if (evaluator != NULL && ! initformP ) {
		    /*
		     *    If an evaluator is supplied, and the
		     *    actual was supplied (e.g. it was from the
		     *    actuals list, not &optional's initform),
		     *    then evaluate it using the given evaluator
		     */
		    actual = (*evaluator)(actual, actuals_env);
		} else if (initformP) {
		    /*
		     *    Otherwise, if it was not supplied (e.g. it was
		     *    from the &optional's initform, always the initform
		     *    LLEvaluator in the context of the lambda.
		     */
		    actual = LLEvaluator(actual, lambda_env);
		}

		if (LL_consp(formal)) {
		    /*
		     *    more destructuring case
		     */
		    lambda_env = LLLambdaArgs(formal, lambda_env,
				    actual, actuals_env,
				    NULL, destructuringP);
		} else {
		    /*
		     *    The normal case
		     */
		    lambda_env = LLExtendEnvironment(formal, LL_valueSlot, actual, lambda_env);
		}

		if ( ! LL_null(suppliedp_formal) ) {
		    lambda_env = LLExtendEnvironment(suppliedp_formal,
			    LL_valueSlot, initformP ? LL_t: LL_nil, lambda_env);
		}
	    }
	}

	next = LL_car(f);
	if (LL_symbolp(next) &&
		(LL_pname(next) == LLAndRest ||
		 LL_pname(next) == LLAndBody)) {
	    /*
	     *   [ &rest var ] [ &... ] )
	     */
	    f = LL_cdr(f);
	    f2 = LL_second(f);	/* goddamnit, this is only for pcc */
	    if (LL_null(f) || ! LL_symbolp(LL_car(f)) ||
		    ! (LL_null(LL_cdr(f)) || LL_pname(f2)[0] == '&')) {
		LLErrorF("\
Badly formed formal list for lambda list\n\
\tbadly formed &rest - must be last and must be followed by a symbol or a lambda-list-keyword\n\
\toffender: ~S\n\
\tformals:  ~S\n\
\tactuals:  ~S\n\
", formal, formals, actuals);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    formal = LL_car(f);
	    f = LL_cdr(f);

	    rest = LL_nil;
	    last = LL_nil;
	    for (/* inherit a */; LL_consp(a); a=LL_cdr(a)) {
		actual = LL_car(a);

		if (evaluator != NULL)
		    actual = (*evaluator)(actual, actuals_env);
		tmp = LLCons(actual, LL_nil);

		if (LL_null(rest))
		    rest = tmp;
		else
		    LL_rplacd(last, tmp);
		last = tmp;
	    }

	    lambda_env = LLExtendEnvironment(formal, LL_valueSlot, rest, lambda_env);
	}

	if ( ! LL_null(f) ) {
	    LLErrorF("\
Badly formed argument list for lambda list\n\
\tnot enough actual arguments\n\
\toffender: ~S\n\
\tformals:  ~S\n\
\tactuals:  ~S\n\
", LL_car(f), formals, actuals);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if ( ! LL_null(a) ) {
	    LLErrorF("\
Badly formed argument list for lambda list\n\
\ttoo many actual arguments\n\
\toffenders: ~S\n\
\tformals:   ~S\n\
\tactuals:   ~S\n\
", a, formals, actuals);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	return (lambda_env);
}
