;;;;(
;;;;    The Official OctLisp Test and Validation Suite
;;;;    

(defun rcs (symbol)
  "use the symbol as a key to what to pick out of rcs-list,
  and then remove the leading and trailing $ signs"
  (let* ((rcs-list '((Revision . "$Revision: 1.1.1.1 $")
		     (Date . "$Date: 2002/12/31 15:51:27 $")
		     (Author . "$Author: saj $")
		     (Source . "$Source: /home/cvs/octtools/src/OctLisp/src/octlisp/validation/comprehensive.ll,v $")))
	 (string (cdr (assoc symbol rcs-list))))
	(subseq string 1 (1- (length string)))))

(defvar validations nil
  "an a-list stating what validations have been done.
  Values for the a-list value slot are
  'not-validated	- indicating that the test was not run
  'success		- indicating a successful test
  'fail			- indicating an unsuccessful test
  'load-failure		- indicating a failure to load the test")

(defun validation-symbolp (symbol)
  "see if the symbol is known to the validation system"
  (assoc symbol validations))

(defun not-validatedp (a-item)
  "take the a-item in validations - if it was not-validated"
  (eq (cdr a-item) 'not-validated))

(defun validation-for (symbol)
  "stick the symbol in the a-list as not-validated"
  (if (not (assoc symbol validations))
      (setq validations (acons symbol 'not-validated validations))))

(defun successp (a-item)
  "take the a-item in validations - if it was successful"
  (eq (cdr a-item) 'success))

(defun invalidated (symbol)
  "note the symbol in the a-list as unsuccessfully"
  (rplacd (assoc symbol validations) 'fail))

(defun validated (symbol)
  "note the symbol in the a-list as successfully validated"
  (rplacd (assoc symbol validations) 'success))

(defun bad-loadp (a-item)
  "take the a-item in validations - if it was a bad load"
  (eq (cdr a-item) 'load-failure))

(defun bad-load (symbol)
  "note the symbol in the a-list as badly-loaded"
  (rplacd (assoc symbol validations) 'load-failure))

(defun select (predp list)
  "select those in the list that are match predp"
  (cond ((null list) nil)
	((funcall predp (car list))
	 (cons (caar list) (select predp (cdr list))))
	(t
	 (select predp (cdr list)))))

(defmacro validation (symbol &rest body)
  "set up a progn that will print out some stuff,
  protect against (throw 'octlisp ... ) (throw 'lightlisp ... ) and 
  if all goes well, declare the symbol as validated"
  `(progn
    (format t "~%~%~%~%--------------------~%~%    The ~S Test~%~%--------------------~%" ',symbol)
    (cond (just-examinep
	   t)
	  ((catch 'lightlisp
		  (catch 'octlisp
			 ,@body))
	   (validated ',symbol))
	  (t
	   (invalidated ',symbol)))))

(defmacro error-recovery (&body body)
  "when OctLisp is bound, try the error-recovery bit"
  `(when OctLisp
	 (catch 'octlisp
		,@body)))

(defun test (list &optional (examine-onlyp nil))
  "Foreach symbol in the list, load the file ``comprehensive/<symbol>.ll''
  and note failures.
  If just-examinep is non-nil, then the validation macro
  will just load, but not execute - allows a test-the-tester feature"
  (dolist (symbol list)
	  (let ((filename (concatenate 'string "validation/comprehensive/"
				       (symbol-name symbol) ".ll")))
	       (validation-for symbol)
	       (setq just-examinep examine-onlyp)
	       (if (not (load filename))
		   (bad-load symbol)))))

(defun successes ()
  "return the successful tests"
  (select #'successp validations))

(defun failures ()
  "return the not-successful tests"
  (select #'(lambda (v) (not (successp v))) validations))

;;;
;;;    Oblist management
;;;
(let ((ignore '(&body &optional &rest :direction :output :write
		      * *error-output* *standard-output* + - / /= 1+ 1-
		      < <= = > >= abs acons acos acosh and append
		      apply asin asinh assoc atan atanh atom boundp caaaar caaadr caaar
		      caadar caaddr caadr caar cadaar cadadr cadar caddar cadddr caddr cadr
		      call-arguments-limit car case catch cdaaar cdaadr cdaar cdadar cdaddr
		      cdadr cdar cddaar cddadr cddar cdddar cddddr cdddr cddr cdr ceiling
		      clear-output close coerce concatenate cond cons consp copy-list
		      copy-seq copy-tree cos cosh defconstant defmacro defparameter defun
		      defvar do do* dolist dotimes e eighth elt eq eql equal equalp eval
		      evenp exit exp expt f fboundp fceiling ffloor fifth finish-output first
		      flet floatp floor fmakunbound force-output format fourth fresh-line
		      fround ftruncate funcall function functionp gensym identity if integerp
		      is-attached isqrt labels lambda-list-keywords lambda-parameters-limit
		      last length let let* lightlisp lisp-implementation-type
		      lisp-implementation-version list list* list-length listp load log
		      macro-function macroexpand macroexpand-1 macrolet make-symbol
		      makunbound map max member min minusp mod name nconc ninth not nth
		      nthcdr null numberp oblist oddp open or pairlis pi plusp pprint prin1
		      princ print prog1 prog2 progn quote random rem rest round rplaca rplacd
		      s second set setq seventh signum sin sinh sixth special-form-p sqrt
		      stringp subseq subshell symbol-function symbol-name symbol-value
		      symbolp t tan tanh tenth terpri third throw truncate unless
		      unwind-protect when zerop)))
  (flet ((ignore-symbolp (symbol)
			 "see if the symbol should be ignored by this validation suite"
			 (member symbol ignore)))
	(defun untested-symbols ()
	       "return a list of untested symbols"
	       (let ((ret nil))
		    (dolist (symbol (oblist) ret)
			    (unless (or (ignore-symbolp symbol)
					(validation-symbolp symbol))
				    (setq ret (cons symbol ret))))))))

;;;
;;;    Useful functions
;;;
(let ((count 0)
      (prefix "n"))
  (defun genname (&optional (x nil))
	 "generate a new name"
	 (cond ((null x)
		(setq count (1+ count)))
	       ((stringp x)
		(setq prefix x))
	       ((integerp x)
		(setq count x)))
	 (format nil "~A~D" prefix count)))

;; make the directory /tmp/validation
(let ((f (open-facet-by-name "/tmp/validation" (genname)
			     "contents" OCT-CURRENT-VERSION :readwrite)))
  (close-facet f))

(defun gencellname (&optional (x nil))
  "generate a cellname"
  (concatenate 'string "/tmp/validation" (if x (genname x) (genname))))

(defun genfacet (mode)
  "generate (and open) a new facet"
  (open-facet-by-name (gencellname) (genname)
		      "contents" OCT-CURRENT-VERSION mode))

(defmacro with-open-facet-by-name ((symbol &rest open-args) &body body)
  "like with-open-file, but a facet"
  `(let (,symbol (apply #'open-facet-by-name open-args))
	(unwind-protect
	 (progn ,@body)
	 (close-facet ,symbol))))

(defmacro with-genfacet ((symbol mode) &body body)
  "like with-open-facet-by-name but use genfacet"
  `(let ((,symbol (genfacet ,mode)))
	(unwind-protect (progn ,@body)
	 (close-facet ,symbol))))

(defmacro do-containers ((symbol content &optional (mask OCT-ALL-MASK)) &body body)
  "foreach container"
  (let ((generator (gensym)))
       `(do* ((,generator (init-gen-containers ,content ,mask))
	      (,symbol (generate ,generator) (generate ,generator)))
	     ((null ,symbol))
	     ,@body)))

(defmacro do-contents ((symbol container &optional (mask OCT-ALL-MASK)) &body body)
  "foreach content"
  (let ((generator (gensym)))
       `(do* ((,generator (init-gen-contents ,container ,mask))
	      (,symbol (generate ,generator) (generate ,generator)))
	     ((null ,symbol))
	     ,@body)))

(defmacro while (test &body body)
  "while test is true evaluate body"
  `(do ()
      (,test)
      ,@body))

(defun count-attachments (center init-gen)
  "return an a list of counts"
  (map 'list #'cdr (do* ((gen (funcall init-gen center OCT-ALL-MASK))
                         (object (generate gen) (generate gen))
                         (attached nil)
                         (lookup))
                        ((null object) . attached)
                        (setq lookup (assoc object attached))
                        (if (null lookup)
                            (setq attached (acons object 0 attached))
                            (rplacd lookup (1+ (cdr lookup)))))))

(defun average-number-of-attachments (center init-gen)
  "return the average number of attachments (a real)"
  (let ((list (count-attachments object #'init-gen-containers))
        (length (length list)))
       (if (= length 0)
           0
           (/ (apply #'+ list) length))))

;;;
;;;    The Validation Parameters
;;;
(defparameter octlisp-oblist
  '(attach-once attach bb box-of box breadth-first-traversal cell-of
		center-of close-facet copy-facet create-and-attach create-not-attached
		create-or-modify create delete depth-first-traversal detach end-of
		endingAngle-of extract-masks facet-of flush-facet free-facet free-generator
		gen-first-container gen-first-content generalTransform-of generate
		get-by-name get-container-by-name get-facet get-name get-or-create
		init-gen-containers init-gen-contents innerRadius-of instance-of
		is-attached label-of lowerLeft-of mask-member master-of merge-masks mode-of
		modify name-of object-of object-type-of oct-error-status oct-error-string
		open-facet-pair open-facet open-facet-by-name open-master outerRadius-of point-of point
		points-of position-of property-type-of start-of startingAngle-of
		transform-of transform-type-of transform translation-of upperRight-of
		value-of version-of view-of width-of x-of y-of
		boxp generatorp objectp object-form-p object-mask-p object-type-p
		oct-status-p pointp property-type-p transformp transform-mask-p transform-type-p)
  "the functions to validate")

(defparameter OctLisp nil
  "if non-nil, then certain OctLisp-specific error recovery
  mechanisms will be tested.  If nil, then this can be run under CommonLisp")

(defparameter just-examinep nil
  "if nil, then run the whole test
  if non-nil, then just load the tester, but don't execute it")

;;;
;;;    The Validation itself ...
;;;
(format t "~
The OctLisp Test and Validation Suite~%~
    Wendell C. Baker~%~
~%~
    ~A - ~A~%~
" (RCS 'Revision) (RCS 'Date))

(test octlisp-oblist)

(format t "~
The following tests succeeded: ~S~%~
The following tests failed:    ~S~%~
The following were untested:   ~S~%~
" (successes) (failures) (untested-symbols))
;)
