;;;;(
;;;;    Regression testing for OctLisp
;;;;

(defparameter catch-tags '(lightlisp octlisp)
  "a list of the catch tags that are important
  for this regression tester")

(let ((last-regression '(0 . 0))
      (regression-sequence nil))
  (macrolet ((major-of (major-minor)
		       `(car ,major-minor))
	     (minor-of (major-minor)
		       `(cdr ,major-minor)))
	    (labels ((make-catches (body &optional (catch-tags catch-tags))
				   `(catch ',(car catch-tags)
					   ,@(if (null (cdr catch-tags))
						 body
						 (list (make-catches body (cdr catch-tags)))))))
		    (defmacro regression (major minor description &body test-forms)
			      "major and minor should be integers
			      test-forms may be any form at all"
			      (let ((this-regression (cons major minor))
				    (codelist nil)
				    (magic-key (gensym))
				    (form (gensym))
				    (result (gensym))
				    (catch-result (gensym))
				    tmp)
				   ;; insure that all regression tests are in the correct order
				   (when (or (< (major-of this-regression) (major-of last-regression))
					     (and (= (major-of this-regression) (major-of last-regression))
						  (<= (minor-of this-regression) (minor-of last-regression))))
					 (format t "Regression Error~%    ~
						 regression ~S is not increasing in the regression sequence~%    ~
						 last regression: ~S~%regressions: ~S~%"
						 this-regression last-regression regression-sequence)
					 (throw 'regression nil))
				   ;; update the bookkeeping
				   (setq last-regression this-regression)
				   (setq regression-sequence (cons last-regression regression-sequence))
				   ;; generate the code
				   (dolist (tf test-forms)
					   (setq tmp (let* ((catch-body `((setq ,result ,tf)
									  ',magic-key))
							    (catch-form (make-catches catch-body)))
							   `(let ((,form ',tf)
								  ,result
								  ,catch-result)
								 (format t "Form:~%    ") (pprint ,form) (terpri)
								 (setq ,catch-result
								       ,catch-form)
								 (format t "Result:~%    ") (pprint ,result) (terpri)
								 (unless (eq ',magic-key ,catch-result)
									 (princ "Warning: a throw occurred (the result is from the throw)")
									 (terpri))
								 (terpri)
								 (terpri))))
					   ;; use nconc to append it to the list
					   (setq codelist (cons tmp codelist)))
				   ;; reverse the codelist
				   (labels ((reverse (list)	; not flet
						     (cond ((null list) nil)
							   ((null (cdr list)) list)
							   (t (nconc (reverse (cdr list)) (cons (car list) nil))))))
					   (setq codelist (reverse codelist)))
				   ;; return the progn
				   `(progn
				     (format t "-----~%Regression Test ~S.~S~%~%Description:~%~A~%~%"
					     ,(major-of this-regression)
					     ,(minor-of this-regression)
					     ,description)
				     ,@codelist
				     (format t "-----~%~%~%~%")))))))

(dotimes (i 2) (terpri))
(princ lisp-implementation-version)
(dotimes (i 3) (terpri))

;;;
;;;    Release 1 --------------------
;;;
(regression 1 0
  "trivial - it ran")

(regression 1 1
  "OLMakeObject now takes the whole octObject (not an objectId)"
  (open-facet-by-name "/tmp/finkie" "symbolic" "contents" "" :readwrite-append))

(regression 1 2
  "OLObjectReform now does a octGetById in order to refresh the fields.
  OLObjectReform and OLObjectForm now behave correctly with respect
  to indexing into the octObject field tables"
  (let* ((facet (open-facet-by-name "/tmp/flop" "symbolic" "contents" "" :readwrite-append))
	 (prop (create facet `(object ,OCT-PROP
				      ((name . "SAM-I-AM")
				       (value . "GREEN-EGGS"))))))
	(object-of facet)))

(regression 1 3
  "added merge-masks extract-masks mask-member"
  (merge-masks OCT-BAG OCT-PROP-MASK (merge-masks OCT-GEO-MASK))
  (extract-masks OCT-ALL-MASK OCT-GEO-MASK)
  (extract-masks OCT-FACET OCT-FACET)
  (merge-masks (extract-masks OCT-ALL-MASK OCT-FACET OCT-INSTANCE)
	       OCT-GEO-MASK)
  (mask-member OCT-CIRCLE OCT-GEO-MASK)
  (mask-member (merge-masks OCT-INSTANCE OCT-FACET)
	       (extract-masks OCT-ALL-MASK OCT-GEO-MASK))
  (let ((all-transform-mask (merge-masks OCT-NO-TRANSFORM OCT-MIRROR-X OCT-MIRROR-Y
					 OCT-ROT90 OCT-ROT180 OCT-ROT270
					 OCT-MX-ROT90 OCT-MY-ROT90 OCT-FULL-TRANSFORM))
	(geo-transform-mask (merge-masks OCT-MIRROR-X OCT-MIRROR-Y
					 OCT-MX-ROT90 OCT-MY-ROT90)))
       (merge-masks OCT-MIRROR-X OCT-MIRROR-Y-MASK geo-transform-mask)
       (extract-masks all-transform-mask geo-transform-mask)
       (extract-masks OCT-ROT90 OCT-ROT90)
       (merge-masks (extract-masks all-transform-mask OCT-MX-ROT90 OCT-MY-ROT90)
		    geo-transform-mask)
       (mask-member OCT-ROT90 geo-transform-mask)
       (mask-member (merge-masks OCT-FULL-TRANSFORM OCT-NO-TRANSFORM)
		    (extract-masks all-transform-mask geo-transform-mask))))

(regression 1 4
  "moved the octGetById from OLObjectReform to OLExtractObject so that
  ALL users of OLExtractObject need not do a octGetById - printers won't
  throw on invalid objectId's though"
  (let ((x (open-facet-by-name "/tmp/fippo" "symbolic"
			       "contents" OCT-CURRENT-VERSION :readwrite-append)))
       (flet ((field-print (object name extractor)
			   (princ name)
			   (princ " of ")
			   (pprint object)
			   (princ " ")
			   (pprint (funcall extractor x))
			   (terpri)))
	     (field-print x "Cell" #'cell-of)
	     (field-print x "View" #'view-of)
	     (field-print x "Facet" #'facet-of)
	     (field-print x "Version" #'version-of)
	     (field-print x "Mode" #'mode-of))))

(regression 1 5
  "fixed the error printf in facet-of to say facet-of, not view-of"
  (let* ((x (open-facet-by-name "/tmp/fowpoidal" "symbolic"
				"contents" OCT-CURRENT-VERSION :readwrite-append))
	 (p (create x `(object ,OCT-PROP ((name . "prop") (value . "eller"))))))
	(flet ((field-print (object name extractor)
			    (princ name)
			    (princ " of ")
			    (pprint object)
			    (princ " ")
			    (pprint (funcall extractor x))
			    (terpri)))
	      (field-print x "Facet" #'facet-of)
	      (catch 'octlisp
		     (field-print p "Facet" #'facet-of)))))

(regression 1 6
  "create, create-not-attached, create-and-attach,
  modify, get-or-create, create-or-modify, if the container
  is not an object, prints correctly"
  (create 4 `(object ,OCT-PROP ((name . "N") (value . 2))))
  (create-not-attached 1.2 `(object ,OCT-PROP ((name . "N") (value . 2))))
  (modify "yes" `(object ,OCT-PROP ((name . "N") (value . 2)))))

(regression 1 7
  "OLObjectForm was not setting the property type"
  (setq facet (open-facet-by-name "/tmp/faddoiy" "physical" "contents" OCT-CURRENT-VERSION :readwrite-append))
  (create facet `(object ,OCT-PROP ((name . "O-string-O") (value . "bing-o"))))
  (create facet `(object ,OCT-PROP ((name . "O-integer-O") (value . 1))))
  (create facet `(object ,OCT-PROP ((name . "O-real-O") (value . 1.1))))
  (create facet `(object ,OCT-PROP ((name . "O-integer-array-O") (value 1 2 3 4 5 6 7 8))))
  (create facet `(object ,OCT-PROP ((name . "O-real-array-O") (value 1 2 3 4 5 6 7 8.0))))
  (create facet `(object ,OCT-PROP ((name . "O-stranger-O") (value 1 "" 2 3 4 5 6 7 8.0))))
  (create facet `(object ,OCT-PROP ((name . "O-object-O") (value . ,facet)))))

(regression 1 8
  "fixed a bug in scan that inserted garbage-collected objects
  into LLMachine.oblist (not directly demonstrable here)")

(regression 1 9
  "OLExtractObject used ~A to print invalid octId's - now uses ~D"
  (let* ((facet (open-facet-by-name "/tmp/popple" "physical" "contents" OCT-CURRENT-VERSION :readwrite))
	 (prop (create facet `(object ,OCT-PROP ((name . "name-o") (value . "bing-o"))))))
	(close-facet facet)
	prop))

(regression 1 10
  "generate would not return nil on OCT_GEN_DONE, only on error"
  (let* ((facet (open-facet-by-name "/tmp/flopple" "physical" "contents" OCT-CURRENT-VERSION :readwrite)))
	(dotimes (i 10)
		 (create facet `(object ,OCT-PROP
					((name . ,(case i
							(0 "zeroth")
							(1 "first")
							(2 "second")
							(3 "third")
							(t (format nil "~D-th" i))))
					 (value . ,i)))))
	(do* ((gen (init-gen-contents facet OCT-PROP-MASK))
	      (p (generate gen) (generate gen)))
	     ((null p))
	     (format t "Property named ~A valued at ~S~%"
		     (name-of p) (value-of p)))
	(close-facet facet)))

(regression 1 11
  "printers now understand about oct-null-id"
  (prin1 oct-null-id)
  (print oct-null-id)
  (princ oct-null-id)
  (pprint oct-null-id))

(regression 1 12
  "octId fields were not being extracted from the object-form
  this affected create operations on OCT-TERM's only"
  (let* ((facet (open-facet-by-name "/tmp/fluid" "physical" "contents" OCT-CURRENT-VERSION :readwrite)))
	(dotimes (i 10)
		 (create facet `(object ,OCT-TERM
					((name . ,(case i
							(0 "zeroth")
							(1 "first")
							(2 "second")
							(3 "third")
							(t (format nil "~D-th" i))))
					 (instance . ,oct-null-id)))))
	(do* ((gen (init-gen-contents facet OCT-TERM-MASK))
	      (term (generate gen) (generate gen)))
	     ((null term))
	     (format t "Term named ~A formality of ~S~%"
		     (name-of term) (instance-of term)))
	(close-facet facet)))

(regression 1 13
  "added boxp, generatorp, objectp, object-form-p, object-mask-p, object-type-p,
  oct-status-p, pointp, property-type-p, transformp, transform-mask-p, transform-type-p"
  (boxp (box (point 3 4) (point 8 2)))
  (generatorp (let ((facet (open-facet-by-name "/tmp/vase" "physical" "contents" OCT-CURRENT-VERSION :readwrite)))
		   (init-gen-contents facet OCT-ALL-MASK)))
  (objectp (let ((facet (open-facet-by-name "/tmp/nice" "physical" "contents" OCT-CURRENT-VERSION :readwrite)))
		(init-gen-contents facet OCT-ALL-MASK)))
  (object-form-p `(object ,OCT-FACET ((cell . "yes")
				      (view . "no")
				      (facet . "contents")
				      (mode . :readwrite))))
  (object-mask-p OCT-GEO-MASK)
  (object-type-p OCT-EDGE)
  (oct-status-p OCT-OK)
  (pointp (point 0 0))
  (property-type-p OCT-NULL)
  (transformp (transform (point 0 0) OCT-NO-TRANSFORM))
  (transform-mask-p OCT-ROT90-MASK)
  (transform-type-p OCT-FULL-TRANSFORM))

(regression 1 14
  "object-mask printer now prints the smallest representation possible"
  (merge-masks OCT-INSTANCE-MASK OCT-GEO-MASK)
  (extract-masks OCT-ALL-MASK OCT-INSTANCE-MASK OCT-PROP-MASK OCT-LAYER-MASK))

(regression 1 15
  "mode-of returns the correct symbols now"
  (mode-of `(object ,OCT-FACET
		    ((cell . "yes")
		     (view . "no")
		     (facet . "maybeso")
		     (version . "")
		     (mode . :readonly))))
  (let* ((facet (open-facet-by-name "/tmp/mound" "symbolic" "contents" "" :readwrite)))
	(prog1 (mode-of facet)
	       (close-facet facet))))

(regression 1 16
  "extract-masks was not masking off, it was producing the mask itself"
  (extract-masks oct-all-mask oct-instance oct-prop oct-bag oct-layer)
  (extract-masks (merge-masks oct-instance oct-prop oct-bag)
		 oct-instance))

(regression 1 17
  "a string can be nil to mean NULL for most strings in object-form"
  (name-of `(object ,OCT-NET ((name))))
  (let* ((facet (open-facet-by-name "/tmp/spillage" "symbolic" "contents" "" :readwrite)))
	(prog1 (name-of (create facet `(object ,OCT-BAG 
					       ((name)))))
	       (close-facet facet)))
  (let* ((facet (open-facet-by-name "/tmp/spillage1" "symbolic" "contents" "" :readwrite)))
	(prog1 (name-of (create facet `(object ,OCT-NET 
					       ((name)))))
	       (close-facet facet)))
  (let* ((facet (open-facet-by-name "/tmp/spillage1" "symbolic" "contents" "" :readwrite)))
	(prog1 (label-of (create facet `(object ,OCT-LABEL 
						((label)
						 (position . ,(point 3 99))))))
	       (close-facet facet))))

(regression 1 18
  "had the policy for transform-cell (full-transform vs other) mixed up"
  (transform (point 1 2) oct-no-transform)
  (transform (point 3 4) oct-mirror-x)
  (transform (point 5 6) oct-mirror-y)
  (transform (point 7 8) oct-rot90)
  (transform (point 8 9) oct-rot180)
  (transform (point 10 11) oct-rot270)
  (transform (point 12 13) oct-mx-rot90)
  (transform (point 14 15) oct-my-rot90)
  (transform (point 16 17) oct-full-transform '(2.2 0.0 0.0 2.2)))

(regression 1 19
  "retrieving points from an object called LLMakeReal, not LLMakePoint"
  (let ((facet (open-facet-by-name "/tmp/mound" "symbolic" "contents" "" :readwrite-append)))
       (create facet `(object ,OCT-PATH ((width . 300)
					 (points . ,(list (point 0 1)
							  (point 2 3)
							  (point 3 4)
							  (point 4 5)
							  (point 5 6)
							  (point 6 5)
							  (point 5 4)
							  (point 4 3)
							  (point 3 2)
							  (point 2 1)
							  (point 1 0)
							  (point 0 0))))))
       (gen-first-content facet OCT-PATH-MASK)))

(regression 1 20
  "attach and get-facet returned fail on <= OCT-OK
  the correct behavior is to return fail on < OCT-OK"
  (let* ((facet (open-facet-by-name "/tmp/between" "symbolic" "contents" "" :readwrite-append))
	 (prop (create-not-attached facet `(object ,OCT-PROP ((name . "pink")
							      (value . "nibble"))))))
	(attach facet prop))
  (let* ((facet (open-facet-by-name "/tmp/them" "symbolic" "contents" "" :readwrite-append))
	 (prop (create-not-attached facet `(object ,OCT-PROP ((name . "pink")
							      (value . "nibble"))))))
	(get-facet prop)))

(regression 1 21
  "gen-first-container/gen-first-content if passed a not-an-object would seg-fault
  in the error printout; also, gen-first-content called octGenFirstContainer"
  (gen-first-container nil oct-term-mask)
  (let* ((facet (open-facet-by-name "/tmp/hang" "symbolic" "contents" OCT-CURRENT-VERSION :readwrite))
	 (term (create facet `(object ,OCT-TERM ((name . "t")
						 (instance . ,oct-null-id))))))
	(gen-first-content facet OCT-TERM-MASK))
  (gen-first-content 45 oct-term-mask)
  (let* ((facet (open-facet-by-name "/tmp/hang" "symbolic" "contents" OCT-CURRENT-VERSION :readwrite))
	 (term (create facet `(object ,OCT-TERM ((name . "t")
						 (instance . ,oct-null-id))))))
	(gen-first-container term OCT-FACET-MASK)))

(regression 1 22
  "master-of was searching for the instance-of (instance key) if given
  an object-form"
  (master-of `(object ,OCT-INSTANCE
		      ((name . "noodle")
		       (transform . ,(transform (point 0 0) oct-no-transform))
		       (master . "/dev/null")
		       (view . "symbolic")
		       (version . "")))))

(regression 1 23
  "view-of was searching for the version-of (view key) if given
  an object-form"
  (view-of `(object ,OCT-FACET
		    ((cell . "cellular")
		     (view . "sunset")
		     (facet . "diamond")
		     (version . "none")
		     (mode . :readwrite))))
  (view-of `(object ,OCT-INSTANCE
		    ((master . "The-Dominator")
		     (view . "from-below")
		     (version . "none")
		     (name . "george")
		     (translation . ,(transform (point 0 0) oct-mirror-x)))))
  (progn
   (setq ship (open-facet-by-name "/tmp/left" "symbolic
				  "contents" OCT-CURRENT-VERSION
				  :readwrite-append))
   (setq ship-instance
	 (let ((facet (open-facet-by-name "/tmp/right" "symbolic"
					  "contents" OCT-CURRENT-VERSION
					  :readwrite-append)))
	      (create facet `(object ,OCT-INSTANCE
				     ((name . "mine")
				      (transform . ,(transform (point 0 0)
							       oct-no-transform))
				      (master . ,(cell-of ship))
				      (view . ,(view-of ship))
				      (version . ,OCT-CURRENT-VERSION)))))))
  (view-of ship)
  (view-of ship-instance))

(regression 1 24
  "get-by-name/get-container-by-name had a bad errorF if
  container/content was not an object as well if name was not a string"
  (get-by-name nil OCT-TERM "YAP")
  (get-container-by-name nil OCT-TERM "YAP")
  (progn
   (setq ship (create
	       (open-facet-by-name "/tmp/massive" "symbolic"
				   "contents" OCT-CURRENT-VERSION
				   :readwrite-append)
	       `(object ,OCT-BAG ((name . "sack")))))
   (create ship `(object ,OCT-PROP ((name . "BELOW")
				    (value . "down"))))
   (create ship `(object ,OCT-PROP ((name . "ABOVE")
				    (value . "up")))))
  (get-by-name ship OCT-PROP 45)
  (get-container-by-name ship OCT-PROP 87)
  (get-by-name ship OCT-PROP "BELOW")
  (get-container-by-name ship OCT-PROP "ABOVE"))

(regression 1 25
  "OLObjectForm would not pull the octBox correctly out of the OCT-BOX object-form"
  (setq boxes (let ((facet (open-facet-by-name "/tmp/rectoid" "physical"
					       "contents" "" :readwrite))
		    (boxes nil))
		   (dotimes (i 10 boxes)
			    (let ((box (create facet
					       `(object ,OCT-BOX
							((box . ,(box (point i i)
								      (point (+ i 10)
									     (+ i 10)))))))))
				 (setq boxes (acons i box boxes))))))
  ;; insure what you get out is what you put in
  (map 'list
       #'(lambda (i-dot-box)
		 (let ((i (car i-dot-box))
		       (box (cdr i-dot-box)))
		      (and (= i (x-of (lowerLeft-of (box-of box))))
			   (= i (y-of (lowerLeft-of (box-of box))))
			   (= (+ i 10) (x-of (upperRight-of (box-of box))))
			   (= (+ i 10) (y-of (upperRight-of (box-of box)))))))
       boxes))

;;;
;;;    Release 2 --------------------
;;;
(regression 2 0
  "open-facet changed to open-facet-by-name and open-facet now takes an object-form or object
  copy-facet now returns new-form on success (not just t).
  The open-facet change is the for going to 2.0"
  (flet ((test-mode (mode)
		    (let* ((cell "/tmp/shipwhick-and-pip)
			   (view "physical")
			   (facet (open-facet `(object ,OCT-FACET
						       ((cell . ,cell)
							(view . ,view)
							(facet . "contents")
							(version . ,OCT-CURRENT-VERSION)
							(mode . ,mode)))
					      mode)))
			  (when (not facet)
				(format t "Could not open ~A:~A:contents as ~S because ~A~%"
					cell view mode (oct-error-string))
				(throw 'octlisp nil))
			  (format t "facet ~a:~a:contents mode is ~S~%"
				  (cell-of facet)
				  (view-of facet)
				  (mode-of facet))
			  (if (not (close-facet facet))
			      (format t "Could not close ~A:~A:contents as ~S because ~A~%"
				      cell view mode (oct-error-string))))))
	
	(test-mode :readonly)
	(test-mode :readwrite)
	(test-mode :readwrite-append))
  (flet ((test-mode (mode)
		    (let* ((cell "/tmp/shipwhick-and-pop")
			   (view "physical")
			   (facet (open-facet-by-name cell view "contents" OCT-CURRENT-VERSION mode)))
			  (when (not facet)
				(format t "Could not open ~A:~A:contents as ~S because ~A~%"
					cell view mode (oct-error-string))
				(throw 'octlisp nil))
			  (format t "facet ~a:~a:contents mode is ~S~%"
				  (cell-of facet)
				  (view-of facet)
				  (mode-of facet))
			  (if (not (close-facet facet))
			      (format t "Could not close ~A:~A:contents as ~S because ~A~%"
				      cell view mode (oct-error-string))))))
	
	(test-mode :readonly)
	(test-mode :readwrite)
	(test-mode :readwrite-append))
  (let ((facet (open-facet-by-name "/tmp/oldfowpoidal-didly" "physical" "contents" "" :readwrite))
	new)
       (macrolet ((p (name value)
		     "a shorthand for property creation"
		     `(create facet '(object ,OCT-FACET
					     ((name . ,name)
					      (value . ,value))))))
		 (p "YES" "no")
		 (p "NO" "yes")
		 (dotimes (i 5)
			  (p "yop" i)))
       (close-facet facet)
       (setq new (open-facet
		  (copy-facet `(object ,OCT-FACET
				       ((cell . "/tmp/fowpoidal-didly")
					(view . "physical")
					(facet . "contents")
					(version . "")
					(mode . :readonly)))
			      `(object ,OCT-FACET
				       ((cell . "/tmp/old-fowpoidal-didly")
					(view . "physical")
					(facet . "contents")
					(version . "")
					(mode . :readonly))))
		  :readwrite-append))
       (pprint new) (terpri)
       (do* ((gen (init-gen-contents new oct-all-mask))
	     (o (generate gen) (generate gen)))
	    ((null o))
	    (princ "    ") (pprint o) (terpri))))

(regression 2 1
  "OLObjectForm had dome serious problems with respect to integer array props
  real array props, stranger props, paths and polygons dealing with the check-for-correctness
  paths and polygons - could not extract the point-list members correctly into the point array"
  (setq fig (open-facet-by-name "/tmp/pigship" "gross" "contents" "" :readwrite))
  ;; this is a repeat of 1.7
  (create fig `(object ,OCT-PROP ((name . "O-nothor-integer-array-O") (value 1 2 3 4 5 6 7 8))))
  (create fig `(object ,OCT-PROP ((name . "O-nothor-real-array-O") (value 1 2 3 4 5 6 7 8.0))))
  (create fig `(object ,OCT-PROP ((name . "O-nothor-stranger-O") (value 1 "" 2 3 4 5 6 7 8.0))))
  ;
  (create fig `(object ,OCT-PATH ((width . 900) (points . ,(list
							    (point 1 2)
							    (point 3 4)
							    (point 5 6)
							    (point 7 8))))))
  (create fig `(object ,OCT-POLYGON ((points . ,(list
						 (point -1 -2)
						 (point -3 -4)
						 (point -5 -6)
						 (point -7 -8)
						 (point -9 -10)))))))

(regression 2 2
  "the argument facet-form to open-facet may be an object or object-form"
  (open-facet (open-facet (open-facet `(object-form ,OCT-FACET
						    ((cell . "/tmp/hype")
						     (view . "symbolic")
						     (facet . "contents")
						     (version . "")
						     (mode . :readonly))
						    :readwrite)))))

(regression 2 3
  "open-facet-pair now checks to see if alternate is a string (not facet)"
  (open-facet-pair (open-facet-by-name "/tmp/slime" "ball" "contents" "" :readwrite)
		   "interface" :readwrite))

(regression 2 4
  "open-facet was broken - it accepted nothing; this is a repeat of 2.2.
  open-facet-pair now accepts an object also"
  (open-facet (open-facet (open-facet `(object ,OCT-FACET
					       ((cell . "/tmp/hype-again")
						(view . "symbolic")
						(facet . "contents")
						(version . "")
						(mode . :readonly))
					       :readwrite))))
  (open-facet-pair (object-of (open-facet `(object ,OCT-FACET
						   ((cell . "/tmp/hop")
						    (view . "symbolic")
						    (facet . "contents")
						    (version . "")
						    (mode . :readonly))
						   :readwrite)))))

(regression 2 5
  "Changed over to LLInternString vs LLInternSymbol (not directly testable)")

(regression 2 6
  "transform-of required 2 args and 1 optional (like transform)
  now it only requires one"
  (transform-of `(object ,OCT-INSTANCE
			 ((version . "")
			  (view . "structure_instance")
			  (master . "PCW_COUNTER")
			  (name . "PCW_COUNTER")
			  (transform . (transform (point 0 0)
						  OCT-NO-TRANSFORM))))))

(regression 2 7
  "helplib had OL_OHelper_Initialized defined instead of OL_Helper_Initialize
  fixed, but not demonstrable from here")

(regression 2 8
  "OLInitializedP was not set to FALSE in OLUninitialize
  nothing demonstrable from here")

(regression 2 9
  "helplib moved to oct_helperlib
  OL_Helper_Initialize transformed to OL_Octh_Initialize
  OL_Helper_Uninitialize transformed to OL_Octh_Uninitialize
  nothing demonstrable from here")

(regression 2 10
  "trivial change in type-of.c
  not demonstrable from here")

(regression 2 11
  "use LL_eq in OL_object_formp instead of ==
  not demonstrable from here")

(regression 2 12
  "OLGetPoints sets *pnarray to zero before calling octGetPoints
  was exhibited by printing out a polygon"
  (let* ((facet (open-facet `(object ,OCT-FACET
				     ((cell . "/tmp/fipper")
				      (view . "symbolic")
				      (facet . "contents")
				      (version . "")
				      (mode . :readwrite)))
			    :readwrite))
	 (polygon (create facet `(object ,OCT-POLYGON
					 ((points . ,(list (point 0 0)
							   (point -1 1)
							   (point -2 2)
							   (point -3 3))))))))
	(format t "polygon ~S~%" polygon)
	(object-of polygon)))

(regression 2 13
  "Bug in OLObjectForm getting array valued object-forms
  such as points, reals and integers"
  (let* ((facet (open-facet `(object ,OCT-FACET
				     ((cell . "/tmp/gop")
				      (view . "symbolic")
				      (facet . "contents")
				      (version . "")
				      (mode . :readwrite)))))
	 (path (create facet `(object ,OCT-PATH
				      ((width . 0)
				       (points . ,(list (point -1 -2)
							(point -3 -4)
							(point -4 -5)))))))
	 (polygon (create facet `(object ,OCT-POLYGON
					 ((points . ,(list (point -1 -2)
							   (point -3 -4)
							   (point -4 -5)))))))
	 (integers (create facet `(object ,OCT-PROP
					  ((name . "integers")
					   (value . ,(list 0 1 2 3 4 5 6))))))
	 (reals (create facet `(object ,OCT-PROP
				       ((name . "integers")
					(value . ,(list 0.0 1 2 3 4 5 6)))))))
	(dolist (i (list facet path polygon integers reals))
		(format t "Object ~S~%" i))
	(close-facet facet)))

(regression 2 14
  "OLGetPoints was passing parray, not *parray to octGetPoints
  OLGlobalOctStatus updated on octGetPoints"
  (let ((facet (open-facet `(object ,OCT-FACET
				    ((cell . "/tmp/polygonal")
				     (view . "physical")
				     (facet . "contents")
				     (version . ,OCT-CURRENT-VERSION)
				     (mode . :readwrite)))
			   :readwrite)))
       (create facet `(object ,OCT-POLYGON
			      ((points . ,(list (point 1 2)
						(point 3 4)
						(point 4 5)
						(point 5 6)
						(point 6 7)
						(point 7 8))))))
       (gen-first-content facet OCT-POLYGON-MASK))
  (let ((facet (open-facet `(object ,OCT-FACET
				    ((cell . "/tmp/pathal")
				     (view . "physical")
				     (facet . "contents")
				     (version . ,OCT-CURRENT-VERSION)
				     (mode . :readwrite)))
			   :readwrite)))
       (create facet `(object ,OCT-PATH
			      ((width . 3)
			       (points . ,(list (point 1 2)
						(point 3 4)
						(point 4 5)
						(point 5 6)
						(point 6 7)
						(point 7 8))))))
       (gen-first-content facet OCT-PATH-MASK)))

(regression 2 15
  " oct_helperlib.h had externs for an OL_Helper_Initialize and
  OL_Helper_Uninitialize instead of OL_Octh_Initialize and OL_Octh_Uninitialize
  not demonstrable from here - just compiles better now")

;)
