#ifndef BOSS_H
#define BOSS_H
#pragma once

#include <stream.h>

#include "oct.h"
#include "symlib.h"

typedef struct octObject bossOctObject;

#define BOSS_PKG_NAME "boss"

class BossFacet : public bossOctObject  {
 public:
    BossFacet() {};
    BossFacet(  char* cell,
	      char* view,
	      char* technology = "scmos",
	      char* facet = "contents", 
	      char* mode = "w",
	      char* version = OCT_CURRENT_VERSION);
    ~BossFacet();
    void flush();
    void createInterface();
    char *cell() 		{return contents.facet.cell;}
    char *view() 		{return contents.facet.view;}
    char *facet() 	{return contents.facet.facet;}
    char *version() 	{return contents.facet.version;}
    char *mode() 		{return contents.facet.mode;}
    void print_name();
};

extern class BossTerm;

class BossNet : public bossOctObject
{
 public:
    BossNet( int i ) { objectId = i; }
    BossNet(BossFacet& facet, char* name = 0);
    char*& net_name() {return contents.net.name;}
    void add_term(BossTerm& term);
    void operator=( int i ) { objectId = i; }
    int  operator!=( int i ) { return objectId != i; }
};

class BossPhysicalObject: public bossOctObject
{
    int& transform_x()   {return (contents.instance.transform.translation.x);}
    int& transform_y()   {return (contents.instance.transform.translation.y);}
 public:
    octBox BB;			/* The abutment box relative to the master (instances only) */

    BossPhysicalObject() {};
    octBox& getBB();
    void move_x( int );
    void move_y( int );
    void transform(octTransformType);
    octTransform& get_transform() {return (contents.instance.transform);}

    void modify();


    void move_to_origin();

    void below   ( BossPhysicalObject&, int = 0 );
    void above   ( BossPhysicalObject&, int = 0 );
    void left_of ( BossPhysicalObject&, int = 0 );
    void right_of( BossPhysicalObject&, int = 0 );
    void align_x ( BossPhysicalObject&, int = 0 );
    void align_y ( BossPhysicalObject&, int = 0 );

    void between_x ( BossPhysicalObject& obj1, BossPhysicalObject& obj2, int = 0 );
    void between_y ( BossPhysicalObject& obj1, BossPhysicalObject& obj2, int = 0 );

    int upperRight_x();
    int upperRight_y();
    int lowerLeft_x();
    int lowerLeft_y();
    int center_x();
    int center_y();
    int extent_x();
    int extent_y();

};

class BossInstance : public BossPhysicalObject {
  void create(bossOctObject& facet);
  friend class BossContact;
 public:
  BossInstance() {};
  BossInstance(bossOctObject& facet,
	       char* name, 
	       char* master,
	       char* view,
	       char* facet_name = "contents", 
	       char* version = OCT_CURRENT_VERSION);
  void offsetBB(int top, int bot, int left, int right);
  BossTerm& term(char *);
  char* format_name();
  void print_name();
};

class BossTerm : public BossPhysicalObject {
  octStatus get(BossInstance& inst, char *name);
public:
  BossTerm() : () {};
  BossTerm(BossInstance& inst, char *name);
  BossTerm(bossOctObject& term);
  char *term_name() {return contents.term.name;}
  octId instId() {return contents.term.instanceId;}
  int num_layers();
  char* layer(int num);
};

class BossContact : public BossInstance {
  void fill_in(bossOctObject* facet, char* layer1, char* layer2, int dim1 = 0, int dim2 = 0);
  void create(bossOctObject& facet);
public:
  BossContact(bossOctObject* facet);
  BossContact(bossOctObject& facet, 
	      char* layer1 = "MET1", 
	      char* layer2 = "MET1", 
	      int dim1 = 80, 
	      int dim2 = 80);
  BossTerm& term();
  void add_layer(char* name);
  void get(octId objectId);
};

class BossLayer : public bossOctObject {
public:
  BossLayer(BossFacet& facet, char *name);
  BossLayer(bossOctObject* facet, char *name);
  BossLayer( int i) { objectId = i;}
  int min_spacing();
  int min_width();
};

class BossBox : public BossPhysicalObject {
 public:
  BossBox(BossFacet& facet, 
	  int height,
	  int width,
	  char* layer);
};

class BossFormalTerm : public bossOctObject {
 public:
  BossFormalTerm( BossTerm& term, char *name);
  BossFormalTerm( BossContact& cont, char *name);
};

class BossInstArray {
  BossInstance** array;
  int first_row;
  int first_col;
  int final_row;
  int final_col;
  int num_rows;
  int num_cols;
  BossInstance*& get_elmt(int row, int col);
  friend class BossSubArray;
public:
  BossInstArray() {};
  BossInstArray( int rows, int cols);
  int last_row() {return final_row - first_row;}
  int last_col() {return final_col - first_col;}
  BossInstance& elmt(int row, int col);
  void move_x( int delta );
  void move_y( int delta );

  void populate_all( BossFacet& facet, char* master, char*view, char* name = 0,char* facet_name = "contents");
  void populate_row( BossFacet& facet, int row_num, char* master, char*view, char* name = 0,char* facet_name = "contents");
  void populate_col( BossFacet& facet, int col_num, char* master, char*view, char* name = 0,char* facet_name = "contents");
  void populate_loc( BossFacet& facet, int row_num, int col_num, char* master, char*view, char* name = 0,char* facet_name = "contents");

  void transform_all(octTransformType trans);
  void transform_row(octTransformType trans, int row );
  void transform_col(octTransformType trans, int col );

  void place_all();
  void place_row( int row_num, int delta_x, int delta_y );
  void place_col( int col_num, int delta_x, int delta_y );

  void promote_all(BossFacet& facet, char* term_name, char*f_term_name);
  void promote_row(BossFacet& facet, int row, char* term_name, char*f_term_name);
  void promote_col(BossFacet& facet, int col, char* term_name, char*f_term_name);

  void connect_all( char* term_name, BossNet& net);
  void connect_row( int row, char* term_name, BossNet& net);
  void connect_col( int col, char* term_name, BossNet& net);

  void offsetBB_all( int top, int bot, int left, int right);
  void offsetBB_row( int row, int top, int bot, int left, int right);
  void offsetBB_col( int col, int top, int bot, int left, int right);

  void zipper_by_row( BossFacet& facet, int row1, char* term1_name, int row2, char* term2_name);
  void zipper_by_col( BossFacet& facet, int col1, char* term1_name, int col2, char* term2_name);
};

class BossSubArray : public BossInstArray {
public:
  BossSubArray( BossInstArray& base_array, 
	       int row_start, 
	       int row_end, 
	       int col_start, 
	       int col_end);
};

class BossTransistor : public BossInstance {
public:
  BossTransistor( BossFacet& facet, char *type, int w, int l, char *name = "trans");
};



//
//
//   exported routines.
//
//

// nets

enum BossNetConnType {BOSS_ALL, BOSS_UNCONNECTED};
extern void bossConnectAll( char *term_name, BossNet& net, enum BossNetConnType type = BOSS_ALL);

// routing

overload bossStraightRoute;
extern void bossStraightRoute( BossTerm &term1, BossTerm &term2, int width, char *layer_name);
extern void bossStraightRoute( BossTerm &term1, BossContact &cont2, int width, char *layer_name);
extern void bossStraightRoute( BossContact &cont1, BossTerm &term2, int width, char *layer_name);
extern void bossStraightRoute( BossContact &cont1, BossContact &cont2, int width, char *layer_name);

enum BossRouteType {BOSS_VERTICAL, BOSS_HORIZONTAL};

overload bossZRoute;

extern void bossZRoute( BossTerm &term1, BossTerm &term2, int width,
		 enum BossRouteType center_seg_dir, 
		 char *layer_name1, 
		 char *layer_name2 = 0, 
		 char *layer_name3 = 0);

extern void bossZRoute(  BossTerm &term1, BossContact &cont2, int width,
		 enum BossRouteType center_seg_dir, 
		 char *layer_name1, 
		 char *layer_name2 = 0, 
		 char *layer_name3 = 0);

extern void bossZRoute(  BossContact &cont1, BossTerm &term2, int width,
		 enum BossRouteType center_seg_dir, 
		 char *layer_name1, 
		 char *layer_name2 = 0, 
		 char *layer_name3 = 0);

extern void bossZRoute(  BossContact &cont1, BossContact &cont2, int width,
		 enum BossRouteType center_seg_dir, 
		 char *layer_name1, 
		 char *layer_name2 = 0, 
		 char *layer_name3 = 0);

enum BossDirection {BOSS_UP, BOSS_DOWN, BOSS_RIGHT, BOSS_LEFT};
overload bossStub;
extern BossContact& bossStub( BossTerm& term, enum BossDirection dir, 
			     int offset = 0, char* layer = 0, int width = 0);
extern BossContact& bossStub( BossContact& cont, enum BossDirection dir, 
			     int offset = 0, char* layer = 0, int width = 0);
extern int bossTryRoute( BossTerm &term1, BossTerm &term2 );

// terminals

extern void bossPromote( BossTerm& term);

// error handling

extern void bossWarning(char *);
extern void bossMessage(char *);

// constraints

extern void bossClearConstraints(BossFacet& facet);
extern void bossSetConstraint(int BossDirection, BossInstance& inst1, BossInstance& inst2, int min, int max);
extern void bossSetSideConstraint(BossInstance& inst, int side, int min, int max);

// search

extern int bossSearchRegion( bossOctObject& facet, octBox& box, char* layer_name );

// adj

extern void bossRouteAdjacent( BossFacet& facet, BossInstance& inst1, BossInstance& inst2);

extern void bossPromoteAll( BossInstance& inst, char* extension = 0);

extern char* bossStrsav( const char* s  );

// properties

overload bossCreateProp;

extern void bossCreateProp( bossOctObject& obj, char* name, int value);
extern void bossCreateProp( bossOctObject& obj, char* name, double value);
extern void bossCreateProp( bossOctObject& obj, char* name, char* value);

extern int bossGetPropInt(bossOctObject& obj, char* name);
extern double bossGetPropReal(bossOctObject& obj, char* name);
extern char* bossGetPropStr(bossOctObject& obj, char* name);

extern void bossDebug_off();
extern void bossDebug_on();


extern void bossBegin( int argc, char** argv );
extern void bossEnd( int status );    			// Does not return.

#endif  // BOSS_H
