
/*
 * text-maker: a program for translating text descriptions of slides
 *             and making slide-maker slides
 *
 * Original Author: Richard Rudell
 * Last Modified By: Rick Spickelmier, 11/5/87
 *
 */

#include "copyright.h"
#include "port.h"
#include "utility.h"
#include <sys/file.h>
#include "smfont.h"
#include "rpc.h"
#include "oh.h"
#include "st.h"
#include "sm.h"
#include "vov.h"

char *readline();
extern char *loadFont();
  
font *fontptr[MAXGLYPH];
char *FontFile;
char *FontDirectory;

int GenLetters = 1;

st_table *fonts = NIL(st_table);

#define MAXLINE		1024

char color[256];
char fontName[256];
char just[256];
double scale;
int indent;
  

main(argc, argv)
int argc;
char **argv;
{
    char *slide, *background;
    char buffer[4096];
    
    if ((argc != 2) && (argc != 4)) {
	(void) fprintf(stderr, "usage: text-maker-driver [-f background] slidename < input file\n");
	exit(2);
    }

    if (argc == 4) {
	if (strcmp(argv[1], "-f") != 0) {
	    (void) fprintf(stderr, "usage: text-maker-driver [-f background] slidename < input file\n");
	    exit(2);
	}
	slide = argv[3];
	background = argv[2];
	fprintf(stderr, "\tusing background facet: %s:slide\n", background);
    } else {
	slide = argv[1];
	background = NIL(char);
    }

    VOVbegin( argc, argv );
    (void) sprintf(buffer, "%s/%s", "$OCTTOOLS", FONTFILE);
    FontFile = util_file_expand(buffer);

    if (access(FontFile, F_OK) != 0) {
        char errmsg[4096];
        (void) sprintf(errmsg, "The letter files (%s, %s) are not installed\n",
                       FontFile, buffer);
	VEMMSG(errmsg);
	VOVend(-1);
    }

    getfonts(FontFile);

    /* set up the hash table */
    if ((fonts = st_init_table(strcmp, st_strhash)) == NULL) {
	VEMMSG("initialization error\n");
	VOVend(-1);
    }

    (void) sprintf(buffer, "%s/%s", "$OCTTOOLS", FONTDIRECTORY);
    FontDirectory = util_file_expand(buffer);

    if (access(FontDirectory, F_OK) != 0) {
        char errmsg[4096];
        (void) sprintf(errmsg, "The font definition files (%s, %s) are not installed\n",
                       FontDirectory, buffer);
	VEMMSG(errmsg);
	VOVend(-1);
    }

    st_insert(fonts, "math",  (char *) loadFont(FontDirectory, "math.hmp"));
    st_insert(fonts, "greekc",  (char *) loadFont(FontDirectory, "greekc.hmp"));
    st_insert(fonts, "italict", (char *) loadFont(FontDirectory, "italict.hmp"));
    st_insert(fonts, "romant",  (char *) loadFont(FontDirectory, "romant.hmp"));
    
    do_slide(stdin, slide, background);
}


do_slide(fp, root, background)
FILE *fp;
char *root, *background;
{
    octObject facet, bgfacet, prop;
    char s[MAXLINE], name[MAXLINE], keyword[MAXLINE], token[MAXLINE];
    char *p, c;
    int xloc, yloc, centerx, rightx, leftx;
    RPCSpot spot;
    struct octPoint point, *savepoint;
    char *ptr;
    st_table *savetable;
    

    savetable = st_init_table(strcmp, st_strhash);
    (void) strcpy(color, "cyan");
    (void) strcpy(fontName, "romant");
    (void) strcpy(just, "left");
    scale = 1.0;
    indent = UNITSPERINCH * 0.5 + OFFSET;
    yloc = (PAGEWIDTH - 1.0) * UNITSPERINCH + OFFSET;
    centerx = (PAGEHEIGHT / 2.0) * UNITSPERINCH + OFFSET;
    leftx = UNITSPERINCH * 0.5 + OFFSET;
    rightx = (PAGEWIDTH - 0.5) * UNITSPERINCH + OFFSET;

    /* Open the fact for write, then close it to force it to be made */
    facet.type = OCT_FACET;
    facet.contents.facet.cell = root;
    facet.contents.facet.view = "slide";
    facet.contents.facet.facet = "contents";
    facet.contents.facet.version = OCT_CURRENT_VERSION;
    facet.contents.facet.mode = "w";

    if (background != NIL(char)) {
	bgfacet.type = OCT_FACET;
	bgfacet.contents.facet.cell = background;
	bgfacet.contents.facet.view = "slide";
	bgfacet.contents.facet.facet = "contents";
	bgfacet.contents.facet.version = OCT_CURRENT_VERSION;
	bgfacet.contents.facet.mode = "r";

	VOVinputFacet( &bgfacet );
	VOVoutputFacet( &facet );
	if (octCopyFacet(&facet, &bgfacet) < OCT_OK) {
	    octError("copy facet failure");
	    VOVend(2);
	}
	facet.contents.facet.mode = "a";
    }

    fprintf(stderr, "\tcreating slide: %s:%s\n", facet.contents.facet.cell,
	    facet.contents.facet.view);

    VOVoutputFacet( &facet );
    OCTASSERT(octOpenFacet(&facet), "can not open the slide");


    OCTASSERT(ohCreateOrModifyPropStr(&facet, &prop, "TECHNOLOGY", "slide"), "tech");
    OCTASSERT(ohCreateOrModifyPropStr(&facet, &prop, "VIEWTYPE", "SLIDE"), "tech");
    OCTASSERT(ohCreateOrModifyPropStr(&facet, &prop, "EDITSTYLE", "PHYSICAL"), "tech");
    
    while (! feof(fp)) {

	xloc = indent;

read_line:
	/* read a line from the file */
	if (readline(s, MAXLINE, fp) == NULL) break;

	if (s[0] == '.') {
	    (void) sscanf(s, ".%s %s", keyword, token);

	    /* backwards compatibility */
	    if (strcmp(keyword, "center") == 0) {
		(void) strcpy(keyword, "margin");
		(void) strcpy(token, "center");
		
		(void) strcpy(just, "center");
		
	    } else if (strcmp(keyword, "right") == 0) {
		(void) strcpy(keyword, "just");
		(void) strcpy(token, "right");
	    } else if (strcmp(keyword, "left") == 0) {
		(void) strcpy(keyword, "just");
		(void) strcpy(token, "left");
	    }
	       
	    if (strcmp(keyword, "scale") == 0) {
		if (token[0] == '*') {
		    scale *= atof(token+1);
		} else {
		    scale = atof(token);
		}
		goto read_line;

	    } else if (strcmp(keyword, "color") == 0) {
		(void) strcpy(color, token);
		goto read_line;

	    } else if (strcmp(keyword, "font") == 0) {
		(void) strcpy(fontName, token);
		goto read_line;

	    } else if (strcmp(keyword, "indent") == 0) {
		if (token[0] == '+') {
		    indent += atof(token+1) * UNITSPERINCH;
		} else if (token[0] == '-') {
		    indent -= atof(token+1) * UNITSPERINCH;
		} else {
		    indent = atof(token) * UNITSPERINCH + OFFSET;
		}
		xloc = indent;
		goto read_line;

	    } else if (strcmp(keyword, "space") == 0) {
		yloc -= atof(token) * UNITSPERINCH;
		goto read_line;

	    } else if (strcmp(keyword, "save") == 0) {
		savepoint = ALLOC(struct octPoint, 1);
		savepoint->x = xloc;
		savepoint->y = yloc;
		st_insert(savetable, token, (char *) savepoint);
		goto read_line;

	    } else if (strcmp(keyword, "restore") == 0) {
		if (st_lookup(savetable, token, &ptr) != 1) {
		    fprintf(stderr, "bad argument to .restore, %s\n", token);
		    VOVend(-1);
		}
		savepoint = (struct octPoint *) ptr;
		xloc = savepoint->x;
		yloc = savepoint->y;
		goto read_line;

	    } else if (strcmp(keyword, "just") == 0) {
		(void) strcpy(just, token);
		goto read_line;
		
	    } else if (strcmp(keyword, "margin") == 0) {
		if (strcmp(token, "left") == 0) {
		    xloc = indent = leftx;
		} else if (strcmp(token, "center") == 0) {
		    xloc = indent = centerx;
		} else if (strcmp(token, "right") == 0) {
		    xloc = indent = rightx;
		} else {
		    fprintf(stderr, "bad argument to .margin, %s\n", token);
		    VOVend(-1);
		}
		goto read_line;

	    } else if (strcmp(keyword, "vertical-box") == 0) {
		spot.theWin = 0;
		spot.facet = facet.objectId;
		spot.thePoint.x = 0;
		spot.thePoint.y = 0;
		(void) vertBox(&spot, lsCreate(), 0);
		centerx = (PAGEWIDTH / 2.0) * UNITSPERINCH + OFFSET;
		leftx = 0.5 * UNITSPERINCH + OFFSET;
		rightx = (PAGEWIDTH - 0.5) * UNITSPERINCH + OFFSET;
		yloc = (PAGEHEIGHT - 1.0) * UNITSPERINCH + OFFSET;
		goto read_line;

	    } else if (strcmp(keyword, "horizontal-box") == 0) {
		spot.theWin = 0;
		spot.facet = facet.objectId;
		spot.thePoint.x = 0;
		spot.thePoint.y = 0;
		(void) horizBox(&spot, lsCreate(), 0);
		centerx = (PAGEHEIGHT / 2.0) * UNITSPERINCH + OFFSET;
		leftx = 0.5 * UNITSPERINCH + OFFSET;
		rightx = (PAGEHEIGHT - 0.5) * UNITSPERINCH + OFFSET;
		yloc = (PAGEWIDTH - 1.0) * UNITSPERINCH + OFFSET;
		goto read_line;
	    }
	}

	/* Check for blank line */
	for(p = s; (c = *p++) && isspace(c); )
	    ;
	if (c != '\0') {
	    point.x = xloc;
	    point.y = yloc;
	    (void) createString(&facet, point, s, fontName, color, scale, just);
	}

	yloc -= scale * 200 * 1.9;
    }

    OCTASSERT(octCloseFacet(&facet), "can not close the slide");

    VOVend(0);
}


char *readline(s, n, fp)
char *s;
int n;
FILE *fp;
{
    int c, i = 0;

    while ((c = getc(fp)) != EOF) {
	if (c == '\\') {
	    /* peek at next char */
	    if (ungetc(getc(fp), fp) == 'n') {
		getc(fp);	/* discard n */
		break;
	    }
	}
	if (c == '\n') {
	    break;
	}
	if (i < n) {
	    s[i++] = c;
	}
    }
    if (c == EOF && i == 0) return NULL;
    s[i++] = '\0';
    return s;
}
