#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "yybdsyn.h"
#include "eval.h"
#include "mult.h"

/*
 * Will return true if expr is a constant, and is not concat'ed to
 * anything.  And is not a wildcard.
 */
#define SIMPLECONST(expr) \
    (((expr) != NULL) && ((expr)->type == T_IS_CONSTANT) && \
    ((expr)->conc == NULL) && (!(expr)->un.constant.dont_care))

node *constant_prop();
node *binary_bitwise();
node *not_bitwise();
node *not_bit();
node *bitblast();
node *one_bit_result();

extern int eval_subst;		/* if true, do simple forward
				 * substitution for variables */
extern int saw_complex;		/* Have there been macro substs */

/*
 * This routine attempts to reduce cnode to its simplest form.
 * It will return a constant if possible.  If it is impossible, it will
 * return the simplest vector of bit operations.
 */
node *node_eval( root )
node *root;
{
    int allconst;		/* is true if all args are constant */
    node *arg1,*arg2,*arg3;	/* operand of bit operator */
    node *concsave;		/* stuff that follows on conc ptr */
    node *lastsave;		/* stuff whose next points here */
    node *nextsave;		/* stuff that follows on next ptr */
    node *cnode;		/* current node to process */
    node *result;
    node *resultend;		/* end of result string */
    int count;			/* number of bits */
    int i;			/* index variable */
    node *nthbit;		/* nth bit of expression */
    node *lastbit;		/* last bit of expression */
    node *temp;			/* intermediate node */
    node *bit1, *bit3;		/* index into arg1 and arg3 */
    node *lastx;		/* assign of variable we looked up */
    int isDC;			/* true if RHS of T_EQ is DONT_CARE */
    node *rhs;

    if ( root == NULL )
	return NULL;
    lastsave = NULL;
    cnode = root;

    while ( cnode != NULL ) {

	nextsave = cnode->next;
	cnode->next = NULL;
	concsave = node_eval( cnode->conc );
	cnode->conc = NULL;

		    /* STAGE 1: eval decendents */
	switch ( cnode->type ) {

	    case T_IS_CONSTANT:
	    case T_IS_STATE:
		break;

	    case T_NOT: case T_BUF: case T_WIDTH: case T_UNARY_OR:
	    case T_UNARY_AND: case T_UNARY_PLUS: case T_UNARY_MINUS:
		arg1 = node_eval( cnode->un.unary.expr );
		arg2 = arg3 = NULL;
		cnode->un.unary.expr = NULL;
		allconst = SIMPLECONST( arg1 );
		break;

	    case T_AND: case T_OR: case T_NAND: case T_NOR: case T_XOR:
	    case T_EQV: case T_EQL: case T_NEQ: case T_LEQ: case T_GEQ:
	    case T_LSS: case T_GTR: case T_SL0: case T_SR0: case T_SL1:
	    case T_SR1: case T_SLR: case T_SRR: case T_SXT: case T_ZXT:
	    case T_OXT: case T_PLUS: case T_MINUS: case T_TIMES:
	    case T_DIV: case T_MOD:
		arg1 = node_eval( cnode->un.binary.lexpr );
		arg2 = node_eval( cnode->un.binary.rexpr );
		arg3 = NULL;
		cnode->un.binary.lexpr = NULL;
		cnode->un.binary.rexpr = NULL;
		allconst = SIMPLECONST( arg1 ) &&
			    SIMPLECONST( arg2 );
		break;

	    case T_BIT:
		arg1 = node_eval( cnode->un.ternary.lexpr );
		arg2 = node_eval( cnode->un.ternary.mexpr );
		arg3 = node_eval( cnode->un.ternary.rexpr );
		cnode->un.ternary.lexpr = NULL;
		cnode->un.ternary.mexpr = NULL;
		cnode->un.ternary.rexpr = NULL;
		if ( arg3 == NULL ) {
		    allconst = SIMPLECONST( arg1 ) &&
			    SIMPLECONST( arg2 );
		} else {
		    allconst = SIMPLECONST( arg1 ) &&
			    SIMPLECONST( arg2 ) &&
			    SIMPLECONST( arg3 );
		}
		break;

	    case T_EQ:
		arg1 = node_eval( cnode->un.assign.lexpr );
		arg3 = node_eval( cnode->un.assign.rexpr );
		cnode->un.assign.lexpr = NULL;
		cnode->un.assign.rexpr = NULL;
		break;

	    case T_IF:
		cnode->un.bdif.expr = bitblast(node_eval(cnode->un.bdif.expr) );
		NODE_DISPOSE( cnode->un.bdif.expr->conc );	/* only first */
		cnode->un.bdif.expr->conc = NULL;		/* bit counts */
		cnode->un.bdif.thenb = node_eval(cnode->un.bdif.thenb);
		cnode->un.bdif.elseb = node_eval(cnode->un.bdif.elseb);
		break;

	    case T_SELECT:
		cnode->un.select.expr = node_eval(cnode->un.select.expr);
		cnode->un.select.bdcase = node_eval(cnode->un.select.bdcase);
		break;
		
	    case T_CASE:
		cnode->un.bdcase.expr = node_eval(cnode->un.bdcase.expr);
		cnode->un.bdcase.stmt = node_eval(cnode->un.bdcase.stmt);
		break;

	    case T_FOR:
	    case T_FOR_DOWN:
		cnode->un.bdfor.stmt = node_eval( cnode->un.bdfor.stmt );
		break;

	    case T_LABEL:
		cnode->un.label.stmt = node_eval( cnode->un.label.stmt );
		break;

	    case T_LEAVE:
	    case T_OTHERWISE:
		break;

	    default:
		node_print(cnode,20,"err ");
		ierror("unknown type in eval", cnode);
	}

			/* STAGE 2: evaluate self */
	/* Do constant propagation if possible */
	if ( canconstprop( cnode->type ) && allconst ) {
	    result = constant_prop( cnode,arg1,arg2,arg3 );
	} else {

	    /* process the cases that we can handle */
	    switch ( cnode->type ) {

		case T_IS_STATE:
		    if (eval_subst && (cnode->un.bdvar.rev != NULL)) {
			lastx = cnode->un.bdvar.rev->assign;
			rhs = lastx->un.assign.rexpr;
			if (((rhs->type == T_IS_STATE) &&
				(upper_strcmp(rhs->un.bdvar.name->name,
				"DONT_CARE") != 0)) ||
				(rhs->type == T_IS_CONSTANT)) {
			    /* change cnode to effect forward subst */
			    cnode->type = rhs->type;
			    if ( cnode->type == T_IS_STATE ) {
				cnode->un.bdvar.name =
					rhs->un.bdvar.name;
				cnode->un.bdvar.lbit =
					rhs->un.bdvar.lbit;
				cnode->un.bdvar.hbit =
					rhs->un.bdvar.hbit;
				cnode->un.bdvar.rev =
					rhs->un.bdvar.rev;
			    } else {
				cnode->un.constant.num =
					rhs->un.constant.num;
				cnode->un.constant.dont_care =
					rhs->un.constant.dont_care;
				cnode->un.constant.hbit =
					rhs->un.constant.hbit;
			    }
			}
		    }
		    result = cnode;
		    break;

		case T_IS_CONSTANT:
		case T_IF:
		case T_CASE:
		case T_OTHERWISE:
		case T_LEAVE:
		case T_LABEL:
		case T_SELECT:
		    result = cnode;
		    break;

		case T_FOR:
		case T_FOR_DOWN:
		    if( cnode->un.bdfor.stmt == NULL ) {
			result = node_create_bit( 0, cnode );
			if ( lastsave == NULL ) {
			    root = result;
			} else {
			    lastsave->next = result;
			}
		    } else {
			result = cnode->un.bdfor.stmt;
			while( result->next != NULL ) {
			    if ( lastsave == NULL ) {
				root = result;
			    } else {
				lastsave->next = result;
			    }
			    lastsave = result;
			    result = result->next;
			}
		    }
		    cnode->un.bdfor.stmt = NULL;
		    node_dispose( cnode );
		    break;

		case T_EQL:
/* NOT (NEQ () )  -- Doesn't seem to be as efficient.
		    cnode->type = T_XOR;
		    arg1 = binary_bitwise( cnode, arg1, arg2 );
		    cnode = node_create_fi( T_UNARY_OR, arg1 );
		    arg1 =  one_bit_result( cnode, arg1 );
		    result =  not_bit( arg1 );
		    break;
*/
		    cnode->type = T_EQV;
		    arg1 = binary_bitwise( cnode, arg1, arg2 );
		    cnode = node_create_fi( T_UNARY_AND, arg1 );
		    result =  one_bit_result( cnode, arg1 );
		    break;

		case T_NEQ:
		    cnode->type = T_XOR;
		    arg1 = binary_bitwise( cnode, arg1, arg2 );
		    cnode = node_create_fi( T_UNARY_OR, arg1 );
		    result =  one_bit_result( cnode, arg1 );
		    break;

		case T_EQ:
		    /* The left and right right arguements of T_EQ will
		     * be size adjusted, but the T_EQ node will not be
		     * bit blasted until multiple assignment.  This will
		     * allow correct processing of "x = x SL0 1" */
		    bit1 = arg1 = bitblast( arg1 );
		    bit3 = arg3 = bitblast( arg3 );
		    isDC = (arg3->type == T_IS_STATE) &&
			(upper_strcmp( arg3->un.bdvar.name->name,
			"DONT_CARE" ) == 0);
		    /* make sure DC is only one bit */
		    if ( isDC ) {
			NODE_DISPOSE( arg3->conc );
			arg3->conc = NULL;
		    }
		    while ( bit1->conc != NULL ) {
			if ( bit3->conc == NULL ) {
			    /* extend DONT_CARE size to match LHS, or ZXT */
			    if ( isDC ) {
				bit3->conc = node_dup( bit3 );
			    } else {
				bit3->conc = node_create_bit(0, cnode);
			    }
			}
			bit3 = bit3->conc;
			bit1 = bit1->conc;
		    }
		    NODE_DISPOSE( bit3->conc );		/* toss excess RHS */
		    bit3->conc = NULL;
		    cnode->un.assign.lexpr = arg1;
		    cnode->un.assign.rexpr = arg3;
		    result = cnode;
		    break;

		 case T_BIT:
		    if ( !SIMPLECONST( arg2 ) || ((arg3 != NULL) &&
			    !SIMPLECONST( arg3 )) ) {
			error("Bit operation is not supported", cnode );
		    } else {
			if ( arg3 == NULL ) {
			    arg3 = node_dup( arg2 );
			}
			if (arg2->un.constant.num < arg3->un.constant.num)
			    error("Backwards bit range",cnode);
			arg1 = bitblast( arg1 );
			count = countbits(arg1, &allconst,
				arg2->un.constant.num, &nthbit,&lastbit);
			if (count < (arg2->un.constant.num + 1) ||
				arg2->un.constant.num < 0) {
			    error("Bit select out of range",cnode);
			}
			/* throw out trailing bits */
			NODE_DISPOSE(nthbit->conc);
			nthbit->conc = NULL;
			/* throw out leading bits */
			if (arg3->un.constant.num != 0) {
			    (void) countbits(arg1,&allconst,
				    arg3->un.constant.num - 1,
				    &nthbit, &lastbit);
			    temp = nthbit->conc;
			    nthbit->conc = NULL;
			    NODE_DISPOSE( arg1 );
			    arg1 = temp;
			}
			NODE_DISPOSE( arg2 );
			NODE_DISPOSE( arg3 );
			NODE_DISPOSE( cnode );
			result = arg1;
		    }
		    break;

		case T_SL0: case T_SL1: case T_SLR:
		case T_SR0: case T_SR1: case T_SRR:
		    if ( !SIMPLECONST(arg2) ) {
			ierror("Bad shift in eval", arg2);
		    } else {
			if ( arg2->un.constant.num < 0 ) {
			    arg2->un.constant.num *= -1;
			    switch ( cnode->type ) {
				case T_SR0:
				    cnode->type = T_SL0;
				    break;
				case T_SR1:
				    cnode->type = T_SL1;
				    break;
				case T_SRR:
				    cnode->type = T_SLR;
				    break;
				case T_SL0:
				    cnode->type = T_SR0;
				    break;
				case T_SL1:
				    cnode->type = T_SR1;
				    break;
				case T_SLR:
				    cnode->type = T_SRR;
				    break;
			    }
			}
			arg1 = bitblast( arg1 );
			count = countbits( arg1, &allconst, 0, &nthbit,
				&lastbit );
			if ( cnode->type == T_SRR || cnode->type == T_SLR ) {
			    arg2->un.constant.num %= count;
			}
			if ( arg2->un.constant.num == 0 ) {
			    NODE_DISPOSE( arg2 );
			    NODE_DISPOSE( cnode );
			    result = arg1;
			    break;
			}
			if ( cnode->type == T_SRR ) {
			    (void) countbits( arg1, &allconst,
				    arg2->un.constant.num - 1,
				    &nthbit, &lastbit );
			    temp = nthbit->conc;
			    nthbit->conc = NULL;
			    lastbit->conc = arg1;
			    result = temp;
			} else if ( cnode->type == T_SLR ) {
			    (void) countbits( arg1, &allconst,
				    count - arg2->un.constant.num - 1,
				    &nthbit, &lastbit );
			    temp = nthbit->conc;
			    nthbit->conc = NULL;
			    lastbit->conc = arg1;
			    result = temp;
			} else {
			    /* make fill of one's or zeros */
			    arg3 = node_create_fi( T_IS_CONSTANT, arg1 );
			    arg3->un.constant.hbit = 0;
			    if ( cnode->type == T_SR0 || cnode->type == T_SL0 ) {
				arg3->un.constant.num = 0;
			    } else {
				arg3->un.constant.num = 1;
			    }
			    if ( arg2->un.constant.num > count ) {
				arg2->un.constant.num = count;
			    }
			    nthbit = arg3;
			    for ( i = 1; i < arg2->un.constant.num; i++ ) {
				nthbit->conc = node_dup( nthbit );
				nthbit = nthbit->conc;
			    }
			    temp = nthbit;		/* save for SL */
			    if ( arg2->un.constant.num == count ) {
				result = arg3;
				NODE_DISPOSE( arg1 );
			    } else if ( cnode->type == T_SR0 ||
					cnode->type == T_SR1 ) {
				(void) countbits( arg1, &allconst,
					arg2->un.constant.num - 1,
					&nthbit, &lastbit );
				result = nthbit->conc;
				nthbit->conc = NULL;
				lastbit->conc = arg3;
				NODE_DISPOSE( arg1 );
			    } else {
				(void) countbits( arg1, &allconst,
					count - arg2->un.constant.num - 1,
					&nthbit, &lastbit );
				result = arg3;
				temp->conc = arg1;
				NODE_DISPOSE( nthbit->conc );
				nthbit->conc = NULL;
			    }
			}
			NODE_DISPOSE(arg2);
			NODE_DISPOSE(cnode);
		    }
		    break;

		case T_SXT: case T_OXT: case T_ZXT:
		    if ( !SIMPLECONST( arg2 ) )
			error("Can't extend by non-constant",cnode);
		    arg1 = bitblast( arg1 );
		    count = countbits(arg1, &allconst, 0, &nthbit, &lastbit);
		    if ( count > arg2->un.constant.num )
			error("Extend will shrink expression",cnode);
		    for ( i=1; i <= arg2->un.constant.num - count; i++ ) {
			if ( cnode->type == T_ZXT ) {
			    lastbit->conc = node_create_bit(0, arg1);
			} else if ( cnode->type == T_OXT ) {
			    lastbit->conc = node_create_bit(1, arg1);
			} else {
			    lastbit->conc = node_dup( lastbit );
			}
			lastbit = lastbit->conc;
		    }
		    result = arg1;
		    NODE_DISPOSE(arg2);
		    NODE_DISPOSE(cnode);
		    break;

		case T_AND: case T_OR: case T_NAND:
		case T_NOR: case T_XOR: case T_EQV:
		    result = binary_bitwise( cnode, arg1, arg2 );
		    break;

		case T_UNARY_OR: case T_UNARY_AND:
		    result = one_bit_result( cnode, arg1 );
		    break;

		case T_BUF:
		    result = arg1;
		    NODE_DISPOSE( cnode );
		    break;

		case T_NOT:
		    result = not_bitwise(arg1);
		    NODE_DISPOSE( cnode );
		    break;

		case T_PLUS: case T_MINUS:
		    ierror("Bad plus or minus in eval", cnode);
		    break;

		case T_LSS: case T_LEQ: case T_GTR: case T_GEQ:
		    ierror("Bad compare in eval", cnode);
		    break;

		default:
		    error("Complex operation on non meta-var",cnode);
	    }
	}
	resultend = result;
	while( resultend->conc != NULL )
	    resultend = resultend->conc;
	resultend->conc = concsave;

	/* Try to merge simple constants */
	/* Questionable code -- probably gives alot of overhead and little
	 * benefit */
	if ( !SIMPLECONST( result ) &&
		( count = countbits(result,&allconst,0,&nthbit,&lastbit) ) <= 32
		&& allconst ) {
	    temp = bitblast(result);
	    result = node_create_fi( T_IS_CONSTANT,temp);
	    result->un.constant.hbit = count - 1;
	    result->un.constant.num = mergeit(temp);
	    NODE_DISPOSE( temp );
	}

	if ( lastsave == NULL ) {
	    root = result;
	} else {
	    lastsave->next = result;
	}
	lastsave = result;
	result->next = nextsave;
	cnode = result->next;
    }
    return root;
}


/*
 * Evals T_NOT bitwise on vectors, and becomes many T_NOT's
 */
node *not_bitwise(arg1 )
node *arg1;
{
    node *result;		/* result string of bits */
    node *lastbit;		/* last bit to be added to result string */
    node *expr;			/* bitblast of arg1 */
    node *resultbit;		/* result of a one bit operator */

    /* blast argument to bits */
    expr = bitblast( arg1 );
    result = lastbit = NULL;

    while( expr != NULL ) {
	arg1 = expr;
	expr = expr->conc;
	arg1->conc = NULL;
	resultbit = not_bit(arg1);

	/* add a bit to result string */
	if (lastbit == NULL) {
	    lastbit = result = resultbit;
	} else {
	    lastbit->conc = resultbit;
	    lastbit = lastbit->conc;
	}
    }
    return result;
}


/*
 * This NOTs one bit.  Constants are flipped, and NOT( NOT(x) ) = x.
 */
node *not_bit(arg1)
node *arg1;
{
    node *resultbit;

    if ( arg1->type == T_IS_CONSTANT ) {
	arg1->un.constant.num ^= 1;
	resultbit = arg1;
    } else if ( arg1->type == T_NOT ) {
	resultbit = arg1->un.unary.expr;
	arg1->un.unary.expr = NULL;
	NODE_DISPOSE( arg1 );
    } else {
	resultbit = node_create_fi(T_NOT,arg1);
	resultbit->un.unary.expr = arg1;
    }
    return resultbit;
}

/*
 * This routine handles binary bitwise operators.
 */
node *binary_bitwise( bnode, arg1, arg2 )
node *bnode;
node *arg1;
node *arg2;
{
    node *result;		/* result string of bits */
    node *lastbit;		/* last bit to be added to result string */
    node *bit1, *bit2;		/* two operands of bit operator */
    node *left, *right;		/* bitblast args */
    node *resultbit;		/* result of a one bit operator */
    node *temp;

    lastbit = result = NULL;
    /* blast arguments to bits */
    left = bitblast( arg1 );
    right = bitblast( arg2 );
    while( left != NULL || right != NULL ) {

	bit1 = left;
	if( bit1 != NULL ) {
	    left = left->conc;
	    bit1->conc = NULL;
	}

	bit2 = right;
	if( bit2 != NULL ) {
	    right = right->conc;
	    bit2->conc = NULL;
	}

	if ( bit1 == NULL )
	    bit1 = node_create_bit ( 0, bnode );
	if ( bit2 == NULL )
	    bit2 = node_create_bit ( 0, bnode );
			
	/*
	 * Reorder the operands for later convenience.
	 * If one of the operands is a simple constant, then put
	 * it in bit1.  If there is no simple constant, but there
	 * is a wildcard, put it in bit1.
	 */
	if ( SIMPLECONST(bit2) || (!SIMPLECONST(bit1) && wildcard(bit2)) ) {
	    temp = bit1;
	    bit1 = bit2;
	    bit2 = temp;
	}

	switch ( bnode->type ) {

	    case T_AND:
		if ( SIMPLECONST(bit1) ) {
		    if (bit1->un.constant.num == 0 ) {
			NODE_DISPOSE(bit2);
			resultbit = bit1;
		    } else {
			NODE_DISPOSE(bit1);
			resultbit = bit2;
		    }
		} else if ( wildcard(bit1) ) {
		    NODE_DISPOSE(bit2);
		    resultbit = bit1;
		} else {
		    resultbit = node_create_fi( T_AND, bnode );
		    resultbit->un.binary.lexpr = bit2;
		    resultbit->un.binary.rexpr = bit1;
		}
		break;

	    case T_OR:
		if ( SIMPLECONST(bit1) ) {
		    if (bit1->un.constant.num == 1 ) {
			NODE_DISPOSE(bit2);
			resultbit = bit1;
		    } else {
			NODE_DISPOSE(bit1);
			resultbit = bit2;
		    }
		} else if ( wildcard(bit1) ) {
		    NODE_DISPOSE(bit2);
		    resultbit = bit1;
		} else {
		    resultbit = node_create_fi( T_OR, bnode );
		    resultbit->un.binary.lexpr = bit2;
		    resultbit->un.binary.rexpr = bit1;
		}
		break;

	    case T_NAND:
		if ( SIMPLECONST(bit1) ) {
		    if (bit1->un.constant.num == 0 ) {
			NODE_DISPOSE(bit2);
			bit1->un.constant.num = 1;
			resultbit = bit1;
		    } else {
			NODE_DISPOSE(bit1);
			resultbit = not_bit(bit2);
		    }
		} else if ( wildcard(bit1) ) {
		    NODE_DISPOSE(bit2);
		    resultbit = bit1;
		} else {
		    resultbit = node_create_fi( T_NAND, bnode );
		    resultbit->un.binary.lexpr = bit2;
		    resultbit->un.binary.rexpr = bit1;
		}
		break;

	    case T_NOR:
		if ( SIMPLECONST(bit1) ) {
		    if (bit1->un.constant.num == 1 ) {
			NODE_DISPOSE(bit2);
			bit1->un.constant.num = 0;
			resultbit = bit1;
		    } else {
			NODE_DISPOSE(bit1);
			resultbit = not_bit(bit2);
		    }
		} else if ( wildcard(bit1) ) {
		    NODE_DISPOSE(bit2);
		    resultbit = bit1;
		} else {
		    resultbit = node_create_fi( T_NOR, bnode );
		    resultbit->un.binary.lexpr = bit2;
		    resultbit->un.binary.rexpr = bit1;
		}
		break;

	    case T_XOR:
		/* If there is a wildcard, answer is zero */
		if ( wildcard(bit1) || wildcard(bit2) ) {
		    NODE_DISPOSE(bit1);
		    NODE_DISPOSE(bit2);
		    resultbit = node_create_bit(0,bnode);
		} else if ( SIMPLECONST(bit1) ) {
		    if (bit1->un.constant.num == 0 ) {
			resultbit = bit2;
		    } else {
			resultbit = not_bit(bit2);
		    }
		    NODE_DISPOSE(bit1);
		} else {
		    resultbit = node_create_fi( T_XOR, bnode );
		    resultbit->un.binary.lexpr = bit2;
		    resultbit->un.binary.rexpr = bit1;
		}
		break;

	    case T_EQV:
		/* If there is a wildcard, resultbit is 1 */
		if ( wildcard(bit1) || wildcard(bit2) ) {
		    NODE_DISPOSE(bit1);
		    NODE_DISPOSE(bit2);
		    resultbit = node_create_bit(1,bnode);
		} else if ( SIMPLECONST(bit1) ) {
		    if (bit1->un.constant.num == 1 ) {
			resultbit = bit2;
		    } else {
			resultbit = not_bit(bit2);
		    }
		    NODE_DISPOSE(bit1);
		} else {
		    resultbit = node_create_fi( T_EQV, bnode );
		    resultbit->un.binary.lexpr = bit2;
		    resultbit->un.binary.rexpr = bit1;
		}
		break;
	}

	/* add a bit to result string */
	if (lastbit == NULL) {
	    lastbit = result = resultbit;
	} else {
	    lastbit->conc = resultbit;
	    lastbit = lastbit->conc;
	}
    }

    NODE_DISPOSE(bnode);
    return( result );
}

/*
 * This routine attempts to reduce cnode to a simple constant
 * by calculating all operators on constants.  (e.g.  34 * 12 AND 51
 * will return a single constant node.
 * The old cnode stuff is NODE_DISPOSE if necessary.
 * This routine assumes that this node has been disconnected from
 * anything that is on its "next" pointer.
 */
node *constant_prop( cnode, arg1, arg2, arg3 )
node *cnode;
node *arg1,*arg2,*arg3;
{
    node *temp;

    /* replace arg1 with result */
    switch ( cnode->type ) {

	case T_NOT:
	    arg1->un.constant.num = ~ arg1->un.constant.num;
	    break;

	case T_BUF:
	    /* arg1 = result is correct */
	    break;

	case T_UNARY_OR:
	    arg1->un.constant.num &= maskbits(arg1->un.constant.hbit);
	    arg1->un.constant.num = arg1->un.constant.num != 0;
	    arg1->un.constant.hbit = 0;
	    break;

	case T_UNARY_AND:
	    /* remember DeMorgan? */
	    arg1->un.constant.num = (-1 ==
		(arg1->un.constant.num | ~maskbits(arg1->un.constant.hbit)) );
	    arg1->un.constant.hbit = 0;
	    break;

	case T_WIDTH:
	    temp = node_mkconst( arg1->un.constant.hbit + 1, arg1 );
	    NODE_DISPOSE(arg1);
	    arg1 = temp;
	    break;

	case T_UNARY_PLUS:
	    break;

	case T_UNARY_MINUS:
	    arg1->un.constant.num = - arg1->un.constant.num;
	    break;

	case T_AND:
	    arg1->un.constant.num &= arg2->un.constant.num;
	    if ( arg2->un.constant.hbit > arg1->un.constant.hbit )
		arg1->un.constant.hbit = arg2->un.constant.hbit;
	    break;

	case T_OR:
	    arg1->un.constant.num |= arg2->un.constant.num;
	    if ( arg2->un.constant.hbit > arg1->un.constant.hbit )
		arg1->un.constant.hbit = arg2->un.constant.hbit;
	    break;

	case T_XOR:
	    arg1->un.constant.num ^= arg2->un.constant.num;
	    if ( arg2->un.constant.hbit > arg1->un.constant.hbit )
		arg1->un.constant.hbit = arg2->un.constant.hbit;
	    break;

	case T_PLUS:
	    arg1->un.constant.num += arg2->un.constant.num;
	    if ( arg2->un.constant.hbit > arg1->un.constant.hbit )
		arg1->un.constant.hbit = arg2->un.constant.hbit;
	    arg1->un.constant.hbit++;
	    break;

	case T_MINUS:
	    arg1->un.constant.num -= arg2->un.constant.num;
	    if ( arg2->un.constant.hbit > arg1->un.constant.hbit )
		arg1->un.constant.hbit = arg2->un.constant.hbit;
	    arg1->un.constant.hbit++;
	    break;

	case T_TIMES:
	    arg1->un.constant.num *= arg2->un.constant.num;
	    arg1->un.constant.hbit += arg2->un.constant.hbit + 1;
	    break;

	case T_DIV:
	    arg1->un.constant.num /= arg2->un.constant.num;
	    break;

	case T_MOD:
	    arg1->un.constant.num %= arg2->un.constant.num;
	    break;

	case T_NAND:
	    arg1->un.constant.num = ~(arg1->un.constant.num & arg2->un.constant.num);
	    if ( arg2->un.constant.hbit > arg1->un.constant.hbit )
		arg1->un.constant.hbit = arg2->un.constant.hbit;
	    break;

	case T_NOR:
	    arg1->un.constant.num = ~(arg1->un.constant.num | arg2->un.constant.num);
	    if ( arg2->un.constant.hbit > arg1->un.constant.hbit )
		arg1->un.constant.hbit = arg2->un.constant.hbit;
	    break;

	case T_EQV:
	    arg1->un.constant.num = ~(arg1->un.constant.num ^ arg2->un.constant.num);
	    if ( arg2->un.constant.hbit > arg1->un.constant.hbit )
		arg1->un.constant.hbit = arg2->un.constant.hbit;
	    break;

	case T_EQL:
	    arg1->un.constant.num =
		    (arg1->un.constant.num == arg2->un.constant.num );
	    arg1->un.constant.hbit = 0;
	    break;

	case T_NEQ:
	    arg1->un.constant.num =
		    (arg1->un.constant.num != arg2->un.constant.num );
	    arg1->un.constant.hbit = 0;
	    break;

	case T_GEQ:
	    arg1->un.constant.num =
		    (arg1->un.constant.num >= arg2->un.constant.num );
	    arg1->un.constant.hbit = 0;
	    break;

	case T_LEQ:
	    arg1->un.constant.num =
		    (arg1->un.constant.num <= arg2->un.constant.num );
	    arg1->un.constant.hbit = 0;
	    break;

	case T_GTR:
	    arg1->un.constant.num =
		    (arg1->un.constant.num > arg2->un.constant.num );
	    arg1->un.constant.hbit = 0;
	    break;

	case T_LSS:
	    arg1->un.constant.num =
		    (arg1->un.constant.num < arg2->un.constant.num );
	    arg1->un.constant.hbit = 0;
	    break;

	case T_SR0:
	    arg1->un.constant.num &= maskbits( arg1->un.constant.hbit );
	    arg1->un.constant.num =
		    (unsigned) arg1->un.constant.num >> arg2->un.constant.num;
	    break;

	case T_SL0:
	    arg1->un.constant.num =
			arg1->un.constant.num << arg2->un.constant.num;
	    arg1->un.constant.hbit =
			const_width((unsigned)arg1->un.constant.num) - 1;
	    /* arg1->un.constant.num &= maskbits( arg1->un.constant.hbit ); */
	    break;

	case T_BIT:
	    if ( arg3 == NULL ) {
		arg3 = node_dup( arg2 );
	    }
	    if (arg2->un.constant.num < arg3->un.constant.num )
		error("Backwards bit range",cnode);
	    if ( arg1->un.constant.hbit <
		    (arg2->un.constant.num - arg3->un.constant.num) )
		error("Bit select out of range",cnode);
	    arg1->un.constant.hbit = arg2->un.constant.num - arg3->un.constant.num;
	    arg1->un.constant.num = arg1->un.constant.num >> arg3->un.constant.num;
	    arg1->un.constant.num &= maskbits( arg1->un.constant.hbit );
	    break;

	default:
	    ierror("Constant_prop", cnode);
    }
    NODE_DISPOSE(cnode);
    NODE_DISPOSE(arg2);
    NODE_DISPOSE(arg3);
    return arg1;
}

/*
 * This routines will take contant and variable vectors and reduce it
 * to a concat string of bits.  Functions encountered by this procedure
 * are assumed to be single bit functions already.
 */
node *bitblast(bnode)
node *bnode;
{
    node *result;		/* result of bitblast */
    node *resultend;		/* end of current result string */
    node *expr;			/* one part of the input */
    node *nextexpr;		/* next expression in conc chain */
    node *string;		/* result string for one part of input */
    node *stringend;		/* current end of string */
    int number;			/* num field of constant */
    int hbit,lbit;		/* bit range to produce */
    int i;

    if( bnode == NULL ) {
	ierror("bitblast of NULL", bnode);
    }
    result = resultend = NULL;
    expr = bnode;
    nextexpr = expr->conc;
    expr->conc = NULL;

    while( expr != NULL ) {

	switch ( expr->type ) {

	    case T_IS_CONSTANT:
		string = stringend = expr;
		number = expr->un.constant.num;
		hbit = expr->un.constant.hbit;
		expr->un.constant.num = number & 1;
		expr->un.constant.hbit = 0;
		for ( i = 1; i <= hbit; i++ ) {
		    stringend->conc = node_create_fi(T_IS_CONSTANT,expr);
		    stringend = stringend->conc;
		    number = (unsigned) number >> 1;
		    stringend->un.constant.num = number & 1;
		    stringend->un.constant.hbit = 0;
		}
		break;

	    case T_IS_STATE:
		string = stringend = expr;
		hbit = expr->un.bdvar.hbit;
		lbit = expr->un.bdvar.lbit;
		expr->un.bdvar.hbit = lbit;
		for ( i = lbit + 1; i <= hbit; i++ ) {
		    stringend->conc = node_create_fi(T_IS_STATE,expr);
		    stringend = stringend->conc;
		    stringend->un.bdvar.hbit = i;
		    stringend->un.bdvar.lbit = i;
		    stringend->un.bdvar.name = expr->un.bdvar.name;
		    stringend->un.bdvar.rev = expr->un.bdvar.rev;
		}
		break;

	    default:
		string = stringend = expr;
	}
	if ( resultend == NULL ) {
	    result = string;
	} else {
	    resultend->conc = string;
	}
	resultend = stringend;

	expr = nextexpr;
	if ( expr != NULL ) {
	    nextexpr = expr->conc;
	    expr->conc = NULL;
	}
    }
    return result;
}

/*
 * Will return a bit mask that is suitable for a constant with hbit
 */
maskbits( hbit )
int hbit;
{
    if (hbit >= 31) {
	return ~0;
    } else {
	return ~(~0 << (hbit+1));
    }
}

/*
 * Returns true if constant prop can handle type of function.
 */
canconstprop( type )
int type;
{
    switch( type ) {
	case T_NOT: case T_BUF: case T_UNARY_OR: case T_UNARY_AND:
	case T_WIDTH: case T_UNARY_PLUS: case T_UNARY_MINUS:
	case T_AND: case T_OR: case T_XOR:
	case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
	case T_NAND: case T_NOR: case T_EQV:
	case T_EQL: case T_NEQ: case T_GEQ: case T_LEQ: case T_GTR: case T_LSS:
	case T_SR0: case T_SL0: case T_BIT:
	    return 1;

	default:
	    return 0;
    }
}

/*
 * Will return true if expr is a wildcard bit
 */
wildcard( expr )
node *expr;
{
    return (expr->type == T_IS_CONSTANT) && (expr->un.constant.dont_care);
}

/*
 * Merges constants for bitglue.
 */
mergeit(expr)
node *expr;
{
    if ( expr == NULL ) {
	return 0;
    } else {
	return expr->un.constant.num +
	    (unsigned) 2 * (unsigned) ( mergeit( expr->conc ));
    }
}

/*
 * This routine simplifies ANDR and ORR on arbitrary expressions.
 * Constants and wildcards will simplify things accordingly.
 */
node *one_bit_result( onode, arg1 )
node *onode;
node *arg1;
{
    node *expr;			/* bit blast of arg1 */
    node *lastbit;		/* last bit processed */
    node *bit;			/* bit to process */

    if ( arg1 == NULL ) {
	expr = bit = NULL;
    } else {
	expr = bit = bitblast( arg1 );
    }
    lastbit = NULL;

    while ( bit != NULL ) {
	if ( bit->type == T_IS_CONSTANT ) {
	    if ( bit->un.constant.dont_care ) {
		NODE_DISPOSE( expr );
		expr = node_create_bit( 0, onode );
		expr->un.constant.dont_care = 1;
		NODE_DISPOSE( onode );
		return expr;
	    }
	    if ( onode->type == T_UNARY_OR && bit->un.constant.num == 1) {
		NODE_DISPOSE( expr );
		expr = node_create_bit( 1, onode );
		NODE_DISPOSE( onode );
		return expr;
	    }
	    if ( onode->type == T_UNARY_AND && bit->un.constant.num == 0) {
		NODE_DISPOSE( expr );
		expr = node_create_bit(0, onode );
		NODE_DISPOSE( onode );
		return expr;
	    }
	    if ( lastbit == NULL ) {
		expr = bit->conc;
		bit->conc = NULL;
		NODE_DISPOSE( bit );
		bit = expr;
	    } else {
		lastbit->conc = bit->conc;
		bit->conc = NULL;
		NODE_DISPOSE( bit );
		bit = lastbit->conc;
	    }
	} else {
	    lastbit = bit;
	    bit = bit->conc;
	}
    }
    /* Take care of all 1 AND or all 0 OR */
    if ( expr == NULL && onode->type == T_UNARY_OR ) {
	expr = node_create_bit(0,onode);
	NODE_DISPOSE(onode);
    } else if ( expr == NULL && onode->type == T_UNARY_AND ) {
	expr = node_create_bit(1,onode);
	NODE_DISPOSE(onode);
    } else if ( expr->conc == NULL ) {
	/* ORR or ANDR of one bit = that bit */
	NODE_DISPOSE(onode);
    } else {
	/* Keep onode if expression under it is > single bit */
	onode->un.unary.expr = expr;
	expr = onode;
    }
    return expr;
}

/*
 * This routine has several uses.  It accepts a bitblasted argument
 * and returns:
 * The number of bits. (function value)
 * If the bits are all constant. (consts)
 * The getbith bit. (returnbit)  -- LSB is 0.
 * The last bit (lastbit)
 * Non-bitblasted constants are allowed, but you may not get a 
 * returnbit or lastbit.
 */
countbits( arg1, consts, getbit, returnbit, lastbit )
node *arg1;
int *consts;
int getbit;
node **returnbit;
node **lastbit;
{
    node *bit;			/* current bit */
    int count;

    bit = arg1;
    count = 0;
    *consts = 1;

    while( bit != NULL ) {
	if (count == getbit)
	    *returnbit = bit;
	if ( bit->type != T_IS_CONSTANT || bit->un.constant.dont_care )
	    *consts = 0;
	if ( bit->type == T_IS_CONSTANT ) {
	    count += bit->un.constant.hbit + 1;
	} else {
	    count++;
	}
	*lastbit = bit;
	bit = bit->conc;
    }
    return count;
}
