#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "yybdsyn.h"
#include "yylex.h"

static int map_local();
static int un_map_local();
static node **expand();
static node **inline_routine();
static char *rootname;

node *inline_func(psym)
symbol *psym;
{
    node *q, *routine;

    if ((routine = psym->u.routine.body) == NIL(node)) {
	return NIL(node);
    }

    rootname = ALLOC(char, 50);
    rootname[0] = '\0';

    q = node_create(T_FCALL, routine->file, routine->line);
    q->un.fcall.def = psym;
    q->un.fcall.actual_args = NIL(node);
    (void) inline_routine(q, &q);
    FREE(rootname);
    node_dispose(routine);
    return q;
}


node *expand_name(tree, name)
node *tree;
char *name;
{
    rootname = ALLOC(char, strlen(name) + 2);
    (void) sprintf(rootname, "%s/", name);
    (void) expand(tree, &tree);
    FREE(rootname);
    return tree;
}

/*
 *  expand -- locate all T_FCALL nodes in the tree starting from p and
 *  replace the T_FCALL node with a reference to the return variable
 *  for the routine.  The actual body of the routine is inserted
 *  following "plast".  The return value is a new value for plast which
 *  still points to p.
 *
 *  The in-line expansion includes replacing all references to local
 *  variables (including the return value and the formal arguments)
 *  with a symbol table pointer to a global variable.  These global
 *  variables are created by inline_routine, and pointers to the global
 *  variable are inserted into the local symbol table for the routine.
 *
 *  We inforce slightly tighter semantics than other recursive routines
 *  walking over the tree:
 *      p->conc must be NIL(node) for "statements"
 */

static node **expand(p, plast)
node *p;		/* pointer to current node */
node **plast;		/* the address of a pointer which points to us */
{
    node *q, *qnext;

    if (p == NIL(node)) {
	return plast;
    }

    switch(p->type) {

	/* most of the work is mapping local state names */
	case T_IS_STATE:
	    if (p->un.bdvar.name->u.state.is_local) {
		/* replace the name pointer with the global variable */
		assert(p->un.bdvar.name->u.state.local_def != 0);
		assert(!p->un.bdvar.name->u.state.local_def->u.state.is_local);
		p->un.bdvar.name = p->un.bdvar.name->u.state.local_def;
	    }
	    plast = expand(p->conc, plast);
	    break;

	/* the rest is inline-expanding routine calls */
	case T_FCALL:
	    /* walk down the list ourselves */
	    for(q = p->un.fcall.actual_args; q != NIL(node); q = q->next) {
		qnext = q->next;
		q->next = NIL(node);
		plast = expand(q, plast);
		q->next = qnext;
	    }
	    plast = inline_routine(p, plast);
	    plast = expand(p->conc, plast);
	    break;

	/* constants are pretty boring */
	case T_IS_CONSTANT:
	    plast = expand(p->conc, plast);
	    break;

	/* unary operators inside expressions */
	case T_WIDTH:      case T_NOT:         case T_BUF: 
	case T_UNARY_OR:   case T_UNARY_AND: 
	case T_UNARY_PLUS: case T_UNARY_MINUS:
	    plast = expand(p->un.unary.expr, plast);
	    plast = expand(p->conc, plast);
	    break;
    
	/* binary operators inside expressions */
	case T_AND:  case T_OR:    case T_NAND:  case T_NOR: 
	case T_XOR:  case T_EQV:   case T_EQL:   case T_NEQ: 
	case T_LEQ:  case T_GEQ:   case T_LSS:   case T_GTR:
	case T_SL0:  case T_SR0:   case T_SL1:   case T_SR1: 
	case T_SLR:  case T_SRR:   case T_SXT:   case T_ZXT: case T_OXT:
	case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
	    plast = expand(p->un.binary.lexpr, plast);
	    plast = expand(p->un.binary.rexpr, plast);
	    plast = expand(p->conc, plast);
	    break;

	/* ternary operators inside expressions */
	case T_BIT:
	    plast = expand(p->un.ternary.lexpr, plast);
	    plast = expand(p->un.ternary.mexpr, plast);
	    plast = expand(p->un.ternary.rexpr, plast);
	    plast = expand(p->conc, plast);
	    break;

	/* 
	 * Now we start with the statements -- they have no p->conc, have a p->next
	 */
	case T_EQ:
	    assert(p->conc == NIL(node));
	    plast = expand(p->un.assign.lexpr, plast);
	    plast = expand(p->un.assign.rexpr, plast);
	    break;

	case T_IF:
	    assert(p->conc == NIL(node));
	    plast = expand(p->un.bdif.expr, plast); 
	    plast = expand(p->un.bdif.thenb, &(p->un.bdif.thenb));
	    plast = expand(p->un.bdif.elseb, &(p->un.bdif.elseb));
	    break;

	case T_SELECT:
	case T_SELECTONE:
	case T_SELECTALL:
	    assert(p->conc == NIL(node));
	    plast = expand(p->un.select.expr, plast);
	    plast = expand(p->un.select.bdcase, plast);
	    break;

	case T_CASE:
	    assert(p->conc == NIL(node));
	    plast = expand(p->un.bdcase.expr, plast);
	    plast = expand(p->un.bdcase.stmt, &(p->un.bdcase.stmt));
	    break;

	case T_FOR:
	case T_FOR_DOWN:
	    assert(p->conc == NIL(node));
	    plast = expand(p->un.bdfor.var, plast);
	    plast = expand(p->un.bdfor.from, plast);
	    plast = expand(p->un.bdfor.to, plast);
	    plast = expand(p->un.bdfor.by, plast);
	    plast = expand(p->un.bdfor.stmt, &(p->un.bdfor.stmt));
	    break;

	case T_LABEL:
	    assert(p->conc == NIL(node));
	    plast = expand(p->un.label.stmt, &(p->un.label.stmt));
	    break;

	case T_OTHERWISE:
	case T_LEAVE:
	    assert(p->conc == NIL(node));
	    break;

	default:
	    (void) fprintf(stderr, "expand: Unexpected node %s\n", 
		    map_token_to_name(p->type));
	    exit(1);
    }

    if (p->next != NIL(node)) {
	plast = expand(p->next, &(p->next));
    }
    return plast;
}

/*
 *   inline_routine -- expand a routine in-line
 *
 *  q is a pointer to an fcall node which we wish to expand in-line
 *
 *  steps are:
 *      1) duplicate the parse tree for the routine
 *
 *      2) make assignment statements from each actual parameter to the
 *      formal parameters, and place this before the duplicated parse
 *      tree.  (The formal parameters are merely local variables to the
 *      routine.)
 *
 *      3) for each local state in the routine, create a global
 *      variable which will replace the local state.  Record in the
 *      symbol table a pointer to the global variable which will take
 *      its place.
 *
 *      3) expand the tree for any deeper function calls; this includes
 *      the transformation which maps references to local states
 *      (including arguments and the return value) into references to
 *      the corresponding global variable created in step 2
 *
 *      4) Insert the tree after "plast"
 *
 *      5) Replace q with a node which references the return value of
 *      the routine.
 *
 */

static node **inline_routine(q, plast)
node *q;
node **plast;
{
    char *saveroot, *buffer;
    node *tree, *qq;
    node *actuals, *formals;
    node *retval, *rformal, *ractual, *rformal_next, *ractual_next;
    symbol *routine;

    /* fetch symbol table entry for the routine call */
    routine = q->un.fcall.def;

    /* The routine named ASSERT is special and ignored ... */
    if (upper_strcmp(routine->name, "ASSERT") == 0) {
	q->type = T_IS_CONSTANT;
	q->un.constant.num = 0;
	q->un.constant.hbit = 0;
	q->un.constant.dont_care = 0;
	return plast;
    }

    /* Make up a new root name */
    saveroot = util_strsav(rootname);
    buffer = ALLOC(char, strlen(routine->name) + 20);
    (void) sprintf(buffer, "%s_%d/", routine->name, 
	++(routine->u.routine.call_count));
    rootname = REALLOC(char, rootname, strlen(rootname)+strlen(buffer)+10);
    (void) strcat(rootname, buffer);
    FREE(buffer);

    /* duplicate the parse tree for the routine we're expanding */
    tree = node_dup(routine->u.routine.body);

    /* grab the actual and formal argument lists (and return value) */
    ractual = q->un.fcall.actual_args;
    retval = routine->u.routine.formal_args;
    rformal = retval->next;

    /* Save a copy of the actuals and formals for return var params */
    formals = node_dup(rformal);
    actuals = node_dup(ractual);

    /* Create assignments from the actuals to the formals */
    while(ractual != NIL(node) && rformal != NIL(node)) {
    	assert(rformal->type == T_IS_STATE);

	/* save next pointers, and unlink the "next" field */ 
	rformal_next = rformal->next;
	rformal->next = NIL(node);
	ractual_next = ractual->next;
	ractual->next = NIL(node);

	/* create an assignment, and add before "tree" */
	qq = node_create(T_EQ, q->file, q->line);
	qq->un.assign.lexpr = node_dup(rformal);
	qq->un.assign.rexpr = ractual;
	qq->next = tree;
	tree = qq;

	/* setup for next loop */
	rformal->next = rformal_next;
	rformal = rformal_next;
	ractual = ractual_next;
    }
    if (ractual != NIL(node) || rformal != NIL(node)) {
	error("Routine call argument count mismatch", q);
    }

    /* Find the last statement of the tree */
    for(qq = tree; qq->next != NIL(node); qq = qq->next) {
    }

    /* Create assignments from the var formals to the actuals
     * and insert them at the end of the tree.
     */
    for(ractual = actuals, rformal = formals;
	ractual != NIL(node) && rformal != NIL(node);
	rformal = rformal_next,	ractual = ractual_next) {

    	assert(rformal->type == T_IS_STATE);

	/* save next pointers, and unlink the "next" field */ 
	rformal_next = rformal->next;
	rformal->next = NIL(node);
	ractual_next = ractual->next;
	ractual->next = NIL(node);

	/* We are looking for var formals */
	if (!rformal->un.bdvar.var) {
	  node_dispose(rformal);
	  node_dispose(ractual);
	  continue;
	}

	/* Make sure the actual is a T_IS_STATE. Var params must not
	 * be passed an expression.
	 */
	if (ractual->type != T_IS_STATE && ractual->type != T_BIT) {
	    char emess[1000];

	    sprintf(emess, "Attempt to pass an expression to %s, a VAR param of %s.",
		    rformal->un.bdvar.name->name, routine->name);
	    error(emess, ractual);
	}

	/* create an assignment, and append to "tree" */
	qq = qq->next = node_create(T_EQ, q->file, q->line);
	qq->un.assign.lexpr = ractual;
	qq->un.assign.rexpr = rformal;
	qq->next = NIL(node);
    }


    /* Map the local variables of the called routine into globals */
    symtab_foreach(routine->u.routine.locals, map_local, -1);

    /*  Now we are ready to recur on this routine which may call other routines
     *  or at least we will insert all of the correct stuff
     */
    (void) expand(tree, &tree);

    /* Insert the expanded tree for the function call 
     * into the statement list
     */
    qq->next = *plast;
    *plast = tree;

    /* Restore the rootname */
    (void) strcpy(rootname, saveroot);
    FREE(saveroot);

    /* Replace the node q (which was the FCALL) with a reference to the
     * return value
     */
    q->type = T_IS_STATE;
    q->un.bdvar.name = retval->un.bdvar.name->u.state.local_def;
    assert(q->un.bdvar.name != 0);
    q->un.bdvar.name->u.state.is_temp = 1;
    q->un.bdvar.hbit = retval->un.bdvar.name->u.state.hibit;
    q->un.bdvar.lbit = retval->un.bdvar.name->u.state.lobit;
    q->un.bdvar.rev = (struct usage_struct *) 0;

    /* For debugging, unmap the locals */
    symtab_foreach(routine->u.routine.locals, un_map_local, -1);

    return &(qq->next);
}
	


/* called once for each local symbol which is a T_IS_STATE */
static int map_local(psym)
symbol *psym;
{
    char name[1024];
    symbol *newsym;

    if (psym->type == T_IS_STATE || psym->type == T_IS_META) {
	(void) strcpy(name, rootname);
	(void) strcat(name, psym->name);
	assert(local_syms == 0);
	newsym = yacc_lookup(name);
	if (newsym->type != T_IS_UNDEFINED) {
	    ierror("expected unique name in map_local", NIL(node));
	}
	newsym->type = psym->type;
	newsym->u.state.nobit = psym->u.state.nobit;
	newsym->u.state.lobit = psym->u.state.lobit;
	newsym->u.state.hibit = psym->u.state.hibit;
	psym->u.state.local_def = newsym;
    }
    return 1;			/* continue the search */
}

static int un_map_local(psym)
symbol *psym;
{
    if (psym->type == T_IS_STATE || psym->type == T_IS_META) {
	psym->u.state.local_def = NIL(symbol);
    }
    return 1;
}

static int clear_call_count();
static int first_pass_process();
static void trace_call_graph();


/*
 *  trace_call_tree -- trace the call graph checking for cycles, and
 *  marking those routines which are never called (i.e., the sources)
 */

void trace_call_tree()
{
    symtab_foreach(global_syms, first_pass_process, T_IS_ROUTINE);
    symtab_foreach(global_syms, clear_call_count, T_IS_ROUTINE);
}


static int first_pass_process(routine)
symbol *routine;
{
    /* check that the routine is not currently active */
    if (routine->u.routine.active) {
	error("Recursive calls not supported", routine->u.routine.body);
    }
    routine->u.routine.active = 1;

    if (routine->u.routine.call_count++ == 0) {
	trace_call_graph(routine->u.routine.body);
    }

    routine->u.routine.active = 0;
    return 1;
}


static int clear_call_count(routine)
symbol *routine;
{
    if (upper_strcmp(routine->name, "ASSERT") == 0) {
	routine->u.routine.active = 0;
    } else {
	routine->u.routine.active = (routine->u.routine.call_count == 1);
    }
    routine->u.routine.call_count = 0;
    return 1;
}


static void trace_call_graph(p)
node *p;
{
    if (p == NIL(node))
	return;

    switch(p->type) {
	case T_OTHERWISE:
	case T_IS_CONSTANT:
	case T_IS_STATE:
	    break;

	case T_FCALL:
	    trace_call_graph(p->un.fcall.actual_args);
	    (void) first_pass_process(p->un.fcall.def);
	    break;

	case T_WIDTH:      case T_NOT:         case T_BUF: 
	case T_UNARY_OR:   case T_UNARY_AND: 
	case T_UNARY_PLUS: case T_UNARY_MINUS:
	    trace_call_graph(p->un.unary.expr);
	    break;
    
	case T_AND:  case T_OR:    case T_NAND:  case T_NOR: 
	case T_XOR:  case T_EQV:   case T_EQL:   case T_NEQ: 
	case T_LEQ:  case T_GEQ:   case T_LSS:   case T_GTR:
	case T_SL0:  case T_SR0:   case T_SL1:   case T_SR1: 
	case T_SLR:  case T_SRR:   case T_SXT:   case T_ZXT: case T_OXT:
	case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
	    trace_call_graph(p->un.binary.lexpr);
	    trace_call_graph(p->un.binary.rexpr);
	    break;

	case T_BIT:
	    trace_call_graph(p->un.ternary.lexpr);
	    trace_call_graph(p->un.ternary.mexpr);
	    trace_call_graph(p->un.ternary.rexpr);
	    break;

	case T_EQ:
	    trace_call_graph(p->un.assign.lexpr);
	    trace_call_graph(p->un.assign.rexpr);
	    break;

	case T_IF:
	    trace_call_graph(p->un.bdif.expr);
	    trace_call_graph(p->un.bdif.thenb);
	    trace_call_graph(p->un.bdif.elseb);
	    break;

	case T_SELECT:
	case T_SELECTONE:
	case T_SELECTALL:
	    trace_call_graph(p->un.select.expr);
	    trace_call_graph(p->un.select.bdcase);
	    break;

	case T_CASE:
	    trace_call_graph(p->un.bdcase.expr);
	    trace_call_graph(p->un.bdcase.stmt);
	    break;

	case T_FOR:
	case T_FOR_DOWN:
	    trace_call_graph(p->un.bdfor.var);
	    trace_call_graph(p->un.bdfor.from);
	    trace_call_graph(p->un.bdfor.to);
	    trace_call_graph(p->un.bdfor.by);
	    trace_call_graph(p->un.bdfor.stmt);
	    break;

	case T_LABEL:
	    trace_call_graph(p->un.label.stmt);
	    break;

	case T_LEAVE:
	    break;

	default:
	    (void) fprintf(stderr, "trace_call_graph: Unexpected node %s\n", 
		    map_token_to_name(p->type));
	    exit(1);
    }

    trace_call_graph(p->conc);
    trace_call_graph(p->next);
}
