#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "yybdsyn.h"
#include "mult.h"


/*
 * Structure to keep track of past revisions to variables
 */
typedef struct case_info {
    struct case_info *next;	/* next cases revinfo list */
    revinfo *list;		/* revlist for a case */
} caseinfo;


usage *alloc_usage();
usage *find_reference();
char *version_string();
char *update_oldx();
char *strip_oldx();
revinfo *fix_revlist();
char *new_skeleton();
node *select_version();
node *if_version();


/*
 * Give the LHS of an assignment a brand new rev string.
 * Also, if keeprevs is true, keep a list of every LHS touched
 * Don't do keeprevs on $$COND variables, since they can be thought of
 * as existing in the root level (not nested).  (They are assigned
 * only once and then used immediately.  Why waste effort in mult
 * assignment? )
 * If lhs does not fan out, replace the lhs var with a constant.
 * This will flag "clean_up" to blow away this assignment.
 */
revinfo *revassign(assign, theBlock, keeprevs, revlist)
node *assign;		/* assign that makes new revision */
block *theBlock;	/* block of assign */
int keeprevs;		/* keep revlist only if true */
revinfo *revlist;	/* last revisions of the vars */
{
    node *lhs;			/* LHS of assign */
    symbol *symbolent;		/* LHS's entry in table */
    int lodecl;			/* lobit with which var was declared */
    revision *vers;		/* point to list of assigns to bits of var */
    revision *bitver;		/* version info for this bit */
    register revinfo *revindex;	/* index into revlist */

    lhs = assign->un.assign.lexpr;
    if ( lhs->type != T_IS_STATE) {
	error( "Assignment to CONSTANT",lhs );
    }
    if ((lhs->type == T_IS_STATE) &&
	    ( lhs->un.bdvar.name->type == T_IS_INPUT_PORT)) {
	error( "Assignment to input port", lhs);
    }
    
    symbolent = lhs->un.bdvar.name;
    lodecl = symbolent->u.state.lobit; 

    /* assign a version number for left side */
    if ( symbolent->u.state.versions == NULL ) {
	vers = symbolent->u.state.versions =
		rev_array( symbolent->u.state.hibit - lodecl + 1 );
    } else {
	vers = symbolent->u.state.versions;
    }

    if ( lhs->un.bdvar.rev == NIL(usage) ) {
	/* get pointer to element of version array */
	bitver = &( vers[lhs->un.bdvar.lbit-lodecl] );
	bitver->assigned.current = 1;

	/* check to see if lhs does not fan out */
	if ((lhs->un.bdvar.name->type != T_IS_OUTPUT_PORT) &&
		(bitver->last == NULL)) {
	    lhs->type = T_IS_CONSTANT;
	    lhs->un.constant.num = lhs->un.constant.hbit = 0;
	    lhs->un.constant.dont_care = 0;
	    return revlist;
	}
	bitver->recent = alloc_usage(bitver->recent, theBlock, assign);
	lhs->un.bdvar.rev = bitver->recent;

	/* don't do mult assign stuff on $$COND (just wasted effort) */
	if ( strncmp( symbolent->name, "$$COND", 6 ) == 0 ) {
	    return revlist;
	}

	/* record which revision */
	if ( keeprevs ) {

	    /* check to see if this bit is already in revlist */
	    for (revindex = revlist; revindex != NIL(revinfo);
		    revindex = revindex->next) {
		if ((revindex->name == lhs->un.bdvar.name) &&
			(revindex->bit == lhs->un.bdvar.lbit)) {
		    revindex->info = bitver->recent;
		    return revlist;
		}
	    }

	    /* create new entry in revlist */
	    if (revlist == NIL(revinfo)) {
		revlist = revindex = ALLOC(revinfo, 1);
		revindex->next = NIL(revinfo);
	    } else {
		revindex = ALLOC(revinfo, 1);
		revindex->next = revlist;
		revlist = revindex;
	    }
	    revindex->info = bitver->recent;
	    revindex->name = lhs->un.bdvar.name;
	    revindex->bit = lhs->un.bdvar.lbit;
	}
    }
    return revlist;
}


/*
 * This will substitute in version numbers for all variables.  It will also
 * record the fact that the variable was used.
 */
node *subst_version( root, theBlock, keeprevs, revlist )
node *root;
block *theBlock;		/* block of structure */
int keeprevs;			/* a revlist will be kept when this is true */
revinfo **revlist;		/* returned list of info about assigned vars */
{
    revinfo *revdummy;
    symbol *varinfo;		/* symtab entry for var */
    revision *vers;		/* array of revisions for a var */
    int bitnum;			/* bit of var we are working on */
    int lodecl;			/* lowest declared bit of var */
    node *exprroot;		/* root of expr on list */
    node *wordroot;		/* root of a conc'ed word */
    node *prevroot;		/* previous in next string */
    node *nextroot;		/* next node to process */
    node *lastofnew;		/* last node in inserted block */
    node *nindex, *nindex2;	/* index into node list */
    node *tempnode;
    block *lastgroup;		/* last encountered IF or SELECT block */

    lastgroup = NIL(block);
    prevroot = NULL;
    for ( exprroot = root; exprroot != NULL; exprroot = nextroot ) {
	nextroot = exprroot->next;
	for ( wordroot=exprroot; wordroot != NULL; wordroot = wordroot->conc ) {
	    switch( wordroot->type ) {

		case T_IS_CONSTANT:
		case T_OTHERWISE:
		    break;

		case T_BUF: case T_NOT: case T_WIDTH:
		case T_UNARY_PLUS: case T_UNARY_MINUS:
		case T_UNARY_AND: case T_UNARY_OR:
		    wordroot->un.unary.expr = 
			subst_version( wordroot->un.unary.expr, theBlock,
			0, &revdummy );
		    break;

		case T_AND: case T_OR: case T_NAND:
		case T_NOR: case T_XOR: case T_EQV:
		case T_EQL: case T_NEQ: case T_LEQ:
		case T_GEQ: case T_LSS: case T_GTR:
		case T_SXT: case T_ZXT: case T_OXT:
		case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
		case T_SL0: case T_SR0: case T_SL1: case T_SR1:
		case T_SLR: case T_SRR:
		    wordroot->un.binary.lexpr = 
			subst_version( wordroot->un.binary.lexpr, theBlock,
			0, &revdummy );
		    wordroot->un.binary.rexpr = 
			subst_version( wordroot->un.binary.rexpr, theBlock,
			0, &revdummy );
		    break;

		case T_BIT:
		    wordroot->un.ternary.lexpr = 
			subst_version( wordroot->un.ternary.lexpr, theBlock,
			0, &revdummy );
		    wordroot->un.ternary.mexpr = 
			subst_version( wordroot->un.ternary.mexpr, theBlock,
			0, &revdummy );
		    wordroot->un.ternary.rexpr = 
			subst_version( wordroot->un.ternary.rexpr, theBlock,
			0, &revdummy );
		    break;

		case T_EQ:
		    wordroot->un.assign.rexpr = 
			subst_version( wordroot->un.assign.rexpr, theBlock,
			0, &revdummy );
		    /* OK, now we break up T_EQ into bitwise T_EQ and
		     * call revassign on each one */
		    /* width of lexpr and rexpr were previously
		     * made equal */
		    nindex = wordroot->un.assign.lexpr->conc;
		    nindex2 = wordroot->un.assign.rexpr->conc;
		    wordroot->un.assign.lexpr->conc = NULL;
		    wordroot->un.assign.rexpr->conc = NULL;
		    *revlist = revassign(wordroot, theBlock,
			    keeprevs, *revlist);
		    while ( nindex != NULL ) {
			tempnode = node_create_fi( T_EQ, wordroot );
			tempnode->next = wordroot;
			if ( prevroot == NULL ) {
			    root = tempnode;
			} else {
			    prevroot->next = tempnode;
			}
			prevroot = tempnode;
			tempnode->un.assign.lexpr = nindex;
			tempnode->un.assign.rexpr = nindex2;
			*revlist = revassign(tempnode, theBlock,
				keeprevs, *revlist);
			nindex = nindex->conc;
			nindex2 = nindex2->conc;
			tempnode->un.assign.lexpr->conc = NULL;
			tempnode->un.assign.rexpr->conc = NULL;
		    }
		    break;

		case T_SELECT:
		    if (lastgroup == NIL(block)) {
			lastgroup = theBlock->child;
		    } else {
			lastgroup = lastgroup->nextgroup;
		    }
		    exprroot = wordroot = 
			    select_version( wordroot, theBlock,
			    lastgroup, keeprevs, revlist, &lastofnew);
		    /* link on new exprroot */
		    if ( prevroot == NULL ) {
			root = exprroot;
		    } else {
			prevroot->next = exprroot;
		    }
		    exprroot = lastofnew;
		    exprroot->next = nextroot;
		    break;

		case T_IF:
		    if (lastgroup == NIL(block)) {
			lastgroup = theBlock->child;
		    } else {
			lastgroup = lastgroup->nextgroup;
		    }
		    exprroot = wordroot = 
			    if_version( wordroot, theBlock, lastgroup,
			    keeprevs, revlist, &lastofnew);
		    /* link on new exprroot */
		    if ( prevroot == NULL ) {
			root = exprroot;
		    } else {
			prevroot->next = exprroot;
		    }
		    exprroot = lastofnew;
		    exprroot->next = nextroot;
		    break;

		case T_IS_STATE:
		    /* assign usage to rev field */
		    if (wordroot->un.bdvar.rev != NIL(usage)) {
			ierror("rev assigned twice", wordroot);
		    }
		    varinfo = wordroot->un.bdvar.name;
		    bitnum = wordroot->un.bdvar.lbit;
		    wordroot->un.bdvar.rev = find_reference(varinfo,
			    bitnum, theBlock);

		    /* flag last reference to var, if this is it */
		    vers = varinfo->u.state.versions;
		    lodecl = varinfo->u.state.lobit;
		    if (vers[bitnum-lodecl].last == wordroot) {
			vers[bitnum-lodecl].last = NIL(node);
		    }
		    break;

		default:
		    (void) fprintf(stderr,"%d",wordroot->type);
		    ierror("bad type in subst_version",wordroot); 
	    }
	}
	prevroot = exprroot;
    }
    return root;
}


/*
 * This allocates an array of revision *, initialized to all NULLs
 */
revision *rev_array(num)
int num;
{
    int i;		/* index */
    revision *array;

    array = ALLOC( revision, num );
    for( i=0; i < num ; i++ ) {
	array[i].recent = NIL(usage);
	array[i].last = NIL(node);
	array[i].assigned.previous = 0;
	array[i].assigned.current = 0;
	array[i].assigned.unspec = 0;
    }
    return array;
}


/*
 * Allocate and initialize a new usage
 */
usage *alloc_usage(prevuse, theBlock, assign)
usage *prevuse;			/* usage of var before this */
block *theBlock;		/* block of usage */
node *assign;			/* where is assign of usage */
{
    usage *newuse;

    newuse = ALLOC(usage, 1);
    newuse->block = theBlock;
    newuse->prev = prevuse;
    newuse->assign = assign;
    newuse->rev = 1;		/* default */
    while (prevuse != NIL(usage)) {
	if (prevuse->block == newuse->block) {
	    newuse->rev = prevuse->rev + 1;
	    break;
	}
	prevuse = prevuse->prev;
    }
    return newuse;
}


/*
 * Return the usage for the last revision of var that effects this part
 * of code.  Return NIL if there is no such usage.
 */
usage *find_reference(name, bit, theBlock)
symbol *name;
int bit;
block *theBlock;	/* current block */
{
    revision *vers;		/* array of revisions */
    int lodecl;			/* lowest declared bit of var */
    usage *uindex;

    vers = name->u.state.versions;
    lodecl = name->u.state.lobit;
    if ( vers == NULL ) {
	ierror("find_reference", (node *) 0);
    }

    for (uindex = vers[bit-lodecl].recent; uindex != NIL(usage); 
	    uindex = uindex->prev) {
	while (uindex->block->level < theBlock->level) {
	    theBlock = theBlock->parent;
	}
	if (uindex->block == theBlock) {
	    return uindex;
	}
    }
    /* use of unspecified variable */
    return NIL(usage);
}


/* 
 * This returns a node that is the last update of x.
 */
node *lastx(revtomatch, addtoor, theBlock)
revinfo *revtomatch;
node *addtoor;			/* for node file info only */
block *theBlock;		/* block result is created */
{
    node *result;

    result = node_create_fi( T_IS_STATE, addtoor );
    result->un.bdvar.name = revtomatch->name;
    result->un.bdvar.lbit = revtomatch->bit;
    result->un.bdvar.hbit = revtomatch->bit;
    result->un.bdvar.rev = find_reference(revtomatch->name,
	    revtomatch->bit, theBlock);
    return result;
}


/*
 * Removes variables that will not be used again.
 */
revinfo *fix_revlist( list )
revinfo *list;		/* revision list */
{
    register revinfo *revlistind;	/* index into revision list */
    revinfo *next;			/* next in list */
    revinfo *last;			/* former in list */
    revision *vers;		/* array of revisions for a var */
    int bitnum;			/* bit of var we are working on */
    int lodecl;			/* lowest declared bit of var */

    revlistind = list;
    last = NULL;
    while ( revlistind != NULL ) {
	next = revlistind->next;
	vers = revlistind->name->u.state.versions;
	lodecl = revlistind->name->u.state.lobit;
	bitnum = revlistind->bit;

	if ((revlistind->name->type != T_IS_OUTPUT_PORT) &&
		(vers[bitnum-lodecl].last == NIL(node))) {
	    FREE( revlistind );
	    if ( last == NULL ) {
		list = next;
	    } else {
		last->next = next;
	    }
	} else {
	    last = revlistind;
	}
	revlistind = next;
    }
    return list;
}


/*
 * Handles version assignment for selects.  This includes generating
 * implicit default assignment for each variable.
 */
node *select_version(exprroot, theBlock, subBlock, keeprevs, revlist, lastofnew)
node *exprroot;
block *theBlock;		/* block of structure */
block *subBlock;		/* first clause of select */
int keeprevs;			/* a revlist will be kept when this is true */
revinfo **revlist;		/* returned list of info about assigned vars */
node **lastofnew;		/* last statement in new block */
{
    int casenum;			/* current case of select */
    node *curcase;			/* current case being processed */
    revinfo *revdummy;			/* never holds anything */
    caseinfo *caselist;			/* list of rev info for all cases */
    caseinfo *curcaselist;		/* revisions for current case */
    revinfo *revlistind;		/* index into rev list */
    node *newassign;			/* new assignment for var */
    node *newassignlist;		/* list of newassigns */
    node *addtoor;			/* where to add a term for newass */
    node *andinor;			/* new term in ORR and'ed with cond */
    revinfo *temprevi;
    node *select;			/* the select to process */
    node *result;
    node *nindex;			/* node index */
    block *bindex;

    for (curcase = exprroot->un.select.bdcase; curcase != NIL(node);
	    curcase = curcase->next) {
	curcase->un.bdcase.expr = subst_version(curcase->un.bdcase.expr,
		theBlock, 0, &revdummy);
    }

    casenum = 1;
    caselist = curcaselist = ALLOC(caseinfo, 1);
    curcaselist->list = NULL;
    curcase = exprroot->un.select.bdcase;
    bindex = subBlock;
    while ( curcase != NULL ) {
	curcase->un.bdcase.stmt = subst_version(curcase->un.bdcase.stmt,
		bindex, 1, &(curcaselist->list));
	curcaselist->list = fix_revlist( curcaselist->list );
	curcase = curcase->next;
	if ( curcase == NULL ) {
	    curcaselist->next = NULL;
	} else {
	    curcaselist->next = ALLOC(caseinfo, 1);
	    curcaselist = curcaselist->next;
	    curcaselist->list = NULL;
	}
	bindex = bindex->nextclause;
	casenum++;
    }

    /* Reduce each x in select to one equation based on other x's */
    newassignlist = NULL;
    select = exprroot;
    curcaselist = caselist;
    casenum = 1;
    for( curcase = select->un.select.bdcase; curcase != NULL;
	    curcase = curcase->next) {
	revlistind = curcaselist->list;
	curcaselist->list = NULL;
	while(revlistind != NULL) {
	    /* Make eqn for this var */
	    newassign = node_create_fi(T_EQ,select);
	    newassign->un.assign.rexpr = node_create_fi(T_UNARY_OR,select);
	    newassign->un.assign.rexpr->un.unary.expr = andinor =
		    node_create_fi( T_AND, select );
	    andinor->un.binary.rexpr = node_dup( curcase->un.bdcase.expr );
	    andinor->un.binary.lexpr = addtoor =
		    node_create_fi( T_IS_STATE,select );
	    addtoor->un.bdvar.name = revlistind->name;
	    addtoor->un.bdvar.lbit = revlistind->bit;
	    addtoor->un.bdvar.hbit = revlistind->bit;
	    addtoor->un.bdvar.rev = revlistind->info;
	    other_cases(andinor, select->un.select.bdcase, casenum,
		    caselist, revlistind, theBlock);

	    newassign->un.assign.lexpr = node_create_fi(T_IS_STATE,select);
	    newassign->un.assign.lexpr->un.bdvar.name = revlistind->name;
	    newassign->un.assign.lexpr->un.bdvar.lbit = revlistind->bit;
	    newassign->un.assign.lexpr->un.bdvar.hbit = revlistind->bit;
	    *revlist = revassign(newassign, theBlock, keeprevs, *revlist);
	    temprevi = revlistind->next;
	    FREE( revlistind );
	    revlistind = temprevi;
	    if (newassignlist == NULL) {
		newassignlist = newassign;
	    } else {
		newassign->next = newassignlist;
		newassignlist = newassign;
	    }
	}
	curcaselist = curcaselist->next;
	casenum++;
    }
    free_caseinfo( caselist );
    /* now throw out the case structure and just keep the assigns */
    exprroot->next = NULL;		/* disconnect select */
    result = nindex = NULL;
    for( curcase = exprroot->un.select.bdcase; curcase != NULL ;
	    curcase = curcase->next ) {
	if ( nindex == NULL ) {
	    result = nindex = curcase->un.bdcase.stmt;
	} else {
	    nindex->next = curcase->un.bdcase.stmt;
	}
	curcase->un.bdcase.stmt = NULL;
	while ( nindex != NULL && nindex->next != NULL ) {
	    nindex = nindex->next;
	}
    }
    node_dispose( exprroot );	/* throw out select structure */
    if ( result == NULL ) {
	/* stuff in no-op where statements have failed to be */
	result = nindex = node_create_bit( 0, exprroot );
    }
    nindex->next = newassignlist;
    while ( nindex->next != NULL ) nindex = nindex->next;
    *lastofnew = nindex;
    return result;
}


/*
 * Searches the case blocks for another use of a variable in thiscase
 * block.  It rounds out the generated assignment appropriately.
 */
other_cases(andinor, cases, thiscasenum, caselist, revtomatch, theBlock)
node *andinor;
node *cases;
int thiscasenum;
caseinfo *caselist;
revinfo *revtomatch;		/* var to look for */
block *theBlock;		/* block of SELECT (not case) */
{
    revinfo *lastrli;			/* last revision in list seen */
    register revinfo *revlistind;	/* index into a revision list */
    caseinfo *curcaselist;		/* which revision list */
    node *curcase;			/* which case */
    node *newtoor;			/* new term in OR for var */
    int casenum;			/* case number */
    node *defaults;			/* condition for all cases
					   where x is unassigned */
    node *temp;

    defaults = node_create_fi( T_AND, andinor );
    defaults->un.binary.rexpr = lastx(revtomatch, andinor, theBlock);
    defaults->un.binary.lexpr = node_create_fi( T_UNARY_OR, andinor );
    curcaselist = caselist;
    curcase = cases;
    casenum = 1;
    while( curcase != NULL ) {
	if ( casenum != thiscasenum ) {

	    revlistind = curcaselist->list;
	    lastrli = NULL;
	    while((revlistind != NIL(revinfo)) &&
		    !((revlistind->name == revtomatch->name) &&
		      (revlistind->bit == revtomatch->bit))) {
		lastrli = revlistind;
		revlistind = revlistind->next;
	    }

	    if ( revlistind == NULL ) {
		if ( defaults->un.binary.lexpr->un.unary.expr == NULL ) {
		    defaults->un.binary.lexpr->un.unary.expr =
			    node_dup( curcase->un.bdcase.expr );
		} else {
		    temp = defaults->un.binary.lexpr->un.unary.expr;
		    defaults->un.binary.lexpr->un.unary.expr =
			    node_dup( curcase->un.bdcase.expr );
		    defaults->un.binary.lexpr->un.unary.expr->conc = temp;
		}
	    } else {
		/* Make eqn for this var */
		andinor->conc = node_create_fi( T_AND, andinor );
		andinor = andinor->conc;
		andinor->un.binary.rexpr = node_dup( curcase->un.bdcase.expr );
		andinor->un.binary.lexpr = newtoor =
			node_create_fi( T_IS_STATE, andinor );
		newtoor->un.bdvar.name = revtomatch->name;
		newtoor->un.bdvar.lbit = revtomatch->bit;
		newtoor->un.bdvar.hbit = revtomatch->bit;
		newtoor->un.bdvar.rev = revlistind->info;

		if( lastrli == NULL ) {
		    curcaselist->list = revlistind->next;
		} else {
		    lastrli->next = revlistind->next;
		}
		FREE( revlistind );
	    }
	}
	andinor->conc = defaults;
	curcase = curcase->next;
	curcaselist = curcaselist->next;
	casenum++;
    }
}


/*
 * Handles version assignment for ifs.  This includes generating
 * implicit default assignment for each variable.
 */
node *if_version(exprroot, theBlock, subBlock, keeprevs, revlist, lastofnew)
node *exprroot;
block *theBlock;		/* block of structure */
block *subBlock;		/* block of the then clause */
int keeprevs;			/* a revlist will be kept when this is true */
revinfo **revlist;		/* returned list of info about assigned vars */
node **lastofnew;		/* last statement in new block */
{
    revinfo *revdummy;			/* never holds anything */
    revinfo *revlistind;		/* index into rev list */
    node *newassign;			/* new assignment for var */
    node *addtoor;			/* where to add a term for newass */
    node *andinor;			/* new term in ORR and'ed with cond */
    node *newassignlist;		/* list of newassigns */

    revinfo *temprevi;
    revinfo *thenrevs;			/* revisions in then clause */
    revinfo *elserevs;			/* revisions in else clause */
    block *bindex;
    int pass;				/* 1 implies then, 2 else */
    node *result;
    node *nindex;

    exprroot->un.bdif.expr = 
	subst_version(exprroot->un.bdif.expr, theBlock, 0, &revdummy);
    thenrevs = elserevs = NULL;
    bindex = subBlock;
    exprroot->un.bdif.thenb = subst_version(exprroot->un.bdif.thenb, 
	    bindex, 1, &thenrevs );
    thenrevs = fix_revlist(thenrevs);
    bindex = bindex->nextclause;
    exprroot->un.bdif.elseb = subst_version( exprroot->un.bdif.elseb,
	    bindex, 1, &elserevs );
    elserevs = fix_revlist(elserevs);

    /* Reduce each x in if to one equation based on other x's */
    newassignlist = NULL;
    revlistind = thenrevs;
    for( pass = 1; pass <=2; pass++ ) {
	while(revlistind != NULL) {
	    /* Make eqn for this var */
	    newassign = node_create_fi( T_EQ, exprroot );
	    newassign->un.assign.rexpr = node_create_fi( T_OR,exprroot );
	    newassign->un.assign.rexpr->un.binary.lexpr = andinor =
		    node_create_fi( T_AND, exprroot );
	    newassign->un.assign.rexpr->un.binary.lexpr->un.binary.rexpr =
	    	    node_dup( exprroot->un.bdif.expr );
	    if ( pass == 1 ) {
		andinor->un.binary.lexpr = (addtoor =
			node_create_fi( T_IS_STATE,exprroot ) );
		addtoor->un.bdvar.name = revlistind->name;
		addtoor->un.bdvar.lbit = revlistind->bit;
		addtoor->un.bdvar.hbit = revlistind->bit;
		addtoor->un.bdvar.rev = revlistind->info;
	    } else {
		andinor->un.binary.lexpr = lastx(revlistind, andinor, theBlock);
	    }

	    newassign->un.assign.rexpr->un.binary.rexpr = (andinor =
		    node_create_fi( T_AND, exprroot ) );
	    andinor->un.binary.rexpr = node_create_fi(T_NOT, exprroot);
	    andinor->un.binary.rexpr->un.unary.expr =
		    node_dup( exprroot->un.bdif.expr );
	    if ( pass == 1 ) {
		else_block( andinor, theBlock, &elserevs, revlistind );
	    } else {
		andinor->un.binary.lexpr = addtoor =
			node_create_fi( T_IS_STATE,exprroot );
		addtoor->un.bdvar.name = revlistind->name;
		addtoor->un.bdvar.lbit = revlistind->bit;
		addtoor->un.bdvar.hbit = revlistind->bit;
		addtoor->un.bdvar.rev = revlistind->info;
	    }
	    newassign->un.assign.lexpr = node_create_fi(T_IS_STATE,exprroot);
	    newassign->un.assign.lexpr->un.bdvar.name = revlistind->name;
	    newassign->un.assign.lexpr->un.bdvar.lbit = revlistind->bit;
	    newassign->un.assign.lexpr->un.bdvar.hbit = revlistind->bit;
	    *revlist = revassign(newassign, theBlock, keeprevs, *revlist);
	    temprevi = revlistind->next;
	    FREE( revlistind );
	    revlistind = temprevi;
	    if (newassignlist == NULL) {
		newassignlist = newassign;
	    } else {
		newassign->next = newassignlist;
		newassignlist = newassign;
	    }
	}
	revlistind = elserevs;
    }
    /* now throw out the if structure and just keep the assigns */
    exprroot->next = NULL;		/* disconnect if */
    result = nindex = exprroot->un.bdif.thenb;
    exprroot->un.bdif.thenb = NULL;
    if ( nindex == NULL ) {
	result = nindex = exprroot->un.bdif.elseb;
    } else {
	while ( nindex->next != NULL ) nindex = nindex->next;
	nindex->next = exprroot->un.bdif.elseb;
    }
    exprroot->un.bdif.elseb = NULL;
    if ( nindex == NULL ) {
	/* stuff in no-op where statements have failed to be */
	result = nindex = node_create_bit( 0, exprroot );
    } else {
	while ( nindex->next != NULL ) nindex = nindex->next;
    }
    node_dispose( exprroot );
    nindex->next = newassignlist;
    while ( nindex->next != NULL ) nindex = nindex->next;
    *lastofnew = nindex;
    return result;
}


/*
 * Searches the else block for another use of a variable in the then
 * block.  It rounds out the generated assignment appropriately.
 */
else_block( andinor, theBlock, elserevs, revtomatch )
node *andinor;			/* new var in or AND condition */
block *theBlock;		/* block of IF statement (not else) */
revinfo **elserevs;		/* revisions in else block */
revinfo *revtomatch;		/* var to look for */
{
    register revinfo *revlistind;	/* index into rev list */
    revinfo *lastrli;			/* last revlistind */
    node *newtoor;			/* last rev of var in else */

    lastrli = NULL;
    revlistind = *elserevs;
    while((revlistind != NIL(revinfo)) &&
	    !((revlistind->name == revtomatch->name) &&
	      (revlistind->bit == revtomatch->bit))) {
	lastrli = revlistind;
	revlistind = revlistind->next;
    }

    if ( revlistind == NULL ) {
	andinor->un.binary.lexpr = lastx(revtomatch, andinor, theBlock);
    } else {
	andinor->un.binary.lexpr = newtoor =
		node_create_fi( T_IS_STATE, andinor );
	newtoor->un.bdvar.name = revtomatch->name;
	newtoor->un.bdvar.lbit = revtomatch->bit;
	newtoor->un.bdvar.hbit = revtomatch->bit;
	newtoor->un.bdvar.rev = revlistind->info;

	if( lastrli == NULL ) {
	    *elserevs = revlistind->next;
	} else {
	    lastrli->next = revlistind->next;
	}
	FREE( revlistind );
    }
}


/*
 * This frees up a revision array attached to a variable
 */
free_revision_struct( p )
symbol *p;
{
    usage *uindex, *temp;
    int i;

    if ( p->u.state.versions != NULL ) {
	for ( i = 0; i < p->u.state.hibit - p->u.state.lobit + 1; i++ ) {
	    uindex = p->u.state.versions[i].recent;
	    while (uindex != NIL(usage)) {
		temp = uindex;
		uindex = uindex->prev;
		FREE(temp);
	    }
	}
	FREE( p->u.state.versions );
    }
}


/*
 * This will free an entire caselist.
 */
free_caseinfo( caselist )
caseinfo *caselist;
{
    caseinfo *temp;

    while ( caselist != NULL ) {
	temp = caselist->next;
	FREE( caselist );
	caselist = temp;
    }
}
