/*
    This file defines the structure of a node in a BDSyn parse tree.
    There are several classes of node.  There are three classes of tree
    arc.  The first points down and indicates arguments of a node.  The
    second points right and implies a concat operation.  The third also
    points right, but it implies the next argument in an argument
    list.

*/

#ifndef symtab_h
typedef struct node_struct node;
#endif

struct node_struct {
    char *file;		/* file of origin for node */
    int line;		/* line number of origin */
    int type;		/* type of node */
    node *next;		/* next statement in list */
    node *conc;		/* link for concatination inside expressions */

    union nodeinf {

	/* T_BUF, T_NOT, T_WIDTH, T_UNARY_PLUS, T_UNARY_MINUS */
	struct unary_inf {			/* unary operator */
	    node *expr;		/* operand */
	} unary;


	/* T_AND, T_NAND, T_OR, T_NOR, T_XOR, T_EQV,
	   T_SL0, T_SR0, T_SL1, T_SR1, T_SLR, T_SRR,
	   T_SXT, T_ZXT, T_OXT,
	   T_PLUS, T_MINUS, T_TIMES, T_DIV, T_MOD,
	   T_EQL, T_NEQ, T_GEQ, T_LEQ, T_GTR, T_LSS,
	   T_COLON (colon operator in a case)
	*/
	struct binary_inf {			/* binary operator */
	    node *lexpr;	/* left operand */
	    node *rexpr;	/* right operand */
	} binary;


	/* T_BIT */
	struct bit_inf {			/* ternary operator */
	    node *lexpr;	/* left expr */
	    node *mexpr;	/* middle expr */
	    node *rexpr;	/* right expr */
	} ternary ;


	/* T_EQ */
	struct assign_inf {			/* assignment operator */
	    node *lexpr;	/* left expr */
	    node *rexpr;	/* right expr */
	    short needlif;	/* lif table required here */
	    short prelifed;	/* pre-lif has seen this node */
	    int lifed;		/* when was node sent to lif */
	} assign ;


	/* T_IF */
	struct if_inf {		/* if statement */
	    node *expr;		/* if condition */
	    node *thenb;	/* then statements */
	    node *elseb;	/* else statements */
	} bdif ;


	/* T_FOR, T_FOR_DOWN */
	struct for_inf {			/* for statement */
	    node *var;		/* variable */
	    node *from;		/* from value */
	    node *to;		/* to value */
	    node *by;		/* by value */
	    node *stmt;		/* statements */
	} bdfor ;


	/* T_CASE */
	struct case_inf {			/* case block in select */
	    node *expr;		/* list of expressions to match */
	    node *stmt;		/* statements to execute on match */
	} bdcase ;


	/* T_SELECT, T_SELECTONE, T_SELECTALL */
	struct select_inf {			/* select statement */
	    node *expr;		/* left side of case tests */
	    node *bdcase;	/* list of cases */
	} select ;


	/* T_IS_STATE */
	struct var_inf {
	    symbol *name;	/* entry in symbol table */
	    int nbit;           /* no declared range */
	    int hbit, lbit;	/* bit range */
	    struct usage_struct *rev;	/* which usage in mult assign */
	    int var;            /* Var param */
	} bdvar ;


	/* T_IS_CONSTANT */
	struct const_inf {
	    int num;		/* constant value */
	    int dont_care;	/* if true, value is dont_care */
	    int hbit;		/* number of MSB */
	} constant ;


	/*  T_IS_ROUTINE  */
	struct fcall_inf {			/* function call */
	    node *actual_args;	/* the actual args to funct */
	    symbol *def;	/* definition for this funct */
	    int hbit;		/* logical hbit of return value */
	} fcall ;

	/* T_LIF */
	struct lif_inf {
	    char *names;	/* .names line */
	    char *table;	/* truth table */
	} lif ;

	/* T_LABEL */
	struct label_inf {
	    symbol *name;	/* label name */
	    node *stmt;		/* label contents */
	} label ;

	/* T_LEAVE */
	struct leave_inf {
	    symbol *name;	/* leave to what label */
	    node *conds;	/* conds that make leave occur */
	} leave ;

	/* Others: T_OTHERWISE */

    } un ;

};

/* Most node "types" come from the yybdsyn.h file; some are not known by the
   parser, but are known in the parse tree.  These are assigned values below.
*/
#define T_LIF 			1000
#define T_IS_STATE		1001
#define T_IS_CONSTANT		1002
#define T_IS_SYNONYM		1003
#define T_IS_ROUTINE		1004
#define T_IS_INPUT_PORT		1005
#define T_IS_OUTPUT_PORT	1006
#define T_IS_MACRO		1007
#define T_IS_UNDEFINED		1008
#define T_DUP			1009
#define T_UNARY_AND		1010
#define T_UNARY_OR		1011
#define T_UNARY_NAND		1012
#define T_UNARY_NOR		1013
#define T_UNARY_XOR		1014
#define T_UNARY_EQV		1015
#define T_UNARY_PLUS		1016
#define T_UNARY_MINUS		1017
#define T_BIT			1018
#define T_FCALL			1019
#define T_CASE			1020
#define T_FOR_DOWN		1021
#define T_IS_ARG		1022
#define T_REQUIRE		1023
#define T_IS_META		1024
#define T_LABEL			1025


void node_init();
node *node_create();
node *node_dup();
node *node_create_bit();
node *node_create_fi();
node *node_eval();
node *node_mkconst();
char *node_lif();
void node_print();
void node_print_expr();
void node_dispose();

#define node_free(x)	FREE(x)
#define node_alloc()	ALLOC(node, 1)
#define NODE_DISPOSE(x) if ( (x) != NULL ) node_dispose(x)
